Package: kerasnip
Type: Package
Title: A Bridge Between 'keras' and 'tidymodels'
Version: 0.1.2
Authors@R: 
    person(
        "David",
        "Díaz",
        role = c("aut", "cph", "cre"),
        email = "daviddrsch@gmail.com",
        comment = c(ORCID = "0000-0002-0927-9795")
    )
Description: Provides a seamless bridge between 'keras' and the 'tidymodels'
    frameworks. It allows for the dynamic creation of 'parsnip' model
    specifications for 'keras' models.
Depends: R (>= 4.1.0)
Encoding: UTF-8
License: MIT + file LICENSE
URL: https://davidrsch.github.io/kerasnip/,
        https://github.com/davidrsch/kerasnip
BugReports: https://github.com/davidrsch/kerasnip/issues
RoxygenNote: 7.3.3
Imports: abind, generics, parsnip (>= 1.0.0), rlang, keras3, tibble,
        purrr, dplyr, cli, recipes, reticulate, lobstr
Config/testthat/edition: 3
Suggests: testthat (>= 3.0.0), bundle, butcher, modeldata, tidymodels,
        finetune, tune, dials, workflows, rsample, knitr, lme4,
        rmarkdown, future, ggplot2, mgcv, probably
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-02 18:54:41 UTC; david
Author: David Díaz [aut, cph, cre] (ORCID:
    <https://orcid.org/0000-0002-0927-9795>)
Maintainer: David Díaz <daviddrsch@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-03 00:30:03 UTC
Built: R 4.6.0; ; 2026-05-03 04:37:35 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(bake,step_collapse)
S3method(butcher::axe_call,kerasnip_model_fit)
S3method(butcher::axe_ctrl,kerasnip_model_fit)
S3method(butcher::axe_data,kerasnip_model_fit)
S3method(butcher::axe_env,kerasnip_model_fit)
S3method(butcher::axe_fitted,kerasnip_model_fit)
S3method(generics::fit,kerasnip_spec)
S3method(generics::fit_xy,kerasnip_spec)
S3method(glance,kerasnip_model_fit)
S3method(parsnip::set_args,kerasnip_spec)
S3method(parsnip::set_engine,kerasnip_spec)
S3method(prep,step_collapse)
S3method(print,step_collapse)
S3method(required_pkgs,step_collapse)
S3method(stats::predict,kerasnip_model_fit)
S3method(tidy,kerasnip_model_fit)
S3method(tidy,step_collapse)
export(compile_keras_grid)
export(create_keras_functional_spec)
export(create_keras_sequential_spec)
export(extract_keras_history)
export(extract_keras_model)
export(extract_valid_grid)
export(generic_functional_fit)
export(generic_sequential_fit)
export(get_model_env)
export(inform_errors)
export(inp_spec)
export(keras_evaluate)
export(keras_losses)
export(keras_metrics)
export(keras_optimizers)
export(loss_function_keras)
export(model_exists)
export(optimizer_function)
export(process_x_functional)
export(process_x_sequential)
export(process_y_functional)
export(process_y_sequential)
export(register_keras_loss)
export(register_keras_metric)
export(register_keras_optimizer)
export(remove_keras_spec)
export(step_collapse)
importFrom(cli,cli_alert_danger)
importFrom(cli,cli_alert_info)
importFrom(cli,cli_alert_success)
importFrom(cli,cli_bullets)
importFrom(cli,cli_code)
importFrom(cli,cli_h1)
importFrom(cli,cli_h2)
importFrom(cli,cli_text)
importFrom(dplyr,bind_rows)
importFrom(dplyr,filter)
importFrom(dplyr,select)
importFrom(generics,fit)
importFrom(generics,fit_xy)
importFrom(generics,glance)
importFrom(generics,required_pkgs)
importFrom(generics,tidy)
importFrom(keras3,to_categorical)
importFrom(lobstr,obj_size)
importFrom(parsnip,set_args)
importFrom(parsnip,set_engine)
importFrom(parsnip,update_dot_check)
importFrom(recipes,bake)
importFrom(recipes,is_trained)
importFrom(recipes,prep)
importFrom(recipes,sel2char)
importFrom(reticulate,py_require)
importFrom(rlang,arg_match)
importFrom(rlang,dots_list)
importFrom(rlang,enquos)
importFrom(rlang,env_poke)
importFrom(stats,predict)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(utils,tail)

# kerasnip

<!-- badges: start -->

[![R-CMD-check](https://github.com/davidrsch/kerasnip/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/davidrsch/kerasnip/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/davidrsch/kerasnip/graph/badge.svg)](https://app.codecov.io/gh/davidrsch/kerasnip)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/kerasnip)](https://cran.r-project.org/package=kerasnip)
[![Downloads](https://cranlogs.r-pkg.org/badges/last-month/kerasnip)](https://cran.r-project.org/package=kerasnip)
[![Downloads](https://cranlogs.r-pkg.org/badges/grand-total/kerasnip)](https://cran.r-project.org/package=kerasnip)

<!-- badges: end -->

The goal of `kerasnip` is to provide a seamless bridge between the `keras` and `tidymodels` frameworks. It allows for the dynamic creation of `parsnip` model specifications for Keras models, making them fully compatible with `tidymodels` workflows.

## Installation

You can install the development version of `kerasnip` from GitHub with:

```r
# install.packages("pak")
pak::pak("davidrsch/kerasnip")
```

## Example

### Example 1: Building a Sequential MLP

This example shows the core workflow for building a simple, linear stack of layers using `create_keras_sequential_spec()`.

```r
library(kerasnip)
library(tidymodels)
library(keras3)

# 1. Define Keras layer blocks
# The first block initializes the model.
input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}
# Subsequent blocks add layers.
dense_block <- function(model, units = 32) {
  model |> layer_dense(units = units, activation = "relu")
}
# The final block creates the output layer.
output_block <- function(model) {
  model |>
    layer_dense(units = 1)
}

# 2. Create a spec from the layer blocks
# This creates a new model function, `basic_mlp()`, in your environment.
create_keras_sequential_spec(
  model_name = "basic_mlp",
  layer_blocks = list(
    input = input_block,
    dense = dense_block,
    output = output_block
  ),
  mode = "regression"
)

# 3. Use the generated spec to define a model.
# We can set the number of dense layers (`num_dense`) and their parameters (`dense_units`).
spec <- basic_mlp(
  num_dense = 2,
  dense_units = 64,
  fit_epochs = 10,
  learn_rate = 0.01
) |>
  set_engine("keras")

# 4. Fit the model within a tidymodels workflow
rec <- recipe(mpg ~ ., data = mtcars) |>
  step_normalize(all_numeric_predictors())

wf <- workflow(rec, spec)

set.seed(123)
fit_obj <- fit(wf, data = mtcars)

# 5. Make predictions
predict(fit_obj, new_data = mtcars[1:5, ])
#> # A tibble: 5 × 1
#>   .pred
#>   <dbl>
#> 1  21.3
#> 2  21.3
#> 3  22.8
#> 4  21.4
#> 5  18.7
```

### Example 2: Building a Functional "Fork-Join" Model

For complex, non-linear architectures, use `create_keras_functional_spec()`. This example builds a model where the input is forked into two paths, which are then concatenated.

```r
library(kerasnip)
library(tidymodels)
library(keras3)

# 1. Define blocks. For the functional API, blocks are nodes in a graph.
input_block <- function(input_shape) layer_input(shape = input_shape)
path_block <- function(tensor, units = 16) tensor |> layer_dense(units = units)
concat_block <- function(input_a, input_b) layer_concatenate(list(input_a, input_b))
output_block <- function(tensor) layer_dense(tensor, units = 1)

# 2. Create the spec. The graph is defined by block names and their arguments.
create_keras_functional_spec(
  model_name = "forked_mlp",
  layer_blocks = list(
    main_input = input_block,
    path_a = inp_spec(path_block, "main_input"),
    path_b = inp_spec(path_block, "main_input"),
    concatenated = inp_spec(concat_block, c(input_a = "path_a", input_b = "path_b")),
    output = inp_spec(output_block, "concatenated")
  ),
  mode = "regression"
)

# 3. Use the new spec. Arguments are prefixed with their block name.
spec <- forked_mlp(path_a_units = 16, path_b_units = 8, fit_epochs = 10) |>
  set_engine("keras")

# Fit and predict as usual
set.seed(123)
fit(spec, mpg ~ ., data = mtcars) |>
  predict(new_data = mtcars[1:5, ])
#> # A tibble: 5 × 1
#>   .pred
#>   <dbl>
#> 1  19.4
#> 2  19.5
#> 3  21.9
#> 4  18.6
#> 5  17.9
```

### Example 3: Tuning a Sequential MLP Architecture

This example demonstrates how to tune the number of dense layers and the rate of a final dropout layer, showcasing how to tune both architecture and block hyperparameters simultaneously.

```r
library(kerasnip)
library(tidymodels)
library(keras3)

# 1. Define Keras layer blocks for a tunable MLP
input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}
dense_block <- function(model, units = 32) {
  model |> layer_dense(units = units, activation = "relu")
}
dropout_block <- function(model, rate = 0.2) {
  model |> layer_dropout(rate = rate)
}
output_block <- function(model) {
  model |> layer_dense(units = 1)
}

# 2. Create a spec from the layer blocks
create_keras_sequential_spec(
  model_name = "tunable_mlp",
  layer_blocks = list(
    input = input_block,
    dense = dense_block,
    dropout = dropout_block,
    output = output_block
  ),
  mode = "regression"
)

# 3. Define a tunable model specification
tune_spec <- tunable_mlp(
  num_dense = tune(),
  dense_units = tune(),
  num_dropout = 1,
  dropout_rate = tune(),
  fit_epochs = 10
) |>
  set_engine("keras")

# 4. Set up and run a tuning workflow
rec <- recipe(mpg ~ ., data = mtcars) |>
  step_normalize(all_numeric_predictors())

wf_tune <- workflow(rec, tune_spec)

# Define the tuning grid.
params <- extract_parameter_set_dials(wf_tune) |>
  update(
    num_dense = dials::num_terms(c(1, 3)),
    dense_units = dials::hidden_units(c(8, 64)),
    dropout_rate = dials::dropout(c(0.1, 0.5))
  )
grid <- grid_regular(params, levels = 2)

# 5. Run the tuning
set.seed(456)
folds <- vfold_cv(mtcars, v = 3)

tune_res <- tune_grid(
  wf_tune,
  resamples = folds,
  grid = grid,
  control = control_grid(verbose = FALSE)
)

# 6. Show the best architecture
show_best(tune_res, metric = "rmse")
#> # A tibble: 5 × 7
#>   num_dense dense_units dropout_rate .metric .estimator .mean .config
#>       <int>       <int>        <dbl> <chr>   <chr>      <dbl> <chr>
#> 1         1          64          0.1 rmse    standard    2.92 Preprocessor1_Model02
#> 2         1          64          0.5 rmse    standard    3.02 Preprocessor1_Model08
#> 3         3          64          0.1 rmse    standard    3.15 Preprocessor1_Model04
#> 4         1           8          0.1 rmse    standard    3.20 Preprocessor1_Model01
#> 5         3           8          0.1 rmse    standard    3.22 Preprocessor1_Model03
```

#' Dials Parameter for Keras Optimizers
#' @param values A character vector of possible optimizers. Defaults to all
#'   known optimizers (keras defaults + custom registered).
#' @keywords internal
#' @export
#' @return A `dials` parameter object for Keras optimizers.
optimizer_function <- function(values = NULL) {
  if (is.null(values)) {
    values <- unique(c(
      keras_optimizers,
      names(.kerasnip_custom_objects$optimizers)
    ))
  }
  dials::new_qual_param(
    type = "character",
    values = values,
    label = c(optimizer_function = "Optimizer Function"),
    finalize = NULL
  )
}

#' Dials Parameter for Keras Loss Functions
#' @param values A character vector of possible loss functions. Defaults to all
#'   known losses (keras defaults + custom registered).
#' @keywords internal
#' @export
#' @return A `dials` parameter object for Keras loss.
loss_function_keras <- function(values = NULL) {
  if (is.null(values)) {
    values <- unique(c(keras_losses, names(.kerasnip_custom_objects$losses)))
  }
  dials::new_qual_param(
    type = "character",
    values = values,
    label = c(loss_function_keras = "Loss Function"),
    finalize = NULL
  )
}

test_that("E2E: Tuning fit_* and compile_* parameters works", {
  skip_if_no_keras()

  # 1. Define a reusable spec
  model_name <- "tune_fit_compile_spec"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)
  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      dense = function(model, units = 10, input_shape) {
        keras3::keras_model_sequential(input_shape = input_shape) |>
          keras3::layer_dense(units = units)
      },
      output = function(model, num_classes) {
        model |>
          keras3::layer_dense(units = num_classes, activation = "softmax")
      }
    ),
    mode = "classification"
  )

  # 2. Define the model with tunable parameters
  tune_spec <- tune_fit_compile_spec(
    dense_units = tune(),
    fit_batch_size = tune(),
    fit_epochs = tune(),
    compile_optimizer = tune(),
    compile_loss = tune(),
    learn_rate = tune()
  ) |>
    set_engine("keras")

  # 3. Set up workflow and tuning grid
  rec <- recipes::recipe(Species ~ ., data = iris)
  tune_wf <- workflows::workflow(rec, tune_spec)

  params <- tune::extract_parameter_set_dials(tune_wf) |>
    update(
      dense_units = dials::hidden_units(c(8L, 16L)),
      fit_batch_size = dials::batch_size(range = c(16L, 32L), trans = NULL),
      fit_epochs = dials::epochs(range = c(5L, 10L)),
      compile_optimizer = optimizer_function(values = c("adam", "sgd")),
      compile_loss = loss_function_keras(
        values = c("categorical_crossentropy", "kl_divergence")
      ),
      learn_rate = dials::learn_rate(range = c(0.001, 0.01), trans = NULL)
    )

  grid <- dials::grid_regular(params, levels = 2)

  control <- tune::control_grid(save_pred = FALSE, verbose = FALSE)

  # 4. Run tuning
  tune_res <- tune::tune_grid(
    tune_wf,
    resamples = rsample::vfold_cv(iris, v = 2),
    grid = grid,
    control = control
  )

  # 5. Assertions
  expect_s3_class(tune_res, "tune_results")
  metrics <- tune::collect_metrics(tune_res)
  expect_true(all(
    c(
      "dense_units",
      "fit_batch_size",
      "fit_epochs",
      "compile_optimizer",
      "compile_loss",
      "learn_rate"
    ) %in%
      names(metrics)
  ))
  expect_equal(sort(unique(metrics$dense_units)), c(8, 16))
  expect_equal(sort(unique(metrics$fit_batch_size)), c(16, 32))
  expect_equal(sort(unique(metrics$fit_epochs)), c(5, 10))
  expect_equal(sort(unique(metrics$compile_optimizer)), c("adam", "sgd"))
  expect_equal(
    sort(unique(metrics$compile_loss)),
    c("categorical_crossentropy", "kl_divergence")
  )
  expect_equal(sort(unique(metrics$learn_rate)), c(0.001, 0.01))
})

#' Collapse Predictors into a single list-column
#'
#' `step_collapse()` creates a a *specification* of a recipe step that will
#'  convert a group of predictors into a single list-column. This is useful
#'  for custom models that need the predictors in a different format.
#'
#' @param recipe A recipe object. The step will be added to the sequence of
#'   operations for this recipe.
#' @param ... One or more selector functions to choose which variables are
#'   affected by the step. See `[selections()]` for more details. For the `tidy`
#'   method, these are not currently used.
#' @param role For model terms created by this step, what analysis role should
#'   they be assigned?. By default, the new columns are used as predictors.
#' @param trained A logical to indicate if the quantities for preprocessing
#'   have been estimated.
#' @param columns A character string of the selected variable names. This is
#'   `NULL` until the step is trained by `[prep.recipe()]`.
#' @param new_col A character string for the name of the new list-column. The
#'   default is "predictor_matrix".
#' @param skip A logical. Should the step be skipped when the recipe is
#'   baked by `[bake.recipe()]`? While all operations are baked when `prep` is
#'   run, skipping when `bake` is run may be other times when it is desirable to
#'   skip a processing step.
#' @param id A character string that is unique to this step to identify it.
#'
#' @return An updated version of `recipe` with the new step added to the
#'   sequence of existing steps (if any). For the `tidy` method, a tibble with
#'   columns `terms` (the selected column names), `value` (the name of the
#'   destination list-column), and `id` (the step identifier).
#'
#' @examples
#' library(recipes)
#'
#' # 2 predictors
#' dat <- data.frame(
#'   x1 = 1:10,
#'   x2 = 11:20,
#'   y = 1:10
#' )
#'
#' rec <- recipe(y ~ ., data = dat) %>%
#'   step_collapse(x1, x2, new_col = "pred") %>%
#'   prep()
#'
#' bake(rec, new_data = NULL)
#' @importFrom recipes prep bake is_trained sel2char
#' @importFrom generics tidy
#' @importFrom tibble tibble
#' @export
step_collapse <- function(
  recipe,
  ...,
  role = "predictor",
  trained = FALSE,
  columns = NULL,
  new_col = "predictor_matrix",
  skip = FALSE,
  id = recipes::rand_id("collapse")
) {
  recipes::add_step(
    recipe,
    step_collapse_new(
      terms = enquos(...),
      role = role,
      trained = trained,
      columns = columns,
      new_col = new_col,
      skip = skip,
      id = id
    )
  )
}

step_collapse_new <- function(
  terms,
  role,
  trained,
  columns,
  new_col,
  skip,
  id
) {
  recipes::step(
    subclass = "collapse",
    terms = terms,
    role = role,
    trained = trained,
    columns = columns,
    new_col = new_col,
    skip = skip,
    id = id
  )
}

#' @export
prep.step_collapse <- function(x, training, info = NULL, ...) {
  col_names <- recipes::recipes_eval_select(x$terms, training, info)

  step_collapse_new(
    terms = x$terms,
    role = x$role,
    trained = TRUE,
    columns = col_names,
    new_col = x$new_col,
    skip = x$skip,
    id = x$id
  )
}

#' @export
bake.step_collapse <- function(object, new_data, ...) {
  if (object$skip) {
    return(new_data)
  }
  if (length(object$columns) == 0) {
    return(new_data)
  }
  recipes::check_new_data(object$columns, object, new_data)

  rows_list <- apply(
    new_data[, object$columns, drop = FALSE],
    1,
    function(row) matrix(row, nrow = 1),
    simplify = FALSE
  )

  new_data[[object$new_col]] <- rows_list

  # drop original predictor columns
  new_data <- new_data[, setdiff(names(new_data), object$columns), drop = FALSE]

  new_data
}

#' @export
print.step_collapse <- function(x, ...) {
  if (is.null(x$columns) || length(x$columns) == 0) {
    cat("Collapse predictors into list-column (unprepped)\n")
  } else {
    cat(
      "Collapse predictors into list-column:",
      paste(x$columns, collapse = ", "),
      " -> ",
      x$new_col,
      "\n"
    )
  }
  invisible(x)
}

#' @importFrom generics required_pkgs
#' @export
required_pkgs.step_collapse <- function(x, ...) {
  c("kerasnip")
}

#' @export
tidy.step_collapse <- function(x, ...) {
  if (recipes::is_trained(x)) {
    if (length(x$columns) > 0) {
      tibble::tibble(
        terms = x$columns,
        value = x$new_col,
        id = x$id
      )
    } else {
      tibble::tibble(
        terms = character(),
        value = character(),
        id = character()
      )
    }
  } else {
    tibble::tibble(
      terms = recipes::sel2char(x$terms),
      value = NA_character_,
      id = x$id
    )
  }
}

dat <- data.frame(
  x1 = 1:10,
  x2 = 11:20,
  x3 = 21:30,
  y = 1:10
)

test_that("step_collapse works with basic selectors", {
  rec <- recipe(y ~ ., data = dat) |>
    step_collapse(x1, x2)

  prepped_rec <- prep(rec)
  baked_dat <- bake(prepped_rec, new_data = NULL)

  expect_equal(ncol(baked_dat), 3)
  expect_true("predictor_matrix" %in% names(baked_dat))
  expect_true(is.list(baked_dat$predictor_matrix))
  expect_equal(length(baked_dat$predictor_matrix), 10)
  expect_equal(vapply(baked_dat$predictor_matrix, nrow, numeric(1)), rep(1, 10))
  expect_equal(vapply(baked_dat$predictor_matrix, ncol, numeric(1)), rep(2, 10))
})

test_that("step_collapse works with tidyselect helpers", {
  rec <- recipe(y ~ ., data = dat) |>
    step_collapse(starts_with("x"))

  prepped_rec <- prep(rec)
  baked_dat <- bake(prepped_rec, new_data = NULL)

  expect_equal(ncol(baked_dat), 2)
  expect_true("predictor_matrix" %in% names(baked_dat))
  expect_true(is.list(baked_dat$predictor_matrix))
  expect_equal(length(baked_dat$predictor_matrix), 10)
  expect_equal(vapply(baked_dat$predictor_matrix, nrow, numeric(1)), rep(1, 10))
  expect_equal(vapply(baked_dat$predictor_matrix, ncol, numeric(1)), rep(3, 10))
})

test_that("new_col argument works", {
  rec <- recipe(y ~ ., data = dat) |>
    step_collapse(x1, x2, new_col = "collapsed_predictors")

  prepped_rec <- prep(rec)
  baked_dat <- bake(prepped_rec, new_data = NULL)

  expect_true("collapsed_predictors" %in% names(baked_dat))
})

test_that("skip argument works", {
  rec <- recipe(y ~ ., data = dat) |>
    step_collapse(x1, x2, skip = TRUE)

  prepped_rec <- prep(rec)
  baked_dat <- bake(prepped_rec, new_data = NULL)

  expect_equal(ncol(baked_dat), 4)
  expect_false("predictor_matrix" %in% names(baked_dat))
})

test_that("print method works", {
  rec <- recipe(y ~ ., data = dat) |>
    step_collapse(x1, x2)

  expect_snapshot(print(rec$steps[[1]]))

  prepped_rec <- prep(rec)
  expect_snapshot(print(prepped_rec$steps[[1]]))
})

test_that("print method works for prepped recipe", {
  rec <- recipe(y ~ ., data = dat) |>
    step_collapse(x1, x2) |>
    prep()

  expect_snapshot(print(rec$steps[[1]]))
})

test_that("step_collapse handles no selectors", {
  rec <- recipe(y ~ ., data = dat) |>
    step_collapse()

  prepped_rec <- prep(rec)
  baked_dat <- bake(prepped_rec, new_data = NULL)

  expect_equal(ncol(baked_dat), 4)
  expect_false("predictor_matrix" %in% names(baked_dat))
})

test_that("step_collapse handles selectors that don't match", {
  rec <- recipe(y ~ ., data = dat) |>
    step_collapse(non_existent_col)

  expect_error(prep(rec))
})

test_that("required_pkgs.step_collapse returns kerasnip", {
  rec <- recipe(y ~ ., data = dat) |>
    step_collapse(x1, x2)
  expect_equal(recipes::required_pkgs(rec$steps[[1]]), "kerasnip")
})

test_that("tidy.step_collapse works before and after prep", {
  rec <- recipe(y ~ ., data = dat) |>
    step_collapse(x1, x2, new_col = "pred")

  # Before prep: terms from selector, NA value
  unprepped_tidy <- tidy(rec, number = 1)
  expect_s3_class(unprepped_tidy, "tbl_df")
  expect_named(unprepped_tidy, c("terms", "value", "id"))
  expect_true(all(is.na(unprepped_tidy$value)))

  # After prep: actual column names and destination
  prepped_rec <- prep(rec)
  prepped_tidy <- tidy(prepped_rec, number = 1)
  expect_s3_class(prepped_tidy, "tbl_df")
  expect_named(prepped_tidy, c("terms", "value", "id"))
  expect_setequal(prepped_tidy$terms, c("x1", "x2"))
  expect_true(all(prepped_tidy$value == "pred"))
})

test_that("tidy.step_collapse returns empty tibble when trained with no columns", {
  # prep() errors if selectors match nothing, so the only way columns can be
  # empty post-prep is via direct construction (defensive branch).
  empty_step <- kerasnip:::step_collapse_new(
    terms = rlang::quos(),
    role = "predictor",
    trained = TRUE,
    columns = character(0),
    new_col = "predictor_matrix",
    skip = FALSE,
    id = "collapse_empty"
  )
  result <- tidy(empty_step)
  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 0L)
  expect_named(result, c("terms", "value", "id"))
})

#' Collect and Finalize Compilation Arguments
#'
#' @description
#' This internal helper extracts all arguments prefixed with `compile_` from a
#' list of arguments, resolves them, and combines them with defaults.
#'
#' @details
#' This function orchestrates the compilation setup. It gives precedence to
#' user-provided arguments (e.g., `compile_optimizer`) over the mode-based
#' defaults. It handles the special logic for the `optimizer`, where a string
#' name (e.g., `"sgd"`) is resolved to a Keras optimizer object, applying the
#' top-level `learn_rate` if necessary. It also resolves string names for `loss`
#' and `metrics` using `get_keras_object()`.
#'
#' @param all_args The list of all arguments passed to the fitting
#'   function's `...`.
#' @param learn_rate The top-level `learn_rate` parameter.
#' @param default_loss The default loss function to use if not provided. Can
#'   be a single value or a named list.
#' @param default_metrics The default metric(s) to use if not provided. Can
#'   be a single value or a named list of vectors/single values.
#' @return A named list of arguments ready to be passed to `keras3::compile()`.
#' @noRd
collect_compile_args <- function(
  all_args,
  learn_rate,
  default_loss,
  default_metrics,
  get_object_fn = get_keras_object
) {
  compile_arg_names <- names(all_args)[startsWith(names(all_args), "compile_")]
  user_compile_args <- all_args[compile_arg_names]
  names(user_compile_args) <- sub("^compile_", "", names(user_compile_args))

  # --- 3a. Resolve and Finalize Compile Arguments ---
  final_compile_args <- list()

  # Determine the final optimizer object, ensuring `learn_rate` is applied.
  optimizer_arg <- user_compile_args$optimizer %||% NULL
  if (!is.null(optimizer_arg)) {
    if (is.character(optimizer_arg)) {
      # Resolve string to object, passing the learn_rate
      final_compile_args$optimizer <- get_object_fn(
        optimizer_arg,
        "optimizer",
        learning_rate = learn_rate
      )
    } else {
      # User passed a pre-constructed optimizer object, use it as is.
      # We assume they have configured the learning rate within it.
      final_compile_args$optimizer <- optimizer_arg
    }
  } else {
    # No optimizer provided, use the default (Adam) with the learn_rate.
    final_compile_args$optimizer <- keras3::optimizer_adam(
      learning_rate = learn_rate
    )
  }

  # Handle loss: can be single or multiple outputs
  if (is.list(default_loss) && !is.null(names(default_loss))) {
    # Multiple outputs
    # User can provide a single loss for all outputs, or a named list
    loss_arg <- user_compile_args$loss %||% default_loss
    if (is.character(loss_arg) && length(loss_arg) == 1) {
      # Single loss string for all outputs
      final_compile_args$loss <- get_object_fn(loss_arg, "loss")
    } else if (is.list(loss_arg) && !is.null(names(loss_arg))) {
      # Named list of losses
      final_compile_args$loss <- lapply(loss_arg, function(l) {
        if (is.character(l)) get_object_fn(l, "loss") else l
      })
    } else {
      stop(
        "For multiple outputs, 'compile_loss' must be a single string",
        " or a named list of losses."
      )
    }
  } else {
    # Single output
    loss_arg <- user_compile_args$loss %||% default_loss
    if (is.character(loss_arg)) {
      final_compile_args$loss <- get_object_fn(loss_arg, "loss")
    } else {
      final_compile_args$loss <- loss_arg
    }
  }

  # Handle metrics: can be single or multiple outputs
  if (is.list(default_metrics) && !is.null(names(default_metrics))) {
    # Multiple outputs
    # User can provide a single metric for all outputs, or a named list
    metrics_arg <- user_compile_args$metrics %||% default_metrics
    if (is.character(metrics_arg) && length(metrics_arg) == 1) {
      # Single metric string for all outputs
      final_compile_args$metrics <- get_object_fn(metrics_arg, "metric")
    } else if (is.list(metrics_arg) && !is.null(names(metrics_arg))) {
      # Named list of metrics
      final_compile_args$metrics <- lapply(metrics_arg, function(m) {
        if (is.character(m)) get_object_fn(m, "metric") else m
      })
    } else {
      stop(
        "For multiple outputs, 'compile_metrics' must be a single string",
        " or a named list of metrics."
      )
    }
  } else {
    # Single output
    metrics_arg <- user_compile_args$metrics %||% default_metrics
    if (is.character(metrics_arg)) {
      final_compile_args$metrics <- lapply(
        metrics_arg,
        get_object_fn,
        "metric"
      )
    } else {
      final_compile_args$metrics <- metrics_arg
    }
  }

  # Add any other user-provided compile arguments (e.g., `weighted_metrics`)
  other_args <- user_compile_args[
    !names(user_compile_args) %in% c("optimizer", "loss", "metrics")
  ]
  final_compile_args <- c(final_compile_args, other_args)
  # Filter out arguments that are NULL or rlang_zap before passing to
  # keras3::compile
  final_compile_args <- final_compile_args[
    !vapply(
      final_compile_args,
      function(x) inherits(x, "rlang_zap"),
      logical(1)
    )
  ]
  final_compile_args
}

#' Collect and Finalize Fitting Arguments
#'
#' @description
#' This internal helper extracts all arguments prefixed with `fit_` from a list
#' of arguments and combines them with the core arguments for `keras3::fit()`.
#'
#' @details
#' It constructs the final list of arguments for `keras3::fit()`. It starts with
#' the required data (`x`, `y`) and the `verbose` setting. It then merges any
#' user-provided arguments from the model specification (e.g., `fit_epochs`,
#' `fit_callbacks`), with the user-provided arguments taking precedence over
#' any defaults.
#'
#' @param x_proc The processed predictor data.
#' @param y_mat The processed outcome data.
#' @param verbose The verbosity level.
#' @param all_args The list of all arguments passed to the fitting
#'   function's `...`.
#' @return A named list of arguments ready to be passed to `keras3::fit()`.
#' @noRd
collect_fit_args <- function(
  x_proc,
  y_mat,
  verbose,
  all_args
) {
  # Collect all arguments starting with "fit_" from `...`
  fit_arg_names <- names(all_args)[startsWith(names(all_args), "fit_")]
  user_fit_args <- all_args[fit_arg_names]
  names(user_fit_args) <- sub("^fit_", "", names(user_fit_args))

  # Build the core argument set. `verbose` can be overridden by `fit_verbose`.
  base_args <- list(
    x = x_proc,
    y = y_mat,
    verbose = verbose
  )

  merged_args <- utils::modifyList(base_args, user_fit_args)

  # Filter out arguments that are NULL or rlang_zap before passing to
  # keras3::fit
  merged_args <- merged_args[
    !vapply(
      merged_args,
      function(x) {
        inherits(x, "rlang_zap")
      },
      logical(1)
    )
  ]
  merged_args
}

# Mock get_keras_object to isolate the logic of collect_compile_args.
# Passed directly as get_object_fn to avoid any namespace manipulation.
mock_get_keras_object <- function(name, type, ...) {
  paste0("mocked_", type, "_", name)
}

test_that("collect_compile_args handles single-output cases correctly", {
  # Case 1: Single output, non-character loss and metrics
  # (get_object_fn not called)
  dummy_loss_obj <- structure(list(), class = "dummy_loss")
  dummy_metric_obj <- structure(list(), class = "dummy_metric")

  args <- collect_compile_args(
    all_args = list(
      compile_loss = dummy_loss_obj,
      compile_metrics = list(dummy_metric_obj)
    ),
    learn_rate = 0.01,
    default_loss = "mse",
    default_metrics = "mae",
    get_object_fn = mock_get_keras_object
  )
  expect_equal(args$loss, dummy_loss_obj)
  expect_equal(args$metrics, list(dummy_metric_obj))
})

test_that("collect_compile_args handles multi-output cases correctly", {
  # Case 2: Multi-output, single string for loss and metrics
  args <- collect_compile_args(
    all_args = list(
      compile_loss = "categorical_crossentropy",
      compile_metrics = "accuracy"
    ),
    learn_rate = 0.01,
    default_loss = list(out1 = "mse", out2 = "mae"),
    default_metrics = list(out1 = "mse", out2 = "mae"),
    get_object_fn = mock_get_keras_object
  )
  expect_equal(args$loss, "mocked_loss_categorical_crossentropy")
  expect_equal(args$metrics, "mocked_metric_accuracy")

  # Case 3: Multi-output, named list with mixed types
  dummy_loss_obj_2 <- structure(list(), class = "dummy_loss_2")
  args_mixed <- collect_compile_args(
    all_args = list(
      compile_loss = list(out1 = "mae", out2 = dummy_loss_obj_2)
    ),
    learn_rate = 0.01,
    default_loss = list(out1 = "mse", out2 = "mae"),
    default_metrics = list(out1 = "mse", out2 = "mae"),
    get_object_fn = mock_get_keras_object
  )
  expect_equal(args_mixed$loss$out1, "mocked_loss_mae")
  expect_equal(args_mixed$loss$out2, dummy_loss_obj_2)
})

test_that("collect_compile_args handles named list of metrics (multi-output)", {
  # Test case: Named list of metrics with mixed types (character and object)
  dummy_metric_obj_3 <- structure(list(), class = "dummy_metric_3")
  args_mixed_metrics <- collect_compile_args(
    all_args = list(
      compile_metrics = list(out1 = "accuracy", out2 = dummy_metric_obj_3)
    ),
    learn_rate = 0.01,
    default_loss = list(out1 = "mse", out2 = "mae"),
    default_metrics = list(out1 = "mse", out2 = "mae"),
    get_object_fn = mock_get_keras_object
  )
  expect_equal(args_mixed_metrics$metrics$out1, "mocked_metric_accuracy")
  expect_equal(args_mixed_metrics$metrics$out2, dummy_metric_obj_3)

  # Test case: Named list of metrics with all characters
  args_all_char_metrics <- collect_compile_args(
    all_args = list(
      compile_metrics = list(out1 = "accuracy", out2 = "mse")
    ),
    learn_rate = 0.01,
    default_loss = list(out1 = "mse", out2 = "mae"),
    default_metrics = list(out1 = "mse", out2 = "mae"),
    get_object_fn = mock_get_keras_object
  )
  expect_equal(args_all_char_metrics$metrics$out1, "mocked_metric_accuracy")
  expect_equal(args_all_char_metrics$metrics$out2, "mocked_metric_mse")

  # Test case: Named list of metrics with all objects
  dummy_metric_obj_4 <- structure(list(), class = "dummy_metric_4")
  dummy_metric_obj_5 <- structure(list(), class = "dummy_metric_5")
  args_all_obj_metrics <- collect_compile_args(
    all_args = list(
      compile_metrics = list(
        out1 = dummy_metric_obj_4,
        out2 = dummy_metric_obj_5
      )
    ),
    learn_rate = 0.01,
    default_loss = list(out1 = "mse", out2 = "mae"),
    default_metrics = list(out1 = "mse", out2 = "mae"),
    get_object_fn = mock_get_keras_object
  )
  expect_equal(args_all_obj_metrics$metrics$out1, dummy_metric_obj_4)
  expect_equal(args_all_obj_metrics$metrics$out2, dummy_metric_obj_5)
})

test_that("collect_compile_args throws errors for invalid multi-output args", {
  # Case 4: Multi-output, invalid loss argument
  expect_error(
    collect_compile_args(
      all_args = list(compile_loss = list("a", "b")), # Unnamed list
      learn_rate = 0.01,
      default_loss = list(out1 = "mse", out2 = "mae"),
      default_metrics = list(out1 = "mse", out2 = "mae"),
      get_object_fn = mock_get_keras_object
    ),
    "For multiple outputs, 'compile_loss' must be a single string"
  )

  # Case 5: Multi-output, invalid metrics argument
  expect_error(
    collect_compile_args(
      all_args = list(compile_metrics = list("a", "b")), # Unnamed list
      learn_rate = 0.01,
      default_loss = list(out1 = "mse", out2 = "mae"),
      default_metrics = list(out1 = "mse", out2 = "mae"),
      get_object_fn = mock_get_keras_object
    ),
    "For multiple outputs, 'compile_metrics' must be a single string"
  )
})

#' @name keras_objects
#' @title Dynamically Discovered Keras Objects
#' @description
#' These exported vectors contain the names of optimizers, losses, and metrics
#' discovered from the installed `keras3` package when `kerasnip` is loaded.
#' This ensures that `kerasnip` is always up-to-date with your Keras version.
#' @details
#' These objects are primarily used to provide the default `values` for the
#' `dials` parameter functions, [optimizer_function()] and
#' [loss_function_keras()]. This allows for tab-completion in IDEs and
#' validation of optimizer and loss names when tuning models.
#'
#' The discovery process in `.onLoad()` scrapes the `keras3` namespace for
#' functions matching `optimizer_*`, `loss_*`, and `metric_*` patterns.
#' @keywords internal
NULL

#' @rdname keras_objects
#' @export
keras_optimizers <- NULL

#' @rdname keras_objects
#' @export
keras_losses <- NULL

#' @rdname keras_objects
#' @export
keras_metrics <- NULL

# These will be populated by .onLoad to hold the names of arguments from
# the keras3::fit and keras3::compile functions.
keras_fit_arg_names <- NULL
keras_compile_arg_names <- NULL


#' Populate Keras Object Lists on Package Load
#'
#' @description
#' This `.onLoad` hook is executed when the `kerasnip` package is loaded. Its
#' main purpose is to inspect the installed `keras3` package and populate the
#' `keras_optimizers`, `keras_losses`, `keras_metrics`, `keras_fit_arg_names`,
#' and `keras_compile_arg_names` vectors.
#'
#' @details
#' The function works by:
#' \enumerate{
#'   \item Checking if `keras3` is installed.
#'   \item Discovering the names of arguments for `keras3::fit()` and
#'     `keras3::compile()`. These are used by `create_keras_*_spec()` to
#'     dynamically generate the `fit_*` and `compile_*` arguments for the
#'     model specification, allowing users to control fitting and compilation
#'     directly from the spec.
#'   \item Listing all functions in the `keras3` namespace that match the
#'     patterns `optimizer_*`, `loss_*`, and `metric_*`.
#'   \item For each function, it attempts to extract the default value of
#'     the `name` argument (e.g., for `keras3::optimizer_adam()`, it extracts
#'     `"adam"`).
#'   \item It populates the exported vectors with these discovered names. For
#'     metrics, it also adds a list of common string aliases that Keras accepts.
#' }
#' This dynamic discovery ensures that `kerasnip` automatically supports all
#' objects available in the user's installed version of Keras.
#' @importFrom reticulate py_require
#' @noRd
.onLoad <- function(libname, pkgname) {
  py_require(c("keras", "pydot", "scipy", "pandas", "Pillow", "ipython"))

  # Helper to get the default string name from a Keras function's `name`
  # argument
  get_keras_default_name <- function(fn_name, keras_ns) {
    fn <- get(fn_name, envir = keras_ns)
    args <- formals(fn)
    if (!is.null(args$name)) {
      # The default value is a string, e.g., name = "adam"
      return(as.character(args$name))
    }
    NA_character_
  }

  if (!requireNamespace("keras3", quietly = TRUE)) {
    return()
  }
  keras_ns <- asNamespace("keras3")

  opt_fns <- ls(keras_ns, pattern = "^optimizer_")
  loss_fns <- ls(keras_ns, pattern = "^loss_")
  metric_fns <- ls(keras_ns, pattern = "^metric_")

  # Find and assign the vectors to the package's namespace
  assign(
    "keras_optimizers",
    stats::na.omit(purrr::map_chr(opt_fns, get_keras_default_name, keras_ns)),
    envir = parent.env(environment())
  )
  assign(
    "keras_losses",
    stats::na.omit(
      purrr::map_chr(loss_fns, get_keras_default_name, keras_ns)
    ),
    envir = parent.env(environment())
  )

  # For metrics, we find the metric_* functions and also add common string
  # aliases that Keras accepts, as they don't all have function wrappers.
  discovered_metrics <- stats::na.omit(purrr::map_chr(
    metric_fns,
    get_keras_default_name,
    keras_ns
  ))
  common_metric_aliases <- c(
    "mae",
    "mape",
    "mse",
    "msle",
    "auc",
    "poisson",
    "cosine_similarity"
  )
  all_metrics <- unique(sort(c(discovered_metrics, common_metric_aliases)))

  assign("keras_metrics", all_metrics, envir = parent.env(environment()))

  # Discover and store fit() and compile() arguments
  fit_args <- names(formals(get(
    "fit.keras.src.models.model.Model",
    asNamespace("keras3")
  )))
  compile_args <- names(formals(get(
    "compile.keras.src.models.model.Model",
    asNamespace("keras3")
  )))

  # Exclude args that are handled specially or don't make sense in the spec
  fit_args_to_exclude <- c("object", "x", "y", "...")
  compile_args_to_exclude <- c("object", "...")

  assign(
    "keras_fit_arg_names",
    setdiff(fit_args, fit_args_to_exclude),
    envir = parent.env(environment())
  )
  assign(
    "keras_compile_arg_names",
    setdiff(compile_args, compile_args_to_exclude),
    envir = parent.env(environment())
  )
}

#' Discover and Collect Model Specification Arguments
#'
#' @description
#' This internal helper introspects the user-provided `layer_blocks` functions
#' to generate a complete list of arguments for the new model specification.
#' The logic for discovering arguments differs for sequential and functional
#' models.
#'
#' @details
#' For **sequential models** (`functional = FALSE`):
#' - It creates `num_{block_name}` arguments to control block repetition.
#' - It inspects the arguments of each block function, skipping the first
#'   (assumed to be the `model` object), to find tunable hyperparameters.
#'
#' For **functional models** (`functional = TRUE`):
#' - It creates `num_{block_name}` arguments to control block repetition.
#' - It inspects the arguments of each block function. Arguments whose names
#'   match other block names are considered graph connections (inputs) and are
#'   ignored. The remaining arguments are treated as tunable hyperparameters.
#'
#' In both cases, it also adds global training parameters (like `epochs`) and
#' filters out special engine-supplied arguments (`input_shape`, `num_classes`).
#'
#' @param layer_blocks A named list of functions defining Keras layer blocks.
#' @param functional A logical. If `TRUE`, uses discovery logic for the
#'   Functional API. If `FALSE`, uses logic for the Sequential API.
#' @return A list containing two elements:
#'
#' @noRd
collect_spec_args <- function(
  layer_blocks,
  functional
) {
  if (any(c("compile", "fit", "optimizer") %in% names(layer_blocks))) {
    stop(
      "`compile`, `fit` and `optimizer` are protected names and\n",
      "cannot be used as layer block names.",
      call. = FALSE
    )
  }

  all_args <- list()
  parsnip_names <- character()

  block_names <- names(layer_blocks)

  # block repetition counts (e.g., num_dense)
  for (block_name in block_names) {
    num_name <- paste0("num_", block_name)
    all_args[[num_name]] <- rlang::zap()
    parsnip_names <- c(parsnip_names, num_name)
  }

  # These args are passed by the fit engine, not set by the user in the spec
  engine_args <- c("input_shape", "num_classes")
  # Discover block-specific hyperparameters
  for (block_name in block_names) {
    block_fmls <- rlang::fn_fmls(layer_blocks[[block_name]])

    if (isTRUE(functional)) {
      # For functional models, hyperparameters are arguments that are NOT
      # names of other blocks (which are graph connections).
      hyperparam_names <- setdiff(
        names(block_fmls),
        c(block_names, engine_args)
      )
    } else {
      # For sequential models, hyperparameters are all args except the first
      # ('model') and special engine args.
      fmls_to_process <- if (length(block_fmls) > 0) block_fmls[-1] else list()
      hyperparam_names <- names(fmls_to_process)[
        !names(fmls_to_process) %in% engine_args
      ]
    }

    for (arg in hyperparam_names) {
      full <- paste0(block_name, "_", arg)
      all_args[[full]] <- rlang::zap()
      parsnip_names <- c(parsnip_names, full)
    }
  }

  # Add global training and compile parameters dynamically
  # These are discovered from keras3::fit and keras3::compile in zzz.R
  fit_params <- if (length(keras_fit_arg_names) > 0) {
    paste0("fit_", keras_fit_arg_names)
  } else {
    character()
  }
  compile_params <- if (length(keras_compile_arg_names) > 0) {
    paste0("compile_", keras_compile_arg_names)
  } else {
    character()
  }

  # learn_rate is a special convenience argument for the default optimizer
  special_params <- "learn_rate"

  dynamic_global_args <- c(special_params, fit_params, compile_params)

  for (g in dynamic_global_args) {
    all_args[[g]] <- rlang::zap()
    parsnip_names <- c(parsnip_names, g)
  }

  list(all_args = all_args, parsnip_names = parsnip_names)
}

#' Remap Layer Block Arguments for Model Specification
#'
#' @description
#' Creates a wrapper function around a Keras layer block to rename its
#' arguments. This is a powerful helper for defining the `layer_blocks` in
#' [create_keras_functional_spec()] and [create_keras_sequential_spec()],
#' allowing you to connect reusable blocks into a model graph without writing
#' verbose anonymous functions.
#'
#' @details
#' `inp_spec()` makes your model definitions cleaner and more readable. It
#' handles the metaprogramming required to create a new function with the
#' correct argument names, while preserving the original block's hyperparameters
#' and their default values.
#'
#' The function supports two modes of operation based on `input_map`:
#' 1.  **Single Input Renaming**: If `input_map` is a single character string,
#'     the wrapper function renames the *first* argument of the `block` function
#'     to the provided string. This is the common case for blocks that take a
#'     single tensor input.
#' 2.  **Multiple Input Mapping**: If `input_map` is a named character vector,
#'     the **names must match the argument names of `block`** and each value
#'     must be the name of an upstream layer block whose output should be fed
#'     into that argument. This orientation matches the
#'     syntax (e.g., `c(numeric = "processed_numerical")`). This is used for
#'     blocks with multiple inputs, like a concatenation layer.
#'
#' _Note_: Prior releases accepted the opposite orientation
#' (`c(processed_numerical = "numeric")`). Existing code written in that style
#' must flip the names/values when upgrading to this version.

#'
#' @param block A function that defines a Keras layer or a set of layers. The
#'   first arguments should be the input tensor(s).
#' @param input_map A single character string or a named character vector that
#'   specifies how to rename/remap the arguments of `block`.
#'
#' @return A new function (a closure) that wraps the `block` function with
#'   renamed arguments, ready to be used in a `layer_blocks` list.
#'
#' @export
#' @examples
#' \donttest{
#' # --- Example Blocks ---
#' # A standard dense block with one input tensor and one hyperparameter.
#' dense_block <- function(tensor, units = 16) {
#'   tensor |> keras3::layer_dense(units = units, activation = "relu")
#' }
#'
#' # A block that takes two tensors as input.
#' concat_block <- function(input_a, input_b) {
#'   keras3::layer_concatenate(list(input_a, input_b))
#' }
#'
#' # An output block with one input.
#' output_block <- function(tensor) {
#'   tensor |> keras3::layer_dense(units = 1)
#' }
#'
#' # --- Usage ---
#' layer_blocks <- list(
#'   main_input = keras3::layer_input,
#'   path_a = inp_spec(dense_block, "main_input"),
#'   path_b = inp_spec(dense_block, "main_input"),
#'   concatenated = inp_spec(
#'     concat_block,
#'     c(input_a = "path_a", input_b = "path_b")
#'   ),
#'   output = inp_spec(output_block, "concatenated")
#' )
#' }
inp_spec <- function(block, input_map) {
  new_fun <- function() {}
  original_formals <- formals(block)
  original_names <- names(original_formals)

  if (length(original_formals) == 0) {
    stop("The 'block' function must have at least one argument.")
  }

  new_formals <- original_formals

  if (
    is.character(input_map) &&
      is.null(names(input_map)) &&
      length(input_map) == 1
  ) {
    # Case 1: Single string, rename first argument
    names(new_formals)[1] <- input_map
  } else if (is.character(input_map) && !is.null(names(input_map))) {
    # Case 2: Named vector for mapping (argument-first)
    if (!all(names(input_map) %in% original_names)) {
      missing_args <- names(input_map)[!names(input_map) %in% original_names]
      stop(paste(
        "Argument(s)",
        paste(shQuote(missing_args), collapse = ", "),
        "not found in the block function."
      ))
    }

    new_names <- original_names
    new_names[match(names(input_map), original_names)] <- unname(input_map)
    names(new_formals) <- new_names
  } else {
    stop("`input_map` must be a single string or a named character vector.")
  }

  formals(new_fun) <- new_formals

  call_args <- lapply(names(new_formals), as.symbol)
  names(call_args) <- original_names

  body(new_fun) <- as.call(c(list(as.symbol("block")), call_args))
  environment(new_fun) <- environment()
  new_fun
}

#' Internal Implementation for Creating Keras Specifications
#'
#' @description
#' This is the core implementation for both `create_keras_sequential_spec()` and
#' `create_keras_functional_spec()`. It orchestrates the argument collection,
#' function building, and `parsnip` registration steps.
#'
#' @inheritParams create_keras_sequential_spec
#' @param functional A logical, if `TRUE`, registers the model to be fit with
#'   the Functional API (`generic_functional_fit`). Otherwise, uses the
#'   Sequential API (`generic_sequential_fit`).
#'
#' @noRd
create_keras_spec_impl <- function(
  model_name,
  layer_blocks,
  mode,
  functional,
  env
) {
  args_info <- collect_spec_args(layer_blocks, functional = functional)
  spec_fun <- build_spec_function(
    model_name,
    mode,
    args_info$all_args,
    args_info$parsnip_names,
    layer_blocks,
    functional = functional
  )

  register_core_model(model_name, mode)
  register_model_args(model_name, args_info$parsnip_names)
  register_fit_predict(model_name, mode, layer_blocks, functional = functional)
  register_update_method(model_name, args_info$parsnip_names, env = env)

  rlang::env_poke(env, model_name, spec_fun)
  invisible(NULL)
}

test_that("E2E: Functional spec (classification) works", {
  skip_if_no_keras()

  # Define blocks for a simple forked functional model
  input_block <- function(input_shape) keras3::layer_input(shape = input_shape)
  # Add a default to `units` to work around a bug in the doc generator
  # when handling args with no default. This doesn't affect runtime as the
  # value is always overridden.
  path_block <- function(tensor, units = 16) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  concat_block <- function(input_a, input_b) {
    keras3::layer_concatenate(list(input_a, input_b))
  }
  output_block_class <- function(tensor, num_classes) {
    tensor |> keras3::layer_dense(units = num_classes, activation = "softmax")
  }

  model_name <- "e2e_func_class"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  # Create a spec with two parallel paths that are then concatenated
  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = list(
      main_input = input_block,
      path_a = inp_spec(path_block, "main_input"),
      path_b = inp_spec(path_block, "main_input"),
      concatenated = inp_spec(
        concat_block,
        c(input_a = "path_a", input_b = "path_b")
      ),
      output = inp_spec(output_block_class, "concatenated")
    ),
    mode = "classification"
  )

  spec <- e2e_func_class(
    path_a_units = 8,
    path_b_units = 4,
    fit_epochs = 2
  ) |>
    set_engine("keras")

  data <- iris
  rec <- recipe(Species ~ ., data = data)
  wf <- workflows::workflow(rec, spec)

  expect_no_error(fit_obj <- parsnip::fit(wf, data = data))
  expect_s3_class(fit_obj, "workflow")

  preds_class <- predict(fit_obj, new_data = data[1:5, ], type = "class")
  expect_s3_class(preds_class, "tbl_df")
  expect_equal(names(preds_class), ".pred_class")
  expect_equal(levels(preds_class$.pred_class), levels(data$Species))

  preds_prob <- predict(fit_obj, new_data = data[1:5, ], type = "prob")
  expect_s3_class(preds_prob, "tbl_df")
  expect_equal(names(preds_prob), paste0(".pred_", levels(data$Species)))
  expect_true(all(abs(rowSums(preds_prob) - 1) < 1e-5))
})

test_that("E2E: Functional spec tuning (including repetition) works", {
  skip_if_no_keras()

  input_block <- function(input_shape) keras3::layer_input(shape = input_shape)
  # Add a default to `units` to work around a bug in the doc generator
  # when handling args with no default. This doesn't affect runtime as the
  # value is always overridden by the tuning grid.
  dense_block <- function(tensor, units = 16) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block_class <- function(tensor, num_classes) {
    tensor |> keras3::layer_dense(units = num_classes, activation = "softmax")
  }

  model_name <- "e2e_func_tune"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = list(
      main_input = input_block,
      # This block has a single input, so it can be repeated
      dense_path = inp_spec(dense_block, "main_input"),
      output = inp_spec(output_block_class, "dense_path")
    ),
    mode = "classification"
  )

  tune_spec <- e2e_func_tune(
    num_dense_path = tune(),
    dense_path_units = tune(),
    fit_epochs = 1
  ) |>
    set_engine("keras")

  rec <- recipe(Species ~ ., data = iris)
  tune_wf <- workflows::workflow(rec, tune_spec)

  folds <- rsample::vfold_cv(iris, v = 2)
  params <- extract_parameter_set_dials(tune_wf) |>
    update(
      num_dense_path = num_terms(c(1, 2)),
      dense_path_units = hidden_units(c(4, 8))
    )
  grid <- grid_regular(params, levels = 2)
  control <- control_grid(save_pred = FALSE, verbose = FALSE)

  tune_res <- try(
    tune_grid(
      tune_wf,
      resamples = folds,
      grid = grid,
      control = control
    ),
    silent = TRUE
  )

  if (inherits(tune_res, "try-error")) {
    testthat::skip(paste("Tuning failed with error:", as.character(tune_res)))
  }

  expect_s3_class(tune_res, "tune_results")

  metrics <- collect_metrics(tune_res)
  expect_s3_class(metrics, "tbl_df")
  expect_true(all(c("num_dense_path", "dense_path_units") %in% names(metrics)))
})

test_that("E2E: Multi-input, single-output functional classification works", {
  skip_if_no_keras()
  options(kerasnip.show_removal_messages = FALSE)
  on.exit(options(kerasnip.show_removal_messages = TRUE), add = TRUE)

  # Define layer blocks
  input_block_1 <- function(input_shape) {
    keras3::layer_input(shape = input_shape, name = "input_1")
  }
  input_block_2 <- function(input_shape) {
    keras3::layer_input(shape = input_shape, name = "input_2")
  }
  flatten_block <- function(tensor) keras3::layer_flatten(tensor)
  dense_path <- function(tensor, units = 16) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  concat_block <- function(in_1, in_2) {
    keras3::layer_concatenate(list(in_1, in_2))
  }
  output_block_class <- function(tensor, num_classes) {
    keras3::layer_dense(tensor, units = num_classes, activation = "softmax")
  }

  model_name <- "multi_in_class"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = list(
      input_a = input_block_1,
      input_b = input_block_2,
      flatten_a = inp_spec(flatten_block, "input_a"),
      flatten_b = inp_spec(flatten_block, "input_b"),
      path_a = inp_spec(dense_path, "flatten_a"),
      path_b = inp_spec(dense_path, "flatten_b"),
      concatenated = inp_spec(
        concat_block,
        c(in_1 = "path_a", in_2 = "path_b")
      ),
      output = inp_spec(output_block_class, "concatenated")
    ),
    mode = "classification"
  )

  spec <- multi_in_class(fit_epochs = 2) |> set_engine("keras")

  # Prepare dummy data
  set.seed(123)
  x1 <- matrix(rnorm(100 * 5), ncol = 5)
  x2 <- matrix(rnorm(100 * 3), ncol = 3)
  y <- factor(sample(c("a", "b"), 100, replace = TRUE))

  train_df <- tibble::tibble(
    input_a = lapply(seq_len(nrow(x1)), function(i) x1[i, , drop = FALSE]),
    input_b = lapply(seq_len(nrow(x2)), function(i) x2[i, , drop = FALSE]),
    outcome = y
  )

  rec <- recipe(outcome ~ input_a + input_b, data = train_df)
  wf <- workflows::workflow(rec, spec)

  expect_no_error(fit_obj <- parsnip::fit(wf, data = train_df))

  new_data_df <- tibble::tibble(
    input_a = lapply(seq_len(5), function(i) matrix(rnorm(5), ncol = 5)),
    input_b = lapply(seq_len(5), function(i) matrix(rnorm(3), ncol = 3))
  )
  preds <- predict(fit_obj, new_data = new_data_df)

  expect_s3_class(preds, "tbl_df")
  expect_equal(names(preds), c(".pred_class"))
  expect_equal(nrow(preds), 5)
  expect_true(is.factor(preds$.pred_class))
})

test_that("E2E: Functional spec with pre-constructed optimizer works", {
  skip_if_no_keras()

  # Define blocks for a simple forked functional model
  input_block <- function(input_shape) keras3::layer_input(shape = input_shape)
  path_block <- function(tensor, units = 16) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  concat_block <- function(input_a, input_b) {
    keras3::layer_concatenate(list(input_a, input_b))
  }
  output_block_class <- function(tensor, num_classes) {
    tensor |> keras3::layer_dense(units = num_classes, activation = "softmax")
  }

  model_name <- "e2e_func_class_optimizer"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  # Create a spec with two parallel paths that are then concatenated
  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = list(
      main_input = input_block,
      path_a = inp_spec(path_block, "main_input"),
      path_b = inp_spec(path_block, "main_input"),
      concatenated = inp_spec(
        concat_block,
        c(input_a = "path_a", input_b = "path_b")
      ),
      output = inp_spec(output_block_class, "concatenated")
    ),
    mode = "classification"
  )

  # Define a pre-constructed optimizer
  my_optimizer <- keras3::optimizer_sgd(learning_rate = 0.001)

  spec <- e2e_func_class_optimizer(
    path_a_units = 8,
    path_b_units = 4,
    fit_epochs = 2,
    compile_optimizer = my_optimizer
  ) |>
    set_engine("keras")

  data <- iris
  rec <- recipe(Species ~ ., data = data)
  wf <- workflows::workflow(rec, spec)

  expect_no_error(fit_obj <- parsnip::fit(wf, data = data))
  expect_s3_class(fit_obj, "workflow")
})

test_that("E2E: Functional spec with string loss works", {
  skip_if_no_keras()

  # Define blocks for a simple forked functional model
  input_block <- function(input_shape) keras3::layer_input(shape = input_shape)
  path_block <- function(tensor, units = 16) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  concat_block <- function(input_a, input_b) {
    keras3::layer_concatenate(list(input_a, input_b))
  }
  output_block_class <- function(tensor, num_classes) {
    tensor |> keras3::layer_dense(units = num_classes, activation = "softmax")
  }

  model_name <- "e2e_func_class_loss_string"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  # Create a spec with two parallel paths that are then concatenated
  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = list(
      main_input = input_block,
      path_a = inp_spec(path_block, "main_input"),
      path_b = inp_spec(path_block, "main_input"),
      concatenated = inp_spec(
        concat_block,
        c(input_a = "path_a", input_b = "path_b")
      ),
      output = inp_spec(output_block_class, "concatenated")
    ),
    mode = "classification"
  )

  spec <- e2e_func_class_loss_string(
    path_a_units = 8,
    path_b_units = 4,
    fit_epochs = 2,
    compile_loss = "categorical_crossentropy"
  ) |>
    set_engine("keras")

  data <- iris
  rec <- recipe(Species ~ ., data = data)
  wf <- workflows::workflow(rec, spec)

  expect_no_error(fit_obj <- parsnip::fit(wf, data = data))
  expect_s3_class(fit_obj, "workflow")
})

# --- Test Data ---
x_train <- as.matrix(iris[, 1:4])
y_train <- iris$Species
train_df <- tibble(x = I(x_train), y = y_train)
# --- Tests ---

test_that("compile_keras_grid errors if spec is not a model_spec", {
  expect_error(
    compile_keras_grid("not_a_spec", data.frame(), NULL, NULL),
    "`spec` must be a `parsnip` model specification."
  )
})

test_that("compile_keras_grid errors if grid is not a data frame", {
  spec <- parsnip::rand_forest()
  expect_error(
    compile_keras_grid(spec, "not_a_df", NULL, NULL),
    "`grid` must be a data frame or tibble."
  )
})

test_that("compile_keras_grid errors if grid has zero rows", {
  spec <- parsnip::rand_forest()
  expect_error(
    compile_keras_grid(spec, tibble::tibble(), NULL, NULL),
    "`grid` must have at least one row"
  )
})

test_that("compile_keras_grid works with single-row empty grid (no tuning columns)", {
  skip_on_cran()

  model_name <- "test_seq_spec_compile_notuning"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    mode = "classification",
    layer_blocks = list(
      dense = function(model, units = 32) {
        if (is.null(model)) {
          keras3::keras_model_sequential(input_shape = 4L) |>
            keras3::layer_dense(units = units, activation = "relu")
        } else {
          model |> keras3::layer_dense(units = units, activation = "relu")
        }
      },
      output = function(model, num_classes) {
        model |>
          keras3::layer_dense(units = num_classes, activation = "softmax")
      }
    )
  )

  spec <- test_seq_spec_compile_notuning(dense_units = 16L) |>
    set_engine("keras")

  # tibble(.rows = 1L) = "build once with the spec's current args, no tuning"
  results <- compile_keras_grid(
    spec,
    tibble::tibble(.rows = 1L),
    select(train_df, x),
    select(train_df, y)
  )

  expect_s3_class(results, "tbl_df")
  expect_equal(nrow(results), 1L)
  expect_true("compiled_model" %in% names(results))
  expect_true("error" %in% names(results))
  expect_true(is.na(results$error[[1L]]))
  expect_true(inherits(
    results$compiled_model[[1L]],
    "keras.src.models.model.Model"
  ))
})

test_that("compile_keras_grid works for sequential models", {
  skip_on_cran()

  model_name <- "test_seq_spec_compile"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    mode = "classification",
    layer_blocks = list(
      dense = function(model, units = 32, activation = "relu") {
        if (is.null(model)) {
          keras3::keras_model_sequential(input_shape = 4) |>
            keras3::layer_dense(units = units, activation = activation)
        } else {
          model |> keras3::layer_dense(units = units, activation = activation)
        }
      },
      output = function(model, num_classes) {
        model |>
          keras3::layer_dense(units = num_classes, activation = "softmax")
      }
    )
  )

  spec <- test_seq_spec_compile() |>
    set_engine("keras")

  grid <- tibble::tibble(
    dense_units = c(16, 32),
    learn_rate = c(0.01, 0.001)
  )

  results <- compile_keras_grid(
    spec,
    grid,
    select(train_df, x),
    select(train_df, y)
  )

  expect_s3_class(results, "tbl_df")
  expect_equal(nrow(results), 2)
  expect_true(all(
    c(
      "dense_units",
      "learn_rate",
      "compiled_model",
      "error"
    ) %in%
      names(results)
  ))
  expect_true(all(is.na(results$error)))
  expect_true(all(sapply(
    results$compiled_model,
    inherits,
    "keras.src.models.model.Model"
  )))
})

test_that("compile_keras_grid works for functional models", {
  skip_on_cran()

  model_name <- "test_func_spec_compile"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  input_block <- function(input_shape) {
    keras3::layer_input(shape = input_shape, name = "x")
  }

  dense_block <- function(tensor, units = 32) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }

  output_block <- function(tensor, num_classes) {
    tensor |>
      keras3::layer_dense(
        units = num_classes,
        activation = "softmax",
        name = "y"
      )
  }

  create_keras_functional_spec(
    model_name = model_name,
    mode = "classification",
    layer_blocks = list(
      input = input_block,
      dense = inp_spec(dense_block, "input"),
      output = inp_spec(output_block, "dense")
    )
  )

  spec <- test_func_spec_compile() |>
    set_engine("keras")

  rec <- recipe(y ~ x, data = train_df) # Recipe for two outputs
  wf <- workflow() |>
    add_recipe(rec) |>
    add_model(spec)

  fit_obj <- fit(wf, data = train_df)

  grid <- tibble::tibble(
    dense_units = c(16, 32),
    learn_rate = c(0.01, 0.001)
  )

  results <- compile_keras_grid(
    spec,
    grid,
    select(train_df, x),
    select(train_df, y)
  )

  expect_s3_class(results, "tbl_df")
  expect_equal(nrow(results), 2)
  expect_true(all(
    c(
      "dense_units",
      "learn_rate",
      "compiled_model",
      "error"
    ) %in%
      names(results)
  ))
  expect_true(all(is.na(results$error)))
  expect_true(all(sapply(
    results$compiled_model,
    inherits,
    "keras.src.models.model.Model"
  )))
})

test_that("compile_keras_grid handles errors gracefully", {
  skip_on_cran()

  model_name <- "test_bad_func_spec_compile"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_functional_spec(
    model_name = model_name,
    mode = "classification",
    layer_blocks = list(
      input = function(input_shape) {
        keras3::layer_input(shape = input_shape, name = "x")
      },
      dense1 = function(input, units = 32) {
        input |> keras3::layer_dense(units = units, activation = "relu")
      },
      dense2 = function(units = 16) {
        # Missing input tensor
        keras3::layer_dense(units = units, activation = "relu")
      },
      output = function(dense2, num_classes) {
        dense2 |>
          keras3::layer_dense(units = num_classes, activation = "softmax")
      }
    )
  )

  spec <- test_bad_func_spec_compile() |>
    set_engine("keras")

  grid <- tibble::tibble(dense1_units = 16)

  expect_warning(
    results <- compile_keras_grid(
      spec,
      grid,
      select(train_df, x),
      select(train_df, y)
    ),
    "Block 'dense2' has no inputs from other blocks."
  )

  expect_s3_class(results, "tbl_df")
  expect_equal(nrow(results), 1)
  expect_false(is.na(results$error[1]))
  expect_true(is.null(results$compiled_model[[1]]))
})

# ============================================================
# Helpers shared across save/load tests
# ============================================================

make_seq_blocks <- function() {
  input_block <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block <- function(model, units = 16) {
    model |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }
  list(input = input_block, dense = dense_block, output = output_block)
}

make_func_blocks <- function() {
  input_block <- function(input_shape) keras3::layer_input(shape = input_shape)
  dense_block <- function(tensor, units = 16) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block <- function(tensor) keras3::layer_dense(tensor, units = 1)
  list(
    input = input_block,
    dense = inp_spec(dense_block, "input"),
    output = inp_spec(output_block, "dense")
  )
}

fit_seq_workflow <- function(model_name) {
  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = make_seq_blocks(),
    mode = "regression"
  )
  spec <- get(model_name)(fit_epochs = 2, compile_loss = "mse") |>
    parsnip::set_engine("keras")
  wf <- workflows::workflow(
    recipes::recipe(mpg ~ ., data = mtcars),
    spec
  )
  workflows::fit(wf, data = mtcars)
}

fit_func_workflow <- function(model_name) {
  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = make_func_blocks(),
    mode = "regression"
  )
  spec <- get(model_name)(fit_epochs = 2, compile_loss = "mse") |>
    parsnip::set_engine("keras")
  wf <- workflows::workflow(
    recipes::recipe(mpg ~ ., data = mtcars),
    spec
  )
  workflows::fit(wf, data = mtcars)
}

# ============================================================
# Test 1: Spec carries kerasnip_spec class and metadata
# ============================================================

test_that("spec object has kerasnip_spec class and metadata attributes", {
  skip_if_no_keras()

  model_name <- "sl_class_seq"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = make_seq_blocks(),
    mode = "regression"
  )

  spec <- get(model_name)(fit_epochs = 2)

  expect_true(inherits(spec, "kerasnip_spec"))
  expect_false(is.null(attr(spec, "kerasnip_layer_blocks")))
  expect_false(is.null(attr(spec, "kerasnip_functional")))
  expect_false(attr(spec, "kerasnip_functional"))
})

# ============================================================
# Test 2: fit() produces kerasnip_model_fit class
# ============================================================

test_that("fit() tags result with kerasnip_model_fit class", {
  skip_if_no_keras()

  model_name <- "sl_class_fit"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  fit_obj <- fit_seq_workflow(model_name)
  mf <- workflows::extract_fit_parsnip(fit_obj)

  expect_true(inherits(mf, "kerasnip_model_fit"))
})

# ============================================================
# Test 3: predict() auto-registers after simulated reload (sequential)
# ============================================================

test_that("predict() auto-registers sequential spec when missing", {
  skip_if_no_keras()

  model_name <- "sl_auto_seq"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  fit_obj <- fit_seq_workflow(model_name)

  rds_path <- tempfile(fileext = ".rds")
  on.exit(unlink(rds_path), add = TRUE)
  saveRDS(fit_obj, rds_path)

  # Simulate new session: remove parsnip registration
  suppressMessages(remove_keras_spec(model_name))
  expect_false(model_exists(model_name))

  loaded <- readRDS(rds_path)

  # predict() should auto-register and succeed
  expect_no_error({
    preds <- predict(loaded, new_data = mtcars[1:3, ])
  })
  expect_equal(nrow(preds), 3)
  expect_true(model_exists(model_name))
})

# ============================================================
# Test 4: predict() auto-registers after simulated reload (functional)
# ============================================================

test_that("predict() auto-registers functional spec when missing", {
  skip_if_no_keras()

  model_name <- "sl_auto_func"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  fit_obj <- fit_func_workflow(model_name)

  rds_path <- tempfile(fileext = ".rds")
  on.exit(unlink(rds_path), add = TRUE)
  saveRDS(fit_obj, rds_path)

  suppressMessages(remove_keras_spec(model_name))
  expect_false(model_exists(model_name))

  loaded <- readRDS(rds_path)

  expect_no_error({
    preds <- predict(loaded, new_data = mtcars[1:3, ])
  })
  expect_equal(nrow(preds), 3)
  expect_true(model_exists(model_name))
})

# ============================================================
# Test 5: bundle round-trip + predict works (requires bundle package)
# ============================================================

test_that("bundle/unbundle round-trip works and predict succeeds", {
  skip_if_no_keras()
  skip_if_not_installed("bundle")

  model_name <- "sl_bundle_seq"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  fit_obj <- fit_seq_workflow(model_name)

  bundled_path <- tempfile(fileext = ".rds")
  on.exit(unlink(bundled_path), add = TRUE)

  b <- bundle::bundle(fit_obj)
  saveRDS(b, bundled_path)

  suppressMessages(remove_keras_spec(model_name))
  expect_false(model_exists(model_name))

  b2 <- readRDS(bundled_path)
  restored <- bundle::unbundle(b2)

  expect_no_error({
    preds <- predict(restored, new_data = mtcars[1:3, ])
  })
  expect_equal(nrow(preds), 3)
})

test_that("inp_spec throws error for block with no arguments", {
  block_no_args <- function() {}
  expect_error(
    kerasnip:::inp_spec(block_no_args, "input"),
    "The 'block' function must have at least one argument."
  )
})

test_that("inp_spec throws error for mismatched input_map names", {
  block_with_args <- function(a, b) {}
  input_map_bad <- c(new_a = "a", new_c = "c") # 'c' does not exist
  expect_error(
    kerasnip:::inp_spec(block_with_args, input_map_bad),
    class = "simpleError"
  )
})


test_that("inp_spec supports argument-first mapping", {
  block_with_args <- function(numeric, categorical) {
    list(numeric = numeric, categorical = categorical)
  }
  mapper <- c(
    numeric = "processed_numeric",
    categorical = "processed_categorical"
  )
  wrapped <- kerasnip:::inp_spec(block_with_args, mapper)

  expect_identical(
    names(formals(wrapped))[1:2],
    c("processed_numeric", "processed_categorical")
  )
  res <- wrapped(processed_numeric = 10, processed_categorical = 20)
  expect_identical(res$numeric, 10)
  expect_identical(res$categorical, 20)
})

test_that("inp_spec rejects the legacy input_map orientation", {
  block_with_args <- function(input_a, input_b) {}
  legacy_mapper <- c(processed_a = "input_a", processed_b = "input_b")
  expect_error(
    kerasnip:::inp_spec(block_with_args, legacy_mapper),
    "not found in the block function"
  )
})


test_that("inp_spec throws error for invalid input_map type", {
  block_with_args <- function(a) {}
  expect_error(
    kerasnip:::inp_spec(block_with_args, 123),
    "`input_map` must be a single string or a named character vector."
  )
  expect_error(
    kerasnip:::inp_spec(block_with_args, list(a = "b")),
    "`input_map` must be a single string or a named character vector."
  )
})

# =============================================================================
# Conformal inference integration tests
#
# Verifies that all three conformal inference methods from the `probably`
# package work end-to-end with kerasnip workflows — for both the sequential
# and functional Keras APIs.
#
# All three methods are tested:
#   - int_conformal_split()  : calibration-set approach, no model refit
#   - int_conformal_cv()     : cross-validation approach, no model refit
#   - int_conformal_full()   : full conformal, refits model per candidate value
#
# int_conformal_full() is kept intentionally small (2 epochs, 5 rows of
# new_data, tight grid) to remain feasible in CI while still exercising
# the full code path.
# =============================================================================

# ---------------------------------------------------------------------------
# Shared sequential layer blocks
# ---------------------------------------------------------------------------
make_conformal_seq_blocks <- function() {
  input_block <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block <- function(model, units = 8) {
    model |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }
  list(input = input_block, dense = dense_block, output = output_block)
}

# ---------------------------------------------------------------------------
# Shared functional layer blocks
# ---------------------------------------------------------------------------
make_conformal_func_blocks <- function() {
  input_block <- function(input_shape) keras3::layer_input(shape = input_shape)
  dense_block <- function(tensor, units = 8) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block <- function(tensor) keras3::layer_dense(tensor, units = 1)
  list(
    main_input = input_block,
    dense = inp_spec(dense_block, "main_input"),
    output = inp_spec(output_block, "dense")
  )
}

# ---------------------------------------------------------------------------
# Helper: assert that a conformal prediction result has the right shape
# ---------------------------------------------------------------------------
expect_valid_intervals <- function(result, n_rows) {
  expect_s3_class(result, "tbl_df")
  expect_true(
    all(c(".pred_lower", ".pred_upper") %in% names(result)),
    info = paste(
      "Expected .pred_lower and .pred_upper, got:",
      paste(names(result), collapse = ", ")
    )
  )
  expect_equal(nrow(result), n_rows)
  expect_true(all(result$.pred_lower <= result$.pred_upper))
}

# =============================================================================
# Sequential API
# =============================================================================

test_that("conformal: int_conformal_split works with sequential kerasnip workflow", {
  skip_if_no_keras()
  skip_if_not_installed("probably")

  model_name <- "conf_split_seq"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = make_conformal_seq_blocks(),
    mode = "regression"
  )

  spec <- conf_split_seq(fit_epochs = 3) |> set_engine("keras")
  data <- mtcars
  rec <- recipe(mpg ~ ., data = data)
  wf <- workflow(rec, spec)

  set.seed(42)
  split <- rsample::initial_split(data, prop = 0.70)
  train_dat <- rsample::training(split)
  cal_dat <- rsample::testing(split)

  fit_obj <- fit(wf, data = train_dat)
  conformal <- probably::int_conformal_split(fit_obj, cal_data = cal_dat)
  result <- predict(conformal, new_data = cal_dat[1:5, ], level = 0.90)

  expect_valid_intervals(result, 5)
})

test_that("conformal: int_conformal_cv works with sequential kerasnip workflow", {
  skip_if_no_keras()
  skip_if_not_installed("probably")
  skip_if_not_installed("tune")

  model_name <- "conf_cv_seq"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = make_conformal_seq_blocks(),
    mode = "regression"
  )

  spec <- conf_cv_seq(fit_epochs = 2) |> set_engine("keras")
  data <- mtcars
  rec <- recipe(mpg ~ ., data = data)
  wf <- workflow(rec, spec)

  set.seed(42)
  folds <- rsample::vfold_cv(data, v = 2)
  fitted_folds <- tune::fit_resamples(
    wf,
    resamples = folds,
    control = tune::control_resamples(
      save_pred = TRUE,
      extract = function(x) x
    )
  )
  conformal <- probably::int_conformal_cv(fitted_folds)
  result <- predict(conformal, new_data = data[1:5, ], level = 0.90)

  expect_valid_intervals(result, 5)
})

test_that("conformal: int_conformal_full works with sequential kerasnip workflow", {
  skip_if_no_keras()
  skip_if_not_installed("probably")

  model_name <- "conf_full_seq"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = make_conformal_seq_blocks(),
    mode = "regression"
  )

  spec <- conf_full_seq(fit_epochs = 2) |> set_engine("keras")
  # Use a small subset to keep the number of model refits manageable in CI.
  # int_conformal_full refits the model for every candidate value of every
  # new test observation, so dataset and new_data size directly controls
  # total training cycles.
  data <- mtcars[1:15, ]
  new_data <- mtcars[16:17, ]
  rec <- recipe(mpg ~ ., data = data)
  wf <- workflow(rec, spec)

  fit_obj <- fit(wf, data = data)
  conformal <- probably::int_conformal_full(
    fit_obj,
    train_data = data,
    control = probably::control_conformal_full(
      method = "grid",
      trial_points = 20
    )
  )
  result <- predict(conformal, new_data = new_data, level = 0.90)

  expect_valid_intervals(result, nrow(new_data))
})

# =============================================================================
# Functional API
# =============================================================================

test_that("conformal: int_conformal_split works with functional kerasnip workflow", {
  skip_if_no_keras()
  skip_if_not_installed("probably")

  model_name <- "conf_split_func"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = make_conformal_func_blocks(),
    mode = "regression"
  )

  spec <- conf_split_func(fit_epochs = 3) |> set_engine("keras")
  data <- mtcars
  rec <- recipe(mpg ~ ., data = data)
  wf <- workflow(rec, spec)

  set.seed(42)
  split <- rsample::initial_split(data, prop = 0.70)
  train_dat <- rsample::training(split)
  cal_dat <- rsample::testing(split)

  fit_obj <- fit(wf, data = train_dat)
  conformal <- probably::int_conformal_split(fit_obj, cal_data = cal_dat)
  result <- predict(conformal, new_data = cal_dat[1:5, ], level = 0.90)

  expect_valid_intervals(result, 5)
})

test_that("conformal: int_conformal_cv works with functional kerasnip workflow", {
  skip_if_no_keras()
  skip_if_not_installed("probably")
  skip_if_not_installed("tune")

  model_name <- "conf_cv_func"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = make_conformal_func_blocks(),
    mode = "regression"
  )

  spec <- conf_cv_func(fit_epochs = 2) |> set_engine("keras")
  data <- mtcars
  rec <- recipe(mpg ~ ., data = data)
  wf <- workflow(rec, spec)

  set.seed(42)
  folds <- rsample::vfold_cv(data, v = 2)
  fitted_folds <- tune::fit_resamples(
    wf,
    resamples = folds,
    control = tune::control_resamples(
      save_pred = TRUE,
      extract = function(x) x
    )
  )
  conformal <- probably::int_conformal_cv(fitted_folds)
  result <- predict(conformal, new_data = data[1:5, ], level = 0.90)

  expect_valid_intervals(result, 5)
})

test_that("conformal: int_conformal_full works with functional kerasnip workflow", {
  skip_if_no_keras()
  skip_if_not_installed("probably")

  model_name <- "conf_full_func"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = make_conformal_func_blocks(),
    mode = "regression"
  )

  spec <- conf_full_func(fit_epochs = 2) |> set_engine("keras")
  data <- mtcars[1:15, ]
  new_data <- mtcars[16:17, ]
  rec <- recipe(mpg ~ ., data = data)
  wf <- workflow(rec, spec)

  fit_obj <- fit(wf, data = data)
  conformal <- probably::int_conformal_full(
    fit_obj,
    train_data = data,
    control = probably::control_conformal_full(
      method = "grid",
      trial_points = 20
    )
  )
  result <- predict(conformal, new_data = new_data, level = 0.90)

  expect_valid_intervals(result, nrow(new_data))
})

test_that("E2E: Functional spec (regression) works", {
  skip_if_no_keras()

  # Define blocks for a simple forked functional model
  input_block <- function(input_shape) keras3::layer_input(shape = input_shape)
  path_block <- function(tensor, units = 8) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  concat_block <- function(input_a, input_b) {
    keras3::layer_concatenate(list(input_a, input_b))
  }
  output_block_reg <- function(tensor) keras3::layer_dense(tensor, units = 1)

  model_name <- "e2e_func_reg"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  # Create a spec with two parallel paths that are then concatenated
  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = list(
      main_input = input_block,
      path_a = inp_spec(path_block, "main_input"),
      path_b = inp_spec(path_block, "main_input"),
      concatenated = inp_spec(
        concat_block,
        c(input_a = "path_a", input_b = "path_b")
      ),
      output = inp_spec(output_block_reg, "concatenated")
    ),
    mode = "regression"
  )

  spec <- e2e_func_reg(
    path_a_units = 32,
    path_b_units = 16,
    fit_epochs = 2
  ) |>
    set_engine("keras")

  data <- mtcars
  rec <- recipe(mpg ~ ., data = data)
  wf <- workflows::workflow(rec, spec)

  expect_no_error(fit_obj <- parsnip::fit(wf, data = data))
  expect_s3_class(fit_obj, "workflow")

  preds <- predict(fit_obj, new_data = data[1:5, ])
  expect_s3_class(preds, "tbl_df")
  expect_equal(names(preds), ".pred")
  expect_equal(nrow(preds), 5)
  expect_true(is.numeric(preds$.pred))
})

test_that("E2E: Functional regression works with named predictors in formula", {
  skip_if_no_keras()

  input_block <- function(input_shape) keras3::layer_input(shape = input_shape)
  path_block <- function(tensor, units = 8) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  concat_block <- function(input_a, input_b) {
    keras3::layer_concatenate(list(input_a, input_b))
  }
  output_block_reg <- function(tensor) keras3::layer_dense(tensor, units = 1)

  model_name <- "e2e_func_reg_named"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = list(
      main_input = input_block,
      path_a = inp_spec(path_block, "main_input"),
      path_b = inp_spec(path_block, "main_input"),
      concatenated = inp_spec(
        concat_block,
        c(input_a = "path_a", input_b = "path_b")
      ),
      output = inp_spec(output_block_reg, "concatenated")
    ),
    mode = "regression"
  )

  spec <- e2e_func_reg_named(
    fit_epochs = 1
  ) |>
    set_engine("keras")

  data <- mtcars
  # Use named predictors to cover the x <- data[, x_names, drop = FALSE] line
  expect_no_error(
    fit_obj <- fit(spec, mpg ~ cyl + disp, data = data)
  )
  expect_s3_class(fit_obj, "model_fit")
})

test_that("E2E: Block repetition works for functional models", {
  skip_if_no_keras()

  input_block <- function(input_shape) keras3::layer_input(shape = input_shape)
  dense_block <- function(tensor, units = 8) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block <- function(tensor) keras3::layer_dense(tensor, units = 1)

  model_name <- "e2e_func_repeat"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = list(
      main_input = input_block,
      dense_path = inp_spec(dense_block, "main_input"),
      output = inp_spec(output_block, "dense_path")
    ),
    mode = "regression"
  )

  # --- Test with 1 repetition ---
  spec_1 <- e2e_func_repeat(num_dense_path = 1, fit_epochs = 1) |>
    set_engine("keras")
  fit_1 <- fit(spec_1, mpg ~ ., data = mtcars)
  model_1_layers <- fit_1 |>
    extract_keras_model() |>
    pluck("layers")

  # Expect 3 layers: Input, Dense, Output
  expect_equal(length(model_1_layers), 3)

  # --- Test with 2 repetitions ---
  spec_2 <- e2e_func_repeat(num_dense_path = 2, fit_epochs = 1) |>
    set_engine("keras")
  fit_2 <- fit(spec_2, mpg ~ ., data = mtcars)
  model_2_layers <- fit_2 |>
    extract_keras_model() |>
    pluck("layers")
  # Expect 4 layers: Input, Dense, Dense, Output
  expect_equal(length(model_2_layers), 4)

  # --- Test with 0 repetitions ---
  spec_3 <- e2e_func_repeat(num_dense_path = 0, fit_epochs = 1) |>
    set_engine("keras")
  fit_3 <- fit(spec_3, mpg ~ ., data = mtcars)
  model_3_layers <- fit_3 |>
    extract_keras_model() |>
    pluck("layers")
  # Expect 2 layers: Input, Output
  expect_equal(length(model_3_layers), 2)
})

test_that("E2E: Multi-input, multi-output functional regression works", {
  skip_if_no_keras()
  options(kerasnip.show_removal_messages = FALSE)
  on.exit(options(kerasnip.show_removal_messages = TRUE), add = TRUE)

  # Define layer blocks
  input_block_1 <- function(input_shape) {
    keras3::layer_input(shape = input_shape, name = "input_1")
  }
  input_block_2 <- function(input_shape) {
    keras3::layer_input(shape = input_shape, name = "input_2")
  }
  dense_path <- function(tensor, units = 16) {
    tensor |> keras3::layer_dense(units = units, activation = "relu")
  }
  concat_block <- function(in_1, in_2) {
    keras3::layer_concatenate(list(in_1, in_2))
  }
  output_block_1 <- function(tensor) {
    keras3::layer_dense(tensor, units = 1, name = "output_1")
  }
  output_block_2 <- function(tensor) {
    keras3::layer_dense(tensor, units = 1, name = "output_2")
  }

  model_name <- "multi_in_out_reg"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_functional_spec(
    model_name = model_name,
    layer_blocks = list(
      input_a = input_block_1,
      input_b = input_block_2,
      path_a = inp_spec(dense_path, "input_a"),
      path_b = inp_spec(dense_path, "input_b"),
      concatenated = inp_spec(
        concat_block,
        c(in_1 = "path_a", in_2 = "path_b")
      ),
      output_1 = inp_spec(output_block_1, "concatenated"),
      output_2 = inp_spec(output_block_2, "concatenated")
    ),
    mode = "regression"
  )

  spec <- multi_in_out_reg(fit_epochs = 2) |> set_engine("keras")

  # Prepare dummy data
  set.seed(123)
  x1 <- matrix(rnorm(100 * 5), ncol = 5)
  x2 <- matrix(rnorm(100 * 3), ncol = 3)
  y1 <- rnorm(100)
  y2 <- rnorm(100)

  train_df <- tibble::tibble(
    input_a = lapply(seq_len(nrow(x1)), function(i) x1[i, , drop = FALSE]),
    input_b = lapply(seq_len(nrow(x2)), function(i) x2[i, , drop = FALSE]),
    output_1 = y1,
    output_2 = y2
  )

  rec <- recipe(output_1 + output_2 ~ input_a + input_b, data = train_df)
  wf <- workflows::workflow(rec, spec)

  expect_no_error(fit_obj <- parsnip::fit(wf, data = train_df))

  new_data_df <- tibble::tibble(
    input_a = lapply(seq_len(5), function(i) matrix(rnorm(5), ncol = 5)),
    input_b = lapply(seq_len(5), function(i) matrix(rnorm(3), ncol = 3))
  )
  preds <- predict(fit_obj, new_data = new_data_df)

  expect_s3_class(preds, "tbl_df")
  expect_equal(names(preds), c(".pred_output_1", ".pred_output_2"))
  expect_equal(nrow(preds), 5)
  expect_true(is.numeric(preds$.pred_output_1))
  expect_true(is.numeric(preds$.pred_output_2))
})

test_that("collect_spec_args handles empty keras arg names", {
  testthat::with_mocked_bindings(
    .env = as.environment("package:kerasnip"),
    keras_fit_arg_names = character(0),
    keras_compile_arg_names = character(0),
    {
      args_info <- kerasnip:::collect_spec_args(
        layer_blocks = list(dense = function(model, units = 10) {}),
        functional = FALSE
      )
      # Expect only num_dense, dense_units, and learn_rate
      expected_args <- c("num_dense", "dense_units", "learn_rate")
      expect_equal(sort(names(args_info$all_args)), sort(expected_args))
      expect_equal(sort(args_info$parsnip_names), sort(expected_args))
    }
  )
})

#' Create a Custom Keras Sequential Model Specification for Tidymodels
#'
#' @description
#' This function acts as a factory to generate a new `parsnip` model
#' specification based on user-defined blocks of Keras layers using the
#' Sequential API. This is the ideal choice for creating models that are a
#' simple, linear stack of layers. For models with complex, non-linear
#' topologies, see [create_keras_functional_spec()].
#'
#' @param model_name A character string for the name of the new model
#'   specification function (e.g., "custom_cnn"). This should be a valid R
#'   function name.
#' @param layer_blocks A named, ordered list of functions. Each function defines
#'   a "block" of Keras layers. The function must take a Keras model object as
#'   its first argument and return the modified model. Other arguments to the
#'   function will become tunable parameters in the final model specification.
#' @param mode A character string, either "regression" or "classification".
#' @param ... Reserved for future use. Currently not used.
#' @param env The environment in which to create the new model specification
#'   function and its associated `update()` method. Defaults to the calling
#'   environment (`parent.frame()`).
#'
#' @details
#' This function generates all the boilerplate needed to create a custom,
#' tunable `parsnip` model specification that uses the Keras Sequential API.
#'
#' The function inspects the arguments of your `layer_blocks` functions
#' (ignoring special arguments like `input_shape` and `num_classes`)
#' and makes them available as arguments in the generated model specification,
#' prefixed with the block's name (e.g., `dense_units`).
#'
#' The new model specification function and its `update()` method are created in
#' the environment specified by the `env` argument.
#'
#' @section Model Architecture (Sequential API):
#' `kerasnip` builds the model by applying the functions in `layer_blocks` in
#' the order they are provided. Each function receives the Keras model built by
#' the previous function and returns a modified version.
#'
#' 1.  The **first block** must initialize the model (e.g., with
#'     `keras_model_sequential()`). It can accept an `input_shape` argument,
#'     which `kerasnip` will provide automatically during fitting.
#' 2.  **Subsequent blocks** add layers to the model.
#' 3.  The **final block** should add the output layer. For classification, it
#'     can accept a `num_classes` argument, which is provided automatically.
#'
#' A key feature of this function is the automatic creation of
#' `num_{block_name}` arguments (e.g., `num_hidden`). This allows you to
#' control how many times each block is repeated, making it easy to tune the
#' depth of your network.
#'
#' @importFrom rlang enquos dots_list arg_match env_poke
#' @importFrom parsnip update_dot_check
#'
#' @return Invisibly returns `NULL`. Its primary side effect is to create a new
#'   model specification function (e.g., `my_mlp()`) in the specified
#'   environment and register the model with `parsnip` so it can be used within
#'   the `tidymodels` framework.
#'
#' @seealso [remove_keras_spec()], [parsnip::new_model_spec()],
#'   [create_keras_functional_spec()]
#'
#' @section Saving and Reloading Models:
#' To save a fitted workflow and reload it in a new R session, use
#' `bundle::bundle()` before saving — this is required to preserve the Keras
#' model weights:
#'
#' ```r
#' library(bundle)
#' bundled <- bundle(fitted_workflow)
#' saveRDS(bundled, "model.rds")
#'
#' # New session:
#' library(kerasnip); library(bundle)
#' fitted_workflow <- unbundle(readRDS("model.rds"))
#' predict(fitted_workflow, new_data = test_data)  # works
#' ```
#'
#' Plain `saveRDS()` without `bundle()` does not preserve Keras weights, but
#' `predict()` will still auto-register the parsnip model type from metadata
#' stored on the spec.
#'
#' @export
#' @examples
#' \donttest{
#' if (requireNamespace("keras3", quietly = TRUE)) {
#' library(keras3)
#' library(parsnip)
#' library(dials)
#'
#' # 1. Define layer blocks for a complete model.
#' # The first block must initialize the model. `input_shape` is passed
#' # automatically.
#' input_block <- function(model, input_shape) {
#'   keras_model_sequential(input_shape = input_shape)
#' }
#' # A block for hidden layers. `units` will become a tunable parameter.
#' hidden_block <- function(model, units = 32) {
#'   model |> layer_dense(units = units, activation = "relu")
#' }
#'
#' # The output block. `num_classes` is passed automatically for classification.
#' output_block <- function(model, num_classes) {
#'   model |> layer_dense(units = num_classes, activation = "softmax")
#' }
#'
#' # 2. Create the spec, providing blocks in the correct order.
#' create_keras_sequential_spec(
#' model_name = "my_mlp_seq_spec",
#'   layer_blocks = list(
#'     input = input_block,
#'     hidden = hidden_block,
#'     output = output_block
#'   ),
#'   mode = "classification"
#' )
#'
#' # 3. Use the newly created specification function!
#' # Note the new arguments `num_hidden` and `hidden_units`.
#' model_spec <- my_mlp_seq_spec(
#'   num_hidden = 2,
#'   hidden_units = 64,
#'   epochs = 10,
#'   learn_rate = 0.01
#' )
#'
#' print(model_spec)
#' remove_keras_spec("my_mlp_seq_spec")
#' }
#' }
create_keras_sequential_spec <- function(
  model_name,
  layer_blocks,
  mode = c("regression", "classification"),
  ...,
  env = parent.frame()
) {
  mode <- arg_match(mode)
  create_keras_spec_impl(
    model_name,
    layer_blocks,
    mode,
    functional = FALSE,
    env
  )
}

test_that("autoplot works with multiple hidden units parameters", {
  skip_if_no_keras()
  skip_if_not_installed("ggplot2")

  # 1. Define a spec with multiple hidden unit parameters
  model_name <- "autoplot_spec"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)
  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = function(model, input_shape) {
        keras3::keras_model_sequential(input_shape = input_shape)
      },
      dense1 = function(model, units = 10) {
        model |> keras3::layer_dense(units = units)
      },
      dense2 = function(model, units = 10) {
        model |> keras3::layer_dense(units = units)
      },
      output = function(model, num_classes) {
        model |>
          keras3::layer_dense(units = num_classes, activation = "softmax")
      }
    ),
    mode = "classification"
  )

  tune_spec <- autoplot_spec(
    dense1_units = tune(id = "denseone"),
    dense2_units = tune(id = "densetwo")
  ) |>
    set_engine("keras")

  # 2. Set up workflow and tuning grid
  rec <- recipes::recipe(Species ~ ., data = iris)
  tune_wf <- workflows::workflow(rec, tune_spec)

  params <- tune::extract_parameter_set_dials(tune_wf)

  # The user code should not need to change.
  # `hidden_units` will be `kerasnip::hidden_units` which auto-detects the id.
  params <- params |>
    update(
      denseone = hidden_units(range = c(4L, 8L)),
      densetwo = hidden_units(range = c(4L, 8L))
    )
  params$name
  params$id
  params$source
  params$component
  params$component_id
  params$object

  grid <- dials::grid_regular(params, levels = 2)
  control <- tune::control_grid(save_pred = FALSE, verbose = FALSE)

  # 3. Run tuning
  tune_res <- tune::tune_grid(
    tune_wf,
    resamples = rsample::vfold_cv(iris, v = 2),
    grid = grid,
    control = control
  )

  # 4. Assert that autoplot works without error
  expect_no_error(
    ggplot2::autoplot(tune_res)
  )
})

test_that("E2E: Classification spec generation, fitting, prediction", {
  skip_if_no_keras()

  input_block_class <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block_class <- function(model, units = 16) {
    model |>
      keras3::layer_dense(units = units, activation = "relu")
  }
  output_block_class <- function(model, num_classes) {
    model |> keras3::layer_dense(units = num_classes, activation = "softmax")
  }

  model_name <- "e2e_mlp_class"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_class,
      dense = dense_block_class,
      output = output_block_class
    ),
    mode = "classification"
  )

  spec <- e2e_mlp_class(
    num_dense = 2,
    dense_units = 8,
    fit_epochs = 2
  ) |>
    set_engine("keras")

  # --- Multiclass test ---
  multi_data <- iris
  rec_multi <- recipe(Species ~ ., data = multi_data)
  wf_multi <- workflow(rec_multi, spec)

  expect_no_error(fit_multi <- fit(wf_multi, data = multi_data))
  expect_s3_class(fit_multi, "workflow")

  preds_class_multi <- predict(
    fit_multi,
    new_data = multi_data[1:5, ],
    type = "class"
  )
  expect_s3_class(preds_class_multi, "tbl_df")
  expect_equal(names(preds_class_multi), ".pred_class")
  expect_equal(nrow(preds_class_multi), 5)
  expect_equal(
    levels(preds_class_multi$.pred_class),
    levels(multi_data$Species)
  )

  preds_prob_multi <- predict(
    fit_multi,
    new_data = multi_data[1:5, ],
    type = "prob"
  )
  expect_s3_class(preds_prob_multi, "tbl_df")
  expect_equal(
    names(preds_prob_multi),
    paste0(".pred_", levels(multi_data$Species))
  )
  expect_equal(nrow(preds_prob_multi), 5)
  expect_true(all(abs(rowSums(preds_prob_multi) - 1) < 1e-5))

  # --- Binary test ---
  binary_data <- modeldata::two_class_dat
  rec_bin <- recipe(Class ~ ., data = binary_data)
  wf_bin <- workflow(rec_bin, spec)

  expect_no_error(fit_bin <- fit(wf_bin, data = binary_data))
  expect_s3_class(fit_bin, "workflow")

  preds_class_bin <- predict(
    fit_bin,
    new_data = binary_data[1:5, ],
    type = "class"
  )
  expect_s3_class(preds_class_bin, "tbl_df")
  expect_equal(names(preds_class_bin), ".pred_class")
  expect_equal(nrow(preds_class_bin), 5)
  expect_equal(levels(preds_class_bin$.pred_class), levels(binary_data$Class))

  preds_prob_bin <- predict(
    fit_bin,
    new_data = binary_data[1:5, ],
    type = "prob"
  )
  expect_s3_class(preds_prob_bin, "tbl_df")
  expect_equal(names(preds_prob_bin), c(".pred_Class1", ".pred_Class2"))
  expect_equal(nrow(preds_prob_bin), 5)
  expect_true(all(abs(rowSums(preds_prob_bin) - 1) < 1e-5))
})

test_that("E2E: Tuning works with a generated spec", {
  skip_if_no_keras()

  input_block_tune <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block_tune <- function(model, units = 16) {
    model |>
      keras3::layer_dense(units = units, activation = "relu")
  }
  output_block_tune <- function(model, num_classes) {
    model |> keras3::layer_dense(units = num_classes, activation = "softmax")
  }

  model_name <- "e2e_mlp_class_tune"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_tune,
      dense = dense_block_tune,
      output = output_block_tune
    ),
    mode = "classification"
  )

  tune_spec <- e2e_mlp_class_tune(
    num_dense = tune(),
    dense_units = tune(),
    fit_epochs = 1
  ) |>
    set_engine("keras")

  rec <- recipe(Species ~ ., data = iris)
  tune_wf <- workflow(rec, tune_spec)

  folds <- rsample::vfold_cv(iris, v = 2)
  params <- extract_parameter_set_dials(tune_wf) |>
    update(
      num_dense = num_terms(c(1, 2)),
      dense_units = hidden_units(c(4, 8))
    )
  grid <- grid_regular(params, levels = 2)
  control <- control_grid(save_pred = FALSE, verbose = FALSE)

  # Use a try block because tuning can sometimes fail for non-package reasons
  tune_res <- try(
    tune_grid(
      tune_wf,
      resamples = folds,
      grid = grid,
      control = control
    ),
    silent = TRUE
  )

  if (inherits(tune_res, "try-error")) {
    testthat::skip(paste("Tuning failed with error:", as.character(tune_res)))
  }

  expect_s3_class(tune_res, "tune_results")

  metrics <- collect_metrics(tune_res)
  expect_s3_class(metrics, "tbl_df")
  expect_true(all(c("num_dense", "dense_units") %in% names(metrics)))
})

test_that("E2E: Multi-block model tuning works", {
  skip_if_no_keras()

  input_block_mb <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }

  starting_layers <- function(model, layer1_units = 16, layer2_units = 32) {
    model |>
      keras3::layer_dense(units = layer1_units, activation = "relu") |>
      keras3::layer_dense(units = layer2_units, activation = "relu")
  }

  ending_layers <- function(model, units = 32, dropout = 0.2) {
    model |>
      keras3::layer_dense(units = units, activation = "relu") |>
      keras3::layer_dropout(rate = dropout)
  }

  output_block_mb <- function(model, num_classes) {
    model |> keras3::layer_dense(units = num_classes, activation = "softmax")
  }

  model_name <- "mb_mt"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_mb,
      start = starting_layers,
      end = ending_layers,
      output = output_block_mb
    ),
    mode = "classification"
  )

  tune_spec <- mb_mt(
    num_start = tune(),
    start_layer1_units = tune(),
    start_layer2_units = tune(),
    end_units = tune(),
    fit_epochs = 1
  ) |>
    set_engine("keras")

  rec <- recipe(Species ~ ., data = iris)
  wf <- workflow(rec) |>
    add_model(tune_spec)

  folds <- rsample::vfold_cv(iris, v = 2)

  params <- extract_parameter_set_dials(wf) |>
    update(
      num_start = dials::num_terms(c(1, 2)),
      start_layer1_units = dials::hidden_units(c(4, 8)),
      start_layer2_units = dials::hidden_units(c(8, 16)),
      end_units = dials::hidden_units(c(4, 8))
    )

  grid <- grid_regular(params, levels = 2)
  control <- control_grid(
    save_pred = FALSE,
    verbose = FALSE,
    save_workflow = TRUE
  )

  # Use a try block because tuning can sometimes fail for non-package reasons
  tune_res <- try(
    tune_grid(
      wf,
      resamples = folds,
      grid = grid,
      control = control
    ),
    silent = TRUE
  )

  if (inherits(tune_res, "try-error")) {
    testthat::skip(paste("Tuning failed with error:", as.character(tune_res)))
  }

  expect_s3_class(tune_res, "tune_results")

  metrics <- collect_metrics(tune_res)
  expect_s3_class(metrics, "tbl_df")
  expect_true(all(
    c("num_start", "start_layer1_units", "start_layer2_units", "end_units") %in%
      names(metrics)
  ))

  expect_no_error(
    best_fit <- tune::fit_best(tune_res)
  )
  expect_s3_class(best_fit, "workflow")
})

test_that("E2E: Regression spec generation, fitting, and prediction works", {
  skip_if_no_keras()

  input_block_reg <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block_reg <- function(model, units = 16, dropout = 0.1) {
    model |>
      keras3::layer_dense(units = units, activation = "relu") |>
      keras3::layer_dropout(rate = dropout)
  }
  output_block_reg <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }

  model_name <- "e2e_mlp_reg"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_reg,
      dense = dense_block_reg,
      output = output_block_reg
    ),
    mode = "regression"
  )

  spec <- e2e_mlp_reg(
    num_dense = 2,
    dense_units = 8,
    fit_epochs = 2,
    learn_rate = 0.01
  ) |>
    set_engine("keras")

  data <- mtcars
  rec <- recipe(mpg ~ ., data = data)
  wf <- workflow(rec, spec)

  expect_no_error(
    fit_obj <- fit(wf, data = data)
  )
  expect_s3_class(fit_obj, "workflow")

  preds <- predict(fit_obj, new_data = data[1:5, ])
  expect_s3_class(preds, "tbl_df")
  expect_equal(names(preds), ".pred")
  expect_equal(nrow(preds), 5)
  expect_true(is.numeric(preds$.pred))
})

test_that("E2E: Sequential regression works with named predictors in formula", {
  skip_if_no_keras()

  input_block_reg <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block_reg <- function(model, units = 16, dropout = 0.1) {
    model |>
      keras3::layer_dense(units = units, activation = "relu") |>
      keras3::layer_dropout(rate = dropout)
  }
  output_block_reg <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }

  model_name <- "e2e_mlp_reg_named"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_reg,
      dense = dense_block_reg,
      output = output_block_reg
    ),
    mode = "regression"
  )

  spec <- e2e_mlp_reg_named(
    fit_epochs = 1
  ) |>
    set_engine("keras")

  data <- mtcars
  # Use named predictors to cover the x <- data[, x_names, drop = FALSE] line
  expect_no_error(
    fit_obj <- fit(spec, mpg ~ cyl + disp, data = data)
  )
  expect_s3_class(fit_obj, "model_fit")
})

test_that("tidy.kerasnip_model_fit returns correct structure", {
  skip_if_not(
    reticulate::py_module_available("keras"),
    "keras not available"
  )

  input_block <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  hidden_block <- function(model, units = 8) {
    model |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block <- function(model, num_classes) {
    model |> keras3::layer_dense(units = 1)
  }

  create_keras_sequential_spec(
    model_name = "tidy_test_model",
    layer_blocks = list(
      input = input_block,
      hidden = hidden_block,
      output = output_block
    ),
    mode = "regression"
  )
  on.exit(remove_keras_spec("tidy_test_model"), add = TRUE)

  x_train <- matrix(rnorm(50 * 3), ncol = 3)
  y_train <- rnorm(50)
  train_df <- data.frame(x = I(x_train), y = y_train)

  spec <- tidy_test_model(fit_epochs = 2L) |>
    parsnip::set_engine("keras")
  fitted <- parsnip::fit(spec, y ~ x, data = train_df)

  result <- generics::tidy(fitted)

  expect_s3_class(result, "tbl_df")
  expect_named(result, c("layer", "class", "n_params"))
  expect_gt(nrow(result), 0L)
  expect_type(result$layer, "character")
  expect_type(result$class, "character")
  expect_type(result$n_params, "integer")
})

test_that("glance.kerasnip_model_fit returns one-row tibble with loss", {
  skip_if_not(
    reticulate::py_module_available("keras"),
    "keras not available"
  )

  input_block <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  hidden_block <- function(model, units = 8) {
    model |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block <- function(model, num_classes) {
    model |> keras3::layer_dense(units = 1)
  }

  create_keras_sequential_spec(
    model_name = "glance_test_model",
    layer_blocks = list(
      input = input_block,
      hidden = hidden_block,
      output = output_block
    ),
    mode = "regression"
  )
  on.exit(remove_keras_spec("glance_test_model"), add = TRUE)

  x_train <- matrix(rnorm(50 * 3), ncol = 3)
  y_train <- rnorm(50)
  train_df <- data.frame(x = I(x_train), y = y_train)

  spec <- glance_test_model(fit_epochs = 2L) |>
    parsnip::set_engine("keras")
  fitted <- parsnip::fit(spec, y ~ x, data = train_df)

  result <- generics::glance(fitted)

  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 1L)
  expect_true("loss" %in% names(result))
})

test_that("glance returns empty tibble when history is NULL", {
  # Simulate a butchered fit where history was stripped
  mock_fit <- structure(
    list(fit = list(history = NULL)),
    class = c("kerasnip_model_fit", "model_fit")
  )
  result <- generics::glance(mock_fit)
  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 0L)
})

test_that("E2E: Model spec removal works", {
  skip_if_no_keras()

  model_name <- "removable_model"

  input_block <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  output_block <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(input = input_block, output = output_block),
    mode = "regression"
  )

  update_method_name <- paste0("update.", model_name)

  expect_true(exists(model_name, inherits = FALSE))
  expect_true(exists(update_method_name, inherits = FALSE))
  expect_error(parsnip:::check_model_doesnt_exist(model_name))

  remove_keras_spec(model_name)

  expect_false(exists(model_name, inherits = FALSE))
  expect_false(exists(update_method_name, inherits = FALSE))
  expect_no_error(parsnip:::check_model_doesnt_exist(model_name))
})

test_that("E2E: Model spec removal is not too aggressive", {
  skip_if_no_keras()

  model_name <- "my_mlp"
  model_name_2 <- "my_mlp_2"

  input_block <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  output_block <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(input = input_block, output = output_block),
    mode = "regression"
  )

  create_keras_sequential_spec(
    model_name = model_name_2,
    layer_blocks = list(input = input_block, output = output_block),
    mode = "regression"
  )

  expect_true(exists(model_name, inherits = FALSE))
  expect_true(exists(model_name_2, inherits = FALSE))
  expect_error(parsnip:::check_model_doesnt_exist(model_name))
  expect_error(parsnip:::check_model_doesnt_exist(model_name_2))

  remove_keras_spec(model_name)

  expect_false(exists(model_name, inherits = FALSE))
  expect_true(exists(model_name_2, inherits = FALSE))
  expect_no_error(parsnip:::check_model_doesnt_exist(model_name))
  expect_error(parsnip:::check_model_doesnt_exist(model_name_2))

  # cleanup
  remove_keras_spec(model_name_2)
})

#' Register Model Arguments with Parsnip and Dials
#'
#' @description
#' Registers each model argument with `parsnip` and maps it to a corresponding
#' `dials` parameter function. This is a crucial step that makes the model's
#' parameters visible to the `tidymodels` ecosystem for tuning.
#'
#' @details
#' This function iterates through each argument name discovered by
#' `collect_spec_args()` and calls `parsnip::set_model_arg()`.
#'
#' The mapping from a `kerasnip` argument to a `dials` function is determined
#' by the following logic:
#' \itemize{
#'   \item Arguments starting with `num_` (e.g., `num_dense`) are mapped to
#'     `dials::num_terms()`.
#'   \item Other arguments are mapped based on their suffix (e.g., `dense_units`
#'     is mapped based on `units`). The internal `keras_dials_map` object
#'     contains common mappings like `units` -> `dials::hidden_units()`.
#'   \item Arguments for `compile_loss` and `compile_optimizer` are mapped to
#'     custom `dials` parameter functions (`loss_function_keras()` and
#'     `optimizer_function()`) that are part of the `kerasnip` package itself.
#'     The function correctly sets the `pkg` for these to `kerasnip`.
#' }
#'
#' @param model_name The name of the new model specification.
#' @param parsnip_names A character vector of all argument names to be
#'   registered.
#' @return Invisibly returns `NULL`. Called for its side effects.
#' @noRd
register_model_args <- function(model_name, parsnip_names) {
  keras_dials_map <- tibble::tribble(
    ~keras_arg,
    ~dials_fun,
    "units",
    "hidden_units",
    "filters",
    "hidden_units",
    "kernel_size",
    "kernel_size",
    "pool_size",
    "pool_size",
    "dropout",
    "dropout",
    "rate",
    "dropout",
    "learn_rate",
    "learn_rate",
    "fit_epochs",
    "epochs",
    "fit_batch_size",
    "batch_size",
    "compile_loss", # parsnip arg
    "loss_function_keras", # dials function from kerasnip
    "compile_optimizer", # parsnip arg
    "optimizer_function" # dials function from kerasnip
  )

  # We now allow optimizer to be tuned. Metrics are for tracking, not training.
  non_tunable <- c("fit_verbose")

  for (arg in parsnip_names) {
    if (arg %in% non_tunable) {
      next
    }

    if (startsWith(arg, "num_")) {
      dials_fun <- "num_terms"
    } else {
      # First, try to match the full argument name
      idx <- match(arg, keras_dials_map$keras_arg)
      if (!is.na(idx)) {
        dials_fun <- keras_dials_map$dials_fun[idx]
      } else {
        # If no full match, try to match the base name (e.g., "units" from
        # "dense_units")
        base_arg <- sub(".*_", "", arg)
        idx <- match(base_arg, keras_dials_map$keras_arg)
        dials_fun <- if (!is.na(idx)) keras_dials_map$dials_fun[idx] else arg
      }
    }

    pkg <- if (dials_fun %in% c("loss_function_keras", "optimizer_function")) {
      "kerasnip"
    } else {
      "dials"
    }

    parsnip::set_model_arg(
      model = model_name,
      eng = "keras",
      parsnip = arg,
      original = arg,
      func = list(pkg = pkg, fun = dials_fun),
      has_submodel = FALSE
    )
  }
}

#' Register Core Model Information with Parsnip
#'
#' @description
#' Sets up the basic model definition with `parsnip`. This function is called
#' once when a new specification is created.
#'
#' @details
#' This function makes a series of calls to `parsnip`'s registration API:
#' - `parsnip::set_new_model()`: Declares the new model.
#' - `parsnip::set_model_mode()`: Sets the mode (e.g., "regression").
#' - `parsnip::set_model_engine()`: Sets the engine to "keras".
#' - `parsnip::set_dependency()`: Declares the dependency on the `keras3`
#'   package.
#' - `parsnip::set_encoding()`: Specifies data preprocessing requirements.
#'
#' @param model_name The name of the new model.
#' @param mode The model mode ("regression" or "classification").
#' @return Invisibly returns `NULL`. Called for its side effects.
#' @noRd
register_core_model <- function(model_name, mode) {
  parsnip::set_new_model(model_name)
  parsnip::set_model_mode(model_name, mode)
  parsnip::set_model_engine(model_name, mode, "keras")
  parsnip::set_dependency(model_name, "keras", "keras3")
  parsnip::set_dependency(model_name, "keras", "kerasnip")

  parsnip::set_encoding(
    model = model_name,
    eng = "keras",
    mode = mode,
    options = list(
      predictor_indicators = "traditional",
      compute_intercept = TRUE,
      remove_intercept = TRUE,
      allow_sparse_x = FALSE
    )
  )
}

#' Generate Roxygen Documentation for a Dynamic Spec Function
#'
#' @description
#' This internal helper constructs a complete Roxygen comment block as a single
#' string. This string is then attached to the dynamically created model
#' specification function, making it self-documenting.
#'
#' @details
#' The function assembles the documentation in a structured way:
#' \itemize{
#'   \item \strong{Title & Description:} A title is generated from the
#'      `model_name`, and the description indicates which `kerasnip` function
#'      created it.
#'   \item \strong{Parameters (`@param`):} It documents several groups of
#'      parameters:
#'     \itemize{
#'       \item Block-specific hyperparameters (e.g., `dense_units`),
#'          introspecting `layer_blocks` to find default values.
#'       \item Architecture parameters (e.g., `num_dense`).
#'       \item Global training parameters (e.g., `epochs`, `learn_rate`).
#'       \item Compilation override parameters (e.g., `compile_loss`).
#'     }
#'   \item \strong{Sections (`@section`):} It creates dedicated sections for:
#'     \itemize{
#'       \item \strong{Model Architecture:} Explains how the model is built,
#'         with different content for the Sequential vs. Functional API
#'         (controlled by the `functional` flag).
#'       \item \strong{Model Fitting:} Explains how to pass arguments to
#'         `keras3::fit()` using the `fit_` prefix.
#'       \item \strong{Model Compilation:} Explains the default compilation
#'         behavior and how to override it using the `compile_` prefix.
#'     }
#'   \item \strong{Other Tags:} Adds `@seealso` to link to relevant `kerasnip`
#'     functions and `@export` to make the generated function available to
#'     users.
#' }
#'
#' @param model_name A character string for the model's name, used to generate
#'   the documentation title.
#' @param layer_blocks The named list of user-provided layer block functions.
#'   This is introspected to find default values for block-specific parameters.
#' @param all_args A named list of all arguments for the new function's
#'   signature, used to determine which `@param` tags to generate.
#' @param functional A logical. If `TRUE`, generates documentation specific to
#'   the Functional API. If `FALSE`, generates documentation for the Sequential
#'   API.
#' @return A single string containing the full Roxygen documentation, ready to
#'   be attached to a function using `comment()`.
#' @noRd
generate_roxygen_docs <- function(
  model_name,
  layer_blocks,
  all_args,
  functional = FALSE
) {
  # Title and Description
  title <- paste(
    gsub("_", " ", tools::toTitleCase(model_name)),
    "Model Specification"
  )
  desc <- paste0(
    "Defines a `parsnip` model specification for a Keras model built with ",
    "custom layer blocks. This function was generated by `kerasnip::",
    if (isTRUE(functional)) {
      "create_keras_functional_spec"
    } else {
      "create_keras_sequential_spec"
    },
    "()`."
  )

  # Parameters
  param_docs <- c()
  arg_names <- names(all_args)

  # Group args for structured documentation
  num_params <- arg_names[startsWith(arg_names, "num_")]
  fit_params <- arg_names[startsWith(arg_names, "fit_")]
  compile_params <- arg_names[startsWith(arg_names, "compile_")]
  # `learn_rate` is a special top-level convenience argument
  special_params <- "learn_rate"

  block_params <- setdiff(
    arg_names,
    c(num_params, fit_params, compile_params, special_params)
  )

  if (length(block_params) > 0) {
    # Sort block names by length descending to handle overlapping names
    # (e.g., "dense" and "dense_layer")
    sorted_block_names <- names(layer_blocks)[
      order(nchar(names(layer_blocks)), decreasing = TRUE)
    ]

    param_docs <- c(
      param_docs,
      purrr::map_chr(block_params, function(p) {
        # Find the block name that is a prefix for this parameter.
        # The `Find` function returns the first match, and since we sorted
        # block names by length, it will find the longest possible match.
        block_name <- Find(
          function(bn) startsWith(p, paste0(bn, "_")),
          sorted_block_names
        )

        if (is.null(block_name)) {
          # This should not happen if collect_spec_args is correct, but as a
          # fallback, we avoid an error.
          return(paste0("@param ", p, " A model parameter."))
        }

        param_name <- sub(paste0(block_name, "_"), "", p, fixed = TRUE)
        block_fn <- layer_blocks[[block_name]]
        default_val <- rlang::fn_fmls(block_fn)[[param_name]]
        default_str <- if (
          !is.null(default_val) && !rlang::is_missing(default_val)
        ) {
          paste0(
            " Defaults to `",
            deparse(default_val, width.cutoff = 500L),
            "`."
          )
        } else {
          ""
        }
        paste0(
          "@param ",
          p,
          " The `",
          param_name,
          "` for the '",
          block_name,
          "' block.",
          default_str
        )
      })
    )
  }

  # Document architecture params
  if (length(num_params) > 0) {
    param_docs <- c(
      param_docs,
      purrr::map_chr(num_params, function(p) {
        block_name <- sub("num_", "", p)
        paste0(
          "@param ",
          p,
          " The number of times to repeat the '",
          block_name,
          "' block. Defaults to 1."
        )
      })
    )
  }

  # Document special `learn_rate` param
  param_docs <- c(
    param_docs,
    "@param learn_rate The learning rate for the default Adam optimizer.",
    "  This is ignored if `compile_optimizer` is provided as a pre-built",
    "  Keras optimizer object."
  )

  # Document compile params
  if (length(compile_params) > 0) {
    param_docs <- c(
      param_docs,
      purrr::map_chr(compile_params, function(p) {
        paste0(
          "@param ",
          p,
          " Argument to `keras3::compile()`. See the ",
          "'Model Compilation' section."
        )
      })
    )
  }

  # Document fit params
  if (length(fit_params) > 0) {
    param_docs <- c(
      param_docs,
      purrr::map_chr(fit_params, function(p) {
        paste0(
          "@param ",
          p,
          " Argument to `keras3::fit()`. See the 'Model Fitting' section."
        )
      })
    )
  }

  # Add ... param
  param_docs <- c(
    param_docs,
    paste0(
      "@param ... Additional arguments passed to the Keras engine. ",
      "Use this for arguments to `keras3::fit()` or ",
      "`keras3::compile()` that are not exposed as top-level ",
      "arguments."
    )
  )

  # Sections
  if (isTRUE(functional)) {
    architecture_section <- c(
      "#' @section Model Architecture (Functional API):",
      "#' The Keras model is constructed using the Functional API.",
      "#' Each layer block function's arguments determine its inputs.",
      "#' For example, a block `function(input_a, input_b, ...)` will be",
      "#' connected to the outputs of the `input_a` and `input_b` blocks.",
      "#' You can also repeat a block by setting the `num_{block_name}`",
      "#' argument, provided the block has a single input tensor.",
      "#' The first block in `layer_blocks` is assumed to be the input",
      "#' layer and should not have inputs from other layers."
    )
    see_also_fit <- "generic_functional_fit()"
    see_also_create <- "create_keras_functional_spec()"
  } else {
    architecture_section <- c(
      "#' @section Model Architecture (Sequential API):",
      "#' The Keras model is constructed by sequentially applying",
      "#' the layer blocks in the order they were provided to",
      "#' `create_keras_sequential_spec()`.",
      "#' You can control the number of times each block is repeated",
      "#' by setting the `num_{block_name}` argument (e.g.,",
      "#' `num_dense = 2`).",
      "#' This allows for dynamically creating deeper or more complex",
      "#' architectures during tuning."
    )
    see_also_fit <- "generic_sequential_fit()"
    see_also_create <- "create_keras_sequential_spec()"
  }

  compilation_section <- c(
    "#' @section Model Compilation:",
    "#' The model is compiled with a default optimizer, loss function,",
    "#' and metric based on the model's mode. You can override these",
    "#' defaults by providing arguments prefixed with `compile_`.",
    "#' \\itemize{",
    "#'   \\item \\strong{Optimizer}: Defaults to `keras3::optimizer_adam()`",
    "#'     using the `learn_rate` argument. Override with",
    "#'     `compile_optimizer` (e.g., `\"sgd\"` or",
    "#'     `keras3::optimizer_sgd(...)`).",
    "#'   \\item \\strong{Loss}: Defaults to `\"mean_squared_error\"` for",
    "#'     regression and `\"categorical_crossentropy\"` or",
    "#'     `\"binary_crossentropy\"` for classification. Override with",
    "#'     `compile_loss`.",
    "#'   \\item \\strong{Metrics}: Defaults to `\"mean_absolute_error\"` for",
    "#'     regression and `\"accuracy\"` for classification. Override with",
    "#'     `compile_metrics` (e.g., `c(\"mae\", \"mape\")`).",
    "#' }",
    paste0(
      "#' For more details, see the documentation for `kerasnip::",
      see_also_fit,
      "`."
    )
  )

  fitting_section <- c(
    "#' @section Model Fitting:",
    "#' The model is fit using `keras3::fit()`. You can pass any",
    "#' argument to this function by prefixing it with `fit_`.",
    "#' For example, to add Keras callbacks, you can pass",
    "#' `fit_callbacks = list(callback_early_stopping())`.",
    "#' Common arguments include `fit_epochs`, `fit_batch_size`,",
    "#' and `fit_validation_split`."
  )

  # Other tags
  other_tags <- c(
    paste0("#' @seealso [", see_also_create, "], [", see_also_fit, "]"),
    "#' @export"
  )

  # Combine all parts
  paste(
    c(
      paste0("#' ", title),
      "#'",
      paste0("#' ", desc),
      "#'",
      paste0("@", param_docs),
      architecture_section,
      fitting_section,
      compilation_section,
      other_tags
    ),
    collapse = "\n"
  )
}

test_that("generate_roxygen_docs handles fallback case for block params", {
  # This test covers the case where a parameter is passed in `all_args`
  # that is not a num_, fit_, or compile_ param, and does not match
  # any of the layer_blocks. This is a fallback case that should not
  # happen in normal operation but is tested for completeness.
  doc_string <- kerasnip:::generate_roxygen_docs(
    model_name = "test_model",
    layer_blocks = list(
      dense = function(units = 10) {}
    ),
    all_args = list(
      unmatched_param = 1
    ),
    functional = FALSE
  )
  expect_true(grepl("@param unmatched_param A model parameter.", doc_string))
})

`%||%` <- function(x, y) {
  # Use y if x is NULL or a parsnip "zapped" argument
  if (is.null(x) || inherits(x, "rlang_zap")) y else x
}

# Serialize a fitted Keras model to a raw byte vector using the native .keras
# format. The bytes are stored alongside the Python object so the model can be
# restored after saveRDS()/readRDS() or bundle()/unbundle() invalidates the
# reticulate external pointer. Issues a warning and returns NULL if save_model
# fails so that fit() is not aborted, but the user is informed that the saved
# object will not support predict() after reload.
keras_model_to_bytes <- function(model) {
  tryCatch(
    {
      tmp <- tempfile(fileext = ".keras")
      on.exit(unlink(tmp), add = TRUE)
      keras3::save_model(model, tmp)
      readBin(tmp, "raw", n = file.size(tmp))
    },
    error = function(e) {
      warning(
        "Could not serialize the Keras model to bytes. ",
        "The fitted model will not support predict() after ",
        "saveRDS()/readRDS() or bundle()/unbundle().\n",
        "Underlying error: ",
        conditionMessage(e),
        call. = FALSE
      )
      NULL
    }
  )
}

# Restore a Keras model from the raw bytes produced by keras_model_to_bytes().
keras_model_from_bytes <- function(bytes) {
  tmp <- tempfile(fileext = ".keras")
  on.exit(unlink(tmp), add = TRUE)
  writeBin(bytes, tmp)
  keras3::load_model(tmp)
}

# Environments to store user-registered custom functions
.kerasnip_custom_objects <- new.env(parent = emptyenv())
.kerasnip_custom_objects$optimizers <- list()
.kerasnip_custom_objects$losses <- list()
.kerasnip_custom_objects$metrics <- list()

#' Register a Custom Keras Optimizer
#'
#' @description
#' Allows users to register a custom optimizer function so it can be used by
#' name within `kerasnip` model specifications and tuned with `dials`.
#'
#' @details
#' Registered optimizers are stored in an internal environment. When a model is
#' compiled, `kerasnip` will first check this internal registry for an optimizer
#' matching the provided name before checking the `keras3` package.
#'
#' The `optimizer_fn` can be a simple function or a partially applied function
#' using `purrr::partial()`. This is useful for creating versions of Keras
#' optimizers with specific settings.
#'
#' @param name The name to register the optimizer under (character).
#' @param optimizer_fn The optimizer function. It should return a Keras
#'   optimizer object.
#' @seealso [register_keras_loss()], [register_keras_metric()]
#' @export
#' @return No return value, called for side effects.
#' @examples
#' if (requireNamespace("keras3", quietly = TRUE)) {
#'   # Register a custom version of Adam with a different default beta_1
#'   my_adam <- purrr::partial(keras3::optimizer_adam, beta_1 = 0.8)
#'   register_keras_optimizer("my_adam", my_adam)
#'
#'   # Now "my_adam" can be used as a string in a model spec, e.g.,
#'   # my_model_spec(compile_optimizer = "my_adam")
#' }
register_keras_optimizer <- function(name, optimizer_fn) {
  .kerasnip_custom_objects$optimizers[[name]] <- optimizer_fn
  invisible()
}

#' Register a Custom Keras Loss
#'
#' @description
#' Allows users to register a custom loss function so it can be used by name
#' within `kerasnip` model specifications and tuned with `dials`.
#'
#' @details
#' Registered losses are stored in an internal environment. When a model is
#' compiled, `kerasnip` will first check this internal registry for a loss
#' matching the provided name before checking the `keras3` package.
#'
#' @param name The name to register the loss under (character).
#' @param loss_fn The loss function.
#' @seealso [register_keras_optimizer()], [register_keras_metric()]
#' @export
#' @return No return value, called for side effects.
register_keras_loss <- function(name, loss_fn) {
  .kerasnip_custom_objects$losses[[name]] <- loss_fn
  invisible()
}

#' Register a Custom Keras Metric
#'
#' @description
#' Allows users to register a custom metric function so it can be used by name
#' within `kerasnip` model specifications.
#'
#' @details
#' Registered metrics are stored in an internal environment. When a model is
#' compiled, `kerasnip` will first check this internal registry for a metric
#' matching the provided name before checking the `keras3` package.
#'
#' @param name The name to register the metric under (character).
#' @param metric_fn The metric function.
#' @seealso [register_keras_optimizer()], [register_keras_loss()]
#' @export
#' @return No return value, called for side effects.
register_keras_metric <- function(name, metric_fn) {
  .kerasnip_custom_objects$metrics[[name]] <- metric_fn
  invisible()
}

#' Internal helper to retrieve a Keras object by name from the registry
#'
#' @description
#' Resolves a string name into a Keras object (optimizer, loss, or metric)
#' by searching in a specific order.
#'
#' @details
#' The lookup order is:
#' 1.  User-registered custom objects via `register_keras_*()`.
#' 2.  Standard Keras constructors in the `keras3` package
#'     (e.g., `optimizer_adam`).
#' 3.  If not found, the original string is returned, assuming Keras can handle
#'     it.
#'
#' For optimizers, it also passes along any `...` arguments
#' (like `learning_rate`) to the constructor function.
#'
#' @param name The string name of the object.
#' @param type The type of object ("optimizer", "loss", or "metric").
#' @param ... Additional arguments passed to the optimizer constructor.
#' @return A Keras object or a string name.
#' @noRd
get_keras_object <- function(
  name,
  type = c("optimizer", "loss", "metric"),
  ...
) {
  type <- rlang::arg_match(type)
  storage_name <- paste0(type, "s") # optimizers, losses, metrics

  # 1. Check user-registered custom objects
  custom_fn <- .kerasnip_custom_objects[[storage_name]][[name]]
  if (!is.null(custom_fn)) {
    # If it's an optimizer, pass extra args like learn_rate
    if (type == "optimizer") {
      return(rlang::exec(custom_fn, !!!list(...)))
    }
    return(custom_fn)
  }

  # 2. Check keras3 namespace for a constructor (e.g., optimizer_adam)
  keras_fn_name <- if (type == "optimizer") {
    paste0("optimizer_", name)
  } else {
    paste0(type, "_", name)
  }
  if (exists(keras_fn_name, envir = asNamespace("keras3"))) {
    keras_fn <- get(keras_fn_name, envir = asNamespace("keras3"))
    if (type == "optimizer") {
      return(rlang::exec(keras_fn, !!!list(...)))
    }
    # For loss and metric, call the constructor to get an instance.
    # Passing the class object (not an instance) compiles fine but causes
    # keras3::save_model() to fail with "Metric.get_config() missing
    # positional argument: self".
    return(keras_fn())
  }

  # 3. If not found, assume it's a string Keras understands directly
  name
}

#' Process Predictor Input for Keras (Functional API)
#'
#' @description
#' Preprocesses predictor data (`x`) into a format suitable for Keras models
#' built with the Functional API. Handles both tabular data and list-columns
#' of arrays (e.g., for images), supporting multiple inputs.
#'
#' @param x A data frame or matrix of predictors.
#' @return A list containing:
#'   - `x_proc`: The processed predictor data (matrix or array, or list of
#'     arrays).
#'   - `input_shape`: The determined input shape(s) for the Keras model.
#' @keywords internal
#' @export
process_x_functional <- function(x) {
  if (is.data.frame(x)) {
    # Check if it's a multi-input scenario (multiple list-columns)
    if (all(sapply(x, is.list)) && ncol(x) > 1) {
      x_proc_list <- lapply(x, function(col) {
        do.call(abind::abind, c(col, list(along = 0)))
      })
      # For multi-input, input_shape should be a list of shapes
      input_shape_list <- lapply(x_proc_list, function(arr) {
        if (length(dim(arr)) > 2) dim(arr)[-1] else ncol(arr)
      })
      # Add names to the lists
      names(x_proc_list) <- names(x)
      names(input_shape_list) <- names(x)
      return(list(x_proc = x_proc_list, input_shape = input_shape_list))
    } else if (ncol(x) == 1 && is.list(x[[1]])) {
      # Original case: single predictor column containing a list of arrays.
      x_proc <- do.call(abind::abind, c(x[[1]], list(along = 0)))
    } else {
      x_proc <- as.matrix(x)
    }
  }
  input_shape <- if (length(dim(x_proc)) > 2) dim(x_proc)[-1] else ncol(x_proc)
  list(x_proc = x_proc, input_shape = input_shape)
}

#' Process Outcome Input for Keras (Functional API)
#'
#' @description
#' Preprocesses outcome data (`y`) into a format suitable for Keras models
#' built with the Functional API. Handles both regression (numeric) and
#' classification (factor) outcomes, including one-hot encoding for
#' classification, and supports multiple outputs.
#'
#' @param y A vector or data frame of outcomes.
#' @param is_classification Logical, optional. If `TRUE`, treats `y` as
#'   classification. If `FALSE`, treats as regression. If `NULL` (default),
#'   it's determined from `is.factor(y)`.
#' @param class_levels Character vector, optional. The factor levels for
#'   classification outcomes. If `NULL` (default), determined from `levels(y)`.
#' @return A list containing:
#'   - `y_proc`: The processed outcome data (matrix or one-hot encoded array,
#'     or list of these for multiple outputs).
#'   - `is_classification`: Logical, indicating if `y` was treated as
#'     classification.
#'   - `num_classes`: Integer, the number of classes for classification, or
#'     `NULL`.
#'   - `class_levels`: Character vector, the factor levels for classification,
#'     or `NULL`.
#' @importFrom keras3 to_categorical
#' @keywords internal
#' @export
process_y_functional <- function(
  y,
  is_classification = NULL,
  class_levels = NULL
) {
  # If y is a data frame/tibble with one column, extract it to ensure it's
  # processed by the single-output logic path.
  if (is.data.frame(y) && ncol(y) == 1) {
    y <- y[[1]]
  }

  if (is.data.frame(y)) {
    # Handle multiple output columns
    y_proc_list <- list() # This will store the processed y for each output
    class_levels_list <- list() # To store class levels for each output

    for (col_name in names(y)) {
      current_y <- y[[col_name]]
      current_is_classification <- is_classification %||% is.factor(current_y)
      current_class_levels <- class_levels %||% levels(current_y)

      y_proc_single <- NULL
      num_classes_single <- NULL

      if (current_is_classification) {
        if (is.null(current_class_levels)) {
          current_class_levels <- levels(current_y)
        }
        num_classes_single <- length(current_class_levels)
        y_factored <- factor(current_y, levels = current_class_levels)
        y_proc_single <- keras3::to_categorical(
          as.numeric(y_factored) - 1,
          num_classes = num_classes_single
        )
      } else {
        y_proc_single <- as.matrix(current_y)
      }
      y_proc_list[[col_name]] <- y_proc_single
      # Store class levels for each output
      class_levels_list[[col_name]] <- current_class_levels
    }
    # Return a list containing y_proc_list and class_levels_list
    list(y_proc = y_proc_list, class_levels = class_levels_list)
  } else {
    # Original single output case
    if (is.null(is_classification)) {
      is_classification <- is.factor(y)
    }

    y_proc <- NULL
    num_classes <- NULL
    if (is_classification) {
      if (is.null(class_levels)) {
        class_levels <- levels(y)
      }
      num_classes <- length(class_levels)
      y_factored <- factor(y, levels = class_levels)
      y_proc <- keras3::to_categorical(
        as.numeric(y_factored) - 1,
        num_classes = num_classes
      )
    } else {
      y_proc <- as.matrix(y)
    }
    list(
      y_proc = y_proc,
      class_levels = class_levels,
      is_classification = is_classification,
      num_classes = num_classes
    )
  }
}


#' Process Predictor Input for Keras
#'
#' @description
#' Preprocesses predictor data (`x`) into a format suitable for Keras models.
#' Handles both tabular data and list-columns of arrays (e.g., for images).
#'
#' @param x A data frame or matrix of predictors.
#' @return A list containing:
#'   - `x_proc`: The processed predictor data (matrix or array).
#'   - `input_shape`: The determined input shape for the Keras model.
#' @keywords internal
#' @export
process_x_sequential <- function(x) {
  if (is.data.frame(x) && ncol(x) == 1 && is.list(x[[1]])) {
    # Assumes a single predictor column containing a list of arrays.
    # We stack them into a single higher-dimensional array.
    x_proc <- do.call(abind::abind, c(x[[1]], list(along = 0)))
  } else {
    x_proc <- as.matrix(x)
  }
  input_shape <- if (length(dim(x_proc)) > 2) dim(x_proc)[-1] else ncol(x_proc)
  list(x_proc = x_proc, input_shape = input_shape)
}

#' Process Outcome Input for Keras
#'
#' @description
#' Preprocesses outcome data (`y`) into a format suitable for Keras models.
#' Handles both regression (numeric) and classification (factor) outcomes,
#' including one-hot encoding for classification.
#'
#' @param y A vector of outcomes.
#' @param is_classification Logical, optional. If `TRUE`, treats `y` as
#'   classification. If `FALSE`, treats as regression. If `NULL` (default),
#'   it's determined from `is.factor(y)`.
#' @param class_levels Character vector, optional. The factor levels for
#'   classification outcomes. If `NULL` (default), determined from `levels(y)`.
#' @return A list containing:
#'   - `y_proc`: The processed outcome data (matrix or one-hot encoded array).
#'   - `is_classification`: Logical, indicating if `y` was treated as
#'     classification.
#'   - `num_classes`: Integer, the number of classes for classification, or
#'     `NULL`.
#'   - `class_levels`: Character vector, the factor levels for classification,
#'     or `NULL`.
#' @importFrom keras3 to_categorical
#' @keywords internal
#' @export
process_y_sequential <- function(
  y,
  is_classification = NULL,
  class_levels = NULL
) {
  # If y is a data frame/tibble, extract the first column
  if (is.data.frame(y)) {
    y <- y[[1]]
  }

  if (is.null(is_classification)) {
    is_classification <- is.factor(y)
  }

  y_proc <- NULL
  num_classes <- NULL
  if (is_classification) {
    if (is.null(class_levels)) {
      class_levels <- levels(y)
    }
    num_classes <- length(class_levels)
    y_factored <- factor(y, levels = class_levels)
    y_proc <- keras3::to_categorical(
      as.numeric(y_factored) - 1,
      num_classes = num_classes
    )
  } else {
    y_proc <- as.matrix(y)
  }
  list(
    y_proc = y_proc,
    is_classification = is_classification,
    num_classes = num_classes,
    class_levels = class_levels
  )
}

##' Get Parsnip's Model Environment
#'
#' @description
#' This is an internal helper function to retrieve the environment where
#' `parsnip` stores its model definitions. It is used to dynamically
#' interact with the `parsnip` infrastructure.
#'
#' @return The `parsnip` model environment.
#' @examples
#' \donttest{
#' model_env <- kerasnip::get_model_env()
#' }
#' @keywords internal
#' @export
get_model_env <- function() {
  current <- utils::getFromNamespace("parsnip", ns = "parsnip")
  current
}

#' Check if a Kerasnip Model Specification Exists
#'
#' @description
#' This is an internal helper function to check if a model specification has
#' been registered in the `parsnip` model environment.
#'
#' @param model_name A character string giving the name of the model
#'   specification function to check (e.g., "my_mlp").
#' @return A logical value, `TRUE` if the model exists, `FALSE` otherwise.
#' @examples
#' \donttest{
#' if (requireNamespace("parsnip", quietly = TRUE)) {
#'   library(parsnip)
#'
#'   # Check for a model that exists in parsnip
#'   model_exists("mlp")
#'
#'   # Check for a model that does not exist
#'   model_exists("non_existent_model")
#' }
#' }
#' @keywords internal
#' @export
model_exists <- function(model_name) {
  model_name %in% ls(get_model_env())
}

test_that("keras_evaluate throws error for missing processing functions", {
  # Create a dummy fit object that is missing the processing functions
  dummy_fit_object <- list(
    fit = list(
      fit = "dummy_keras_model"
      # process_x and process_y are missing
    )
  )
  class(dummy_fit_object) <- "model_fit"

  expect_error(
    keras_evaluate(dummy_fit_object, x = mtcars[, -1], y = mtcars$mpg),
    "Could not find processing functions in the model fit object."
  )
})

test_that("keras_model_to_bytes returns raw vector for a valid model", {
  skip_if_no_keras()

  model <- keras3::keras_model_sequential(input_shape = 1L) |>
    keras3::layer_dense(units = 1L)
  keras3::compile(model, optimizer = "adam", loss = "mse")

  bytes <- keras_model_to_bytes(model)
  expect_type(bytes, "raw")
  expect_gt(length(bytes), 0L)
})

test_that("keras_model_from_bytes restores a valid model", {
  skip_if_no_keras()

  model <- keras3::keras_model_sequential(input_shape = 1L) |>
    keras3::layer_dense(units = 1L)
  keras3::compile(model, optimizer = "adam", loss = "mse")

  bytes <- keras_model_to_bytes(model)
  restored <- keras_model_from_bytes(bytes)

  expect_true(inherits(restored, "keras.src.models.sequential.Sequential"))
  x <- matrix(1:3, ncol = 1)
  expect_equal(dim(predict(restored, x)), c(3L, 1L))
})

test_that("get_keras_object returns instances not constructors", {
  skip_if_no_keras()

  loss_obj <- get_keras_object("mse", "loss")
  metric_obj <- get_keras_object("mean_squared_error", "metric")

  # Must be instances (Python objects), not the bare constructor functions
  expect_false(identical(loss_obj, keras3::loss_mean_squared_error))
  expect_false(identical(metric_obj, keras3::metric_mean_squared_error))

  # Must be serialisable (the original bug: class objects fail get_config)
  model <- keras3::keras_model_sequential(input_shape = 1L) |>
    keras3::layer_dense(units = 1L)
  keras3::compile(
    model,
    optimizer = "adam",
    loss = loss_obj,
    metrics = list(metric_obj)
  )
  expect_type(keras_model_to_bytes(model), "raw")
})

test_that("keras_model_to_bytes warns when save_model fails", {
  skip_if_no_keras()

  # Passing a plain R object (not a Keras model) forces save_model() to
  # throw, which exercises the warning path in keras_model_to_bytes()
  # independently of Keras version or compilation arguments.
  expect_warning(
    keras_model_to_bytes(1L),
    "Could not serialize"
  )
})

test_that("keras_evaluate with y=NULL skips y-processing and re-throws on bad xptr", {
  skip_if_no_keras()

  # Mock fit object: valid processing functions, invalid Python object, no bytes.
  # This exercises:
  #   - line 98: y_proc <- NULL  (y = NULL path taken)
  #   - line 99: if (!is.null(y)) FALSE  (y branch skipped)
  #   - line 118: stop(e)  (xptr invalid, no keras_bytes to restore from)
  mock_fit <- list(
    fit = list(
      fit = 1L,
      keras_bytes = NULL,
      process_x = function(x, ...) list(x_proc = as.matrix(x)),
      process_y = function(y, ...) list(y_proc = as.integer(y))
    )
  )
  class(mock_fit) <- "model_fit"

  x_new <- as.data.frame(matrix(rnorm(4), nrow = 1))

  expect_error(keras_evaluate(mock_fit, x_new, y = NULL))
})

#' Register the `update()` S3 Method
#'
#' @description
#' Creates and registers an `update()` S3 method for the new model
#' specification. This method is essential for tuning with `dials`
#' and `tune`, as it allows the tuning machinery to modify model
#' parameters after the spec has been created.
#'
#' @details
#' This function uses `rlang` metaprogramming to dynamically
#' construct a complete `update.{{model_name}}` function.
#' The process involves:
#' \enumerate{
#'   \item Building a function signature that includes `object`, `parameters`,
#'     `...`, `fresh`, and all the tunable parameters from `parsnip_names`.
#'   \item Creating a function body that captures all the arguments into
#'     quosures and passes them to `parsnip::update_spec()`.
#'   \item Registering this new function as an S3 method for the generic
#'     `update()` in the specified environment, so S3 dispatch can find it.
#' }
#'
#' @param model_name The name of the new model specification (e.g., "my_mlp").
#' @param parsnip_names A character vector of all argument names that the
#'   `update()` method should be able to modify.
#' @param env The environment in which to create the `update()` S3 method.
#' @return Invisibly returns `NULL`. Called for its side effects.
#' @noRd
register_update_method <- function(model_name, parsnip_names, env) {
  # Build function signature
  update_args_list <- c(
    list(object = rlang::missing_arg(), parameters = rlang::expr(NULL)),
    purrr::map(parsnip_names, ~ rlang::expr(NULL)),
    list(... = rlang::missing_arg(), fresh = rlang::expr(FALSE))
  )
  names(update_args_list)[3:(2 + length(parsnip_names))] <- parsnip_names

  # Create a list of expressions like `arg_name = rlang::enquo(arg_name)`
  args_enquo_exprs <- purrr::map(
    parsnip_names,
    ~ rlang::expr(rlang::enquo(!!rlang::sym(.x)))
  )
  names(args_enquo_exprs) <- parsnip_names

  # Create the expression that builds this list inside the function body
  args_enquo_list_expr <- rlang::expr(
    args <- rlang::list2(!!!args_enquo_exprs)
  )

  # Create the call to `parsnip::update_spec`
  update_spec_call <- rlang::call2(
    "update_spec",
    object = rlang::sym("object"),
    parameters = rlang::sym("parameters"),
    args_enquo_list = rlang::sym("args"),
    fresh = rlang::sym("fresh"),
    cls = model_name,
    rlang::sym("..."),
    .ns = "parsnip"
  )

  # parsnip::update_spec() ends with new_model_spec(), which strips any extra
  # classes and attributes from the spec. Re-attach kerasnip_spec and the
  # metadata attrs that predict.kerasnip_model_fit needs for auto-registration.
  result_assign <- rlang::expr(result <- !!update_spec_call)

  restore_class_expr <- rlang::expr(
    class(result) <- c(class(result)[1L], "kerasnip_spec", class(result)[-1L])
  )
  restore_blocks_expr <- rlang::expr(
    attr(result, "kerasnip_layer_blocks") <- attr(
      object,
      "kerasnip_layer_blocks"
    )
  )
  restore_functional_expr <- rlang::expr(
    attr(result, "kerasnip_functional") <- attr(object, "kerasnip_functional")
  )

  # Combine them into the final body
  update_body <- rlang::call2(
    "{",
    args_enquo_list_expr,
    result_assign,
    restore_class_expr,
    restore_blocks_expr,
    restore_functional_expr,
    rlang::sym("result")
  )

  # Create and register the S3 method
  update_func <- rlang::new_function(
    args = update_args_list,
    body = update_body
  )
  method_name <- paste0("update.", model_name)
  # Poke the function into the target environment (e.g., .GlobalEnv) so that
  # S3 dispatch can find it.
  rlang::env_poke(env, method_name, update_func)
  registerS3method("update", model_name, update_func, envir = env)
}

#' Register Fit and Prediction Methods with Parsnip
#'
#' @description
#' This function registers the methods that `parsnip` will use to fit the model
#' and generate predictions.
#'
#' @details
#' This function makes calls to `parsnip::set_fit()` and `parsnip::set_pred()`:
#' - `set_fit()`: Links the model specification to the appropriate generic
#'   fitting engine (`generic_sequential_fit()` or `generic_functional_fit()`).
#'   It also passes the user's `layer_blocks` list as a default argument to
#'   the fitting function.
#' - `set_pred()`: Defines how to generate predictions for different types
#'   ("numeric", "class", "prob"). It specifies the underlying `predict()`
#'   method and the post-processing functions (`keras_postprocess_*`) needed
#'   to format the output into a standardized `tidymodels` tibble.
#'
#' @param model_name The name of the new model.
#' @param mode The model mode ("regression" or "classification").
#' @param layer_blocks The named list of layer block functions, which is passed
#'   as a default argument to the fit function.
#' @param functional A logical. If `TRUE`, registers `generic_functional_fit` as
#'   the fitting engine. Otherwise, registers `generic_sequential_fit`.
#' @return Invisibly returns `NULL`. Called for its side effects.
#' @noRd
register_fit_predict <- function(model_name, mode, layer_blocks, functional) {
  # Fit method
  parsnip::set_fit(
    model = model_name,
    eng = "keras",
    mode = mode,
    value = list(
      interface = "formula",
      protect = c("formula", "data"),
      func = c(
        pkg = "kerasnip",
        fun = if (functional) {
          "generic_functional_fit"
        } else {
          "generic_sequential_fit"
        }
      ),
      defaults = list(layer_blocks = layer_blocks)
    )
  )

  # Regression prediction
  if (mode == "regression") {
    parsnip::set_pred(
      model = model_name,
      eng = "keras",
      mode = "regression",
      type = "numeric",
      value = list(
        pre = NULL,
        post = keras_postprocess_numeric,
        func = c(fun = "predict"),
        args = list(
          object = rlang::expr(object$fit$fit),
          x = if (functional) {
            rlang::expr(process_x_functional(new_data)$x_proc)
          } else {
            rlang::expr(process_x_sequential(new_data)$x_proc)
          }
        )
      )
    )
  } else {
    # Classification predictions
    parsnip::set_pred(
      model = model_name,
      eng = "keras",
      mode = "classification",
      type = "class",
      value = list(
        pre = NULL,
        post = keras_postprocess_classes,
        func = c(fun = "predict"),
        args = list(
          object = rlang::expr(object$fit$fit),
          x = if (functional) {
            rlang::expr(process_x_functional(new_data)$x_proc)
          } else {
            rlang::expr(process_x_sequential(new_data)$x_proc)
          }
        )
      )
    )
    parsnip::set_pred(
      model = model_name,
      eng = "keras",
      mode = "classification",
      type = "prob",
      value = list(
        pre = NULL,
        post = keras_postprocess_probs,
        func = c(fun = "predict"),
        args = list(
          object = rlang::expr(object$fit$fit),
          x = if (functional) {
            rlang::expr(process_x_functional(new_data)$x_proc)
          } else {
            rlang::expr(process_x_sequential(new_data)$x_proc)
          }
        )
      )
    )
  }
}

##' Post-process Keras Numeric Predictions
#'
#' @description
#' Formats raw numeric predictions from a Keras model into a tibble with the
#' standardized `.pred` column, as required by `tidymodels`.
#'
#' @details
#' This function simply takes the matrix output from `keras3::predict()` and
#' converts it to a single-column tibble.
#' @param results A matrix of numeric predictions from `predict()`.
#' @param object The `parsnip` model fit object.
#' @return A tibble with a `.pred` column.
#' @noRd
keras_postprocess_numeric <- function(results, object) {
  if (is.list(results) && !is.null(names(results))) {
    # Multi-output case: results is a named list of arrays/matrices
    # Combine them into a single tibble with appropriate column names
    combined_preds <- tibble::as_tibble(results)
    # Rename columns to .pred_output_name if there are multiple outputs
    if (length(results) > 1) {
      colnames(combined_preds) <- paste0(".pred_", names(results))
    } else {
      # If only one output, but still a list, name it .pred
      colnames(combined_preds) <- ".pred"
    }
    combined_preds
  } else {
    # Single output case: results is a matrix/array
    tibble::tibble(.pred = as.vector(results))
  }
}

#' Post-process Keras Probability Predictions
#'
#' @description
#' Formats raw probability predictions from a Keras model into a tibble with
#' class-specific column names (e.g., `.pred_class1`, `.pred_class2`).
#'
#' @details
#' This function retrieves the original factor levels from `object$fit$lvl`
#' (which was stored by the fitting engine) and uses them to name the columns.
#' @param results A matrix of probability predictions from `predict()`.
#' @param object The `parsnip` model fit object.
#' @return A tibble with named columns for each class probability.
#' @noRd
keras_postprocess_probs <- function(results, object) {
  if (is.list(results) && !is.null(names(results))) {
    # Multi-output case: results is a named list of arrays/matrices
    combined_preds <- purrr::map2_dfc(
      results,
      names(results),
      function(res, name) {
        lvls <- object$fit$lvl[[name]]
        # Assuming object$fit$lvl is a named list of levels
        if (is.null(lvls)) {
          # Fallback if levels are not specifically named for this output
          lvls <- paste0("class", seq_len(ncol(res)))
        }
        colnames(res) <- lvls
        tibble::as_tibble(res, .name_repair = "unique") |>
          dplyr::rename_with(~ paste0(".pred_", name, "_", .x))
      }
    )
    combined_preds
  } else {
    # Single output case: results is a matrix/array
    # The levels are now nested inside the fit object
    colnames(results) <- object$fit$lvl
    tibble::as_tibble(results)
  }
}

#' Post-process Keras Class Predictions
#'
#' @description
#' Converts raw probability predictions from a Keras model into a single
#' `.pred_class` column of factor predictions.
#'
#' @details
#' For multiclass models, it finds the class with the highest probability
#' (`which.max`). For binary models, it applies a 0.5 threshold. It uses the
#' levels stored in `object$fit$lvl` to ensure the output factor is correct.
#' @param results A matrix of probability predictions from `predict()`.
#' @param object The `parsnip` model fit object.
#' @return A tibble with a `.pred_class` column containing factor predictions.
#' @noRd
keras_postprocess_classes <- function(results, object) {
  if (is.list(results) && !is.null(names(results))) {
    # Multi-output case: results is a named list of arrays/matrices
    combined_preds <- purrr::map2_dfc(
      results,
      names(results),
      function(res, name) {
        lvls <- object$fit$lvl[[name]]
        # Assuming object$fit$lvl is a named list of levels
        if (is.null(lvls)) {
          # Fallback if levels are not specifically named for this output
          lvls <- paste0("class", seq_len(ncol(res)))
          # This might not be correct for classes, but a placeholder
        }

        if (ncol(res) == 1) {
          # Binary classification
          pred_class <- ifelse(res[, 1] > 0.5, lvls[2], lvls[1])
          pred_class <- factor(pred_class, levels = lvls)
        } else {
          # Multiclass classification
          pred_class_int <- apply(res, 1, which.max)
          pred_class <- lvls[pred_class_int]
          pred_class <- factor(pred_class, levels = lvls)
        }
        tibble::tibble(.pred_class = pred_class) |>
          dplyr::rename_with(~ paste0(".pred_class_", name))
      }
    )
    combined_preds
  } else {
    # Single output case: results is a matrix/array
    # The levels are now nested inside the fit object
    lvls <- object$fit$lvl
    if (ncol(results) == 1) {
      # Binary classification
      pred_class <- ifelse(results[, 1] > 0.5, lvls[2], lvls[1])
      pred_class <- factor(pred_class, levels = lvls)
    } else {
      # Multiclass classification
      pred_class_int <- apply(results, 1, which.max)
      pred_class <- lvls[pred_class_int]
      pred_class <- factor(pred_class, levels = lvls)
    }
    tibble::tibble(.pred_class = pred_class)
  }
}

skip_if_no_keras()

# Mock object for post-processing functions
mock_object_single_output <- list(
  fit = list(
    lvl = c("setosa", "versicolor", "virginica") # For classification levels
  )
)
class(mock_object_single_output) <- "model_fit"

mock_object_multi_output <- list(
  fit = list(
    lvl = list(
      output1 = c("classA", "classB"),
      output2 = c("typeX", "typeY", "typeZ")
    )
  )
)
class(mock_object_multi_output) <- "model_fit"

# --- Tests for keras_postprocess_numeric ---

test_that("keras_postprocess_numeric handles single output (matrix)", {
  results <- matrix(c(0.1, 0.2, 0.3), ncol = 1)
  processed <- keras_postprocess_numeric(results, mock_object_single_output)
  expect_s3_class(processed, "tbl_df")
  expect_equal(names(processed), ".pred")
  expect_equal(processed$.pred, c(0.1, 0.2, 0.3))
})

test_that("keras_postprocess_numeric handles single output (named list)", {
  results <- list(output1 = matrix(c(0.1, 0.2, 0.3), ncol = 1))
  names(results) <- "output1"
  processed <- keras_postprocess_numeric(results, mock_object_multi_output)
  expect_s3_class(processed, "tbl_df")
  expect_equal(names(processed), ".pred")
  expect_equal(processed$.pred, matrix(c(0.1, 0.2, 0.3), ncol = 1))
})


test_that("keras_postprocess_numeric handles multi-output (named list)", {
  results <- list(
    output1 = matrix(c(0.1, 0.2), ncol = 1),
    output2 = matrix(c(0.4, 0.5), ncol = 1)
  )
  names(results) <- c("output1", "output2")
  processed <- keras_postprocess_numeric(results, mock_object_multi_output)
  expect_s3_class(processed, "tbl_df")
  expect_equal(names(processed), c(".pred_output1", ".pred_output2"))
  # Change expected values to 1-column matrices
  expect_equal(processed$.pred_output1, matrix(c(0.1, 0.2), ncol = 1))
  expect_equal(processed$.pred_output2, matrix(c(0.4, 0.5), ncol = 1))
})

# --- Tests for keras_postprocess_probs ---

test_that("keras_postprocess_probs handles single output (matrix) correctly", {
  results <- matrix(
    c(
      0.1,
      0.9,
      0.0, # Example probabilities for 3 classes
      0.2,
      0.1,
      0.7,
      0.3,
      0.3,
      0.4
    ),
    ncol = 3,
    byrow = TRUE
  )
  processed <- keras_postprocess_probs(results, mock_object_single_output)
  expect_s3_class(processed, "tbl_df")
  expect_equal(names(processed), c("setosa", "versicolor", "virginica"))
  expect_equal(processed$setosa, c(0.1, 0.2, 0.3))
  expect_equal(processed$versicolor, c(0.9, 0.1, 0.3))
  expect_equal(processed$virginica, c(0.0, 0.7, 0.4))
})

test_that("keras_postprocess_probs handles multi-output (named list)", {
  results <- list(
    output1 = matrix(c(0.1, 0.9, 0.2, 0.8), ncol = 2, byrow = TRUE),
    output2 = matrix(c(0.3, 0.4, 0.3, 0.5, 0.2, 0.3), ncol = 3, byrow = TRUE)
  )
  names(results) <- c("output1", "output2")
  processed <- keras_postprocess_probs(results, mock_object_multi_output)
  expect_s3_class(processed, "tbl_df")
  expect_equal(
    names(processed),
    c(
      ".pred_output1_classA",
      ".pred_output1_classB",
      ".pred_output2_typeX",
      ".pred_output2_typeY",
      ".pred_output2_typeZ"
    )
  )
  expect_equal(processed$.pred_output1_classA, c(0.1, 0.2))
  expect_equal(processed$.pred_output2_typeX, c(0.3, 0.5))
})

test_that("keras_postprocess_probs handles multi-output with NULL levels", {
  results <- list(
    output1 = matrix(c(0.1, 0.9, 0.2, 0.8), ncol = 2, byrow = TRUE)
  )
  names(results) <- "output1"
  mock_object_null_lvl <- list(
    fit = list(
      lvl = list(output1 = NULL) # Simulate NULL levels for this output
    )
  )
  class(mock_object_null_lvl) <- "model_fit"
  processed <- keras_postprocess_probs(results, mock_object_null_lvl)
  expect_s3_class(processed, "tbl_df")
  expect_equal(
    names(processed),
    c(".pred_output1_class1", ".pred_output1_class2")
  )
})

# --- Tests for keras_postprocess_classes ---

test_that("keras_postprocess_classes handles single output (multiclass)", {
  results <- matrix(c(0.1, 0.8, 0.1, 0.2, 0.1, 0.7), ncol = 3, byrow = TRUE)
  processed <- keras_postprocess_classes(results, mock_object_single_output)
  expect_s3_class(processed, "tbl_df")
  expect_equal(names(processed), ".pred_class")
  expect_equal(
    as.character(processed$.pred_class),
    c("versicolor", "virginica")
  )
  expect_true(is.factor(processed$.pred_class))
  expect_equal(
    levels(processed$.pred_class),
    c("setosa", "versicolor", "virginica")
  )
})

test_that("keras_postprocess_classes handles single output (binary)", {
  results <- matrix(c(0.6, 0.4), ncol = 1) # Changed to single column
  mock_object_binary_lvl <- list(
    fit = list(
      lvl = c("negative", "positive")
    )
  )
  class(mock_object_binary_lvl) <- "model_fit"
  processed <- keras_postprocess_classes(results, mock_object_binary_lvl)
  expect_s3_class(processed, "tbl_df")
  expect_equal(names(processed), ".pred_class")
  expect_equal(as.character(processed$.pred_class), c("positive", "negative"))
  expect_true(is.factor(processed$.pred_class))
  expect_equal(levels(processed$.pred_class), c("negative", "positive"))
})

test_that("keras_postprocess_classes handles multi-output (named list)", {
  results <- list(
    output1 = matrix(c(0.1, 0.9, 0.2, 0.8), ncol = 2, byrow = TRUE),
    output2 = matrix(c(0.3, 0.4, 0.3, 0.5, 0.2, 0.3), ncol = 3, byrow = TRUE)
  )
  names(results) <- c("output1", "output2")
  processed <- keras_postprocess_classes(results, mock_object_multi_output)
  expect_s3_class(processed, "tbl_df")
  expect_equal(
    names(processed),
    c(".pred_class_output1", ".pred_class_output2")
  )
  expect_equal(
    as.character(processed$.pred_class_output1),
    c("classB", "classB")
  )
  expect_equal(as.character(processed$.pred_class_output2), c("typeY", "typeX"))
  expect_true(is.factor(processed$.pred_class_output1))
  expect_true(is.factor(processed$.pred_class_output2))
  expect_equal(levels(processed$.pred_class_output1), c("classA", "classB"))
  expect_equal(
    levels(processed$.pred_class_output2),
    c("typeX", "typeY", "typeZ")
  )
})

test_that("keras_postprocess_classes handles multi-output with NULL levels", {
  results <- list(
    output1 = matrix(c(0.6, 0.4, 0.2, 0.8), ncol = 2, byrow = TRUE) # Binary
  )
  names(results) <- "output1"
  mock_object_null_lvl <- list(
    fit = list(
      lvl = list(output1 = NULL) # Simulate NULL levels for this output
    )
  )
  class(mock_object_null_lvl) <- "model_fit"
  processed <- keras_postprocess_classes(results, mock_object_null_lvl)
  expect_s3_class(processed, "tbl_df")
  expect_equal(names(processed), c(".pred_class_output1"))
  expect_equal(
    as.character(processed$.pred_class_output1),
    c("class1", "class2")
  ) # Changed expected
  expect_true(is.factor(processed$.pred_class_output1))
  expect_equal(levels(processed$.pred_class_output1), c("class1", "class2"))
})

test_that("keras_postprocess_classes handles multi-output binary single col", {
  results <- list(
    output1 = matrix(c(0.6, 0.4, 0.2, 0.8), ncol = 1, byrow = TRUE)
  )
  names(results) <- "output1"
  mock_multi_output_bin <- list(
    fit = list(
      lvl = list(output1 = c("negative", "positive")) # Levels for binary output
    )
  )
  class(mock_multi_output_bin) <- "model_fit"
  processed <- keras_postprocess_classes(
    results,
    mock_multi_output_bin
  )
  expect_s3_class(processed, "tbl_df")
  expect_equal(names(processed), c(".pred_class_output1"))
  expect_equal(
    as.character(processed$.pred_class_output1),
    c("positive", "negative", "negative", "positive")
  ) # Expected based on 0.5 threshold
  expect_true(is.factor(processed$.pred_class_output1))
  expect_equal(levels(processed$.pred_class_output1), c("negative", "positive"))
})

#' Remove a Keras Model Specification and its Registrations
#'
#' @description
#' This function completely removes a model specification that was previously
#' created by [create_keras_sequential_spec()] or
#' [create_keras_functional_spec()]. It cleans up both the function in the
#' user's environment and all associated registrations within the `parsnip`
#' package.
#'
#' @details
#' This function is essential for cleanly unloading a dynamically created model.
#' It performs three main actions:
#' \enumerate{
#'   \item It removes the model specification function (e.g., `my_mlp()`) and
#'     its corresponding `update()` method from the specified environment.
#'   \item It searches `parsnip`'s internal model environment for all objects
#'     whose names start with the `model_name` and removes them. This purges
#'     the fit methods, argument definitions, and other registrations.
#'   \item It removes the model's name from `parsnip`'s master list of models.
#' }
#' This function uses the un-exported `get_model_env()` to perform
#' the cleanup.
#'
#' @param model_name A character string giving the name of the model
#'   specification function to remove (e.g., "my_mlp").
#' @param env The environment from which to remove the function and its
#'   `update()` method. Defaults to the calling environment (`parent.frame()`).
#' @return Invisibly returns `TRUE` after attempting to remove the objects.
#' @seealso [create_keras_sequential_spec()], [create_keras_functional_spec()]
#' @export
#' @examples
#' \donttest{
#' if (requireNamespace("keras3", quietly = TRUE)) {
#'   # First, create a dummy spec
#'   input_block <- function(model, input_shape) {
#'     keras3::keras_model_sequential(input_shape = input_shape)
#'   }
#'   dense_block <- function(model, units = 16) {
#'     model |> keras3::layer_dense(units = units)
#'   }
#'   create_keras_sequential_spec(
#'     "my_temp_model",
#'     list(
#'       input = input_block,
#'       dense = dense_block
#'     ),
#'     "regression"
#'   )
#'
#'   # Check it exists in the environment and in parsnip
#'   exists("my_temp_model")
#'   "my_temp_model" %in% parsnip::show_engines("my_temp_model")$model
#'
#'   # Now remove it
#'   remove_keras_spec("my_temp_model")
#'
#'   # Check it's gone
#'   !exists("my_temp_model")
#'   !model_exists("my_temp_model")
#' }
#' }
remove_keras_spec <- function(model_name, env = parent.frame()) {
  # 1. Remove the spec + update fn from the user env
  if (
    exists(model_name, envir = env, inherits = FALSE) &&
      is.function(get(model_name, envir = env))
  ) {
    remove(list = model_name, envir = env)
  }
  update_fn <- paste0("update.", model_name)
  if (exists(update_fn, envir = env, inherits = FALSE)) {
    remove(list = update_fn, envir = env)
  }

  # 2. Nuke every parsnip object whose name starts with model_name
  model_env <- get_model_env()
  all_regs <- ls(envir = model_env)
  to_kill <- intersect(
    all_regs,
    paste0(
      model_name,
      c("", "_args", "_encoding", "_fit", "_modes", "_pkgs", "_predict")
    )
  )
  if (length(to_kill)) {
    rm(list = to_kill, envir = model_env)
    message(
      "Removed from parsnip registry objects: ",
      paste(to_kill, collapse = ", ")
    )
  }

  # 3. Remove the entry in get_model_env()$models
  if ("models" %in% all_regs && model_name %in% model_env$models) {
    model_env$models <- model_env$models[-which(model_name == model_env$models)]
    message("Removed '", model_name, "' from parsnip:::get_model_env()$models")
  }

  invisible(TRUE)
}

#' Create a Custom Keras Functional API Model Specification for Tidymodels
#'
#' This function acts as a factory to generate a new `parsnip` model
#' specification based on user-defined blocks of Keras layers using the
#' Functional API. This allows for creating complex, tunable architectures
#' with non-linear topologies that integrate seamlessly with the `tidymodels`
#' ecosystem.
#'
#' @param model_name A character string for the name of the new model
#'   specification function (e.g., "custom_resnet"). This should be a valid R
#'   function name.
#' @param layer_blocks A named list of functions where each function defines a
#'   "block" (a node) in the model graph. The list names are crucial as they
#'   define the names of the nodes. The arguments of each function define how
#'   the nodes are connected. See the "Model Graph Connectivity" section for
#'   details.
#' @param mode A character string, either "regression" or "classification".
#' @param ... Reserved for future use. Currently not used.
#' @param env The environment in which to create the new model specification
#'   function and its associated `update()` method. Defaults to the calling
#'   environment (`parent.frame()`).
#'
#' @details
#' This function generates all the boilerplate needed to create a custom,
#' tunable `parsnip` model specification that uses the Keras Functional API.
#' This is ideal for models with complex, non-linear topologies, such as
#' networks with multiple inputs/outputs or residual connections.
#'
#' The function inspects the arguments of your `layer_blocks` functions and
#' makes them available as tunable parameters in the generated model
#' specification, prefixed with the block's name (e.g., `dense_units`).
#' Common training parameters such as `epochs` and `learn_rate` are also added.
#'
#' @section Model Graph Connectivity:
#' `kerasnip` builds the model's directed acyclic graph by inspecting the
#' arguments of each function in the `layer_blocks` list. The connection logic
#' is as follows:
#'
#' 1.  The **names of the elements** in the `layer_blocks` list define the names
#'     of the nodes in your graph (e.g., `main_input`, `dense_path`, `output`).
#' 2.  The **names of the arguments** in each block function specify its inputs.
#'     A block function like `my_block <- function(input_a, input_b, ...)`
#'     declares that it needs input from the nodes named `input_a` and
#'     `input_b`. `kerasnip` will automatically supply the output tensors from
#'     those nodes when calling `my_block`.
#'
#' There are two special requirements:
#' *   **Input Block**: The first block in the list is treated as the input
#'     node. Its function should not take other blocks as input, but it can have
#'     an `input_shape` argument, which is supplied automatically during
#'     fitting.
#' *   **Output Block**: Exactly one block must be named `"output"`. The tensor
#'     returned by this block is used as the final output of the Keras model.
#'
#' A key feature is the automatic creation of `num_{block_name}` arguments
#' (e.g., `num_dense_path`). This allows you to control how many times a block
#' is repeated, making it easy to tune the depth of your network. A block can
#' only be repeated if it has exactly one input from another block in the graph.
#'
#' The new model specification function and its `update()` method are created
#' in the environment specified by the `env` argument.
#'
#' @importFrom rlang enquos dots_list arg_match env_poke
#' @importFrom parsnip update_dot_check
#'
#' @return Invisibly returns `NULL`. Its primary side effect is to create a
#'   new model specification function (e.g., `custom_resnet()`) in the
#'   specified environment and register the model with `parsnip` so it can be
#'   used within the `tidymodels` framework.
#'
#' @seealso [remove_keras_spec()], [parsnip::new_model_spec()],
#'   [create_keras_sequential_spec()]
#'
#' @section Saving and Reloading Models:
#' To save a fitted workflow and reload it in a new R session, use
#' `bundle::bundle()` before saving — this is required to preserve the Keras
#' model weights:
#'
#' ```r
#' library(bundle)
#' bundled <- bundle(fitted_workflow)
#' saveRDS(bundled, "model.rds")
#'
#' # New session:
#' library(kerasnip); library(bundle)
#' fitted_workflow <- unbundle(readRDS("model.rds"))
#' predict(fitted_workflow, new_data = test_data)  # works
#' ```
#'
#' Plain `saveRDS()` without `bundle()` does not preserve Keras weights, but
#' `predict()` will still auto-register the parsnip model type from metadata
#' stored on the spec.
#'
#' @export
#' @examples
#' \donttest{
#' if (requireNamespace("keras3", quietly = TRUE)) {
#'   library(keras3)
#'   library(parsnip)
#'
#'   # 1. Define block functions. These are the building blocks of our model.
#'   # An input block that receives the data's shape automatically.
#'   input_block <- function(input_shape) layer_input(shape = input_shape)
#'
#'   # A dense block with a tunable `units` parameter.
#'   dense_block <- function(tensor, units) {
#'     tensor |> layer_dense(units = units, activation = "relu")
#'   }
#'
#'   # A block that adds two tensors together (for the residual connection).
#'   add_block <- function(input_a, input_b) layer_add(list(input_a, input_b))
#'
#'   # An output block for regression.
#'   output_block_reg <- function(tensor) layer_dense(tensor, units = 1)
#'
#'   # 2. Create the spec. The `layer_blocks` list defines the graph.
#'   create_keras_functional_spec(
#'     model_name = "my_resnet_spec",
#'     layer_blocks = list(
#'       # The names of list elements are the node names.
#'       main_input = input_block,
#'
#'       # The argument `main_input` connects this block to the input node.
#'       dense_path = function(main_input, units = 32) {
#'         dense_block(main_input, units)
#'       },
#'
#'       # This block's arguments connect it to the original input AND the dense
#'       # layer.
#'       add_residual = function(main_input, dense_path) {
#'         add_block(main_input, dense_path)
#'       },
#'
#'       # This block must be named 'output'. It connects to the residual add
#'       # layer.
#'       output = function(add_residual) output_block_reg(add_residual)
#'     ),
#'     mode = "regression"
#'   )
#'
#'   # 3. Use the newly created specification function!
#'   # The `dense_path_units` argument was created automatically.
#'   model_spec <- my_resnet_spec(dense_path_units = 64, epochs = 10)
#'
#'   # You could also tune the number of dense layers since it has a single
#'   # input:
#'   # model_spec <- my_resnet_spec(num_dense_path = 2, dense_path_units = 32)
#'
#'   print(model_spec)
#'   remove_keras_spec("my_resnet_spec")
#'   # tune::tunable(model_spec)
#' }
#' }
create_keras_functional_spec <- function(
  model_name,
  layer_blocks,
  mode = c("regression", "classification"),
  ...,
  env = parent.frame()
) {
  mode <- rlang::arg_match(mode)
  # 1. Argument Validation
  create_keras_spec_impl(
    model_name,
    layer_blocks,
    mode,
    functional = TRUE,
    env
  )
}

#' Evaluate a Fitted Kerasnip Model on New Data
#'
#' @title Evaluate a Kerasnip Model
#' @description
#' This function provides an `kera_evaluate()` method for `model_fit` objects
#' created by `kerasnip`. It preprocesses the new data into the format expected
#' by Keras and then calls `keras3::evaluate()` on the underlying model to
#' compute the loss and any other metrics.
#'
#' @param object A `model_fit` object produced by a `kerasnip` specification.
#' @param x A data frame or matrix of new predictor data.
#' @param y A vector or data frame of new outcome data corresponding to `x`.
#' @param ... Additional arguments passed on to `keras3::evaluate()` (e.g.,
#'   `batch_size`).
#'
#' @return A named list containing the evaluation results (e.g., `loss`,
#'   `accuracy`). The names are determined by the metrics the model was compiled
#'   with.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("keras3", quietly = TRUE)) {
#' library(keras3)
#' library(parsnip)
#'
#' # 1. Define layer blocks
#' input_block <- function(model, input_shape) {
#'   keras_model_sequential(input_shape = input_shape)
#' }
#' hidden_block <- function(model, units = 32) {
#'   model |> layer_dense(units = units, activation = "relu")
#' }
#' output_block <- function(model, num_classes) {
#'   model |> layer_dense(units = num_classes, activation = "softmax")
#' }
#'
#' # 2. Define and fit a model ----
#' create_keras_sequential_spec(
#'   model_name = "my_mlp_tools",
#'   layer_blocks = list(
#'     input = input_block,
#'     hidden = hidden_block,
#'     output = output_block
#'   ),
#'   mode = "classification"
#' )
#'
#' mlp_spec <- my_mlp_tools(
#'   hidden_units = 32,
#'   compile_loss = "categorical_crossentropy",
#'   compile_optimizer = "adam",
#'   compile_metrics = "accuracy",
#'   fit_epochs = 5
#' ) |> set_engine("keras")
#'
#' x_train <- matrix(rnorm(100 * 10), ncol = 10)
#' y_train <- factor(sample(0:1, 100, replace = TRUE))
#' train_df <- data.frame(x = I(x_train), y = y_train)
#'
#' fitted_mlp <- fit(mlp_spec, y ~ x, data = train_df)
#'
#' # 3. Evaluate the model on new data ----
#' x_test <- matrix(rnorm(50 * 10), ncol = 10)
#' y_test <- factor(sample(0:1, 50, replace = TRUE))
#'
#' eval_metrics <- keras_evaluate(fitted_mlp, x_test, y_test)
#' print(eval_metrics)
#'
#' # 4. Extract the Keras model object ----
#' keras_model <- extract_keras_model(fitted_mlp)
#' summary(keras_model)
#'
#' # 5. Extract the training history ----
#' history <- extract_keras_history(fitted_mlp)
#' plot(history)
#' remove_keras_spec("my_mlp_tools")
#' }
#' }
#' @export
keras_evaluate <- function(object, x, y = NULL, ...) {
  # 1. Get the correct processing functions from the fit object
  process_x_fun <- object$fit$process_x
  process_y_fun <- object$fit$process_y

  if (is.null(process_x_fun) || is.null(process_y_fun)) {
    stop(
      "Could not find processing functions in the model fit object. ",
      "Please ensure the model was fitted with a recent version of kerasnip.",
      call. = FALSE
    )
  }

  # 2. Preprocess predictor data (x)
  x_processed <- process_x_fun(x)
  x_proc <- x_processed$x_proc

  # 3. Preprocess outcome data (y)
  y_proc <- NULL
  if (!is.null(y)) {
    # Note: For evaluation, we pass the class levels from the trained model
    # to ensure consistent encoding of the new data.
    y_processed <- process_y_fun(
      y,
      is_classification = !is.null(object$fit$lvl),
      class_levels = object$fit$lvl
    )
    y_proc <- y_processed$y_proc
  }

  # 4. Call the underlying Keras evaluate method
  keras_model <- tryCatch(
    {
      reticulate::py_validate_xptr(object$fit$fit)
      object$fit$fit
    },
    error = function(e) {
      if (is.null(object$fit$keras_bytes)) {
        stop(e)
      }
      keras_model_from_bytes(object$fit$keras_bytes)
    }
  )
  keras3::evaluate(keras_model, x = x_proc, y = y_proc, ...)
}

#' Extract the Raw Keras Model from a Kerasnip Fit
#'
#' @title Extract Keras Model from a Fitted Kerasnip Object
#' @description
#' Extracts and returns the underlying Keras model object from a `parsnip`
#' `model_fit` object created by `kerasnip`.
#'
#' @details
#' This is useful when you need to work directly with the Keras model object for
#' tasks like inspecting layer weights, creating custom plots, or passing it to
#' other Keras-specific functions.
#'
#' @param object A `model_fit` object produced by a `kerasnip` specification.
#'
#' @return The raw Keras model object (`keras_model`).
#' @seealso keras_evaluate, extract_keras_history
#' @export
extract_keras_model <- function(object) {
  tryCatch(
    {
      reticulate::py_validate_xptr(object$fit$fit)
      object$fit$fit
    },
    error = function(e) {
      if (is.null(object$fit$keras_bytes)) {
        stop(e)
      }
      keras_model_from_bytes(object$fit$keras_bytes)
    }
  )
}

#' Extract Keras Training History
#'
#' @title Extract Keras Training History
#' @description
#' Extracts and returns the training history from a `parsnip` `model_fit` object
#' created by `kerasnip`.
#'
#' @details
#' The history object contains the metrics recorded during model training, such
#' as loss and accuracy, for each epoch. This is highly useful for visualizing
#' the training process and diagnosing issues like overfitting.
#' The returned object can be plotted directly.
#'
#' @param object A `model_fit` object produced by a `kerasnip` specification.
#' @return A `keras_training_history` object. You can call `plot()` on this
#'   object to visualize the learning curves.
#' @seealso keras_evaluate, extract_keras_model
#' @export
extract_keras_history <- function(object) {
  object$fit$history
}

#' Tidy a Fitted Kerasnip Model
#'
#' @description
#' Returns a tibble with one row per layer of the underlying Keras model,
#' summarising the layer name, Python class, and parameter count.
#'
#' @param x A `kerasnip_model_fit` object.
#' @param ... Not used.
#' @return A tibble with columns `layer` (character), `class` (character),
#'   and `n_params` (integer).
#' @importFrom generics tidy
#' @importFrom tibble tibble
#' @importFrom utils tail
#' @keywords internal
#' @export
tidy.kerasnip_model_fit <- function(x, ...) {
  model <- extract_keras_model(x)
  layers <- model$layers

  layer_name <- vapply(layers, function(l) l$name, character(1))
  layer_class <- vapply(
    layers,
    function(l) {
      parts <- strsplit(class(l)[[1L]], ".", fixed = TRUE)[[1L]]
      tail(parts, 1L)
    },
    character(1)
  )
  n_params <- vapply(
    layers,
    function(l) {
      tryCatch(as.integer(l$count_params()), error = function(e) NA_integer_)
    },
    integer(1)
  )

  tibble(layer = layer_name, class = layer_class, n_params = n_params)
}

#' Glance at a Fitted Kerasnip Model
#'
#' @description
#' Returns a one-row tibble of summary statistics from the final training epoch:
#' every metric the model was compiled with (e.g. `loss`, `accuracy`).
#'
#' @param x A `kerasnip_model_fit` object.
#' @param ... Not used.
#' @return A one-row tibble with one column per compiled metric. Returns an
#'   empty tibble if training history has been stripped (e.g. by butcher).
#' @importFrom generics glance
#' @importFrom tibble as_tibble tibble
#' @keywords internal
#' @export
glance.kerasnip_model_fit <- function(x, ...) {
  history <- x$fit$history
  if (is.null(history) || is.null(history$metrics)) {
    return(tibble::tibble())
  }
  last_vals <- lapply(history$metrics, function(m) m[[length(m)]])
  as_tibble(last_vals)
}

test_that("E2E: Customizing main arguments works", {
  skip_if_no_keras()

  input_block_feat <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block_feat <- function(model, units = 16) {
    model |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block_feat <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }

  model_name <- "e2e_mlp_feat"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_feat,
      dense = dense_block_feat,
      output = output_block_feat
    ),
    mode = "regression"
  )

  # Main arguments (like compile_*) should be set in the spec function,
  # not in set_engine().
  spec <- e2e_mlp_feat(
    fit_epochs = 2,
    compile_optimizer = "sgd",
    compile_loss = "mae",
    compile_metrics = c("mean_squared_error", "root_mean_squared_error")
  ) |>
    parsnip::set_engine("keras")

  # This should now run without the parsnip warning about removing arguments
  fit_obj <- NULL
  expect_no_warning(
    fit_obj <- parsnip::fit(spec, mpg ~ ., data = mtcars)
  )

  # Also verify the arguments were correctly used during compilation
  keras_model <- fit_obj$fit$fit
  compiled_loss <- keras_model$loss
  compiled_optimizer <- tolower(keras_model$optimizer$name)
  compiled_metrics <- sapply(
    keras_model$metrics[[2]]$metrics,
    function(m) {
      m$name
    }
  )

  # Keras might add suffixes or use different casings, so check flexibly
  expect_true(grepl("mae", compiled_loss))
  expect_true(grepl("sgd", tolower(compiled_optimizer)))
  expect_true("mean_squared_error" %in% compiled_metrics)
  expect_true("root_mean_squared_error" %in% compiled_metrics)
})

test_that("E2E: Customizing fit arguments works", {
  skip_if_no_keras()

  input_block_fit <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block_fit <- function(model, units = 16) {
    model |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block_fit <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }

  model_name <- "e2e_mlp_fit"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_fit,
      dense = dense_block_fit,
      output = output_block_fit
    ),
    mode = "regression"
  )

  # Fit arguments (like validation_split, callbacks) should be set in the
  # spec function, not in set_engine().
  spec <- e2e_mlp_fit(
    fit_validation_split = 0.2,
    fit_callbacks = list(keras3::callback_early_stopping(patience = 1)),
    fit_epochs = 3,
    compile_metrics = "mean_squared_error"
  ) |>
    parsnip::set_engine("keras")

  # This will run without error if the arguments are passed correctly
  fit_obj <- NULL
  expect_no_error(fit_obj <- parsnip::fit(spec, mpg ~ ., data = mtcars))

  # Check that the callback was used (model should stop early)
  expect_lt(length(fit_obj$fit$history$metrics$loss), 5)
})

test_that("E2E: Setting num_blocks = 0 works for sequential models", {
  skip_if_no_keras()

  input_block_zero <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block_zero <- function(model, units = 16) {
    model |>
      keras3::layer_dense(
        units = units,
        activation = "relu",
        name = "i_should_not_exist"
      )
  }
  output_block_zero <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }

  model_name <- "e2e_mlp_zero"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_zero,
      dense = dense_block_zero,
      output = output_block_zero
    ),
    mode = "regression"
  )

  spec <- e2e_mlp_zero(num_dense = 0, fit_epochs = 1) |>
    parsnip::set_engine("keras")

  fit_obj <- parsnip::fit(spec, mpg ~ ., data = mtcars)

  # Check that the dense layer is NOT in the model
  keras_model <- fit_obj |> extract_keras_model()
  expect_equal(length(keras_model$layers), 1) # Output layers only

  # Check layer names explicitly
  layer_names <- sapply(keras_model$layers, function(l) l$name)
  expect_false("i_should_not_exist" %in% layer_names)
})

test_that("E2E: Error handling for reserved names works", {
  model_name <- "bad_spec"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  bad_blocks <- list(
    compile = function(model) model, # "compile" is a reserved name
    dense = function(model, u = 1) model |> keras3::layer_dense(units = u)
  )

  expect_error(
    create_keras_sequential_spec(model_name, bad_blocks),
    regexp = "`compile`, `fit` and `optimizer` are protected names"
  )
})

test_that("E2E: extract_keras_model works", {
  skip_if_no_keras()

  # Reuse model setup from previous tests
  input_block_feat <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block_feat <- function(model, units = 16) {
    model |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block_feat <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }

  model_name <- "e2e_mlp_summary_test"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_feat,
      dense = dense_block_feat,
      output = output_block_feat
    ),
    mode = "regression"
  )

  spec <- e2e_mlp_summary_test(fit_epochs = 1) |>
    parsnip::set_engine("keras")

  fit_obj <- parsnip::fit(spec, mpg ~ ., data = mtcars)

  summary_output <- extract_keras_model(fit_obj)

  expect_type(summary_output, "closure")
  expect_true(any(grepl("Layer ", summary_output)))
  expect_true(any(grepl("Output Shape", summary_output)))
  expect_true(any(grepl("Param #", summary_output)))
})

test_that("E2E: extract_keras_history works", {
  skip_if_no_keras()

  # Reuse model setup from previous tests
  input_block_feat <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block_feat <- function(model, units = 16) {
    model |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block_feat <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }

  model_name <- "e2e_mlp_history_test"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_feat,
      dense = dense_block_feat,
      output = output_block_feat
    ),
    mode = "regression"
  )

  epochs_to_train <- 2
  spec <- e2e_mlp_history_test(fit_epochs = epochs_to_train) |>
    parsnip::set_engine("keras")

  fit_obj <- parsnip::fit(spec, mpg ~ ., data = mtcars)

  history_output <- extract_keras_history(fit_obj)

  expect_s3_class(history_output, "keras_training_history")
})

test_that("E2E: keras_evaluate works", {
  skip_if_no_keras()

  # Reuse model setup from previous tests
  input_block_eval <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block_eval <- function(model, units = 16) {
    model |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block_eval <- function(model) {
    model |> keras3::layer_dense(units = 1)
  }

  model_name <- "e2e_mlp_evaluate_test"
  on.exit(suppressMessages(remove_keras_spec(model_name)), add = TRUE)

  create_keras_sequential_spec(
    model_name = model_name,
    layer_blocks = list(
      input = input_block_eval,
      dense = dense_block_eval,
      output = output_block_eval
    ),
    mode = "regression"
  )

  spec <- e2e_mlp_evaluate_test(fit_epochs = 1) |>
    parsnip::set_engine("keras")

  fit_obj <- parsnip::fit(spec, mpg ~ ., data = mtcars)

  # Evaluate the model
  eval_output <- keras_evaluate(fit_obj, x = mtcars[, -1], y = mtcars$mpg)

  expect_true(class(eval_output) == "list")
  expect_true("loss" %in% names(eval_output))
  expect_true("mean_absolute_error" %in% names(eval_output))
})

#' Generic Fitting Function for Functional Keras Models
#'
#' @title Internal Fitting Engine for Functional API Models
#' @description
#' This function serves as the internal engine for fitting `kerasnip` models
#' that are based on the Keras functional API. It is not intended to be called
#' directly by the user. The function is invoked by `parsnip::fit()` when a
#' `kerasnip` functional model specification is used.
#'
#' @details
#' The function orchestrates the three main steps of the model fitting process:
#' \enumerate{
#'   \item \strong{Build and Compile:} It calls
#'     `build_compile_func_model()` to construct the Keras model
#'     architecture based on the provided `layer_blocks` and hyperparameters.
#'   \item \strong{Process Data:} It preprocesses the input (`x`) and output
#'     (`y`) data into the format expected by Keras.
#'   \item \strong{Fit Model:} It calls `keras3::fit()` with the compiled model
#'     and processed data, passing along any fitting-specific arguments (e.g.,
#'     `epochs`, `batch_size`, `callbacks`).
#' }
#'
#' @param formula A formula specifying the predictor and outcome variables,
#'   passed down from the `parsnip::fit()` call.
#' @param data A data frame containing the training data, passed down from the
#'   `parsnip::fit()` call.
#' @param layer_blocks A named list of layer block functions. This is passed
#'   internally from the `parsnip` model specification.
#' @param ... Additional arguments passed down from the model specification.
#'   These can include:
#'   \itemize{
#'     \item \strong{Layer Parameters:} Arguments for the layer blocks, prefixed
#'       with the block name (e.g., `dense_units = 64`).
#'     \item \strong{Architecture Parameters:} Arguments to control the number
#'       of times a block is repeated, in the format `num_{block_name}` (e.g.,
#'       `num_dense = 2`).
#'     \item \strong{Compile Parameters:} Arguments to customize model
#'       compilation, prefixed with `compile_` (e.g., `compile_loss = "mae"`,
#'       `compile_optimizer = "sgd"`).
#'     \item \strong{Fit Parameters:} Arguments to customize model fitting,
#'       prefixed with `fit_` (e.g., `fit_callbacks = list(...)`,
#'       `fit_class_weight = list(...)`).
#'   }
#'
#' @return A list containing the fitted model and other metadata. This list is
#'   stored in the `fit` slot of the `parsnip` model fit object. The list
#'   contains the following elements:
#'   \itemize{
#'     \item `fit`: The raw, fitted Keras model object.
#'     \item `history`: The Keras training history object.
#'     \item `lvl`: A character vector of the outcome factor levels (for
#'       classification) or `NULL` (for regression).
#'   }
#'
#' @examples
#' # This function is not called directly by users.
#' # It is called internally by `parsnip::fit()`.
#' # For example:
#' \donttest{
#' library(parsnip)
#' # create_keras_functional_spec(...) defines my_functional_model
#'
#' # spec <- my_functional_model(hidden_units = 128, fit_epochs = 10) |>
#' #   set_engine("keras")
#'
#' # # This call to fit() would invoke generic_functional_fit() internally
#' # fitted_model <- fit(spec, y ~ x, data = training_data)
#' }
#' @keywords internal
#' @export
generic_functional_fit <- function(
  formula,
  data,
  layer_blocks,
  ...
) {
  # Separate predictors and outcomes from the processed data frame provided by
  # parsnip
  y_names <- all.vars(formula[[2]])
  x_names <- all.vars(formula[[3]])

  # Handle the `.` case for predictors
  if ("." %in% x_names) {
    x <- data[, !(names(data) %in% y_names), drop = FALSE]
  } else {
    x <- data[, x_names, drop = FALSE]
  }
  y <- data[, y_names, drop = FALSE]
  # --- 1. Build and Compile Model ---
  model <- build_compile_func_model(x, y, layer_blocks, ...)

  # --- 2. Model Fitting ---
  all_args <- list(...)
  verbose <- all_args$verbose %||% 0
  x_processed <- process_x_functional(x)
  x_proc <- x_processed$x_proc
  y_processed <- process_y_functional(y)
  y_mat <- y_processed$y_proc

  fit_args <- collect_fit_args(
    x_proc,
    y_mat,
    verbose,
    all_args
  )

  # Fit the model using the constructed arguments
  history <- rlang::exec(keras3::fit, model, !!!fit_args)

  # --- 3. Return value ---
  list(
    fit = model, # The raw Keras model object
    keras_bytes = keras_model_to_bytes(model), # Bytes for RDS-safe restore
    history = history, # The training history
    # Factor levels for classification, NULL for regression
    lvl = y_processed$class_levels,
    process_x = process_x_functional,
    process_y = process_y_functional
  )
}

# Helper to skip tests if Keras is not configured
library(parsnip)
library(recipes)
library(workflows)
library(modeldata)
library(rsample)
library(dials)
library(tune)
library(purrr)
library(dplyr)

skip_if_no_keras <- function() {
  testthat::skip_if_not_installed("keras3")

  # is_keras_available() checks for the python 'keras' module and a backend.
  # This is the most reliable way to check for a working installation.
  # testthat::skip_if_not(
  #   keras3::is_keras_available(),
  #   "Keras 3 and a backend (e.g., tensorflow) are not available for testing"
  # )
}

utils::globalVariables(
  c(
    "compiled_model",
    "engine",
    "error",
    "fresh",
    "keras_compile_arg_names",
    "keras_fit_arg_names",
    "new_data",
    "object",
    "parameters"
  )
)

#' set_args Method for kerasnip Spec Objects
#'
#' @description
#' S3 method for `set_args()` dispatched on `kerasnip_spec` objects.
#' `parsnip::set_args.model_spec()` calls `new_model_spec()`, which strips any
#' extra classes and attributes. This wrapper saves and re-attaches the
#' `kerasnip_layer_blocks` and `kerasnip_functional` metadata attributes (and
#' the `kerasnip_spec` class) after `NextMethod()` has done its work.
#'
#' @param object A `kerasnip_spec` model specification.
#' @param ... Named model arguments to update, passed to
#'   `parsnip::set_args.model_spec()`.
#' @return A `model_spec` object with the `kerasnip_spec` class and metadata
#'   attributes re-attached.
#' @keywords internal
#' @importFrom parsnip set_args
#' @exportS3Method parsnip::set_args
set_args.kerasnip_spec <- function(object, ...) {
  layer_blocks <- attr(object, "kerasnip_layer_blocks")
  functional <- attr(object, "kerasnip_functional")
  result <- NextMethod()
  class(result) <- c(class(result)[1L], "kerasnip_spec", class(result)[-1L])
  attr(result, "kerasnip_layer_blocks") <- layer_blocks
  attr(result, "kerasnip_functional") <- functional
  result
}

#' set_engine Method for kerasnip Spec Objects
#'
#' @description
#' S3 method for `set_engine()` dispatched on `kerasnip_spec` objects.
#' `parsnip::set_engine.model_spec()` internally calls `new_model_spec()`,
#' which re-creates the spec from scratch with only `c(cls, "model_spec")` as
#' the class vector — stripping `kerasnip_spec` and any custom attributes. This
#' wrapper preserves the `kerasnip_layer_blocks` and `kerasnip_functional`
#' metadata attributes and re-attaches them (along with the `kerasnip_spec`
#' class) after `NextMethod()` has done its work.
#'
#' @param object A `kerasnip_spec` model specification.
#' @param engine A character string naming the engine (e.g., `"keras"`).
#' @param ... Additional engine-specific arguments passed to
#'   `parsnip::set_engine.model_spec()`.
#' @return A `model_spec` object with the `kerasnip_spec` class and metadata
#'   attributes re-attached.
#' @keywords internal
#' @importFrom parsnip set_engine
#' @exportS3Method parsnip::set_engine
set_engine.kerasnip_spec <- function(object, engine, ...) {
  layer_blocks <- attr(object, "kerasnip_layer_blocks")
  functional <- attr(object, "kerasnip_functional")
  result <- NextMethod()
  class(result) <- c(class(result)[1L], "kerasnip_spec", class(result)[-1L])
  attr(result, "kerasnip_layer_blocks") <- layer_blocks
  attr(result, "kerasnip_functional") <- functional
  result
}

#' Fit Method for kerasnip Spec Objects
#'
#' @description
#' S3 method for `fit()` dispatched on `kerasnip_spec` objects. Delegates to
#' the standard parsnip `fit.model_spec()` and then tags the result with the
#' `kerasnip_model_fit` class so that `predict.kerasnip_model_fit()` is
#' dispatched on subsequent calls.
#'
#' `kerasnip_spec` is stripped from the class before `NextMethod()` to prevent
#' parsnip's internal `specific_model()` helper from returning more than one
#' model-class entry, which would break registry lookups. The custom metadata
#' attributes remain on the object and are thus stored inside the resulting
#' `model_fit$spec`.
#'
#' @param object A `kerasnip_spec` model specification.
#' @param ... Passed to `parsnip::fit.model_spec()`.
#' @return A `model_fit` object with the additional `kerasnip_model_fit` class
#'   prepended to its class vector.
#' @keywords internal
#' @importFrom generics fit
#' @exportS3Method generics::fit
fit.kerasnip_spec <- function(object, ...) {
  class(object) <- class(object)[class(object) != "kerasnip_spec"]
  result <- NextMethod()
  class(result) <- c("kerasnip_model_fit", class(result))
  result
}

#' fit_xy Method for kerasnip Spec Objects
#'
#' @description
#' S3 method for `fit_xy()` dispatched on `kerasnip_spec` objects. Workflows
#' route through `fit_xy` rather than `fit`, so this method ensures the
#' `kerasnip_model_fit` class is attached in the workflow fitting path as well.
#'
#' `kerasnip_spec` is stripped from the class before `NextMethod()` to prevent
#' parsnip's internal `specific_model()` helper from returning more than one
#' model-class entry, which would break registry lookups. The custom metadata
#' attributes remain on the object and are thus stored inside the resulting
#' `model_fit$spec`.
#'
#' @param object A `kerasnip_spec` model specification.
#' @param ... Passed to `parsnip::fit_xy.model_spec()`.
#' @return A `model_fit` object with the additional `kerasnip_model_fit` class
#'   prepended to its class vector.
#' @keywords internal
#' @importFrom generics fit_xy
#' @exportS3Method generics::fit_xy
fit_xy.kerasnip_spec <- function(object, ...) {
  class(object) <- class(object)[class(object) != "kerasnip_spec"]
  result <- NextMethod()
  class(result) <- c("kerasnip_model_fit", class(result))
  result
}

#' Predict Method for kerasnip Model Fits
#'
#' @description
#' S3 method for `predict()` dispatched on `kerasnip_model_fit` objects.
#' Before delegating to the standard parsnip predict machinery, it checks
#' whether the underlying model type is registered in the current parsnip
#' session. If not (e.g. after loading a saved workflow in a new R session),
#' it transparently replays the full parsnip registration using metadata stored
#' on the spec object — requiring no manual step from the user.
#'
#' @details
#' The metadata needed for re-registration (`kerasnip_layer_blocks`,
#' `kerasnip_functional`) is embedded on the spec object by the spec
#' constructor function at call time. This means it is preserved across
#' `saveRDS()`/`readRDS()` and `bundle()`/`unbundle()` round-trips.
#'
#' For full model weight portability (i.e. to be able to `predict()` on new
#' data in a new R session), use `bundle::bundle()` before saving. Plain
#' `saveRDS()` preserves the spec structure and will auto-register, but the
#' underlying Keras model weights are not portable without bundling.
#'
#' @param object A `kerasnip_model_fit` object.
#' @param new_data A data frame of predictors.
#' @param ... Passed to the parsnip predict method.
#' @return A tibble of predictions.
#' @keywords internal
#' @importFrom stats predict
#' @exportS3Method stats::predict
predict.kerasnip_model_fit <- function(object, new_data, ...) {
  model_name <- class(object$spec)[1L]

  if (!model_exists(model_name)) {
    spec <- object$spec
    layer_blocks <- attr(spec, "kerasnip_layer_blocks")
    functional <- attr(spec, "kerasnip_functional") %||% FALSE
    mode <- spec$mode
    args_info <- collect_spec_args(layer_blocks, functional)

    register_core_model(model_name, mode)
    register_model_args(model_name, args_info$parsnip_names)
    register_fit_predict(model_name, mode, layer_blocks, functional)
    register_update_method(
      model_name,
      args_info$parsnip_names,
      env = globalenv()
    )
  }

  # Restore Keras model from serialized bytes if the Python object has been
  # invalidated (e.g. by saveRDS()/readRDS() within the same session, or after
  # bundle()/unbundle()). reticulate::py_validate_xptr() performs a direct
  # C-level check and throws the same "Unable to access object" error that
  # would otherwise surface later during predict(); catching it here lets us
  # restore the model transparently before dispatching to parsnip.
  if (!is.null(object$fit$keras_bytes)) {
    is_valid <- tryCatch(
      {
        reticulate::py_validate_xptr(object$fit$fit)
        TRUE
      },
      error = function(e) FALSE
    )
    if (!is_valid) {
      object$fit$fit <- keras_model_from_bytes(object$fit$keras_bytes)
    }
  }

  # Strip our class and dispatch explicitly. NextMethod() re-uses the original
  # call arguments, not locally modified ones, so the restored model above
  # would not be forwarded via NextMethod().
  class(object) <- class(object)[class(object) != "kerasnip_model_fit"]
  predict(object, new_data = new_data, ...)
}

#' Build and Compile a Keras Sequential Model
#'
#' @description
#' This internal helper function constructs and compiles a Keras sequential
#' model based on a list of layer blocks and other parameters. It handles data
#' processing, dynamic architecture construction, and model compilation.
#'
#' @param x A data frame or matrix of predictors.
#' @param y A vector or data frame of outcomes.
#' @param layer_blocks A named list of functions that define the layers of the
#'   model. The order of the list determines the order of the layers.
#' @param ... Additional arguments passed to the function, including layer
#'   hyperparameters, repetition counts for blocks, and compile/fit arguments.
#'
#' @return A compiled Keras model object.
#' @noRd
build_compile_seq_model <- function(
  x,
  y,
  layer_blocks,
  ...
) {
  # --- 0. Argument & Data Preparation ---
  all_args <- list(...)
  learn_rate <- all_args$learn_rate %||% 0.01

  # Process x input
  x_processed <- process_x_sequential(x)
  input_shape <- x_processed$input_shape

  # Process y input
  y_processed <- process_y_sequential(y)

  # Determine is_classification and num_classes
  is_classification <- y_processed$is_classification
  num_classes <- y_processed$num_classes

  # Determine default compile arguments based on mode
  default_loss <- if (is_classification) {
    if (num_classes > 2) {
      "categorical_crossentropy"
    } else {
      "binary_crossentropy"
    }
  } else {
    "mean_squared_error"
  }
  default_metrics <- if (is_classification) {
    "accuracy"
  } else {
    "mean_absolute_error"
  }

  # --- 2. Dynamic Model Architecture Construction ---
  # The model is initialized as NULL. The first layer_block is expected to
  # create the model (e.g., by defining an input layer). Subsequent blocks
  # will receive and modify the model object. The order is critical.
  model <- NULL

  for (block_name in names(layer_blocks)) {
    block_fn <- layer_blocks[[block_name]]
    block_fmls <- rlang::fn_fmls(block_fn)

    num_repeats_arg <- paste0("num_", block_name)
    num_repeats_val <- all_args[[num_repeats_arg]]

    # If num_repeats_val is NULL or zapped, default to 1.
    # Otherwise, use the value provided by the user.
    if (is.null(num_repeats_val) || inherits(num_repeats_val, "rlang_zap")) {
      num_repeats <- 1
    } else {
      num_repeats <- as.integer(num_repeats_val)
    }

    # Get the arguments for this specific block from `...`
    block_arg_names <- names(block_fmls)[-1] # Exclude 'model'
    user_args <- list()
    for (arg_name in block_arg_names) {
      full_arg_name <- paste(block_name, arg_name, sep = "_")
      arg_val <- all_args[[full_arg_name]]
      # Only use the argument if it was actually provided by the user
      if (!is.null(arg_val) && !inherits(arg_val, "rlang_zap")) {
        user_args[[arg_name]] <- arg_val
      }
    }

    # Combine user-provided args with the block's defaults
    block_args <- utils::modifyList(as.list(block_fmls[-1]), user_args)

    # If the block function can accept these, provide them. This is useful for
    # the user-defined input and output layers.
    if ("input_shape" %in% names(block_fmls)) {
      block_args$input_shape <- input_shape
    }
    if (is_classification && "num_classes" %in% names(block_fmls)) {
      block_args$num_classes <- num_classes
    }

    # Add the block(s) to the model
    for (i in seq_len(num_repeats)) {
      # The first argument to the block function is the model itself
      # On the first iteration, `model` will be NULL.
      call_args <- c(list(model), block_args)
      model <- rlang::exec(block_fn, !!!call_args)
    }
  }

  # --- 3. Model Compilation ---
  # Collect all arguments starting with "compile_" from `...`
  compile_args <- collect_compile_args(
    all_args,
    learn_rate,
    default_loss,
    default_metrics
  )
  rlang::exec(keras3::compile, model, !!!compile_args)

  model
}

#' Build and Compile a Keras Functional Model
#'
#' @description
#' This internal helper function constructs and compiles a Keras functional
#' model based on a list of layer blocks and other parameters. It handles data
#' processing, dynamic architecture construction (including multiple inputs and
#' branches), and model compilation.
#'
#' @param x A data frame or matrix of predictors. For multiple inputs, this is
#'   often a data frame with list-columns.
#' @param y A vector or data frame of outcomes. Can handle multiple outputs if
#'   provided as a data frame with multiple columns.
#' @param layer_blocks A named list of functions that define the building blocks
#'   of the model graph. Connections are defined by referencing other block
#'   names.
#' @param ... Additional arguments passed to the function, including layer
#'   hyperparameters, repetition counts for blocks, and compile/fit arguments.
#'
#' @return A compiled Keras model object.
#' @noRd
build_compile_func_model <- function(
  x,
  y,
  layer_blocks,
  ...
) {
  # --- 0. Argument & Data Preparation ---
  all_args <- list(...)
  learn_rate <- all_args$learn_rate %||% 0.01

  # Process x input
  x_processed <- process_x_functional(x)
  input_shape <- x_processed$input_shape

  # Process y input
  y_processed <- process_y_functional(y)

  # Determine default compile arguments based on mode
  default_losses <- list()
  default_metrics_list <- list()

  # Check if y_processed$y_proc is a list (indicating multiple outputs)
  if (is.list(y_processed$y_proc) && !is.null(names(y_processed$y_proc))) {
    # Multiple outputs
    for (output_name in names(y_processed$y_proc)) {
      # We need to determine is_classification and num_classes for each output
      # based on the class_levels for that output.
      current_class_levels <- y_processed$class_levels[[output_name]]
      current_is_classification <- !is.null(current_class_levels) &&
        length(current_class_levels) > 0
      current_num_classes <- if (current_is_classification) {
        length(current_class_levels)
      } else {
        NULL
      }

      default_losses[[output_name]] <- if (current_is_classification) {
        if (current_num_classes > 2) {
          "categorical_crossentropy"
        } else {
          "binary_crossentropy"
        }
      } else {
        "mean_squared_error"
      }
      default_metrics_list[[output_name]] <- if (current_is_classification) {
        "accuracy"
      } else {
        "mean_absolute_error"
      }
    }
  } else {
    # Single output case
    # Determine is_classification and num_classes from the top-level class
    # levels
    is_classification <- !is.null(y_processed$class_levels) &&
      length(y_processed$class_levels) > 0
    num_classes <- if (is_classification) {
      length(y_processed$class_levels)
    } else {
      NULL
    }

    default_losses <- if (is_classification) {
      if (num_classes > 2) {
        "categorical_crossentropy"
      } else {
        "binary_crossentropy"
      }
    } else {
      "mean_squared_error"
    }
    default_metrics_list <- if (is_classification) {
      "accuracy"
    } else {
      "mean_absolute_error"
    }
  }

  # --- 2. Dynamic Model Architecture Construction (DIFFERENT from sequential)
  # Create a list to store the output tensors of each block.  The names of the
  # list elements correspond to the block names.
  block_outputs <- list()
  model_input_tensors <- list() # To collect all input tensors for keras_model

  # Identify and process input layers based on names matching input_shape
  # This assumes that if input_shape is a named list, the corresponding
  # input blocks in layer_blocks will have matching names.
  if (is.list(input_shape) && !is.null(names(input_shape))) {
    input_block_names_in_spec <- intersect(
      names(layer_blocks),
      names(input_shape)
    )

    if (length(input_block_names_in_spec) != length(input_shape)) {
      stop(
        "Mismatch between named inputs from process_x and ",
        "named input blocks in layer_blocks. ",
        "Ensure all processed inputs have a corresponding ",
        "named input block in your model specification."
      )
    }

    for (block_name in input_block_names_in_spec) {
      block_fn <- layer_blocks[[block_name]]
      current_input_tensor <- block_fn(input_shape = input_shape[[block_name]])
      block_outputs[[block_name]] <- current_input_tensor
      model_input_tensors[[block_name]] <- current_input_tensor
    }
    remaining_layer_blocks_names <- names(layer_blocks)[
      !(names(layer_blocks) %in% input_block_names_in_spec)
    ]
  } else {
    # Single input case (now also collecting for model_input_tensors)
    first_block_name <- names(layer_blocks)[1]
    first_block_fn <- layer_blocks[[first_block_name]]
    current_input_tensor <- first_block_fn(input_shape = input_shape)
    block_outputs[[first_block_name]] <- current_input_tensor
    model_input_tensors[[first_block_name]] <- current_input_tensor
    remaining_layer_blocks_names <- names(layer_blocks)[-1]
  }

  # Iterate through the remaining blocks, connecting and repeating them
  # as needed.
  for (block_name in remaining_layer_blocks_names) {
    block_fn <- layer_blocks[[block_name]]
    block_fmls <- rlang::fn_fmls(block_fn)
    block_fml_names <- names(block_fmls)

    # --- Get Repetition Count ---
    num_repeats_arg <- paste0("num_", block_name)
    num_repeats_val <- all_args[[num_repeats_arg]]

    # If num_repeats_val is NULL or zapped, default to 1.
    # Otherwise, use the value provided by the user.
    if (is.null(num_repeats_val) || inherits(num_repeats_val, "rlang_zap")) {
      num_repeats <- 1
    } else {
      num_repeats <- as.integer(num_repeats_val)
    }

    # --- Get Hyperparameters for this block ---
    # Hyperparameters are formals that are NOT other block names
    # (graph connections)
    hyperparam_names <- setdiff(block_fml_names, names(layer_blocks))
    user_hyperparams <- list()
    for (hp_name in hyperparam_names) {
      full_arg_name <- paste(block_name, hp_name, sep = "_")
      arg_val <- all_args[[full_arg_name]]
      if (!is.null(arg_val) && !inherits(arg_val, "rlang_zap")) {
        user_hyperparams[[hp_name]] <- arg_val
      }
    }
    # Combine user args with the block's defaults for those hyperparameters
    block_hyperparams <- utils::modifyList(
      as.list(block_fmls[hyperparam_names]),
      user_hyperparams
    )

    # Add special engine-supplied arguments if the block can accept them
    # Add special engine-supplied arguments if the block can accept them
    # This is primarily for output layers that might need num_classes
    if ("num_classes" %in% block_fml_names) {
      # Check if this block is an output block and if it's a classification task
      if (is.list(y_processed$y_proc) && !is.null(names(y_processed$y_proc))) {
        # Multi-output case
        # Find the corresponding output in y_processed based on block_name
        y_names <- names(y_processed$y_proc)
        # If there is only one output, and this block is named 'output',
        # connect them automatically.
        if (length(y_names) == 1 && block_name == "output") {
          y_name <- y_names[1]
          is_cls <- !is.null(y_processed$class_levels[[y_name]]) &&
            length(y_processed$class_levels[[y_name]]) > 0
          if (is_cls) {
            block_hyperparams$num_classes <- length(y_processed$class_levels[[
              y_name
            ]])
          }
        } else if (block_name %in% y_names) {
          # Standard case: block name matches an output name
          current_y_info <- list(
            is_classification = !is.null(y_processed$class_levels[[
              block_name
            ]]) &&
              length(y_processed$class_levels[[block_name]]) > 0,
            num_classes = if (
              !is.null(y_processed$class_levels[[block_name]])
            ) {
              length(y_processed$class_levels[[block_name]])
            } else {
              NULL
            }
          )
          if (current_y_info$is_classification) {
            block_hyperparams$num_classes <- current_y_info$num_classes
          }
        }
      } else {
        # Single output case
        if (is_classification) {
          block_hyperparams$num_classes <- num_classes
        }
      }
    }

    # --- Get Input Tensors for this block ---
    input_tensor_names <- intersect(block_fml_names, names(block_outputs))
    if (length(input_tensor_names) == 0 && block_name != "output") {
      warning("Block '", block_name, "' has no inputs from other blocks.")
    }

    # --- Repetition Loop ---
    if (num_repeats > 1 && length(input_tensor_names) != 1) {
      stop(
        "Block '",
        block_name,
        "' cannot be repeated because it has ",
        length(input_tensor_names),
        " inputs (",
        paste(input_tensor_names, collapse = ", "),
        "). Only blocks with exactly one input tensor can be repeated."
      )
    }

    # The initial input(s) for the first iteration
    input_args <- purrr::map(input_tensor_names, ~ block_outputs[[.x]])
    names(input_args) <- input_tensor_names

    # The tensor that will be updated and passed through the loop
    current_tensor <- input_args[[1]]

    for (i in seq_len(num_repeats)) {
      # For repetitions after the first, update the input tensor
      if (i > 1) {
        input_args[[input_tensor_names[1]]] <- current_tensor
      }
      call_args <- c(input_args, block_hyperparams)
      current_tensor <- rlang::exec(block_fn, !!!call_args)
    }

    # Store the final output of the (possibly repeated) block
    block_outputs[[block_name]] <- current_tensor
  }

  # The last layer must be named 'output' or match the names of
  # y_processed outputs
  final_output_tensors <- list()

  # Check if y_processed$y_proc is a named list, indicating multiple outputs)
  if (is.list(y_processed$y_proc) && !is.null(names(y_processed$y_proc))) {
    # Multiple outputs
    for (output_name in names(y_processed$y_proc)) {
      # Iterate over the names of the actual outputs
      if (is.null(block_outputs[[output_name]])) {
        stop(paste0(
          "An output block named '",
          output_name,
          "' must be defined in layer_blocks for multi-output models."
        ))
      }
      final_output_tensors[[output_name]] <- block_outputs[[output_name]]
    }
  } else {
    # Single output case
    output_tensor <- block_outputs[["output"]]
    if (is.null(output_tensor)) {
      stop("An 'output' block must be defined in layer_blocks.")
    }
    final_output_tensors <- output_tensor
  }

  # If there's only one input, it shouldn't be a list for keras_model
  final_model_inputs <- if (length(model_input_tensors) == 1) {
    model_input_tensors[[1]]
  } else {
    model_input_tensors
  }

  model <- keras3::keras_model(
    inputs = final_model_inputs,
    outputs = final_output_tensors # This will now be a list if multiple outputs
  )

  # --- 3. Model Compilation ---
  # Collect all arguments starting with "compile_" from `...`
  compile_args <- collect_compile_args(
    all_args,
    learn_rate,
    default_losses,
    default_metrics_list
  )
  rlang::exec(keras3::compile, model, !!!compile_args)

  model
}

#' Butcher axe methods for kerasnip_model_fit
#'
#' @description
#' These methods allow `butcher::butcher()` to reduce the memory footprint of
#' fitted kerasnip model objects. The Keras model itself (stored as raw bytes
#' in `$fit$keras_bytes`) is always preserved so that `predict()` continues
#' to work after butchering.
#'
#' The main saving comes from `axe_data()`, which removes the training history
#' object (`$fit$history`). For long training runs this can be several MB.
#'
#' @param x A `kerasnip_model_fit` object.
#' @param verbose Logical. Print information about memory released and
#'   disabled functions. Default is `FALSE`.
#' @param ... Not used.
#' @return An axed `kerasnip_model_fit` object with the `butcher_kerasnip_model_fit`
#'   class prepended.
#' @name axe-kerasnip_model_fit
NULL

#' @rdname axe-kerasnip_model_fit
#' @exportS3Method butcher::axe_data
axe_data.kerasnip_model_fit <- function(x, verbose = FALSE, ...) {
  old <- x
  x$fit$history <- NULL
  add_butcher_attributes(
    x,
    old,
    verbose = verbose,
    disabled = c("extract_keras_history")
  )
}

#' @rdname axe-kerasnip_model_fit
#' @exportS3Method butcher::axe_env
axe_env.kerasnip_model_fit <- function(x, verbose = FALSE, ...) {
  # Intentional no-op: Keras R6 objects rely on Python environments.
  # Stripping R environments from them is unsafe and would break predict().
  add_butcher_attributes(x, x, verbose = verbose)
}

#' @rdname axe-kerasnip_model_fit
#' @exportS3Method butcher::axe_call
axe_call.kerasnip_model_fit <- function(x, verbose = FALSE, ...) {
  # No-op: kerasnip fit objects do not store a call component.
  add_butcher_attributes(x, x, verbose = verbose)
}

#' @rdname axe-kerasnip_model_fit
#' @exportS3Method butcher::axe_ctrl
axe_ctrl.kerasnip_model_fit <- function(x, verbose = FALSE, ...) {
  # No-op: kerasnip fit objects do not store training controls.
  add_butcher_attributes(x, x, verbose = verbose)
}

#' @rdname axe-kerasnip_model_fit
#' @exportS3Method butcher::axe_fitted
axe_fitted.kerasnip_model_fit <- function(x, verbose = FALSE, ...) {
  # No-op: kerasnip does not store fitted values separately from the model.
  add_butcher_attributes(x, x, verbose = verbose)
}

# Copied from tidymodels/butcher R/ui.R (internal, not exported by that package).
# Using ::: on unexported functions is disallowed by CRAN policy, so these
# helpers are reproduced here under butcher's MIT licence.
#' @importFrom lobstr obj_size
#' @importFrom cli cli_alert_info cli_alert_success cli_alert_danger

get_object_size <- function(x, attempts = 5) {
  for (i in seq_len(attempts)) {
    res <- try(lobstr::obj_size(x), silent = TRUE)
    if (!inherits(res, "try-error")) {
      break()
    }
  }
  if (inherits(res, "try-error")) {
    cli::cli_inform(
      "{.fn lobstr::obj_size} failed after {attempts} attempts. Falling back on {.fn object.size}."
    )
    res <- utils::object.size(x)
  }

  res
}

#' Console Messages
#'
#' These console messages are created such that the user is
#' aware of the effects of removing specific components from
#' the model object.
#'
#' @param og Original model object.
#' @param butchered Butchered model object.
#'
#' @keywords internal
#' @name ui

#' @noRd
memory_released <- function(og, butchered) {
  old <- get_object_size(og)
  new <- get_object_size(butchered)
  rel <- old - new
  if (length(rel) == 1) {
    if (isTRUE(all.equal(old, new))) {
      return(NULL)
    }
    return(rel)
  } else {
    return(NULL)
  }
}

#' @noRd
assess_object <- function(og, butchered) {
  mem <- memory_released(og, butchered)
  disabled <- attr(butchered, "butcher_disabled")
  class_added <- grep("butcher", class(butchered)[1])
  if (is.null(mem)) {
    cli::cli_alert_danger("No memory released. Do not butcher.")
  } else {
    abs_mem <- format(abs(mem), big.mark = ",", scientific = FALSE)
    if (mem < 0) {
      cli::cli_alert_danger(
        "The butchered object is {.field {abs_mem}} larger than the original. Do not butcher."
      )
    } else {
      cli::cli_alert_success("Memory released: {.field {abs_mem}}")
      if (!is.null(disabled)) {
        cli::cli_alert_danger("Disabled: {.code {disabled}}")
      }
      if (length(class_added) == 0) {
        class_name <- "butchered"
        cli::cli_alert_danger("Could not add {.cls {class_name}} class")
      }
    }
  }
}

# butcher attributes helper
add_butcher_disabled <- function(x, disabled = NULL) {
  current <- attr(x, "butcher_disabled")
  if (!is.null(disabled)) {
    disabled <- union(current, disabled)
    attr(x, "butcher_disabled") <- disabled
  }
  x
}

# class assignment helper
add_butcher_class <- function(x) {
  if (!any(grepl("butcher", class(x)))) {
    class(x) <- append(paste0("butchered_", class(x)[1]), class(x))
  }
  x
}

# butcher attributes wrapper
add_butcher_attributes <- function(
  x,
  old,
  disabled = NULL,
  add_class = TRUE,
  verbose = FALSE
) {
  if (!identical(x, old)) {
    x <- add_butcher_disabled(x, disabled)
    if (add_class) {
      x <- add_butcher_class(x)
    }
  }
  if (verbose & !missing(old)) {
    assess_object(old, x)
  }
  x
}

test_that("axe_data strips history and predict still works", {
  skip_if_not_installed("butcher")
  skip_if_not(
    reticulate::py_module_available("keras"),
    "keras not available"
  )

  input_block <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  hidden_block <- function(model, units = 8) {
    model |> keras3::layer_dense(units = units, activation = "relu")
  }
  output_block <- function(model, num_classes) {
    model |> keras3::layer_dense(units = 1)
  }

  create_keras_sequential_spec(
    model_name = "butcher_test_model",
    layer_blocks = list(
      input = input_block,
      hidden = hidden_block,
      output = output_block
    ),
    mode = "regression"
  )
  on.exit(remove_keras_spec("butcher_test_model"), add = TRUE)

  x_train <- matrix(rnorm(50 * 3), ncol = 3)
  y_train <- rnorm(50)
  train_df <- data.frame(x = I(x_train), y = y_train)

  spec <- butcher_test_model(fit_epochs = 2L) |>
    parsnip::set_engine("keras")
  fitted <- parsnip::fit(spec, y ~ x, data = train_df)

  # History exists before axing
  expect_false(is.null(fitted$fit$history))

  axed <- butcher::axe_data(fitted)

  # History is gone
  expect_null(axed$fit$history)

  # predict still works
  new_df <- data.frame(x = I(matrix(rnorm(5 * 3), ncol = 3)))
  preds <- predict(axed, new_data = new_df)
  expect_s3_class(preds, "tbl_df")
  expect_equal(nrow(preds), 5L)
})

test_that("butcher helper utilities behave correctly", {
  skip_if_not_installed("butcher")

  old <- list(a = rnorm(1000), history = rep("x", 1000))
  butchered <- old
  butchered$history <- NULL

  size <- kerasnip:::get_object_size(old)
  expect_type(size, "double")
  expect_gt(size, 0)

  mem <- kerasnip:::memory_released(old, butchered)
  expect_type(mem, "double")
  expect_gt(mem, 0)

  expect_no_error(kerasnip:::assess_object(old, butchered))

  axed <- kerasnip:::add_butcher_attributes(
    butchered,
    old,
    disabled = "extract_keras_history",
    verbose = TRUE
  )
  expect_s3_class(axed, "butchered_list")
  expect_equal(attr(axed, "butcher_disabled"), "extract_keras_history")
})

test_that("all axe_* methods are callable without error", {
  skip_if_not_installed("butcher")
  skip_if_not(
    reticulate::py_module_available("keras"),
    "keras not available"
  )

  input_block <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  output_block <- function(model, num_classes) {
    model |> keras3::layer_dense(units = 1)
  }

  create_keras_sequential_spec(
    model_name = "butcher_noop_model",
    layer_blocks = list(
      input = input_block,
      output = output_block
    ),
    mode = "regression"
  )
  on.exit(remove_keras_spec("butcher_noop_model"), add = TRUE)

  x_train <- matrix(rnorm(50 * 3), ncol = 3)
  y_train <- rnorm(50)
  train_df <- data.frame(x = I(x_train), y = y_train)

  spec <- butcher_noop_model(fit_epochs = 2L) |>
    parsnip::set_engine("keras")
  fitted <- parsnip::fit(spec, y ~ x, data = train_df)

  expect_no_error(butcher::axe_env(fitted))
  expect_no_error(butcher::axe_call(fitted))
  expect_no_error(butcher::axe_ctrl(fitted))
  expect_no_error(butcher::axe_fitted(fitted))
})

test_that("axe_env no-op preserves predict functionality", {
  skip_if_not_installed("butcher")
  skip_if_not(
    reticulate::py_module_available("keras"),
    "keras not available"
  )

  input_block <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  output_block <- function(model, num_classes) {
    model |> keras3::layer_dense(units = 1)
  }

  create_keras_sequential_spec(
    model_name = "butcher_env_model",
    layer_blocks = list(
      input = input_block,
      output = output_block
    ),
    mode = "regression"
  )
  on.exit(remove_keras_spec("butcher_env_model"), add = TRUE)

  x_train <- matrix(rnorm(50 * 3), ncol = 3)
  y_train <- rnorm(50)
  train_df <- data.frame(x = I(x_train), y = y_train)

  spec <- butcher_env_model(fit_epochs = 2L) |>
    parsnip::set_engine("keras")
  fitted <- parsnip::fit(spec, y ~ x, data = train_df)

  axed <- butcher::axe_env(fitted)
  new_df <- data.frame(x = I(matrix(rnorm(5 * 3), ncol = 3)))
  preds <- predict(axed, new_data = new_df)
  expect_s3_class(preds, "tbl_df")
  expect_equal(nrow(preds), 5L)
})

#' Compile and Validate Keras Model Architectures
#'
#' @title Compile Keras Models Over a Grid of Hyperparameters
#' @description
#' Pre-compiles Keras models for each hyperparameter combination in a grid.
#'
#' This function is a powerful debugging tool to use before running a full
#' `tune::tune_grid()`. It allows you to quickly validate multiple model
#' architectures, ensuring they can be successfully built and compiled without
#' the time-consuming process of actually fitting them. It helps catch common
#' errors like incompatible layer shapes or invalid argument values early.
#'
#' @details
#' The function iterates through each row of the provided `grid`. For each
#' hyperparameter combination, it attempts to build and compile the Keras model
#' defined by the `spec`. The process is wrapped in a `try-catch` block to
#' gracefully handle and report any errors that occur during model instantiation
#' or compilation.
#'
#' The output is a tibble that mirrors the input `grid`, with additional columns
#' containing the compiled model object or the error message, making it easy to
#' inspect which architectures are valid.
#'
#' @param spec A `parsnip` model specification created by
#'   `create_keras_sequential_spec()` or `create_keras_functional_spec()`.
#' @param grid A `tibble` or `data.frame` containing the grid of hyperparameters
#'   to evaluate. Each row represents a unique model architecture to be
#'   compiled. Must have at least one row. To build the model once using only
#'   the arguments already set on `spec` (e.g. for architecture inspection),
#'   pass `tibble::tibble(.rows = 1L)`.
#' @param x A data frame or matrix of predictors. This is used to infer the
#'   `input_shape` for the Keras model.
#' @param y A vector or factor of outcomes. This is used to infer the output
#'   shape and the default loss function for the Keras model.
#'
#' @return A `tibble` with the following columns:
#'   \itemize{
#'     \item Columns from the input `grid`.
#'     \item `compiled_model`: A list-column containing the compiled Keras model
#'       objects. If compilation failed, the element will be `NULL`.
#'     \item `error`: A list-column containing `NA` for successes or a
#'       character string with the error message for failures.
#'   }
#'
#' @examples
#' \donttest{
#' if (requireNamespace("keras3", quietly = TRUE)) {
#' library(keras3)
#' library(parsnip)
#' library(dials)
#'
#' # 1. Define layer blocks
#' input_block <- function(model, input_shape) {
#'   keras_model_sequential(input_shape = input_shape)
#' }
#' hidden_block <- function(model, units = 32) {
#'   model |> layer_dense(units = units, activation = "relu")
#' }
#' output_block <- function(model, num_classes) {
#'   model |> layer_dense(units = num_classes, activation = "softmax")
#' }
#'
#' # 2. Define a kerasnip model specification
#' create_keras_sequential_spec(
#'   model_name = "my_mlp_grid",
#'   layer_blocks = list(
#'     input = input_block,
#'     hidden = hidden_block,
#'     output = output_block
#'   ),
#'   mode = "classification"
#' )
#'
#' mlp_spec <- my_mlp_grid(
#'   hidden_units = tune(),
#'   compile_loss = "categorical_crossentropy",
#'   compile_optimizer = "adam"
#' )
#'
#' # 3. Create a hyperparameter grid
#' # Include an invalid value (-10) to demonstrate error handling
#' param_grid <- tibble::tibble(
#'   hidden_units = c(32, 64, -10)
#' )
#'
#' # 4. Prepare dummy data
#' x_train <- matrix(rnorm(100 * 10), ncol = 10)
#' y_train <- factor(sample(0:1, 100, replace = TRUE))
#'
#' # 5. Compile models over the grid
#' compiled_grid <- compile_keras_grid(
#'   spec = mlp_spec,
#'   grid = param_grid,
#'   x = x_train,
#'   y = y_train
#' )
#'
#' print(compiled_grid)
#' remove_keras_spec("my_mlp_grid")
#'
#' # 6. Inspect the results
#' # The row with `hidden_units = -10` will show an error.
#' }
#' }
#' @importFrom dplyr bind_rows filter select
#' @importFrom cli cli_h1 cli_alert_danger cli_h2 cli_text cli_bullets
#' @importFrom cli cli_code cli_alert_info cli_alert_success
#' @export
compile_keras_grid <- function(spec, grid, x, y) {
  # Input validation
  if (!inherits(spec, "model_spec")) {
    stop("`spec` must be a `parsnip` model specification.")
  }
  if (!is.data.frame(grid)) {
    stop("`grid` must be a data frame or tibble.")
  }
  if (nrow(grid) == 0L) {
    stop(
      "`grid` must have at least one row. Use `tibble::tibble(.rows = 1)` to build the model with the spec's current arguments and no hyperparameter variation."
    )
  }

  model_env <- get_model_env()
  model_name <- class(spec)[1]

  fit_info_name <- paste0(model_name, "_fit")
  model_info <- model_env[[fit_info_name]]

  if (is.null(model_info)) {
    stop("Could not find model information for this specification.")
  }

  fit_fun_char <- model_info |>
    purrr::pluck("value") |>
    purrr::pluck(1) |>
    purrr::pluck("func") |>
    purrr::pluck(2)

  build_fn <- if (any(grepl("sequential", fit_fun_char))) {
    build_compile_seq_model
  } else if (any(grepl("functional", fit_fun_char))) {
    build_compile_func_model
  } else {
    stop("Unsupported fit function in model spec.")
  }

  layer_blocks <- model_info |>
    purrr::pluck("value") |>
    purrr::pluck(1) |>
    purrr::pluck("defaults") |>
    purrr::pluck("layer_blocks")
  # Prepare to store results
  results <- purrr::map(seq_len(nrow(grid)), function(i) {
    params <- as.list(grid[i, ])
    active_args <- purrr::discard(spec$args, function(arg) {
      inherits(rlang::get_expr(arg), "rlang_zap")
    })

    evaluated_args <- purrr::map(active_args, rlang::eval_tidy)

    args <- list()
    args$x <- x
    args$y <- y
    args$layer_blocks <- layer_blocks

    for (name in names(evaluated_args)) {
      args[[name]] <- evaluated_args[[name]]
    }

    for (name in names(params)) {
      args[[name]] <- params[[name]]
    }

    # Use tryCatch to handle potential errors in model building/compilation
    result <- tryCatch(
      {
        model <- do.call(build_fn, args)
        # Capture the model summary
        list(
          compiled_model = list(model),
          error = NA_character_
        )
      },
      error = function(e) {
        list(
          compiled_model = list(NULL),
          error = as.character(e$message)
        )
      }
    )

    # Combine grid params with the result
    tibble::as_tibble(c(params, result))
  })

  # Combine all results into a single tibble
  dplyr::bind_rows(results)
}

#' Filter a Grid to Only Valid Hyperparameter Sets
#'
#' @title Extract Valid Grid from Compilation Results
#' @description
#' This helper function filters the results from `compile_keras_grid()` to
#' return a new hyperparameter grid containing only the combinations that
#' compiled successfully.
#'
#' @details
#' After running `compile_keras_grid()`, you can use this function to remove
#' problematic hyperparameter combinations before proceeding to the full
#' `tune::tune_grid()`.
#'
#' @param compiled_grid A tibble, the result of a call to
#' `compile_keras_grid()`.
#'
#' @return A tibble containing the subset of the original grid that resulted in
#'   a successful model compilation. The `compiled_model` and `error` columns
#'   are removed, leaving a clean grid ready for tuning.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("keras3", quietly = TRUE)) {
#'   library(keras3)
#'   library(parsnip)
#'   library(dials)
#'
#'   # 1. Define layer blocks
#'   input_block <- function(model, input_shape) {
#'     keras_model_sequential(input_shape = input_shape)
#'   }
#'   hidden_block <- function(model, units = 32) {
#'     model |> layer_dense(units = units, activation = "relu")
#'   }
#'   output_block <- function(model, num_classes) {
#'     model |> layer_dense(units = num_classes, activation = "softmax")
#'   }
#'
#'   # 2. Define a kerasnip model specification
#'   create_keras_sequential_spec(
#'     model_name = "my_mlp_grid_2",
#'     layer_blocks = list(
#'       input = input_block,
#'       hidden = hidden_block,
#'       output = output_block
#'     ),
#'     mode = "classification"
#'   )
#'
#'   mlp_spec <- my_mlp_grid_2(
#'     hidden_units = tune(),
#'     compile_loss = "categorical_crossentropy",
#'     compile_optimizer = "adam"
#'   )
#'
#'   # 3. Create a hyperparameter grid
#'   param_grid <- tibble::tibble(
#'     hidden_units = c(32, 64, -10)
#'   )
#'
#'   # 4. Prepare dummy data
#'   x_train <- matrix(rnorm(100 * 10), ncol = 10)
#'   y_train <- factor(sample(0:1, 100, replace = TRUE))
#'
#'   # 5. Compile models over the grid
#'   compiled_grid <- compile_keras_grid(
#'     spec = mlp_spec,
#'     grid = param_grid,
#'     x = x_train,
#'     y = y_train
#'   )
#'
#'   # 6. Extract the valid grid
#'   valid_grid <- extract_valid_grid(compiled_grid)
#'   print(valid_grid)
#'   remove_keras_spec("my_mlp_grid_2")
#' }
#' }
#' @export
extract_valid_grid <- function(compiled_grid) {
  if (
    !is.data.frame(compiled_grid) ||
      !all(
        c("error", "compiled_model") %in% names(compiled_grid)
      )
  ) {
    stop(
      "`compiled_grid` must be a data frame produced by `compile_keras_grid()`."
    )
  }
  compiled_grid |>
    dplyr::filter(is.na(error)) |>
    dplyr::select(-c(compiled_model, error))
}

#' Display a Summary of Compilation Errors
#'
#' @title Inform About Compilation Errors
#' @description
#' This helper function inspects the results from `compile_keras_grid()` and
#' prints a formatted, easy-to-read summary of any compilation errors that
#' occurred.
#'
#' @details
#' This is most useful for interactive debugging of complex tuning grids where
#' some hyperparameter combinations may lead to invalid Keras models.
#'
#' @param compiled_grid A tibble, the result of a call to
#' `compile_keras_grid()`.
#' @param n A single integer for the maximum number of distinct errors to
#'   display in detail.
#'
#' @return Invisibly returns the input `compiled_grid`. Called for its side
#'   effect of printing a summary to the console.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("keras3", quietly = TRUE)) {
#'   library(keras3)
#'   library(parsnip)
#'   library(dials)
#'
#'   # 1. Define layer blocks
#'   input_block <- function(model, input_shape) {
#'     keras_model_sequential(input_shape = input_shape)
#'   }
#'   hidden_block <- function(model, units = 32) {
#'     model |> layer_dense(units = units, activation = "relu")
#'   }
#'   output_block <- function(model, num_classes) {
#'     model |> layer_dense(units = num_classes, activation = "softmax")
#'   }
#'
#'   # 2. Define a kerasnip model specification
#'   create_keras_sequential_spec(
#'     model_name = "my_mlp_grid_3",
#'     layer_blocks = list(
#'       input = input_block,
#'       hidden = hidden_block,
#'       output = output_block
#'     ),
#'     mode = "classification"
#'   )
#'
#'   mlp_spec <- my_mlp_grid_3(
#'     hidden_units = tune(),
#'     compile_loss = "categorical_crossentropy",
#'     compile_optimizer = "adam"
#'   )
#'
#'   # 3. Create a hyperparameter grid
#'   param_grid <- tibble::tibble(
#'     hidden_units = c(32, 64, -10)
#'   )
#'
#'   # 4. Prepare dummy data
#'   x_train <- matrix(rnorm(100 * 10), ncol = 10)
#'   y_train <- factor(sample(0:1, 100, replace = TRUE))
#'
#'   # 5. Compile models over the grid
#'   compiled_grid <- compile_keras_grid(
#'     spec = mlp_spec,
#'     grid = param_grid,
#'     x = x_train,
#'     y = y_train
#'   )
#'
#'   # 6. Inform about errors
#'   inform_errors(compiled_grid)
#'   remove_keras_spec("my_mlp_grid_3")
#' }
#' }
#' @export
inform_errors <- function(compiled_grid, n = 10) {
  if (
    !is.data.frame(compiled_grid) || !all(c("error") %in% names(compiled_grid))
  ) {
    stop(
      "`compiled_grid` must be a data frame produced by `compile_keras_grid()`."
    )
  }
  error_grid <- compiled_grid |>
    dplyr::filter(!is.na(error))
  if (nrow(error_grid) > 0) {
    cli::cli_h1("Compilation Errors Summary")
    cli::cli_alert_danger(
      "{nrow(error_grid)} of {nrow(compiled_grid)} models failed to compile."
    )

    for (i in seq_len(min(nrow(error_grid), n))) {
      row <- error_grid[i, ]
      params <- row |> dplyr::select(-c(compiled_model, error))
      cli::cli_h2("Error {i}/{nrow(error_grid)}")
      cli::cli_text("Hyperparameters:")
      cli::cli_bullets(paste0(names(params), ": ", as.character(params)))
      cli::cli_text("Error Message:")
      cli::cli_code(row$error)
    }
    if (nrow(error_grid) > n) {
      cli::cli_alert_info("... and {nrow(error_grid) - n} more errors.")
    }
  } else {
    cli::cli_alert_success("All models compiled successfully!")
  }
  invisible(compiled_grid)
}

#' Build the Model Specification Function
#'
#' @description
#' This internal helper uses metaprogramming to construct a complete R function
#' that acts as a `parsnip` model specification (e.g., `my_mlp()`).
#'
#' @details
#' The process involves three main steps:
#' 1.  **Function Body Construction**: An expression for the function body is
#'     created. This body uses `rlang::enquo()` and `rlang::enquos()` to
#'     capture all user-provided arguments (both named and via `...`) into a
#'     list of quosures. This list is then passed to
#'     `parsnip::new_model_spec()`.
#' 2.  **Function Signature Construction**: A formal argument list is created
#'     from `all_args`, and `...` is added to allow passthrough arguments.
#'     `rlang::new_function()` combines the signature and body into a new
#'     function object.
#' 3.  **Documentation Attachment**: `generate_roxygen_docs()` creates a
#'     comprehensive Roxygen comment block as a string, which is then attached
#'     to the new function using `comment()`.
#'
#' @param model_name The name of the model specification function to create
#'   (e.g., "my_mlp").
#' @param mode The model mode ("regression" or "classification").
#' @param all_args A named list of formal arguments for the new function's
#'   signature, as generated by `collect_spec_args()`. The values are typically
#'   `rlang::missing_arg()` or `rlang::zap()`.
#' @param parsnip_names A character vector of all argument names that should be
#'   captured as quosures and passed to `parsnip::new_model_spec()`.
#' @param layer_blocks The user-provided list of layer block functions. This is
#'   passed directly to `generate_roxygen_docs()` to create documentation for
#'   block-specific parameters.
#' @param functional A logical indicating if the model is functional
#'   (for `create_keras_functional_spec()`) or sequential. This is passed to
#'   `generate_roxygen_docs()` to tailor the documentation.
#' @return A new function object with attached Roxygen comments, ready to be
#'   placed in the user's environment.
#' @noRd
build_spec_function <- function(
  model_name,
  mode,
  all_args,
  parsnip_names,
  layer_blocks,
  functional = FALSE
) {
  quos_exprs <- purrr::map(
    parsnip_names,
    ~ rlang::expr(rlang::enquo(!!rlang::sym(.x)))
  )
  names(quos_exprs) <- parsnip_names

  quos_call <- rlang::call2("quos", rlang::sym("..."), .ns = "rlang")

  # Capture at spec-creation time so every instance carries the data needed
  # for auto-registration in predict.kerasnip_model_fit.
  captured_layer_blocks <- layer_blocks
  captured_functional <- functional

  body <- rlang::expr({
    # Capture both explicit args and ... to pass to the fit impl
    # Named arguments are captured into a list of quosures.
    main_args <- rlang::list2(!!!quos_exprs)
    # ... arguments are captured into a separate list of quosures.
    dot_args <- !!quos_call
    args <- c(main_args, dot_args)
    result <- parsnip::new_model_spec(
      !!model_name,
      args = args,
      eng_args = NULL,
      mode = !!mode,
      method = NULL,
      engine = NULL
    )
    class(result) <- c(class(result)[1L], "kerasnip_spec", class(result)[-1L])
    attr(result, "kerasnip_layer_blocks") <- captured_layer_blocks
    attr(result, "kerasnip_functional") <- captured_functional
    result
  })

  # Add ... to the function signature to capture any other compile arguments
  fn_args <- c(all_args, list(... = rlang::missing_arg()))

  fn <- rlang::new_function(args = fn_args, body = body)

  docs <- generate_roxygen_docs(
    model_name,
    layer_blocks,
    all_args,
    functional = functional
  )
  comment(fn) <- docs
  fn
}

---
title: "Tuning Multiple Similar Parameters: Ensuring `autoplot` Uniqueness"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tuning Multiple Similar Parameters: Ensuring `autoplot` Uniqueness}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



## Introduction

When using `kerasnip` to define and tune Keras models within the `tidymodels` framework, you might encounter situations where you want to tune multiple parameters that, by default, map to the same underlying `dials` parameter type. A common example is tuning the number of units in multiple `layer_dense` blocks within the same model.

While `kerasnip` intelligently maps these parameters (e.g., `dense1_units` and `dense2_units` both map to `dials::hidden_units()`), this can lead to ambiguity when visualizing tuning results with `ggplot2::autoplot()`. Without a way to distinguish between these otherwise identical parameter types, `autoplot()` may produce errors or misleading plots.

This vignette demonstrates how to explicitly provide unique identifiers to your tuned parameters, ensuring `autoplot()` can correctly visualize the results for each distinct parameter.

## The Problem (Implicit)

Consider a model with two dense layers, each with a `units` parameter. If you were to define them for tuning without unique `id`s, `autoplot()` would encounter an issue because it cannot distinguish between the two parameters.

For example, if you were to run `ggplot2::autoplot(tune_res)` without unique `id`s, you might encounter an error similar to this:


``` r
#> Error in `dplyr::rename()`:
#> ! Names must be unique.
#> ✖ These names are duplicated:
#>   * "# Hidden Units" at locations 1 and 2.
```

This error clearly indicates that `autoplot()` is trying to rename columns for plotting, but it finds duplicate names like "# Hidden Units" because both `dense1_units` and `dense2_units` are generically identified as `hidden_units` by `dials` without further distinction. This makes it impossible for `autoplot()` to differentiate their tuning results.

## The Solution: Using Unique `id`s with `tune()`

The solution is to provide a unique `id` argument to the `tune()` function for each parameter you wish to distinguish.

Let's define a simple sequential Keras model with two dense layers:


``` r
library(kerasnip)
library(keras3)
library(parsnip)
#> 
#> Attaching package: 'parsnip'
#> The following object is masked from 'package:kerasnip':
#> 
#>     get_model_env
library(dials)
#> Loading required package: scales
library(workflows)
library(recipes)
#> Loading required package: dplyr
#> 
#> Attaching package: 'dplyr'
#> The following objects are masked from 'package:stats':
#> 
#>     filter, lag
#> The following objects are masked from 'package:base':
#> 
#>     intersect, setdiff, setequal, union
#> 
#> Attaching package: 'recipes'
#> The following object is masked from 'package:stats':
#> 
#>     step
library(rsample)
library(tune)
library(ggplot2)

# Define a spec with multiple hidden unit parameters
model_name <- "autoplot_unique_spec"
# Clean up the spec if it already exists from a previous run
if (exists(model_name, mode = "function")) {
  suppressMessages(remove_keras_spec(model_name))
}

input_block <- function(model, input_shape) {
  keras3::keras_model_sequential(input_shape = input_shape)
}

dense_block <- function(model, units = 10) {
  model |> keras3::layer_dense(units = units)
}

output_block <- function(model, num_classes) {
  model |>
    keras3::layer_dense(units = num_classes, activation = "softmax")
}

create_keras_sequential_spec(
  model_name = model_name,
  layer_blocks = list(
    input = input_block,
    dense1 = dense_block,
    dense2 = dense_block,
    output = output_block
  ),
  mode = "classification"
)

# Now, create the model specification and assign unique IDs for tuning
tune_spec <- autoplot_unique_spec(
  dense1_units = tune(id = "dense_layer_one_units"),
  dense2_units = tune(id = "dense_layer_two_units")
) |>
  set_engine("keras")

print(tune_spec)
#> autoplot unique spec Model Specification (classification)
#> 
#> Main Arguments:
#>   num_input = structure(list(), class = "rlang_zap")
#>   num_dense1 = structure(list(), class = "rlang_zap")
#>   num_dense2 = structure(list(), class = "rlang_zap")
#>   num_output = structure(list(), class = "rlang_zap")
#>   dense1_units = tune(id = "dense_layer_one_units")
#>   dense2_units = tune(id = "dense_layer_two_units")
#>   learn_rate = structure(list(), class = "rlang_zap")
#>   fit_batch_size = structure(list(), class = "rlang_zap")
#>   fit_epochs = structure(list(), class = "rlang_zap")
#>   fit_callbacks = structure(list(), class = "rlang_zap")
#>   fit_validation_split = structure(list(), class = "rlang_zap")
#>   fit_validation_data = structure(list(), class = "rlang_zap")
#>   fit_shuffle = structure(list(), class = "rlang_zap")
#>   fit_class_weight = structure(list(), class = "rlang_zap")
#>   fit_sample_weight = structure(list(), class = "rlang_zap")
#>   fit_initial_epoch = structure(list(), class = "rlang_zap")
#>   fit_steps_per_epoch = structure(list(), class = "rlang_zap")
#>   fit_validation_steps = structure(list(), class = "rlang_zap")
#>   fit_validation_batch_size = structure(list(), class = "rlang_zap")
#>   fit_validation_freq = structure(list(), class = "rlang_zap")
#>   fit_verbose = structure(list(), class = "rlang_zap")
#>   fit_view_metrics = structure(list(), class = "rlang_zap")
#>   compile_optimizer = structure(list(), class = "rlang_zap")
#>   compile_loss = structure(list(), class = "rlang_zap")
#>   compile_metrics = structure(list(), class = "rlang_zap")
#>   compile_loss_weights = structure(list(), class = "rlang_zap")
#>   compile_weighted_metrics = structure(list(), class = "rlang_zap")
#>   compile_run_eagerly = structure(list(), class = "rlang_zap")
#>   compile_steps_per_execution = structure(list(), class = "rlang_zap")
#>   compile_jit_compile = structure(list(), class = "rlang_zap")
#>   compile_auto_scale_loss = structure(list(), class = "rlang_zap")
#> 
#> Computational engine: keras
```

Notice how `dense1_units` and `dense2_units` are both passed to `tune()`, but each is given a distinct `id`. This `id` acts as a label that `autoplot()` can use to differentiate the parameters.

### Setting up the Tuning Workflow

Next, we'll set up a `tidymodels` workflow, define the parameter ranges, and create a tuning grid.


``` r
# Set up workflow and tuning grid
rec <- recipes::recipe(Species ~ ., data = iris)
tune_wf <- workflows::workflow(rec, tune_spec)

params <- tune::extract_parameter_set_dials(tune_wf)

# Update the parameter ranges using kerasnip::hidden_units
# The `id`s provided in tune() are automatically detected and used here.
params <- params |>
  update(
    dense_layer_one_units = hidden_units(range = c(4L, 8L)),
    dense_layer_two_units = hidden_units(range = c(4L, 8L))
  )

grid <- dials::grid_regular(params, levels = 2)
control <- tune::control_grid(save_pred = FALSE, verbose = FALSE)

print(grid)
#> # A tibble: 4 × 2
#>   dense_layer_one_units dense_layer_two_units
#>                   <int>                 <int>
#> 1                     4                     4
#> 2                     8                     4
#> 3                     4                     8
#> 4                     8                     8
```

### Running the Tuning Process

Now, we run `tune::tune_grid` to perform the actual tuning. For demonstration purposes, we'll use a small number of resamples and a simple dataset.


``` r
# Run tuning
tune_res <- tune::tune_grid(
  tune_wf,
  resamples = rsample::vfold_cv(iris, v = 2),
  grid = grid,
  control = control
)
#> 3/3 - 0s - 79ms/step
#> 3/3 - 0s - 29ms/step
#> 3/3 - 0s - 67ms/step
#> 3/3 - 0s - 24ms/step
#> 3/3 - 0s - 60ms/step
#> 3/3 - 0s - 24ms/step
#> 3/3 - 0s - 94ms/step
#> 3/3 - 0s - 26ms/step
#> 3/3 - 0s - 75ms/step
#> 3/3 - 0s - 36ms/step
#> 3/3 - 0s - 90ms/step
#> 3/3 - 0s - 34ms/step
#> 3/3 - 0s - 73ms/step
#> 3/3 - 0s - 22ms/step
#> 3/3 - 0s - 85ms/step
#> 3/3 - 0s - 33ms/step

print(tune_res)
#> # Tuning results
#> # 2-fold cross-validation 
#> # A tibble: 2 × 4
#>   splits          id    .metrics          .notes          
#>   <list>          <chr> <list>            <list>          
#> 1 <split [75/75]> Fold1 <tibble [12 × 6]> <tibble [0 × 4]>
#> 2 <split [75/75]> Fold2 <tibble [12 × 6]> <tibble [0 × 4]>
```

### Visualizing Results with `autoplot()`

With the tuning complete, we can now use `ggplot2::autoplot()` to visualize the results. Because we provided unique `id`s, `autoplot()` can correctly generate separate plots for each tuned parameter.


``` r
# Assert that autoplot works without error
ggplot2::autoplot(tune_res)
```

![plot of chunk autoplot_results](figure/autoplot_results-1.png)

As you can see, `autoplot()` successfully generates a plot showing the performance across the different values for `dense_layer_one_units` and `dense_layer_two_units` independently.

## Why Unique `id`s are Necessary

Internally, `kerasnip` maps arguments like `units` from your `layer_blocks` functions to appropriate `dials` parameter objects (e.g., `dials::hidden_units()`). When multiple such arguments exist, they all point to the *same type* of `dials` parameter.

The `id` argument in `tune()` serves as a unique identifier that `tune` and `ggplot2::autoplot()` use to distinguish between different instances of these parameter types. Without it, `autoplot()` would see multiple parameters of type `hidden_units` and wouldn't know how to plot them separately, leading to errors or combining them incorrectly.

## Best Practices

*   **Always use unique `id`s:** When tuning multiple parameters that are conceptually similar (e.g., `units` in different layers, `rate` in different dropout layers), always provide a descriptive and unique `id` to the `tune()` function.
*   **Descriptive `id`s:** Choose `id`s that clearly indicate which part of the model the parameter belongs to (e.g., `dense_layer_one_units`, `conv_filter_size`). This improves readability and understanding of your tuning results.

By following this practice, you ensure that your `kerasnip` models are robustly tunable and that their results can be clearly visualized using the `tidymodels` ecosystem.

---
title: "Tuning Fit and Compile Arguments"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tuning Fit and Compile Arguments}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



## Introduction

While `kerasnip` makes it easy to tune the architecture of a Keras model (e.g., the number of layers or the number of units in a layer), it is often just as important to tune the parameters that control the training process itself. `kerasnip` exposes these parameters through special `fit_*` and `compile_*` arguments in the model specification.

This vignette provides a comprehensive example of how to tune these arguments within a `tidymodels` workflow. We will tune:

*   **`fit_epochs`**: The number of training epochs.
*   **`fit_batch_size`**: The number of samples per gradient update.
*   **`compile_optimizer`**: The optimization algorithm (e.g., "adam", "sgd").
*   **`compile_loss`**: The loss function used for training.
*   **`learn_rate`**: The learning rate for the optimizer.

## Setup

First, we load the necessary packages.


``` r
library(kerasnip)
library(tidymodels)
#> ── Attaching packages ────────────────────────────────────────────────────────────────────────────── tidymodels 1.5.0 ──
#> ✔ broom        1.0.12     ✔ recipes      1.3.2 
#> ✔ dials        1.4.3      ✔ rsample      1.3.2 
#> ✔ dplyr        1.2.1      ✔ tailor       0.1.0 
#> ✔ ggplot2      4.0.3      ✔ tidyr        1.3.2 
#> ✔ infer        1.1.0      ✔ tune         2.1.0 
#> ✔ modeldata    1.5.1      ✔ workflows    1.3.0 
#> ✔ parsnip      1.5.0      ✔ workflowsets 1.1.1 
#> ✔ purrr        1.2.2      ✔ yardstick    1.4.0
#> ── Conflicts ───────────────────────────────────────────────────────────────────────────────── tidymodels_conflicts() ──
#> ✖ purrr::discard() masks scales::discard()
#> ✖ dplyr::filter()  masks stats::filter()
#> ✖ dplyr::lag()     masks stats::lag()
#> ✖ recipes::step()  masks stats::step()
library(keras3)
#> 
#> Attaching package: 'keras3'
#> The following object is masked from 'package:yardstick':
#> 
#>     get_weights
#> The following object is masked from 'package:infer':
#> 
#>     generate
```

## Data Preparation

We will use the classic `iris` dataset for this example. It's a simple, small dataset, which is ideal for demonstrating the tuning process without long training times.


``` r
# Split data into training and testing sets
set.seed(123)
iris_split <- initial_split(iris, prop = 0.8, strata = Species)
iris_train <- training(iris_split)
iris_test <- testing(iris_split)

# Create cross-validation folds for tuning
iris_folds <- vfold_cv(iris_train, v = 3, strata = Species)
```

## Define a `kerasnip` Model

We'll create a very simple sequential model with a single dense layer. This keeps the focus on tuning the `fit_*` and `compile_*` arguments rather than the model architecture.


``` r
# Define layer blocks
input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}
dense_block <- function(model, units = 10) {
  model |> layer_dense(units = units, activation = "relu")
}
output_block <- function(model, num_classes) {
  model |> layer_dense(units = num_classes, activation = "softmax")
}

# Create the kerasnip model specification function
create_keras_sequential_spec(
  model_name = "iris_mlp",
  layer_blocks = list(
    input = input_block,
    dense = dense_block,
    output = output_block
  ),
  mode = "classification"
)
```

## Define the Tunable Specification

Now, we create an instance of our `iris_mlp` model. We set the arguments we want to optimize to `tune()`.


``` r
# Define the tunable model specification
tune_spec <- iris_mlp(
  dense_units = 16, # Keep architecture fixed for this example
  fit_epochs = tune(),
  fit_batch_size = tune(),
  compile_optimizer = tune(),
  compile_loss = tune(),
  learn_rate = tune()
) |>
  set_engine("keras")

print(tune_spec)
#> iris mlp Model Specification (classification)
#> 
#> Main Arguments:
#>   num_input = structure(list(), class = "rlang_zap")
#>   num_dense = structure(list(), class = "rlang_zap")
#>   num_output = structure(list(), class = "rlang_zap")
#>   dense_units = 16
#>   learn_rate = tune()
#>   fit_batch_size = tune()
#>   fit_epochs = tune()
#>   fit_callbacks = structure(list(), class = "rlang_zap")
#>   fit_validation_split = structure(list(), class = "rlang_zap")
#>   fit_validation_data = structure(list(), class = "rlang_zap")
#>   fit_shuffle = structure(list(), class = "rlang_zap")
#>   fit_class_weight = structure(list(), class = "rlang_zap")
#>   fit_sample_weight = structure(list(), class = "rlang_zap")
#>   fit_initial_epoch = structure(list(), class = "rlang_zap")
#>   fit_steps_per_epoch = structure(list(), class = "rlang_zap")
#>   fit_validation_steps = structure(list(), class = "rlang_zap")
#>   fit_validation_batch_size = structure(list(), class = "rlang_zap")
#>   fit_validation_freq = structure(list(), class = "rlang_zap")
#>   fit_verbose = structure(list(), class = "rlang_zap")
#>   fit_view_metrics = structure(list(), class = "rlang_zap")
#>   compile_optimizer = tune()
#>   compile_loss = tune()
#>   compile_metrics = structure(list(), class = "rlang_zap")
#>   compile_loss_weights = structure(list(), class = "rlang_zap")
#>   compile_weighted_metrics = structure(list(), class = "rlang_zap")
#>   compile_run_eagerly = structure(list(), class = "rlang_zap")
#>   compile_steps_per_execution = structure(list(), class = "rlang_zap")
#>   compile_jit_compile = structure(list(), class = "rlang_zap")
#>   compile_auto_scale_loss = structure(list(), class = "rlang_zap")
#> 
#> Computational engine: keras
```

## Create Workflow and Tuning Grid

Next, we create a `workflow` and define the search space for our hyperparameters using `dials`. `kerasnip` provides special `dials` parameter functions for `optimizer` and `loss`.


``` r
# Create a simple recipe
iris_recipe <- recipe(Species ~ ., data = iris_train) |>
  step_normalize(all_numeric_predictors())

# Create the workflow
tune_wf <- workflow() |>
  add_recipe(iris_recipe) |>
  add_model(tune_spec)

# Define the tuning grid
params <- extract_parameter_set_dials(tune_wf) |>
  update(
    fit_epochs = epochs(c(10, 30)),
    fit_batch_size = batch_size(c(16, 64), trans = NULL),
    compile_optimizer = optimizer_function(values = c("adam", "sgd", "rmsprop")),
    compile_loss = loss_function_keras(values = c("categorical_crossentropy", "kl_divergence")),
    learn_rate = learn_rate(c(0.001, 0.01), trans = NULL)
  )

set.seed(456)
tuning_grid <- grid_regular(params, levels = 2)

tuning_grid
#> # A tibble: 32 × 5
#>    learn_rate fit_batch_size fit_epochs compile_optimizer compile_loss            
#>         <dbl>          <int>      <int> <chr>             <chr>                   
#>  1      0.001             16         10 adam              categorical_crossentropy
#>  2      0.01              16         10 adam              categorical_crossentropy
#>  3      0.001             64         10 adam              categorical_crossentropy
#>  4      0.01              64         10 adam              categorical_crossentropy
#>  5      0.001             16         30 adam              categorical_crossentropy
#>  6      0.01              16         30 adam              categorical_crossentropy
#>  7      0.001             64         30 adam              categorical_crossentropy
#>  8      0.01              64         30 adam              categorical_crossentropy
#>  9      0.001             16         10 sgd               categorical_crossentropy
#> 10      0.01              16         10 sgd               categorical_crossentropy
#> # ℹ 22 more rows
```

## Tune the Model

With the workflow and grid defined, we can now run the hyperparameter tuning using `tune_grid()`.


``` r
tune_res <- tune_grid(
  tune_wf,
  resamples = iris_folds,
  grid = tuning_grid,
  metrics = metric_set(accuracy, roc_auc),
  control = control_grid(save_pred = FALSE, save_workflow = TRUE, verbose = FALSE)
)
#> 2/2 - 0s - 83ms/step
#> 2/2 - 0s - 39ms/step
#> 2/2 - 0s - 74ms/step
#> 2/2 - 0s - 34ms/step
#> 2/2 - 0s - 172ms/step
#> 2/2 - 0s - 59ms/step
#> 2/2 - 0s - 80ms/step
#> 2/2 - 0s - 47ms/step
#> 2/2 - 0s - 105ms/step
#> 2/2 - 0s - 52ms/step
#> 2/2 - 0s - 105ms/step
#> 2/2 - 0s - 45ms/step
#> 2/2 - 0s - 120ms/step
#> 2/2 - 0s - 58ms/step
#> 2/2 - 0s - 98ms/step
#> 2/2 - 0s - 32ms/step
#> 2/2 - 0s - 104ms/step
#> 2/2 - 0s - 32ms/step
#> 2/2 - 0s - 99ms/step
#> 2/2 - 0s - 42ms/step
#> 2/2 - 0s - 107ms/step
#> 2/2 - 0s - 60ms/step
#> 2/2 - 0s - 92ms/step
#> 2/2 - 0s - 46ms/step
#> 2/2 - 0s - 202ms/step
#> 2/2 - 0s - 125ms/step
#> 2/2 - 0s - 122ms/step
#> 2/2 - 0s - 31ms/step
#> 2/2 - 0s - 72ms/step
#> 2/2 - 0s - 33ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 28ms/step
#> 2/2 - 0s - 118ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 60ms/step
#> 2/2 - 0s - 35ms/step
#> 2/2 - 0s - 65ms/step
#> 2/2 - 0s - 33ms/step
#> 2/2 - 0s - 67ms/step
#> 2/2 - 0s - 31ms/step
#> 2/2 - 0s - 84ms/step
#> 2/2 - 0s - 34ms/step
#> 2/2 - 0s - 63ms/step
#> 2/2 - 0s - 34ms/step
#> 2/2 - 0s - 62ms/step
#> 2/2 - 0s - 33ms/step
#> 2/2 - 0s - 62ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 72ms/step
#> 2/2 - 0s - 34ms/step
#> 2/2 - 0s - 72ms/step
#> 2/2 - 0s - 67ms/step
#> 2/2 - 0s - 65ms/step
#> 2/2 - 0s - 33ms/step
#> 2/2 - 0s - 61ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 61ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 67ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 59ms/step
#> 2/2 - 0s - 28ms/step
#> 2/2 - 0s - 71ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 68ms/step
#> 2/2 - 0s - 34ms/step
#> 2/2 - 0s - 58ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 68ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 77ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 73ms/step
#> 2/2 - 0s - 26ms/step
#> 2/2 - 0s - 69ms/step
#> 2/2 - 0s - 32ms/step
#> 2/2 - 0s - 69ms/step
#> 2/2 - 0s - 28ms/step
#> 2/2 - 0s - 58ms/step
#> 2/2 - 0s - 27ms/step
#> 2/2 - 0s - 55ms/step
#> 2/2 - 0s - 31ms/step
#> 2/2 - 0s - 60ms/step
#> 2/2 - 0s - 27ms/step
#> 2/2 - 0s - 62ms/step
#> 2/2 - 0s - 33ms/step
#> 2/2 - 0s - 61ms/step
#> 2/2 - 0s - 32ms/step
#> 2/2 - 0s - 61ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 63ms/step
#> 2/2 - 0s - 31ms/step
#> 2/2 - 0s - 67ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 57ms/step
#> 2/2 - 0s - 27ms/step
#> 2/2 - 0s - 65ms/step
#> 2/2 - 0s - 31ms/step
#> 2/2 - 0s - 75ms/step
#> 2/2 - 0s - 35ms/step
#> 2/2 - 0s - 71ms/step
#> 2/2 - 0s - 34ms/step
#> 2/2 - 0s - 69ms/step
#> 2/2 - 0s - 33ms/step
#> 2/2 - 0s - 67ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 32ms/step
#> 2/2 - 0s - 59ms/step
#> 2/2 - 0s - 32ms/step
#> 2/2 - 0s - 57ms/step
#> 2/2 - 0s - 27ms/step
#> 2/2 - 0s - 56ms/step
#> 2/2 - 0s - 36ms/step
#> 2/2 - 0s - 59ms/step
#> 2/2 - 0s - 31ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 36ms/step
#> 2/2 - 0s - 72ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 65ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 53ms/step
#> 2/2 - 0s - 24ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 28ms/step
#> 2/2 - 0s - 85ms/step
#> 2/2 - 0s - 38ms/step
#> 2/2 - 0s - 61ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 75ms/step
#> 2/2 - 0s - 31ms/step
#> 2/2 - 0s - 69ms/step
#> 2/2 - 0s - 36ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 65ms/step
#> 2/2 - 0s - 31ms/step
#> 2/2 - 0s - 73ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 58ms/step
#> 2/2 - 0s - 26ms/step
#> 2/2 - 0s - 77ms/step
#> 2/2 - 0s - 26ms/step
#> 2/2 - 0s - 76ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 61ms/step
#> 2/2 - 0s - 27ms/step
#> 2/2 - 0s - 67ms/step
#> 2/2 - 0s - 33ms/step
#> 2/2 - 0s - 53ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 32ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 38ms/step
#> 2/2 - 0s - 68ms/step
#> 2/2 - 0s - 31ms/step
#> 2/2 - 0s - 75ms/step
#> 2/2 - 0s - 28ms/step
#> 2/2 - 0s - 59ms/step
#> 2/2 - 0s - 33ms/step
#> 2/2 - 0s - 65ms/step
#> 2/2 - 0s - 26ms/step
#> 2/2 - 0s - 59ms/step
#> 2/2 - 0s - 28ms/step
#> 2/2 - 0s - 71ms/step
#> 2/2 - 0s - 28ms/step
#> 2/2 - 0s - 66ms/step
#> 2/2 - 0s - 36ms/step
#> 2/2 - 0s - 66ms/step
#> 2/2 - 0s - 28ms/step
#> 2/2 - 0s - 66ms/step
#> 2/2 - 0s - 34ms/step
#> 2/2 - 0s - 68ms/step
#> 2/2 - 0s - 32ms/step
#> 2/2 - 0s - 61ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 67ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 29ms/step
#> 2/2 - 0s - 76ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 65ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 62ms/step
#> 2/2 - 0s - 30ms/step
#> 2/2 - 0s - 69ms/step
#> 2/2 - 0s - 31ms/step
```

## Inspect the Results

Let's examine the results to see how the different combinations of fitting and compilation parameters performed.


``` r
# Show the best performing models based on accuracy
show_best(tune_res, metric = "accuracy")
#> # A tibble: 5 × 11
#>   learn_rate fit_batch_size fit_epochs compile_optimizer compile_loss     .metric .estimator  mean     n std_err .config
#>        <dbl>          <int>      <int> <chr>             <chr>            <chr>   <chr>      <dbl> <int>   <dbl> <chr>  
#> 1       0.01             64         30 adam              kl_divergence    accura… multiclass 0.935     3  0.0279 pre0_m…
#> 2       0.01             16         30 adam              categorical_cro… accura… multiclass 0.934     3  0.0147 pre0_m…
#> 3       0.01             16         30 adam              kl_divergence    accura… multiclass 0.934     3  0.0147 pre0_m…
#> 4       0.01             64         30 adam              categorical_cro… accura… multiclass 0.901     3  0.0265 pre0_m…
#> 5       0.01             16         10 adam              kl_divergence    accura… multiclass 0.893     3  0.0299 pre0_m…

# Plot the results
autoplot(tune_res) + theme_minimal()
```

![plot of chunk inspect-results](figure/inspect-results-1.png)

``` r

# Select the best hyperparameters
best_params <- select_best(tune_res, metric = "accuracy")
print(best_params)
#> # A tibble: 1 × 6
#>   learn_rate fit_batch_size fit_epochs compile_optimizer compile_loss  .config         
#>        <dbl>          <int>      <int> <chr>             <chr>         <chr>           
#> 1       0.01             64         30 adam              kl_divergence pre0_mod30_post0
```

The results show that `tune` has successfully explored different optimizers, loss functions, learning rates, epochs, and batch sizes, identifying the combination that yields the best accuracy.

## Finalize and Fit

Finally, we finalize our workflow with the best-performing hyperparameters and fit the model one last time on the full training dataset.


``` r
# Finalize the workflow
final_wf <- finalize_workflow(tune_wf, best_params)

# Fit the final model
final_fit <- fit(final_wf, data = iris_train)

print(final_fit)
#> ══ Workflow [trained] ══════════════════════════════════════════════════════════════════════════════════════════════════
#> Preprocessor: Recipe
#> Model: iris_mlp()
#> 
#> ── Preprocessor ────────────────────────────────────────────────────────────────────────────────────────────────────────
#> 1 Recipe Step
#> 
#> • step_normalize()
#> 
#> ── Model ───────────────────────────────────────────────────────────────────────────────────────────────────────────────
#> $fit
#> Model: "sequential_96"
#> ┌─────────────────────────────────────────────────────┬────────────────────────────────────────┬───────────────────────
#> │ Layer (type)                                        │ Output Shape                           │               Param # 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_192 (Dense)                                   │ (None, 16)                             │                    80 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_193 (Dense)                                   │ (None, 3)                              │                    51 
#> └─────────────────────────────────────────────────────┴────────────────────────────────────────┴───────────────────────
#>  Total params: 395 (1.55 KB)
#>  Trainable params: 131 (524.00 B)
#>  Non-trainable params: 0 (0.00 B)
#>  Optimizer params: 264 (1.04 KB)
#> 
#> $keras_bytes
#>     [1] 50 4b 03 04 14 00 00 00 00 00 00 00 21 00 71 a1 30 04 40 00 00 00 40 00 00 00 0d 00 00 00 6d 65 74 61 64 61 74
#>    [38] 61 2e 6a 73 6f 6e 7b 22 6b 65 72 61 73 5f 76 65 72 73 69 6f 6e 22 3a 20 22 33 2e 31 34 2e 30 22 2c 20 22 64 61
#>    [75] 74 65 5f 73 61 76 65 64 22 3a 20 22 32 30 32 36 2d 30 35 2d 30 31 40 31 31 3a 34 38 3a 33 37 22 7d 50 4b 03 04
#>   [112] 14 00 00 00 00 00 00 00 21 00 12 e4 b0 c3 27 0b 00 00 27 0b 00 00 0b 00 00 00 63 6f 6e 66 69 67 2e 6a 73 6f 6e
#>   [149] 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 22 2c 20 22 63 6c 61 73 73 5f 6e 61 6d 65 22 3a 20 22 53 65
#>   [186] 71 75 65 6e 74 69 61 6c 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 6e 61 6d 65 22 3a 20 22 73 65 71 75 65 6e
#>   [223] 74 69 61 6c 5f 39 36 22 2c 20 22 74 72 61 69 6e 61 62 6c 65 22 3a 20 74 72 75 65 2c 20 22 64 74 79 70 65 22 3a
#>   [260] 20 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 22 2c 20 22 63 6c 61 73 73 5f 6e 61 6d 65 22 3a 20 22 44
#>   [297] 54 79 70 65 50 6f 6c 69 63 79 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 6e 61 6d 65 22 3a 20 22 66 6c 6f 61
#>   [334] 74 33 32 22 7d 2c 20 22 72 65 67 69 73 74 65 72 65 64 5f 6e 61 6d 65 22 3a 20 6e 75 6c 6c 2c 20 22 73 68 61 72
#>   [371] 65 64 5f 6f 62 6a 65 63 74 5f 69 64 22 3a 20 32 39 33 38 33 35 36 38 34 36 34 31 36 7d 2c 20 22 6c 61 79 65 72
#>   [408] 73 22 3a 20 5b 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 2e 6c 61 79 65 72 73 22 2c 20 22 63 6c 61 73
#>   [445] 73 5f 6e 61 6d 65 22 3a 20 22 49 6e 70 75 74 4c 61 79 65 72 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 62 61
#>   [482] 74 63 68 5f 73 68 61 70 65 22 3a 20 5b 6e 75 6c 6c 2c 20 34 5d 2c 20 22 64 74 79 70 65 22 3a 20 22 66 6c 6f 61
#>   [519] 74 33 32 22 2c 20 22 73 70 61 72 73 65 22 3a 20 66 61 6c 73 65 2c 20 22 72 61 67 67 65 64 22 3a 20 66 61 6c 73
#>   [556] 65 2c 20 22 6e 61 6d 65 22 3a 20 22 69 6e 70 75 74 5f 6c 61 79 65 72 5f 39 36 22 2c 20 22 6f 70 74 69 6f 6e 61
#>   [593] 6c 22 3a 20 66 61 6c 73 65 7d 2c 20 22 72 65 67 69 73 74 65 72 65 64 5f 6e 61 6d 65 22 3a 20 6e 75 6c 6c 7d 2c
#>   [630] 20 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 2e 6c 61 79 65 72 73 22 2c 20 22 63 6c 61 73 73 5f 6e 61
#>   [667] 6d 65 22 3a 20 22 44 65 6e 73 65 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 6e 61 6d 65 22 3a 20 22 64 65 6e
#>   [704] 73 65 5f 31 39 32 22 2c 20 22 74 72 61 69 6e 61 62 6c 65 22 3a 20 74 72 75 65 2c 20 22 64 74 79 70 65 22 3a 20
#>   [741] 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 22 2c 20 22 63 6c 61 73 73 5f 6e 61 6d 65 22 3a 20 22 44 54
#>   [778] 79 70 65 50 6f 6c 69 63 79 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 6e 61 6d 65 22 3a 20 22 66 6c 6f 61 74
#>   [815] 33 32 22 7d 2c 20 22 72 65 67 69 73 74 65 72 65 64 5f 6e 61 6d 65 22 3a 20 6e 75 6c 6c 7d 2c 20 22 75 6e 69 74
#>   [852] 73 22 3a 20 31 36 2c 20 22 61 63 74 69 76 61 74 69 6f 6e 22 3a 20 22 72 65 6c 75 22 2c 20 22 75 73 65 5f 62 69
#>   [889] 61 73 22 3a 20 74 72 75 65 2c 20 22 6b 65 72 6e 65 6c 5f 69 6e 69 74 69 61 6c 69 7a 65 72 22 3a 20 7b 22 6d 6f
#>   [926] 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 2e 69 6e 69 74 69 61 6c 69 7a 65 72 73 22 2c 20 22 63 6c 61 73 73 5f 6e
#>   [963] 61 6d 65 22 3a 20 22 47 6c 6f 72 6f 74 55 6e 69 66 6f 72 6d 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 73 65
#>  [1000] 65 64 22 3a 20 6e 75 6c 6c 7d 2c 20 22 72 65 67 69 73 74 65 72 65 64 5f 6e 61 6d 65 22 3a 20 6e 75 6c 6c 7d 2c
#>  [1037] 20 22 62 69 61 73 5f 69 6e 69 74 69 61 6c 69 7a 65 72 22 3a 20 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61
#>  [1074] 73 2e 69 6e 69 74 69 61 6c 69 7a 65 72 73 22 2c 20 22 63 6c 61 73 73 5f 6e 61 6d 65 22 3a 20 22 5a 65 72 6f 73
#>  [1111] 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 7d 2c 20 22 72 65 67 69 73 74 65 72 65 64 5f 6e 61 6d 65 22 3a 20 6e
#>  [1148] 75 6c 6c 7d 2c 20 22 6b 65 72 6e 65 6c 5f 72 65 67 75 6c 61 72 69 7a 65 72 22 3a 20 6e 75 6c 6c 2c 20 22 62 69
#>  [1185] 61 73 5f 72 65 67 75 6c 61 72 69 7a 65 72 22 3a 20 6e 75 6c 6c 2c 20 22 6b 65 72 6e 65 6c 5f 63 6f 6e 73 74 72
#>  [1222] 61 69 6e 74 22 3a 20 6e 75 6c 6c 2c 20 22 62 69 61 73 5f 63 6f 6e 73 74 72 61 69 6e 74 22 3a 20 6e 75 6c 6c 2c
#>  [1259] 20 22 71 75 61 6e 74 69 7a 61 74 69 6f 6e 5f 63 6f 6e 66 69 67 22 3a 20 6e 75 6c 6c 7d 2c 20 22 72 65 67 69 73
#> 
#> ...
#> and 634 more lines.
```

We can now use this `final_fit` object to make predictions on the test set.


``` r
# Make predictions
predictions <- predict(final_fit, new_data = iris_test)
#> 1/1 - 0s - 105ms/step

# Evaluate performance
bind_cols(predictions, iris_test) |>
  accuracy(truth = Species, estimate = .pred_class)
#> # A tibble: 1 × 3
#>   .metric  .estimator .estimate
#>   <chr>    <chr>          <dbl>
#> 1 accuracy multiclass     0.933
```

## Conclusion

This vignette demonstrated how to tune the crucial `fit_*` and `compile_*` arguments of a Keras model within the `tidymodels` framework using `kerasnip`. By exposing these as tunable parameters, `kerasnip` gives you full control over the training process, allowing you to optimize not just the model's architecture, but also how it learns.

---
title: "Saving and Reloading Fitted Kerasnip Workflows"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Saving and Reloading Fitted Kerasnip Workflows}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



## Overview

Keras models are backed by Python objects managed by TensorFlow/JAX. These objects live in the current R session and are represented as external pointers (`externalptr`) that become invalid as soon as the session ends, or even within the same session after `saveRDS()` / `readRDS()`.

`kerasnip` handles this transparently so that fitted workflows can be saved, reloaded, and used for prediction without any manual restoration steps.

## Quick workflow example

Before discussing the details, here is the full persistence workflow:


``` r
library(kerasnip)
library(tidymodels)
#> ── Attaching packages ────────────────────────────────────────────────────────────────────────────── tidymodels 1.5.0 ──
#> ✔ broom        1.0.12     ✔ recipes      1.3.2 
#> ✔ dials        1.4.3      ✔ rsample      1.3.2 
#> ✔ dplyr        1.2.1      ✔ tailor       0.1.0 
#> ✔ ggplot2      4.0.3      ✔ tidyr        1.3.2 
#> ✔ infer        1.1.0      ✔ tune         2.1.0 
#> ✔ modeldata    1.5.1      ✔ workflows    1.3.0 
#> ✔ parsnip      1.5.0      ✔ workflowsets 1.1.1 
#> ✔ purrr        1.2.2      ✔ yardstick    1.4.0
#> ── Conflicts ───────────────────────────────────────────────────────────────────────────────── tidymodels_conflicts() ──
#> ✖ purrr::discard() masks scales::discard()
#> ✖ dplyr::filter()  masks stats::filter()
#> ✖ dplyr::lag()     masks stats::lag()
#> ✖ recipes::step()  masks stats::step()
library(keras3)
#> 
#> Attaching package: 'keras3'
#> The following object is masked from 'package:yardstick':
#> 
#>     get_weights
#> The following object is masked from 'package:infer':
#> 
#>     generate

# 1. Define Layer Blocks (Required by kerasnip)
# The first block must initialize the sequential model
input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}

# Hidden layer block
dense_block <- function(model, units = 32) {
  model |> layer_dense(units = units, activation = "relu")
}

# Output layer block (units = 1 for regression)
output_block <- function(model, num_classes) {
  model |> layer_dense(units = 1)
}

# 2. Generate the parsnip specification
create_keras_sequential_spec(
  model_name = "my_mlp",
  layer_blocks = list(
    input = input_block,
    hidden = dense_block,
    output = output_block
  ),
  mode = "regression"
)

# 3. Use the newly created 'my_mlp' function
mod_spec <- my_mlp(fit_epochs = 10) |> 
  set_engine("keras")

# 4. Standard tidymodels workflow
rec_spec <- recipe(mpg ~ ., data = mtcars) |> 
  step_normalize(all_predictors())

fit_wf <- workflow() |> 
  add_recipe(rec_spec) |> 
  add_model(mod_spec) |> 
  fit(data = mtcars)

# Predict
new_data <- mtcars[1:3, ]
predict(fit_wf, new_data)
#> 1/1 - 0s - 116ms/step
#> # A tibble: 3 × 1
#>   .pred
#>   <dbl>
#> 1 1.00 
#> 2 0.956
#> 3 3.03
```

The first call to predict() detects that the Python pointer is invalid and restores the model from the stored bytes automatically.

## What kerasnip does behind the scenes

`kerasnip` handles persistence automatically:

- At **fit time**, the Keras model is serialized to a raw byte vector (`.keras`format) and stored alongside the parsnip `model_fit` object.
- At **predict time**, if the Python pointer is detected as invalid, `predict()` automatically restores the model from those bytes before dispatching.
- The parsnip model specification is also re-registered if it is missing from the session (e.g. after a fresh start).

This means you can use the persistence strategy that best suits your workflow without any extra boilerplate.

## Strategy 1: Plain `saveRDS()` / `readRDS()`

For most use cases: sharing a model file with a colleague, caching a fit between R sessions, or checkpointing during development; plain RDS is the simplest approach.


``` r
library(kerasnip)
library(workflows)
library(parsnip)
library(recipes)

# --- Save ---
saveRDS(fit_wf, "my_model.rds")

# --- Reload in the same or a new R session ---
library(kerasnip)
fit_wf <- readRDS("my_model.rds")

# predict() restores the Keras model from bytes automatically
predictions <- predict(fit_wf, new_data = new_data)
#> 1/1 - 0s - 118ms/step
predictions
#> # A tibble: 3 × 1
#>   .pred
#>   <dbl>
#> 1 1.00 
#> 2 0.956
#> 3 3.03
```

There is nothing special to do after `readRDS()`. The first call to `predict()` detects the invalid pointer, restores the model from the stored bytes, and then proceeds normally.

## Strategy 2: `bundle` / `unbundle`

The [`bundle`](https://rstudio.github.io/bundle/) package provides a standardized serialization interface used by `vetiver`, `plumber`, and other MLOps tools. It is the right choice when:

- You are deploying a model to a `vetiver` API or a Docker container.
- You want a self-contained, version-controlled artifact that does not rely on any R session state.
- You are sharing a model across machines with different R library paths.


``` r
library(kerasnip)
library(bundle)
library(workflows)

# --- Save ---
bundled <- bundle(fit_wf)
saveRDS(bundled, "my_model_bundle.rds")

# --- Reload in any R session ---
library(kerasnip)
library(bundle)
bundled <- readRDS("my_model_bundle.rds")
fit_wf <- unbundle(bundled)
predictions <- predict(fit_wf, new_data = new_data)
#> 1/1 - 0s - 110ms/step
predictions
#> # A tibble: 3 × 1
#>   .pred
#>   <dbl>
#> 1 1.00 
#> 2 0.956
#> 3 3.03
```

## Comparison

| | `saveRDS` / `readRDS` | `bundle` / `unbundle` |
|---|---|---|
| Works across sessions | ✅ | ✅ |
| Works across machines | ✅ (same R library) | ✅ |
| `vetiver` / Docker compatible | ❌ | ✅ |
| Extra dependency needed | ❌ | `bundle` package |
| Code complexity | Minimal | Minimal |

## What happens under the hood

When `kerasnip` fits a model, the generic fit function calls `keras_model_to_bytes()`, which writes the model to a temporary `.keras` file using `keras3::save_model()` and reads the bytes back into R:


``` r
# Simplified version of what happens inside generic_sequential_fit()
keras_bytes <- keras_model_to_bytes(model)
# keras_bytes is a raw vector stored in object$fit$keras_bytes
```

When `predict()` is called on a reloaded object, `predict.kerasnip_model_fit()` runs:


``` r
# Simplified version of predict.kerasnip_model_fit()
if (!is.null(object$fit$keras_bytes)) {
  is_valid <- tryCatch(
    {
      reticulate::py_validate_xptr(object$fit$fit)
      TRUE
    },
    error = function(e) FALSE
  )
  if (!is_valid) {
    object$fit$fit <- keras_model_from_bytes(object$fit$keras_bytes)
  }
}
```

If `keras_model_to_bytes()` fails (e.g. if the model was compiled with a non-serialisable custom object), a warning is issued at fit time and `keras_bytes` is set to `NULL`. In that case, `predict()` after reload will fail with a clear error.

---
title: "Tidymodels Workflow with Functional Keras Models (Multi-Input)"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tidymodels Workflow with Functional Keras Models (Multi-Input)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



## Introduction

This vignette demonstrates a complete `tidymodels` workflow for a regression task using a Keras functional model defined with `kerasnip`. We will use the Ames Housing dataset to predict house prices. A key feature of this example is the use of a multi-input Keras model, where numerical and categorical features are processed through separate input branches.

`kerasnip` allows you to define complex Keras architectures, including those with multiple inputs, and integrate them seamlessly into the `tidymodels` ecosystem for robust modeling and tuning.

## Setup

First, we load the necessary packages.


``` r
library(kerasnip)
library(tidymodels)
#> ── Attaching packages ────────────────────────────────────────────────────────────────────────────── tidymodels 1.5.0 ──
#> ✔ broom        1.0.12     ✔ recipes      1.3.2 
#> ✔ dials        1.4.3      ✔ rsample      1.3.2 
#> ✔ dplyr        1.2.1      ✔ tailor       0.1.0 
#> ✔ ggplot2      4.0.3      ✔ tidyr        1.3.2 
#> ✔ infer        1.1.0      ✔ tune         2.1.0 
#> ✔ modeldata    1.5.1      ✔ workflows    1.3.0 
#> ✔ parsnip      1.5.0      ✔ workflowsets 1.1.1 
#> ✔ purrr        1.2.2      ✔ yardstick    1.4.0
#> ── Conflicts ───────────────────────────────────────────────────────────────────────────────── tidymodels_conflicts() ──
#> ✖ purrr::discard() masks scales::discard()
#> ✖ dplyr::filter()  masks stats::filter()
#> ✖ dplyr::lag()     masks stats::lag()
#> ✖ recipes::step()  masks stats::step()
library(keras3)
#> 
#> Attaching package: 'keras3'
#> The following object is masked from 'package:yardstick':
#> 
#>     get_weights
#> The following object is masked from 'package:infer':
#> 
#>     generate
library(dplyr) # For data manipulation
library(ggplot2) # For plotting
library(future) # For parallel processing
#> 
#> Attaching package: 'future'
#> The following object is masked from 'package:keras3':
#> 
#>     %<-%
library(finetune) # For racing
```

## Data Preparation

We'll use the Ames Housing dataset, which is available in the `modeldata` package. We will then split the data into training and testing sets.


``` r
# Select relevant columns and remove rows with missing values
ames_df <- ames |>
  select(
    Sale_Price,
    Gr_Liv_Area,
    Year_Built,
    Neighborhood,
    Bldg_Type,
    Overall_Cond,
    Total_Bsmt_SF,
    contains("SF")
  ) |>
  na.omit()

# Split data into training and testing sets
set.seed(123)
ames_split <- initial_split(ames_df, prop = 0.8, strata = Sale_Price)
ames_train <- training(ames_split)
ames_test <- testing(ames_split)

# Create cross-validation folds for tuning
ames_folds <- vfold_cv(ames_train, v = 5, strata = Sale_Price)
```

## Recipe for Preprocessing

We will create a `recipes` object to preprocess our data. This recipe will:
*   Predict `Sale_Price` using all other variables.
*   Normalize all numerical predictors.
*   Create dummy variables for categorical predictors.
*   Collapse each group of predictors into a single matrix column using `step_collapse()`.

This final step is crucial for the multi-input Keras model, as the `kerasnip` functional API expects a list of matrices for multiple inputs, where each matrix corresponds to a distinct input layer.


``` r
ames_recipe <- recipe(Sale_Price ~ ., data = ames_train) |>
  step_normalize(all_numeric_predictors()) |>
  step_collapse(all_numeric_predictors(), new_col = "numerical_input") |>
  step_dummy(Neighborhood) |>
  step_collapse(starts_with("Neighborhood"), new_col = "neighborhood_input") |>
  step_dummy(Bldg_Type) |>
  step_collapse(starts_with("Bldg_Type"), new_col = "bldg_input") |>
  step_dummy(Overall_Cond) |>
  step_collapse(starts_with("Overall_Cond"), new_col = "condition_input")
```

## Define Keras Functional Model with `kerasnip`

Now, we define our Keras functional model using `kerasnip`'s layer blocks. This model will have four distinct input layers: one for numerical features and three for categorical features. These branches will be processed separately and then concatenated before the final output layer.


``` r
# Define layer blocks for multi-input functional model

# Input blocks for numerical and categorical features
input_numerical <- function(input_shape) {
  layer_input(shape = input_shape, name = "numerical_input")
}

input_neighborhood <- function(input_shape) {
  layer_input(shape = input_shape, name = "neighborhood_input")
}

input_bldg <- function(input_shape) {
  layer_input(shape = input_shape, name = "bldg_input")
}

input_condition <- function(input_shape) {
  layer_input(shape = input_shape, name = "condition_input")
}

# Processing blocks for each input type
dense_numerical <- function(tensor, units = 32, activation = "relu") {
  tensor |>
    layer_dense(units = units, activation = activation)
}

dense_categorical <- function(tensor, units = 16, activation = "relu") {
  tensor |>
    layer_dense(units = units, activation = activation)
}

# Concatenation block
concatenate_features <- function(numeric, neighborhood, bldg, condition) {
  layer_concatenate(list(numeric, neighborhood, bldg, condition))
}

# Output block for regression
output_regression <- function(tensor) {
  layer_dense(tensor, units = 1, name = "output")
}

# Create the kerasnip model specification function
create_keras_functional_spec(
  model_name = "ames_functional_mlp",
  layer_blocks = list(
    numerical_input = input_numerical,
    neighborhood_input = input_neighborhood,
    bldg_input = input_bldg,
    condition_input = input_condition,
    processed_numerical = inp_spec(dense_numerical, "numerical_input"),
    processed_neighborhood = inp_spec(dense_categorical, "neighborhood_input"),
    processed_bldg = inp_spec(dense_categorical, "bldg_input"),
    processed_condition = inp_spec(dense_categorical, "condition_input"),
    combined_features = inp_spec(
      concatenate_features,
      c(
        numeric = "processed_numerical",
        neighborhood = "processed_neighborhood",
        bldg = "processed_bldg",
        condition = "processed_condition"
      )
    ),
    output = inp_spec(output_regression, "combined_features")
  ),
  mode = "regression"
)
```

## Model Specification

We'll define our `ames_functional_mlp` model specification and set some hyperparameters to `tune()`. Note how the arguments are prefixed with their corresponding block names (e.g., `processed_numerical_units`).


``` r
# Define the tunable model specification
functional_mlp_spec <- ames_functional_mlp(
  # Tunable parameters for numerical branch
  processed_numerical_units = tune(),
  # Tunable parameters for categorical branch
  processed_neighborhood_units = tune(),
  processed_bldg_units = tune(),
  processed_condition_units = tune(),
  # Fixed compilation and fitting parameters
  compile_loss = "mean_squared_error",
  compile_optimizer = "adam",
  compile_metrics = c("mean_absolute_error"),
  fit_epochs = 50,
  fit_batch_size = 32,
  fit_validation_split = 0.2,
  fit_callbacks = list(
    callback_early_stopping(monitor = "val_loss", patience = 5)
  )
) |>
  set_engine("keras")

print(functional_mlp_spec)
#> ames functional mlp Model Specification (regression)
#> 
#> Main Arguments:
#>   num_numerical_input = structure(list(), class = "rlang_zap")
#>   num_neighborhood_input = structure(list(), class = "rlang_zap")
#>   num_bldg_input = structure(list(), class = "rlang_zap")
#>   num_condition_input = structure(list(), class = "rlang_zap")
#>   num_processed_numerical = structure(list(), class = "rlang_zap")
#>   num_processed_neighborhood = structure(list(), class = "rlang_zap")
#>   num_processed_bldg = structure(list(), class = "rlang_zap")
#>   num_processed_condition = structure(list(), class = "rlang_zap")
#>   num_combined_features = structure(list(), class = "rlang_zap")
#>   num_output = structure(list(), class = "rlang_zap")
#>   processed_numerical_units = tune()
#>   processed_numerical_activation = structure(list(), class = "rlang_zap")
#>   processed_neighborhood_units = tune()
#>   processed_neighborhood_activation = structure(list(), class = "rlang_zap")
#>   processed_bldg_units = tune()
#>   processed_bldg_activation = structure(list(), class = "rlang_zap")
#>   processed_condition_units = tune()
#>   processed_condition_activation = structure(list(), class = "rlang_zap")
#>   learn_rate = structure(list(), class = "rlang_zap")
#>   fit_batch_size = 32
#>   fit_epochs = 50
#>   fit_callbacks = list(callback_early_stopping(monitor = "val_loss", patience = 5))
#>   fit_validation_split = 0.2
#>   fit_validation_data = structure(list(), class = "rlang_zap")
#>   fit_shuffle = structure(list(), class = "rlang_zap")
#>   fit_class_weight = structure(list(), class = "rlang_zap")
#>   fit_sample_weight = structure(list(), class = "rlang_zap")
#>   fit_initial_epoch = structure(list(), class = "rlang_zap")
#>   fit_steps_per_epoch = structure(list(), class = "rlang_zap")
#>   fit_validation_steps = structure(list(), class = "rlang_zap")
#>   fit_validation_batch_size = structure(list(), class = "rlang_zap")
#>   fit_validation_freq = structure(list(), class = "rlang_zap")
#>   fit_verbose = structure(list(), class = "rlang_zap")
#>   fit_view_metrics = structure(list(), class = "rlang_zap")
#>   compile_optimizer = adam
#>   compile_loss = mean_squared_error
#>   compile_metrics = c("mean_absolute_error")
#>   compile_loss_weights = structure(list(), class = "rlang_zap")
#>   compile_weighted_metrics = structure(list(), class = "rlang_zap")
#>   compile_run_eagerly = structure(list(), class = "rlang_zap")
#>   compile_steps_per_execution = structure(list(), class = "rlang_zap")
#>   compile_jit_compile = structure(list(), class = "rlang_zap")
#>   compile_auto_scale_loss = structure(list(), class = "rlang_zap")
#> 
#> Computational engine: keras
```

## Create Workflow

A `workflow` combines the recipe and the model specification. 


``` r
ames_wf <- workflow() |>
  add_recipe(ames_recipe) |>
  add_model(functional_mlp_spec)

print(ames_wf)
#> ══ Workflow ════════════════════════════════════════════════════════════════════════════════════════════════════════════
#> Preprocessor: Recipe
#> Model: ames_functional_mlp()
#> 
#> ── Preprocessor ────────────────────────────────────────────────────────────────────────────────────────────────────────
#> 8 Recipe Steps
#> 
#> • step_normalize()
#> • step_collapse()
#> • step_dummy()
#> • step_collapse()
#> • step_dummy()
#> • step_collapse()
#> • step_dummy()
#> • step_collapse()
#> 
#> ── Model ───────────────────────────────────────────────────────────────────────────────────────────────────────────────
#> ames functional mlp Model Specification (regression)
#> 
#> Main Arguments:
#>   num_numerical_input = structure(list(), class = "rlang_zap")
#>   num_neighborhood_input = structure(list(), class = "rlang_zap")
#>   num_bldg_input = structure(list(), class = "rlang_zap")
#>   num_condition_input = structure(list(), class = "rlang_zap")
#>   num_processed_numerical = structure(list(), class = "rlang_zap")
#>   num_processed_neighborhood = structure(list(), class = "rlang_zap")
#>   num_processed_bldg = structure(list(), class = "rlang_zap")
#>   num_processed_condition = structure(list(), class = "rlang_zap")
#>   num_combined_features = structure(list(), class = "rlang_zap")
#>   num_output = structure(list(), class = "rlang_zap")
#>   processed_numerical_units = tune()
#>   processed_numerical_activation = structure(list(), class = "rlang_zap")
#>   processed_neighborhood_units = tune()
#>   processed_neighborhood_activation = structure(list(), class = "rlang_zap")
#>   processed_bldg_units = tune()
#>   processed_bldg_activation = structure(list(), class = "rlang_zap")
#>   processed_condition_units = tune()
#>   processed_condition_activation = structure(list(), class = "rlang_zap")
#>   learn_rate = structure(list(), class = "rlang_zap")
#>   fit_batch_size = 32
#>   fit_epochs = 50
#>   fit_callbacks = list(callback_early_stopping(monitor = "val_loss", patience = 5))
#>   fit_validation_split = 0.2
#>   fit_validation_data = structure(list(), class = "rlang_zap")
#>   fit_shuffle = structure(list(), class = "rlang_zap")
#>   fit_class_weight = structure(list(), class = "rlang_zap")
#>   fit_sample_weight = structure(list(), class = "rlang_zap")
#>   fit_initial_epoch = structure(list(), class = "rlang_zap")
#>   fit_steps_per_epoch = structure(list(), class = "rlang_zap")
#>   fit_validation_steps = structure(list(), class = "rlang_zap")
#>   fit_validation_batch_size = structure(list(), class = "rlang_zap")
#>   fit_validation_freq = structure(list(), class = "rlang_zap")
#>   fit_verbose = structure(list(), class = "rlang_zap")
#>   fit_view_metrics = structure(list(), class = "rlang_zap")
#>   compile_optimizer = adam
#>   compile_loss = mean_squared_error
#>   compile_metrics = c("mean_absolute_error")
#>   compile_loss_weights = structure(list(), class = "rlang_zap")
#>   compile_weighted_metrics = structure(list(), class = "rlang_zap")
#>   compile_run_eagerly = structure(list(), class = "rlang_zap")
#>   compile_steps_per_execution = structure(list(), class = "rlang_zap")
#>   compile_jit_compile = structure(list(), class = "rlang_zap")
#>   compile_auto_scale_loss = structure(list(), class = "rlang_zap")
#> 
#> Computational engine: keras
```

## Define Tuning Grid

We will create a regular grid for our hyperparameters.


``` r
# Define the tuning grid
params <- extract_parameter_set_dials(ames_wf) |>
  update(
    processed_numerical_units = hidden_units(range = c(32, 128)),
    processed_neighborhood_units = hidden_units(range = c(16, 64)),
    processed_bldg_units = hidden_units(range = c(16, 64)),
    processed_condition_units = hidden_units(range = c(16, 64))
  )
functional_mlp_grid <- grid_regular(params, levels = 3)

print(functional_mlp_grid)
#> # A tibble: 81 × 4
#>    processed_numerical_units processed_neighborhood_units processed_bldg_units processed_condition_units
#>                        <int>                        <int>                <int>                     <int>
#>  1                        32                           16                   16                        16
#>  2                        80                           16                   16                        16
#>  3                       128                           16                   16                        16
#>  4                        32                           40                   16                        16
#>  5                        80                           40                   16                        16
#>  6                       128                           40                   16                        16
#>  7                        32                           64                   16                        16
#>  8                        80                           64                   16                        16
#>  9                       128                           64                   16                        16
#> 10                        32                           16                   40                        16
#> # ℹ 71 more rows
```

## Tune Model

Now, we'll use `tune_race_anova()` to perform cross-validation and find the best hyperparameters.


``` r
# Note: Parallel processing with `plan(multisession)` is currently not working
# with Keras models due to backend conflicts

set.seed(123)
ames_tune_results <- tune_race_anova(
  ames_wf,
  resamples = ames_folds,
  grid = functional_mlp_grid,
  metrics = metric_set(rmse, mae, rsq),
  control = control_race(save_pred = TRUE, save_workflow = TRUE)
)
```

## Inspect Tuning Results

We can inspect the tuning results to see which hyperparameter combinations performed best.


``` r
# Show the best performing models based on RMSE
show_best(ames_tune_results, metric = "rmse", n = 5)
#> # A tibble: 2 × 10
#>   processed_numerical_units processed_neighborho…¹ processed_bldg_units processed_condition_…² .metric .estimator   mean
#>                       <int>                  <int>                <int>                  <int> <chr>   <chr>       <dbl>
#> 1                       128                     64                   64                     64 rmse    standard   53524.
#> 2                       128                     64                   40                     64 rmse    standard   54215.
#> # ℹ abbreviated names: ¹​processed_neighborhood_units, ²​processed_condition_units
#> # ℹ 3 more variables: n <int>, std_err <dbl>, .config <chr>

# Autoplot the results
# Currently does not work due to a label issue: autoplot(ames_tune_results)

# Select the best hyperparameters
best_functional_mlp_params <- select_best(ames_tune_results, metric = "rmse")
print(best_functional_mlp_params)
#> # A tibble: 1 × 5
#>   processed_numerical_units processed_neighborhood_units processed_bldg_units processed_condition_units .config         
#>                       <int>                        <int>                <int>                     <int> <chr>           
#> 1                       128                           64                   64                        64 pre0_mod81_post0
```

## Finalize Workflow and Fit Model

Once we have the best hyperparameters, we finalize the workflow and fit the model on the entire training dataset.


``` r
# Finalize the workflow with the best hyperparameters
final_ames_wf <- finalize_workflow(ames_wf, best_functional_mlp_params)

# Fit the final model on the full training data
final_ames_fit <- fit(final_ames_wf, data = ames_train)

print(final_ames_fit)
#> ══ Workflow [trained] ══════════════════════════════════════════════════════════════════════════════════════════════════
#> Preprocessor: Recipe
#> Model: ames_functional_mlp()
#> 
#> ── Preprocessor ────────────────────────────────────────────────────────────────────────────────────────────────────────
#> 8 Recipe Steps
#> 
#> • step_normalize()
#> • step_collapse()
#> • step_dummy()
#> • step_collapse()
#> • step_dummy()
#> • step_collapse()
#> • step_dummy()
#> • step_collapse()
#> 
#> ── Model ───────────────────────────────────────────────────────────────────────────────────────────────────────────────
#> $fit
#> Model: "functional_262"
#> ┌───────────────────────────────────┬──────────────────────────────┬───────────────────┬───────────────────────────────
#> │ Layer (type)                      │ Output Shape                 │           Param # │ Connected to                  
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ numerical_input (InputLayer)      │ (None, 1, 10)                │                 0 │ -                             
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ neighborhood_input (InputLayer)   │ (None, 1, 28)                │                 0 │ -                             
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ bldg_input (InputLayer)           │ (None, 1, 4)                 │                 0 │ -                             
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ condition_input (InputLayer)      │ (None, 1, 9)                 │                 0 │ -                             
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ dense_1033 (Dense)                │ (None, 1, 128)               │             1,408 │ numerical_input[0][0]         
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ dense_1034 (Dense)                │ (None, 1, 64)                │             1,856 │ neighborhood_input[0][0]      
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ dense_1035 (Dense)                │ (None, 1, 64)                │               320 │ bldg_input[0][0]              
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ dense_1036 (Dense)                │ (None, 1, 64)                │               640 │ condition_input[0][0]         
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ concatenate_258 (Concatenate)     │ (None, 1, 320)               │                 0 │ dense_1033[0][0],             
#> │                                   │                              │                   │ dense_1034[0][0],             
#> │                                   │                              │                   │ dense_1035[0][0],             
#> │                                   │                              │                   │ dense_1036[0][0]              
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ output (Dense)                    │ (None, 1, 1)                 │               321 │ concatenate_258[0][0]         
#> └───────────────────────────────────┴──────────────────────────────┴───────────────────┴───────────────────────────────
#>  Total params: 13,637 (53.27 KB)
#>  Trainable params: 4,545 (17.75 KB)
#>  Non-trainable params: 0 (0.00 B)
#>  Optimizer params: 9,092 (35.52 KB)
#> 
#> $keras_bytes
#>     [1] 50 4b 03 04 14 00 00 00 00 00 00 00 21 00 39 22 4e 35 40 00 00 00 40 00 00 00 0d 00 00 00 6d 65 74 61 64 61 74
#>    [38] 61 2e 6a 73 6f 6e 7b 22 6b 65 72 61 73 5f 76 65 72 73 69 6f 6e 22 3a 20 22 33 2e 31 34 2e 30 22 2c 20 22 64 61
#>    [75] 74 65 5f 73 61 76 65 64 22 3a 20 22 32 30 32 36 2d 30 35 2d 30 31 40 31 32 3a 35 38 3a 31 31 22 7d 50 4b 03 04
#>   [112] 14 00 00 00 00 00 00 00 21 00 c3 b3 02 c8 41 20 00 00 41 20 00 00 0b 00 00 00 63 6f 6e 66 69 67 2e 6a 73 6f 6e
#>   [149] 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 2e 73 72 63 2e 6d 6f 64 65 6c 73 2e 66 75 6e 63 74 69 6f 6e
#>   [186] 61 6c 22 2c 20 22 63 6c 61 73 73 5f 6e 61 6d 65 22 3a 20 22 46 75 6e 63 74 69 6f 6e 61 6c 22 2c 20 22 63 6f 6e
#>   [223] 66 69 67 22 3a 20 7b 22 6e 61 6d 65 22 3a 20 22 66 75 6e 63 74 69 6f 6e 61 6c 5f 32 36 32 22 2c 20 22 74 72 61
#>   [260] 69 6e 61 62 6c 65 22 3a 20 74 72 75 65 2c 20 22 6c 61 79 65 72 73 22 3a 20 5b 7b 22 6d 6f 64 75 6c 65 22 3a 20
#>   [297] 22 6b 65 72 61 73 2e 6c 61 79 65 72 73 22 2c 20 22 63 6c 61 73 73 5f 6e 61 6d 65 22 3a 20 22 49 6e 70 75 74 4c
#>   [334] 61 79 65 72 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 62 61 74 63 68 5f 73 68 61 70 65 22 3a 20 5b 6e 75 6c
#>   [371] 6c 2c 20 31 2c 20 31 30 5d 2c 20 22 64 74 79 70 65 22 3a 20 22 66 6c 6f 61 74 33 32 22 2c 20 22 73 70 61 72 73
#>   [408] 65 22 3a 20 66 61 6c 73 65 2c 20 22 72 61 67 67 65 64 22 3a 20 66 61 6c 73 65 2c 20 22 6e 61 6d 65 22 3a 20 22
#>   [445] 6e 75 6d 65 72 69 63 61 6c 5f 69 6e 70 75 74 22 2c 20 22 6f 70 74 69 6f 6e 61 6c 22 3a 20 66 61 6c 73 65 7d 2c
#>   [482] 20 22 72 65 67 69 73 74 65 72 65 64 5f 6e 61 6d 65 22 3a 20 6e 75 6c 6c 2c 20 22 6e 61 6d 65 22 3a 20 22 6e 75
#>   [519] 6d 65 72 69 63 61 6c 5f 69 6e 70 75 74 22 2c 20 22 69 6e 62 6f 75 6e 64 5f 6e 6f 64 65 73 22 3a 20 5b 5d 7d 2c
#>   [556] 20 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 2e 6c 61 79 65 72 73 22 2c 20 22 63 6c 61 73 73 5f 6e 61
#> 
#> ...
#> and 2790 more lines.
```

### Inspect Final Model

You can extract the underlying Keras model and its training history for further inspection.


``` r
# Extract the Keras model summary
final_ames_fit |>
  extract_fit_parsnip() |>
  extract_keras_model() |>
  summary()
#> Model: "functional_262"
#> ┌───────────────────────────────────┬──────────────────────────────┬───────────────────┬───────────────────────────────
#> │ Layer (type)                      │ Output Shape                 │           Param # │ Connected to                  
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ numerical_input (InputLayer)      │ (None, 1, 10)                │                 0 │ -                             
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ neighborhood_input (InputLayer)   │ (None, 1, 28)                │                 0 │ -                             
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ bldg_input (InputLayer)           │ (None, 1, 4)                 │                 0 │ -                             
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ condition_input (InputLayer)      │ (None, 1, 9)                 │                 0 │ -                             
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ dense_1033 (Dense)                │ (None, 1, 128)               │             1,408 │ numerical_input[0][0]         
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ dense_1034 (Dense)                │ (None, 1, 64)                │             1,856 │ neighborhood_input[0][0]      
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ dense_1035 (Dense)                │ (None, 1, 64)                │               320 │ bldg_input[0][0]              
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ dense_1036 (Dense)                │ (None, 1, 64)                │               640 │ condition_input[0][0]         
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ concatenate_258 (Concatenate)     │ (None, 1, 320)               │                 0 │ dense_1033[0][0],             
#> │                                   │                              │                   │ dense_1034[0][0],             
#> │                                   │                              │                   │ dense_1035[0][0],             
#> │                                   │                              │                   │ dense_1036[0][0]              
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ output (Dense)                    │ (None, 1, 1)                 │               321 │ concatenate_258[0][0]         
#> └───────────────────────────────────┴──────────────────────────────┴───────────────────┴───────────────────────────────
#>  Total params: 13,637 (53.27 KB)
#>  Trainable params: 4,545 (17.75 KB)
#>  Non-trainable params: 0 (0.00 B)
#>  Optimizer params: 9,092 (35.52 KB)
```


``` r
# Plot the Keras model
final_ames_fit |>
  extract_fit_parsnip() |>
  extract_keras_model() |>
  plot(show_shapes = TRUE)
```

![Model](images/model_plot_shapes_wf.png){fig-alt="A picture showing the model shape"}


``` r
# Plot the training history
final_ames_fit |>
  extract_fit_parsnip() |>
  extract_keras_history() |>
  plot()
```

![plot of chunk inspect-final-keras-model-history](figure/inspect-final-keras-model-history-1.png)

## Make Predictions and Evaluate

Finally, we will make predictions on the test set and evaluate the model's performance.


``` r
# Make predictions on the test set
ames_test_pred <- predict(final_ames_fit, new_data = ames_test)
#> 19/19 - 0s - 10ms/step

# Combine predictions with actuals
ames_results <- tibble::tibble(
  Sale_Price = ames_test$Sale_Price,
  .pred = ames_test_pred$.pred
)

print(head(ames_results))
#> # A tibble: 6 × 2
#>   Sale_Price   .pred
#>        <int>   <dbl>
#> 1     189900 193909.
#> 2     195500 195484.
#> 3     236500 234049.
#> 4     212000 217096.
#> 5     210000 241706.
#> 6     142000 126019.

# Evaluate performance using yardstick metrics
metrics_results <- metric_set(
  rmse,
  mae,
  rsq
)(
  ames_results,
  truth = Sale_Price,
  estimate = .pred
)

print(metrics_results)
#> # A tibble: 3 × 3
#>   .metric .estimator .estimate
#>   <chr>   <chr>          <dbl>
#> 1 rmse    standard   44687.   
#> 2 mae     standard   27910.   
#> 3 rsq     standard       0.767
```

## Saving and Reloading Your Model

`kerasnip` serializes the Keras model weights to bytes at fit time and stores them alongside the workflow object. This means plain `saveRDS()` / `readRDS()` **works out of the box** — the underlying Keras model is restored automatically the first time `predict()` is called on the reloaded object.


``` r
# Save the FINAL fitted workflow
saveRDS(final_ames_fit, "ames_model.rds")

# Reload — no extra steps needed
final_ames_fit_loaded <- readRDS("ames_model.rds")

# Make predictions again to prove it works
predict(final_ames_fit_loaded, new_data = ames_test) |> head()
#> 19/19 - 0s - 11ms/step
#> # A tibble: 6 × 1
#>     .pred
#>     <dbl>
#> 1 193909.
#> 2 195484.
#> 3 234049.
#> 4 217096.
#> 5 241706.
#> 6 126019.
```

If you need a fully self-contained bundle suitable for deployment with `vetiver` or other MLOps tools, use `bundle::bundle()` instead:


``` r
library(bundle)

# Save as a portable bundle
bundled <- bundle(final_ames_fit)
saveRDS(bundled, "ames_model_bundle.rds")

# Reload in any R session
library(kerasnip)
library(bundle)
final_ames_fit_loaded <- unbundle(readRDS("ames_model_bundle.rds"))

predict(final_ames_fit_loaded, new_data = ames_test) |> head()
#> 19/19 - 0s - 9ms/step
#> # A tibble: 6 × 1
#>     .pred
#>     <dbl>
#> 1 193909.
#> 2 195484.
#> 3 234049.
#> 4 217096.
#> 5 241706.
#> 6 126019.
```

See `vignette("saving_and_reloading")` for a detailed comparison of both approaches.

---
title: The Sequential Model with kerasnip
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{The Sequential Model with kerasnip}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction

This vignette provides a comprehensive guide to using `kerasnip` to define sequential Keras models within the `tidymodels` ecosystem. `kerasnip` bridges the gap between the imperative, layer-by-layer construction of Keras models and the declarative, specification-based approach of `tidymodels`.

Here, we will focus on `create_keras_sequential_spec()`, which is ideal for models where layers form a plain stack, with each layer having exactly one input tensor and one output tensor.

## Setup

We'll start by loading the necessary packages:




``` r
library(kerasnip)
library(tidymodels)
#> ── Attaching packages ────────────────────────────────────────────────────────────────────────────── tidymodels 1.5.0 ──
#> ✔ broom        1.0.12     ✔ recipes      1.3.2 
#> ✔ dials        1.4.3      ✔ rsample      1.3.2 
#> ✔ dplyr        1.2.1      ✔ tailor       0.1.0 
#> ✔ ggplot2      4.0.3      ✔ tidyr        1.3.2 
#> ✔ infer        1.1.0      ✔ tune         2.1.0 
#> ✔ modeldata    1.5.1      ✔ workflows    1.3.0 
#> ✔ parsnip      1.5.0      ✔ workflowsets 1.1.1 
#> ✔ purrr        1.2.2      ✔ yardstick    1.4.0
#> ── Conflicts ───────────────────────────────────────────────────────────────────────────────── tidymodels_conflicts() ──
#> ✖ purrr::discard() masks scales::discard()
#> ✖ dplyr::filter()  masks stats::filter()
#> ✖ dplyr::lag()     masks stats::lag()
#> ✖ recipes::step()  masks stats::step()
library(keras3)
#> 
#> Attaching package: 'keras3'
#> The following object is masked from 'package:yardstick':
#> 
#>     get_weights
#> The following object is masked from 'package:infer':
#> 
#>     generate
```

## When to use `create_keras_sequential_spec()`

A `Sequential` model in Keras is appropriate for a plain stack of layers where each layer has exactly one input tensor and one output tensor. `kerasnip`'s `create_keras_sequential_spec()` function is designed to define such models in a `tidymodels`-compatible way.

Instead of building the model layer-by-layer imperatively, you define a named, ordered list of R functions called `layer_blocks`. Each `layer_block` function takes a Keras model object as its first argument and returns the modified model. `kerasnip` then uses these blocks to construct the full Keras Sequential model.

For models with more complex, non-linear topologies (e.g., multiple inputs/outputs, residual connections, or multi-branch models), you should use `create_keras_functional_spec()`.

## Creating a `kerasnip` Sequential Model Specification

Let's define a simple sequential model with three dense layers.

First, we define our `layer_blocks`:


``` r
# The first block must initialize the model. `input_shape`
# is passed automatically.
input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}

# A reusable block for hidden layers. `units` will become a tunable parameter.
hidden_block <- function(model, units = 32, activation = "relu") {
  model |> layer_dense(units = units, activation = activation)
}

# The output block. `num_classes` is passed automatically for classification.
output_block <- function(model, num_classes, activation = "softmax") {
  model |> layer_dense(units = num_classes, activation = activation)
}
```

Now, we use `create_keras_sequential_spec()` to generate our `parsnip` model specification function. We'll name our model `my_simple_mlp`.


``` r
create_keras_sequential_spec(
  model_name = "my_simple_mlp",
  layer_blocks = list(
    input = input_block,
    hidden_1 = hidden_block,
    hidden_2 = hidden_block,
    output = output_block
  ),
  mode = "classification"
)
```

## A common debugging workflow: `compile_keras_grid()`

In the original Keras guide, a common workflow is to incrementally add layers and call `summary()` to inspect the architecture. With `kerasnip`, the model is defined declaratively, so we can't inspect it layer-by-layer in the same way.

However, `kerasnip` provides a powerful equivalent: `compile_keras_grid()`. This function checks if your `layer_blocks` define a valid Keras model and returns the compiled model structure, all without running a full training cycle. This is perfect for debugging your architecture.

Let's see this in action with a CNN architecture:


``` r
# Define CNN layer blocks
cnn_input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}
cnn_conv_block <- function(
  model,
  filters = 32,
  kernel_size = 3,
  activation = "relu"
) {
  model |>
    layer_conv_2d(
      filters = filters,
      kernel_size = kernel_size,
      activation = activation
    )
}
cnn_pool_block <- function(model, pool_size = 2) {
  model |> layer_max_pooling_2d(pool_size = pool_size)
}
cnn_flatten_block <- function(model) {
  model |> layer_flatten()
}
cnn_output_block <- function(model, num_classes, activation = "softmax") {
  model |> layer_dense(units = num_classes, activation = activation)
}

# Create the kerasnip spec function
create_keras_sequential_spec(
  model_name = "my_cnn",
  layer_blocks = list(
    input = cnn_input_block,
    conv1 = cnn_conv_block,
    pool1 = cnn_pool_block,
    flatten = cnn_flatten_block,
    output = cnn_output_block
  ),
  mode = "classification"
)

# Create a spec instance for a 28x28x1 image
cnn_spec <- my_cnn(
  conv1_filters = 32, conv1_kernel_size = 5,
  compile_loss = "categorical_crossentropy",
  compile_optimizer = "adam"
)

# Prepare dummy data with the correct shape.
# We create a list of 28x28x1 arrays.
x_dummy_list <- lapply(
  1:10,
  function(i) array(runif(28 * 28 * 1), dim = c(28, 28, 1))
)
x_dummy_df <- tibble::tibble(x = x_dummy_list)
y_dummy <- factor(sample(0:9, 10, replace = TRUE), levels = 0:9)
y_dummy_df <- tibble::tibble(y = y_dummy)


# Use compile_keras_grid to get the model summary
# A one-row grid with no extra hyperparameters builds the spec as-is.
compilation_results <- compile_keras_grid(
  spec = cnn_spec,
  grid = tibble::tibble(.rows = 1L),
  x = x_dummy_df,
  y = y_dummy_df
)

# Print the summary
compilation_results |>
  select(compiled_model) |>
  pull() |>
  pluck(1) |>
  summary()
#> Model: "sequential"
#> ┌─────────────────────────────────────────────────────┬────────────────────────────────────────┬───────────────────────
#> │ Layer (type)                                        │ Output Shape                           │               Param # 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ conv2d (Conv2D)                                     │ (None, 24, 24, 32)                     │                   832 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ max_pooling2d (MaxPooling2D)                        │ (None, 12, 12, 32)                     │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ flatten (Flatten)                                   │ (None, 4608)                           │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense (Dense)                                       │ (None, 10)                             │                46,090 
#> └─────────────────────────────────────────────────────┴────────────────────────────────────────┴───────────────────────
#>  Total params: 46,922 (183.29 KB)
#>  Trainable params: 46,922 (183.29 KB)
#>  Non-trainable params: 0 (0.00 B)
```


``` r
compilation_results |>
  select(compiled_model) |>
  pull() |>
  pluck(1) |>
  plot(show_shapes = TRUE)
```

![model](images/model_plot_shapes_s.png){fig-alt="A picture showing the model shape"}

---
title: "Getting Started with kerasnip"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with kerasnip}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



## The Core Idea: From Keras Layers to Tidymodels Specs

The `keras3` package allows for building deep learning models layer-by-layer, which is a powerful and flexible approach. However, the `tidymodels` ecosystem is designed around declarative model specifications, where you define *what* model you want and which of its parameters you want to tune, rather than building it imperatively.

`kerasnip` bridges this gap with a simple but powerful concept: **layer blocks**. You define the components of your neural network (e.g., an input block, a dense block, a dropout block) as simple R functions. `kerasnip` then uses these blocks as building materials to create a brand new `parsnip` model specification function for you.

This new function behaves just like any other `parsnip` model (e.g., `rand_forest()` or `linear_reg()`), making it easy to integrate into `tidymodels` workflows.

## Installation

You can install the development version of `kerasnip` from GitHub. You will also need `keras3`.


``` r
install.packages("pak")
pak::pak("davidrsch/kerasnip")
pak::pak("rstudio/keras3")

# Install the backend
keras3::install_keras()
```

We’ll start by loading `kerasnip`, `tidymodels` and `keras3`:


``` r
library(kerasnip)
library(tidymodels)
#> ── Attaching packages ────────────────────────────────────────────────────────────────────────────── tidymodels 1.5.0 ──
#> ✔ broom        1.0.12     ✔ recipes      1.3.2 
#> ✔ dials        1.4.3      ✔ rsample      1.3.2 
#> ✔ dplyr        1.2.1      ✔ tailor       0.1.0 
#> ✔ ggplot2      4.0.3      ✔ tidyr        1.3.2 
#> ✔ infer        1.1.0      ✔ tune         2.1.0 
#> ✔ modeldata    1.5.1      ✔ workflows    1.3.0 
#> ✔ parsnip      1.5.0      ✔ workflowsets 1.1.1 
#> ✔ purrr        1.2.2      ✔ yardstick    1.4.0
#> ── Conflicts ───────────────────────────────────────────────────────────────────────────────── tidymodels_conflicts() ──
#> ✖ purrr::discard() masks scales::discard()
#> ✖ dplyr::filter()  masks stats::filter()
#> ✖ dplyr::lag()     masks stats::lag()
#> ✖ recipes::step()  masks stats::step()
library(keras3)
#> 
#> Attaching package: 'keras3'
#> The following object is masked from 'package:yardstick':
#> 
#>     get_weights
#> The following object is masked from 'package:infer':
#> 
#>     generate
```

## A `kerasnip` MNIST Example

Let’s replicate the classic Keras introductory example, training a simple MLP on the MNIST dataset, but using the `kerasnip` workflow. This will demonstrate how to translate a standard Keras model into a reusable, modular `parsnip` specification.

If you’re familiar with Keras, you’ll recognize the structure; if not, this is a perfect place to start. We’ll begin by learning the basics through a simple task: recognizing handwritten digits from the MNIST dataset.

The MNIST dataset contains 28×28 pixel grayscale images of handwritten digits, like these:

![MINIST](images/MNIST.png){fig-alt="A picture showing grayscale images of handwritten digits (5, 0, 4 and 1)"}

Each image comes with a label indicating which digit it represents. For example, the labels for the images above might be 5, 0, 4, and 1.

### Preparing the Data

This step is identical to any other Keras model. We load the MNIST dataset, reshape the predictors, and convert the outcome to a factor for `tidymodels`.


``` r
mnist <- dataset_mnist()
x_train <- mnist$train$x
y_train <- mnist$train$y
x_test <- mnist$test$x
y_test <- mnist$test$y

# Reshape
x_train <- array_reshape(x_train, c(nrow(x_train), 784))
x_test <- array_reshape(x_test, c(nrow(x_test), 784))
# Rescale
x_train <- x_train / 255
x_test <- x_test / 255

# Convert outcomes to factors for tidymodels
# kerasnip will handle y convertion internally using keras3::to_categorical()
y_train_factor <- factor(y_train)
y_test_factor <- factor(y_test)

# For tidymodels, it's best to work with data frames
# Use I() to keep the matrix structure of x within the data frame
train_df <- data.frame(x = I(x_train), y = y_train_factor)
test_df <- data.frame(x = I(x_test), y = y_test_factor)
```

### The Standard Keras Approach (for comparison)

Before diving into the `kerasnip` workflow, let's quickly look at how this same model is built using standard `keras3` code. This will help highlight the different approach `kerasnip` enables.


``` r
# The standard Keras3 approach
model <- keras_model_sequential(input_shape = 784) |>
  layer_dense(units = 256, activation = "relu") |>
  layer_dropout(rate = 0.4) |>
  layer_dense(units = 128, activation = "relu") |>
  layer_dropout(rate = 0.3) |>
  layer_dense(units = 10, activation = "softmax")

summary(model)

model |>
  compile(
    loss = "categorical_crossentropy",
    optimizer = optimizer_rmsprop(),
    metrics = "accuracy"
  )

# The model would then be trained with model |> fit(...)
```

The code above is imperative: you define each layer and add it to the model step-by-step. Now, let's see how `kerasnip` approaches this by defining reusable components for a declarative, `tidymodels`-friendly workflow.

### Defining the Model with Reusable Blocks

The original Keras example interleaves `layer_dense()` and `layer_dropout()`. With `kerasnip`, we can encapsulate this pattern into a single, reusable block. This makes the overall architecture cleaner and more modular.


``` r
# An input block to initialize the model.
# The 'model' argument is supplied implicitly by the kerasnip backend.
mlp_input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}

# A reusable "module" that combines a dense layer and a dropout layer.
# All arguments that should be tunable need a default value.
dense_dropout_block <- function(model, units = 128, rate = 0.1) {
  model |>
    layer_dense(units = units, activation = "relu") |>
    layer_dropout(rate = rate)
}

# The output block for classification.
mlp_output_block <- function(model, num_classes) {
  model |> layer_dense(units = num_classes, activation = "softmax")
}
```

Now, we use `create_keras_sequential_spec()` to generate our `parsnip` model function.


``` r
create_keras_sequential_spec(
  model_name = "mnist_mlp",
  layer_blocks = list(
    input = mlp_input_block,
    hidden_1 = dense_dropout_block,
    hidden_2 = dense_dropout_block,
    output = mlp_output_block
  ),
  mode = "classification"
)
```

### Building and Fitting the Model

We can now use our new `mnist_mlp()` function. Notice how its arguments, such as `hidden_1_units` and `hidden_1_rate`, were automatically generated by `kerasnip`. The names are created by combining the name of the layer block (e.g., `hidden_1`) with the arguments of that block's function (e.g., `units`, `rate`).

To replicate the `keras3` example, we'll use both `hidden` blocks and provide their parameters.


``` r
mlp_spec <- mnist_mlp(
  hidden_1_units = 256,
  hidden_1_rate = 0.4,
  hidden_2_rate = 0.3,
  hidden_2_units =  128,
  compile_loss = "categorical_crossentropy",
  compile_optimizer = optimizer_rmsprop(),
  compile_metrics = c("accuracy"),
  fit_epochs = 30,
  fit_batch_size = 128,
  fit_validation_split = 0.2
) |>
  set_engine("keras")

# Fit the model
mlp_fit <- fit(mlp_spec, y ~ x, data = train_df)
```


``` r
mlp_fit |>
  extract_keras_model() |>
  summary()
#> Model: "sequential"
#> ┌─────────────────────────────────────────────────────┬────────────────────────────────────────┬───────────────────────
#> │ Layer (type)                                        │ Output Shape                           │               Param # 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense (Dense)                                       │ (None, 256)                            │               200,960 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dropout (Dropout)                                   │ (None, 256)                            │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_1 (Dense)                                     │ (None, 128)                            │                32,896 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dropout_1 (Dropout)                                 │ (None, 128)                            │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_2 (Dense)                                     │ (None, 10)                             │                 1,290 
#> └─────────────────────────────────────────────────────┴────────────────────────────────────────┴───────────────────────
#>  Total params: 470,294 (1.79 MB)
#>  Trainable params: 235,146 (918.54 KB)
#>  Non-trainable params: 0 (0.00 B)
#>  Optimizer params: 235,148 (918.55 KB)
```


``` r
mlp_fit |>
  extract_keras_model() |>
  plot(show_shapes = TRUE)
```

![model](images/model_plot_shapes_gs.png){fig-alt="A picture showing the model shape"}


``` r
mlp_fit |>
  extract_keras_history() |>
  plot()
```

![plot of chunk model-fit-history](figure/model-fit-history-1.png)

### Evaluating Model Performance

The `keras_evaluate()` function provides a straightforward way to assess the model's performance on a test set, using the underlying `keras3::evaluate()` method. It returns the loss and any other metrics that were specified during the model compilation step.


``` r
mlp_fit |> keras_evaluate(x_test, y_test)
#> 313/313 - 1s - 2ms/step - accuracy: 0.9816 - loss: 0.0933
#> $accuracy
#> [1] 0.9816
#> 
#> $loss
#> [1] 0.09331781
```

### Making Predictions

Once the model is trained, we can use the standard `tidymodels` `predict()` function to generate predictions on new data. By default, `predict()` on a `parsnip` classification model returns the predicted class labels.


``` r
# Predict the class for the first 5 images in the test set
class_preds <- mlp_fit |>
  predict(new_data = head(select(test_df, x)))
#> 1/1 - 0s - 112ms/step
class_preds
#> # A tibble: 6 × 1
#>   .pred_class
#>   <fct>      
#> 1 7          
#> 2 2          
#> 3 1          
#> 4 0          
#> 5 4          
#> 6 1
```

To get the underlying probabilities for each class, we can set `type = "prob"`. This returns a tibble with a probability column for each of the 10 classes (0-9).


``` r
# Predict probabilities for the first 5 images
prob_preds <- mlp_fit |>
  predict(new_data = head(select(test_df, x)), type = "prob")
#> 1/1 - 0s - 79ms/step
prob_preds
#> # A tibble: 6 × 10
#>     .pred_0   .pred_1  .pred_2  .pred_3   .pred_4  .pred_5  .pred_6  .pred_7  .pred_8  .pred_9
#>       <dbl>     <dbl>    <dbl>    <dbl>     <dbl>    <dbl>    <dbl>    <dbl>    <dbl>    <dbl>
#> 1 5.50 e-17 2.57 e-15 2.56e-11 1.41e- 9 3.52 e-19 1.40e-14 3.42e-27 1   e+ 0 7.46e-19 1.19e-10
#> 2 7.14 e-19 1.02 e-12 1   e+ 0 2.66e-13 4.80 e-25 3.38e-22 8.05e-18 8.01e-20 2.42e-18 2.33e-27
#> 3 1.17 e-11 1.000e+ 0 8.86e- 9 1.80e-11 4.10 e- 8 3.54e-12 2.33e- 9 7.20e- 8 1.40e- 8 8.54e-11
#> 4 1.000e+ 0 5.15 e-11 8.54e- 9 3.70e-12 3.02 e-11 8.76e-10 1.26e- 7 1.23e-10 8.62e-12 9.42e-10
#> 5 3.89 e-11 3.38 e-11 1.01e-11 3.33e-14 1.000e+ 0 1.75e-13 2.33e-11 1.73e- 7 5.30e-14 2.48e- 7
#> 6 1.88 e-12 1.000e+ 0 1.17e-10 1.25e-12 5.62 e- 8 3.72e-14 2.97e-10 2.18e- 8 3.60e- 9 2.09e-11
```

We can then compare the predicted class to the actual class for these images to see how the model is performing.


``` r
# Combine predictions with actuals for comparison
comparison <- bind_cols(
  class_preds,
  prob_preds
) |>
  bind_cols(
    head(test_df[, "y", drop = FALSE])
  )
comparison
#> # A tibble: 6 × 12
#>   .pred_class   .pred_0   .pred_1  .pred_2  .pred_3   .pred_4  .pred_5  .pred_6  .pred_7  .pred_8  .pred_9 y    
#>   <fct>           <dbl>     <dbl>    <dbl>    <dbl>     <dbl>    <dbl>    <dbl>    <dbl>    <dbl>    <dbl> <fct>
#> 1 7           5.50 e-17 2.57 e-15 2.56e-11 1.41e- 9 3.52 e-19 1.40e-14 3.42e-27 1   e+ 0 7.46e-19 1.19e-10 7    
#> 2 2           7.14 e-19 1.02 e-12 1   e+ 0 2.66e-13 4.80 e-25 3.38e-22 8.05e-18 8.01e-20 2.42e-18 2.33e-27 2    
#> 3 1           1.17 e-11 1.000e+ 0 8.86e- 9 1.80e-11 4.10 e- 8 3.54e-12 2.33e- 9 7.20e- 8 1.40e- 8 8.54e-11 1    
#> 4 0           1.000e+ 0 5.15 e-11 8.54e- 9 3.70e-12 3.02 e-11 8.76e-10 1.26e- 7 1.23e-10 8.62e-12 9.42e-10 0    
#> 5 4           3.89 e-11 3.38 e-11 1.01e-11 3.33e-14 1.000e+ 0 1.75e-13 2.33e-11 1.73e- 7 5.30e-14 2.48e- 7 4    
#> 6 1           1.88 e-12 1.000e+ 0 1.17e-10 1.25e-12 5.62 e- 8 3.72e-14 2.97e-10 2.18e- 8 3.60e- 9 2.09e-11 1
```

## Example 2: Tuning the Model Architecture

Now we’ll showcase the main strength of `kerasnip`: tuning the network architecture itself. We can treat the number of layers, and the parameters of those layers, as hyperparameters to be optimized by `tune`.

Using the `mnist_mlp` spec we just created, let's define a tunable model.


``` r
# Define a tunable specification
# We set num_hidden_2 = 0 to disable the second hidden block
# for this tuning example
tune_spec <- mnist_mlp(
  num_hidden_1 = tune(),
  hidden_1_units = tune(),
  hidden_1_rate = tune(),
  num_hidden_2 = 0,
  compile_loss = "categorical_crossentropy",
  compile_optimizer = optimizer_rmsprop(),
  compile_metrics = c("accuracy"),
  fit_epochs = 30,
  fit_batch_size = 128,
  fit_validation_split = 0.2
) |>
  set_engine("keras")

# Create a workflow
tune_wf <- workflow(y ~ x, tune_spec)
```

Next, we define the search space for our tunable parameters using `dials`.


``` r
# Define the tuning grid
params <- extract_parameter_set_dials(tune_wf) |>
  update(
    num_hidden_1 = dials::num_terms(c(1, 3)),
    hidden_1_units = dials::hidden_units(c(64, 256)),
    hidden_1_rate = dials::dropout(c(0.2, 0.4))
  )
grid <- grid_regular(params, levels = 3)
grid
#> # A tibble: 27 × 3
#>    num_hidden_1 hidden_1_units hidden_1_rate
#>           <int>          <int>         <dbl>
#>  1            1             64           0.2
#>  2            2             64           0.2
#>  3            3             64           0.2
#>  4            1            160           0.2
#>  5            2            160           0.2
#>  6            3            160           0.2
#>  7            1            256           0.2
#>  8            2            256           0.2
#>  9            3            256           0.2
#> 10            1             64           0.3
#> # ℹ 17 more rows
```


``` r
# Using only the first 100 rows for speed. The real call
# should be: folds <- vfold_cv(train_df, v = 3)
folds <- vfold_cv(train_df[1:100, ], v = 3)

tune_res <- tune_grid(
  tune_wf,
  resamples = folds,
  grid = grid,
  metrics = metric_set(accuracy),
  control = control_grid(save_pred = FALSE, save_workflow = TRUE)
)
#> 2/2 - 0s - 67ms/step
#> 2/2 - 0s - 82ms/step
#> 2/2 - 0s - 85ms/step
#> 2/2 - 0s - 73ms/step
#> 2/2 - 0s - 66ms/step
#> 2/2 - 0s - 100ms/step
#> 2/2 - 0s - 67ms/step
#> 2/2 - 0s - 79ms/step
#> 2/2 - 0s - 92ms/step
#> 2/2 - 0s - 67ms/step
#> 2/2 - 0s - 80ms/step
#> 2/2 - 0s - 75ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 77ms/step
#> 2/2 - 0s - 96ms/step
#> 2/2 - 0s - 60ms/step
#> 2/2 - 0s - 88ms/step
#> 2/2 - 0s - 87ms/step
#> 2/2 - 0s - 70ms/step
#> 2/2 - 0s - 78ms/step
#> 2/2 - 0s - 96ms/step
#> 2/2 - 0s - 63ms/step
#> 2/2 - 0s - 72ms/step
#> 2/2 - 0s - 110ms/step
#> 2/2 - 0s - 79ms/step
#> 2/2 - 0s - 76ms/step
#> 2/2 - 0s - 94ms/step
#> 2/2 - 0s - 70ms/step
#> 2/2 - 0s - 71ms/step
#> 2/2 - 0s - 89ms/step
#> 2/2 - 0s - 70ms/step
#> 2/2 - 0s - 76ms/step
#> 2/2 - 0s - 82ms/step
#> 2/2 - 0s - 62ms/step
#> 2/2 - 0s - 86ms/step
#> 2/2 - 0s - 85ms/step
#> 2/2 - 0s - 68ms/step
#> 2/2 - 0s - 76ms/step
#> 2/2 - 0s - 85ms/step
#> 2/2 - 0s - 70ms/step
#> 2/2 - 0s - 74ms/step
#> 2/2 - 0s - 78ms/step
#> 2/2 - 0s - 75ms/step
#> 2/2 - 0s - 81ms/step
#> 2/2 - 0s - 81ms/step
#> 2/2 - 0s - 55ms/step
#> 2/2 - 0s - 77ms/step
#> 2/2 - 0s - 86ms/step
#> 2/2 - 0s - 53ms/step
#> 2/2 - 0s - 79ms/step
#> 2/2 - 0s - 82ms/step
#> 2/2 - 0s - 60ms/step
#> 2/2 - 0s - 82ms/step
#> 2/2 - 0s - 93ms/step
#> 2/2 - 0s - 57ms/step
#> 2/2 - 0s - 69ms/step
#> 2/2 - 0s - 85ms/step
#> 2/2 - 0s - 60ms/step
#> 2/2 - 0s - 65ms/step
#> 2/2 - 0s - 89ms/step
#> 2/2 - 0s - 59ms/step
#> 2/2 - 0s - 62ms/step
#> 2/2 - 0s - 98ms/step
#> 2/2 - 0s - 63ms/step
#> 2/2 - 0s - 71ms/step
#> 2/2 - 0s - 93ms/step
#> 2/2 - 0s - 63ms/step
#> 2/2 - 0s - 69ms/step
#> 2/2 - 0s - 84ms/step
#> 2/2 - 0s - 65ms/step
#> 2/2 - 0s - 73ms/step
#> 2/2 - 0s - 85ms/step
#> 2/2 - 0s - 61ms/step
#> 2/2 - 0s - 66ms/step
#> 2/2 - 0s - 77ms/step
#> 2/2 - 0s - 59ms/step
#> 2/2 - 0s - 68ms/step
#> 2/2 - 0s - 64ms/step
#> 2/2 - 0s - 58ms/step
#> 2/2 - 0s - 68ms/step
#> 2/2 - 0s - 68ms/step
```

Finally, we can inspect the results to find which architecture performed the best. First, a summary table:


``` r
# Show the summary table of the best models
show_best(tune_res, metric = "accuracy")
#> # A tibble: 5 × 9
#>   num_hidden_1 hidden_1_units hidden_1_rate .metric  .estimator  mean     n std_err .config         
#>          <int>          <int>         <dbl> <chr>    <chr>      <dbl> <int>   <dbl> <chr>           
#> 1            3            256         0.2   accuracy multiclass 0.800     3  0.0211 pre0_mod25_post0
#> 2            2            256         0.400 accuracy multiclass 0.780     3  0.0220 pre0_mod18_post0
#> 3            3            256         0.400 accuracy multiclass 0.779     3  0.0379 pre0_mod27_post0
#> 4            1            160         0.2   accuracy multiclass 0.770     3  0.0213 pre0_mod04_post0
#> 5            1            160         0.400 accuracy multiclass 0.770     3  0.0213 pre0_mod06_post0
```

Now that we've identified the best-performing hyperparameters, our final step is to create and train the final model. We use `select_best()` to get the top parameters, `finalize_workflow()` to update our workflow with them, and then `fit()` one last time on our full training dataset.


``` r
# Select the best hyperparameters
best_hps <- select_best(tune_res, metric = "accuracy")

# Finalize the workflow with the best hyperparameters
final_wf <- finalize_workflow(tune_wf, best_hps)

# Fit the final model on the full training data
final_fit <- fit(final_wf, data = train_df)
```

We can now inspect our final, tuned model.


``` r
# Print the model summary
final_fit |>
  extract_fit_parsnip() |>
  extract_keras_model() |>
  summary()
#> Model: "sequential_82"
#> ┌─────────────────────────────────────────────────────┬────────────────────────────────────────┬───────────────────────
#> │ Layer (type)                                        │ Output Shape                           │               Param # 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_246 (Dense)                                   │ (None, 256)                            │               200,960 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dropout_164 (Dropout)                               │ (None, 256)                            │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_247 (Dense)                                   │ (None, 256)                            │                65,792 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dropout_165 (Dropout)                               │ (None, 256)                            │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_248 (Dense)                                   │ (None, 256)                            │                65,792 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dropout_166 (Dropout)                               │ (None, 256)                            │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_249 (Dense)                                   │ (None, 10)                             │                 2,570 
#> └─────────────────────────────────────────────────────┴────────────────────────────────────────┴───────────────────────
#>  Total params: 670,230 (2.56 MB)
#>  Trainable params: 335,114 (1.28 MB)
#>  Non-trainable params: 0 (0.00 B)
#>  Optimizer params: 335,116 (1.28 MB)

# Plot the training history
final_fit |>
  extract_fit_parsnip() |>
  extract_keras_history() |>
  plot()
```

![plot of chunk inspect-final-model](figure/inspect-final-model-1.png)

This result shows that `tune` has tested various network depths, widths, and dropout rates, successfully finding the best-performing combination within the search space. By using `kerasnip`, we were able to integrate this complex architectural tuning directly into a standard `tidymodels` workflow.

---
title: "Tidymodels Workflow with Sequential Keras Models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tidymodels Workflow with Sequential Keras Models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



## Introduction

This vignette demonstrates a complete `tidymodels` workflow for a classification task using a Keras sequential model defined with `kerasnip`. We will use the Palmer Penguins dataset to predict penguin species based on physical measurements.

The `kerasnip` package allows you to define Keras models using a modular "layer block" approach, which then integrates seamlessly with the `parsnip` and `tune` packages for model specification and hyperparameter tuning.

## Setup

First, we load the necessary packages.


``` r
library(kerasnip)
library(tidymodels)
#> ── Attaching packages ────────────────────────────────────────────────────────────────────────────── tidymodels 1.5.0 ──
#> ✔ broom        1.0.12     ✔ recipes      1.3.2 
#> ✔ dials        1.4.3      ✔ rsample      1.3.2 
#> ✔ dplyr        1.2.1      ✔ tailor       0.1.0 
#> ✔ ggplot2      4.0.3      ✔ tidyr        1.3.2 
#> ✔ infer        1.1.0      ✔ tune         2.1.0 
#> ✔ modeldata    1.5.1      ✔ workflows    1.3.0 
#> ✔ parsnip      1.5.0      ✔ workflowsets 1.1.1 
#> ✔ purrr        1.2.2      ✔ yardstick    1.4.0
#> ── Conflicts ───────────────────────────────────────────────────────────────────────────────── tidymodels_conflicts() ──
#> ✖ purrr::discard() masks scales::discard()
#> ✖ dplyr::filter()  masks stats::filter()
#> ✖ dplyr::lag()     masks stats::lag()
#> ✖ recipes::step()  masks stats::step()
library(keras3)
#> 
#> Attaching package: 'keras3'
#> The following object is masked from 'package:yardstick':
#> 
#>     get_weights
#> The following object is masked from 'package:infer':
#> 
#>     generate
library(dplyr) # For data manipulation
library(ggplot2) # For plotting
library(future) # For parallel processing
#> 
#> Attaching package: 'future'
#> The following object is masked from 'package:keras3':
#> 
#>     %<-%
library(finetune) # For racing
```

## Data Preparation

We'll use the `penguins` dataset from the `modeldata` package. We will clean it by removing rows with missing values and ensuring the `species` column is a factor.


``` r
# Remove rows with missing values
penguins_df <- penguins |>
  na.omit() |>
  # Convert species to factor for classification
  mutate(species = factor(species))

# Split data into training and testing sets
set.seed(123)
penguin_split <- initial_split(penguins_df, prop = 0.8, strata = species)
penguin_train <- training(penguin_split)

penguin_test <- testing(penguin_split)

# Create cross-validation folds for tuning
penguin_folds <- vfold_cv(penguin_train, v = 5, strata = species)
```

## Recipe for Preprocessing

We will create a `recipes` object to preprocess our data. This recipe will:
*   Predict `species` using all other variables.
*   Normalize all numeric predictors.
*   Create dummy variables for all categorical predictors.


``` r
penguin_recipe <- recipe(species ~ ., data = penguin_train) |>
  step_normalize(all_numeric_predictors()) |>
  step_dummy(all_nominal_predictors())
```

## Define Keras Sequential Model with `kerasnip`

Now, we define our Keras sequential model using `kerasnip`'s layer blocks. We'll create a simple Multi-Layer Perceptron (MLP) with two hidden layers.

For a sequential Keras model with tabular data, all preprocessed input features are typically combined into a single input layer. The `recipes` package handles this preprocessing, transforming predictors into a single matrix that serves as the input to the Keras model.


``` r
# Define layer blocks
input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}

hidden_block <- function(model, units = 32, activation = "relu", rate = 0.2) {
  model |>
    layer_dense(units = units, activation = activation) |>
    layer_dropout(rate = rate)
}

output_block <- function(model, num_classes, activation = "softmax") {
  model |>
    layer_dense(units = num_classes, activation = activation)
}

# Create the kerasnip model specification function
create_keras_sequential_spec(
  model_name = "penguin_mlp",
  layer_blocks = list(
    input = input_block,
    hidden_1 = hidden_block,
    hidden_2 = hidden_block,
    output = output_block
  ),
  mode = "classification"
)
```

## Model Specification

We'll define our `penguin_mlp` model specification and set some hyperparameters to `tune()`, indicating that they should be optimized. We will also set fixed parameters for compilation and fitting.


``` r
# Define the tunable model specification
mlp_spec <- penguin_mlp(
  # Tunable parameters for hidden layers
  hidden_1_units = tune(),
  hidden_1_rate = tune(),
  hidden_2_units = tune(),
  hidden_2_rate = tune(),
  # Fixed compilation and fitting parameters
  compile_loss = "categorical_crossentropy",
  compile_optimizer = "adam",
  compile_metrics = c("accuracy"),
  fit_epochs = 20,
  fit_batch_size = 32,
  fit_validation_split = 0.2,
  fit_callbacks = list(
    callback_early_stopping(monitor = "val_loss", patience = 5)
  )
) |>
  set_engine("keras")

print(mlp_spec)
#> penguin mlp Model Specification (classification)
#> 
#> Main Arguments:
#>   num_input = structure(list(), class = "rlang_zap")
#>   num_hidden_1 = structure(list(), class = "rlang_zap")
#>   num_hidden_2 = structure(list(), class = "rlang_zap")
#>   num_output = structure(list(), class = "rlang_zap")
#>   hidden_1_units = tune()
#>   hidden_1_activation = structure(list(), class = "rlang_zap")
#>   hidden_1_rate = tune()
#>   hidden_2_units = tune()
#>   hidden_2_activation = structure(list(), class = "rlang_zap")
#>   hidden_2_rate = tune()
#>   output_activation = structure(list(), class = "rlang_zap")
#>   learn_rate = structure(list(), class = "rlang_zap")
#>   fit_batch_size = 32
#>   fit_epochs = 20
#>   fit_callbacks = list(callback_early_stopping(monitor = "val_loss", patience = 5))
#>   fit_validation_split = 0.2
#>   fit_validation_data = structure(list(), class = "rlang_zap")
#>   fit_shuffle = structure(list(), class = "rlang_zap")
#>   fit_class_weight = structure(list(), class = "rlang_zap")
#>   fit_sample_weight = structure(list(), class = "rlang_zap")
#>   fit_initial_epoch = structure(list(), class = "rlang_zap")
#>   fit_steps_per_epoch = structure(list(), class = "rlang_zap")
#>   fit_validation_steps = structure(list(), class = "rlang_zap")
#>   fit_validation_batch_size = structure(list(), class = "rlang_zap")
#>   fit_validation_freq = structure(list(), class = "rlang_zap")
#>   fit_verbose = structure(list(), class = "rlang_zap")
#>   fit_view_metrics = structure(list(), class = "rlang_zap")
#>   compile_optimizer = adam
#>   compile_loss = categorical_crossentropy
#>   compile_metrics = c("accuracy")
#>   compile_loss_weights = structure(list(), class = "rlang_zap")
#>   compile_weighted_metrics = structure(list(), class = "rlang_zap")
#>   compile_run_eagerly = structure(list(), class = "rlang_zap")
#>   compile_steps_per_execution = structure(list(), class = "rlang_zap")
#>   compile_jit_compile = structure(list(), class = "rlang_zap")
#>   compile_auto_scale_loss = structure(list(), class = "rlang_zap")
#> 
#> Computational engine: keras
```

## Create Workflow

A `workflow` combines the recipe and the model specification.


``` r
penguin_wf <- workflow() |>
  add_recipe(penguin_recipe) |>
  add_model(mlp_spec)

print(penguin_wf)
#> ══ Workflow ════════════════════════════════════════════════════════════════════════════════════════════════════════════
#> Preprocessor: Recipe
#> Model: penguin_mlp()
#> 
#> ── Preprocessor ────────────────────────────────────────────────────────────────────────────────────────────────────────
#> 2 Recipe Steps
#> 
#> • step_normalize()
#> • step_dummy()
#> 
#> ── Model ───────────────────────────────────────────────────────────────────────────────────────────────────────────────
#> penguin mlp Model Specification (classification)
#> 
#> Main Arguments:
#>   num_input = structure(list(), class = "rlang_zap")
#>   num_hidden_1 = structure(list(), class = "rlang_zap")
#>   num_hidden_2 = structure(list(), class = "rlang_zap")
#>   num_output = structure(list(), class = "rlang_zap")
#>   hidden_1_units = tune()
#>   hidden_1_activation = structure(list(), class = "rlang_zap")
#>   hidden_1_rate = tune()
#>   hidden_2_units = tune()
#>   hidden_2_activation = structure(list(), class = "rlang_zap")
#>   hidden_2_rate = tune()
#>   output_activation = structure(list(), class = "rlang_zap")
#>   learn_rate = structure(list(), class = "rlang_zap")
#>   fit_batch_size = 32
#>   fit_epochs = 20
#>   fit_callbacks = list(callback_early_stopping(monitor = "val_loss", patience = 5))
#>   fit_validation_split = 0.2
#>   fit_validation_data = structure(list(), class = "rlang_zap")
#>   fit_shuffle = structure(list(), class = "rlang_zap")
#>   fit_class_weight = structure(list(), class = "rlang_zap")
#>   fit_sample_weight = structure(list(), class = "rlang_zap")
#>   fit_initial_epoch = structure(list(), class = "rlang_zap")
#>   fit_steps_per_epoch = structure(list(), class = "rlang_zap")
#>   fit_validation_steps = structure(list(), class = "rlang_zap")
#>   fit_validation_batch_size = structure(list(), class = "rlang_zap")
#>   fit_validation_freq = structure(list(), class = "rlang_zap")
#>   fit_verbose = structure(list(), class = "rlang_zap")
#>   fit_view_metrics = structure(list(), class = "rlang_zap")
#>   compile_optimizer = adam
#>   compile_loss = categorical_crossentropy
#>   compile_metrics = c("accuracy")
#>   compile_loss_weights = structure(list(), class = "rlang_zap")
#>   compile_weighted_metrics = structure(list(), class = "rlang_zap")
#>   compile_run_eagerly = structure(list(), class = "rlang_zap")
#>   compile_steps_per_execution = structure(list(), class = "rlang_zap")
#>   compile_jit_compile = structure(list(), class = "rlang_zap")
#>   compile_auto_scale_loss = structure(list(), class = "rlang_zap")
#> 
#> Computational engine: keras
```

## Define Tuning Grid

We will create a regular grid for our hyperparameters.


``` r
# Define the tuning grid
params <- extract_parameter_set_dials(penguin_wf) |>
  update(
    hidden_1_units = hidden_units(range = c(32, 128)),
    hidden_1_rate = dropout(range = c(0.1, 0.4)),
    hidden_2_units = hidden_units(range = c(16, 64)),
    hidden_2_rate = dropout(range = c(0.1, 0.4))
  )
mlp_grid <- grid_regular(params, levels = 3)

print(mlp_grid)
#> # A tibble: 81 × 4
#>    hidden_1_units hidden_1_rate hidden_2_units hidden_2_rate
#>             <int>         <dbl>          <int>         <dbl>
#>  1             32         0.1               16           0.1
#>  2             80         0.1               16           0.1
#>  3            128         0.1               16           0.1
#>  4             32         0.25              16           0.1
#>  5             80         0.25              16           0.1
#>  6            128         0.25              16           0.1
#>  7             32         0.400             16           0.1
#>  8             80         0.400             16           0.1
#>  9            128         0.400             16           0.1
#> 10             32         0.1               40           0.1
#> # ℹ 71 more rows
```

## Tune Model

Now, we'll use `tune_race_anova()` to perform cross-validation and find the best hyperparameters.


``` r
# Note: Parallel processing with `plan(multisession)` is currently not working
# with Keras models due to backend conflicts

set.seed(123)
penguin_tune_results <- tune_race_anova(
  penguin_wf,
  resamples = penguin_folds,
  grid = mlp_grid,
  metrics = metric_set(accuracy, roc_auc, f_meas),
  control = control_race(save_pred = TRUE, save_workflow = TRUE)
)
```

## Inspect Tuning Results

We can inspect the tuning results to see which hyperparameter combinations performed best.


``` r
# Show the best performing models based on accuracy
show_best(penguin_tune_results, metric = "accuracy", n = 5)
#> # A tibble: 5 × 10
#>   hidden_1_units hidden_1_rate hidden_2_units hidden_2_rate .metric  .estimator  mean     n std_err .config         
#>            <int>         <dbl>          <int>         <dbl> <chr>    <chr>      <dbl> <int>   <dbl> <chr>           
#> 1             32         0.1               64         0.25  accuracy multiclass     1     5       0 pre0_mod08_post0
#> 2             32         0.400             40         0.400 accuracy multiclass     1     5       0 pre0_mod24_post0
#> 3             32         0.400             64         0.1   accuracy multiclass     1     5       0 pre0_mod25_post0
#> 4             80         0.1               16         0.1   accuracy multiclass     1     5       0 pre0_mod28_post0
#> 5             80         0.1               16         0.25  accuracy multiclass     1     5       0 pre0_mod29_post0

# Autoplot the results
# Currently does not work due to a label issue: autoplot(penguin_tune_results)

# Select the best hyperparameters
best_mlp_params <- select_best(penguin_tune_results, metric = "accuracy")
print(best_mlp_params)
#> # A tibble: 1 × 5
#>   hidden_1_units hidden_1_rate hidden_2_units hidden_2_rate .config         
#>            <int>         <dbl>          <int>         <dbl> <chr>           
#> 1             32           0.1             64          0.25 pre0_mod08_post0
```

## Finalize Workflow and Fit Model

Once we have the best hyperparameters, we finalize the workflow and fit the model on the entire training dataset.


``` r
# Finalize the workflow with the best hyperparameters
final_penguin_wf <- finalize_workflow(penguin_wf, best_mlp_params)

# Fit the final model on the full training data
final_penguin_fit <- fit(final_penguin_wf, data = penguin_train)

print(final_penguin_fit)
#> ══ Workflow [trained] ══════════════════════════════════════════════════════════════════════════════════════════════════
#> Preprocessor: Recipe
#> Model: penguin_mlp()
#> 
#> ── Preprocessor ────────────────────────────────────────────────────────────────────────────────────────────────────────
#> 2 Recipe Steps
#> 
#> • step_normalize()
#> • step_dummy()
#> 
#> ── Model ───────────────────────────────────────────────────────────────────────────────────────────────────────────────
#> $fit
#> Model: "sequential_376"
#> ┌─────────────────────────────────────────────────────┬────────────────────────────────────────┬───────────────────────
#> │ Layer (type)                                        │ Output Shape                           │               Param # 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_1127 (Dense)                                  │ (None, 32)                             │                   256 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dropout_750 (Dropout)                               │ (None, 32)                             │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_1128 (Dense)                                  │ (None, 64)                             │                 2,112 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dropout_751 (Dropout)                               │ (None, 64)                             │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_1129 (Dense)                                  │ (None, 3)                              │                   195 
#> └─────────────────────────────────────────────────────┴────────────────────────────────────────┴───────────────────────
#>  Total params: 7,691 (30.05 KB)
#>  Trainable params: 2,563 (10.01 KB)
#>  Non-trainable params: 0 (0.00 B)
#>  Optimizer params: 5,128 (20.04 KB)
#> 
#> $keras_bytes
#>     [1] 50 4b 03 04 14 00 00 00 00 00 00 00 21 00 f1 8f 83 4b 40 00 00 00 40 00 00 00 0d 00 00 00 6d 65 74 61 64 61 74
#>    [38] 61 2e 6a 73 6f 6e 7b 22 6b 65 72 61 73 5f 76 65 72 73 69 6f 6e 22 3a 20 22 33 2e 31 34 2e 30 22 2c 20 22 64 61
#>    [75] 74 65 5f 73 61 76 65 64 22 3a 20 22 32 30 32 36 2d 30 35 2d 30 31 40 31 32 3a 31 35 3a 34 30 22 7d 50 4b 03 04
#>   [112] 14 00 00 00 00 00 00 00 21 00 fa 5d 54 2d 1d 11 00 00 1d 11 00 00 0b 00 00 00 63 6f 6e 66 69 67 2e 6a 73 6f 6e
#>   [149] 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 22 2c 20 22 63 6c 61 73 73 5f 6e 61 6d 65 22 3a 20 22 53 65
#>   [186] 71 75 65 6e 74 69 61 6c 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 6e 61 6d 65 22 3a 20 22 73 65 71 75 65 6e
#>   [223] 74 69 61 6c 5f 33 37 36 22 2c 20 22 74 72 61 69 6e 61 62 6c 65 22 3a 20 74 72 75 65 2c 20 22 64 74 79 70 65 22
#>   [260] 3a 20 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 22 2c 20 22 63 6c 61 73 73 5f 6e 61 6d 65 22 3a 20 22
#>   [297] 44 54 79 70 65 50 6f 6c 69 63 79 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 6e 61 6d 65 22 3a 20 22 66 6c 6f
#>   [334] 61 74 33 32 22 7d 2c 20 22 72 65 67 69 73 74 65 72 65 64 5f 6e 61 6d 65 22 3a 20 6e 75 6c 6c 2c 20 22 73 68 61
#>   [371] 72 65 64 5f 6f 62 6a 65 63 74 5f 69 64 22 3a 20 31 39 38 30 33 37 32 35 38 30 35 36 30 7d 2c 20 22 6c 61 79 65
#>   [408] 72 73 22 3a 20 5b 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 2e 6c 61 79 65 72 73 22 2c 20 22 63 6c 61
#>   [445] 73 73 5f 6e 61 6d 65 22 3a 20 22 49 6e 70 75 74 4c 61 79 65 72 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 62
#>   [482] 61 74 63 68 5f 73 68 61 70 65 22 3a 20 5b 6e 75 6c 6c 2c 20 37 5d 2c 20 22 64 74 79 70 65 22 3a 20 22 66 6c 6f
#>   [519] 61 74 33 32 22 2c 20 22 73 70 61 72 73 65 22 3a 20 66 61 6c 73 65 2c 20 22 72 61 67 67 65 64 22 3a 20 66 61 6c
#>   [556] 73 65 2c 20 22 6e 61 6d 65 22 3a 20 22 69 6e 70 75 74 5f 6c 61 79 65 72 5f 33 37 36 22 2c 20 22 6f 70 74 69 6f
#>   [593] 6e 61 6c 22 3a 20 66 61 6c 73 65 7d 2c 20 22 72 65 67 69 73 74 65 72 65 64 5f 6e 61 6d 65 22 3a 20 6e 75 6c 6c
#>   [630] 7d 2c 20 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 2e 6c 61 79 65 72 73 22 2c 20 22 63 6c 61 73 73 5f
#>   [667] 6e 61 6d 65 22 3a 20 22 44 65 6e 73 65 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 6e 61 6d 65 22 3a 20 22 64
#>   [704] 65 6e 73 65 5f 31 31 32 37 22 2c 20 22 74 72 61 69 6e 61 62 6c 65 22 3a 20 74 72 75 65 2c 20 22 64 74 79 70 65
#>   [741] 22 3a 20 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 22 2c 20 22 63 6c 61 73 73 5f 6e 61 6d 65 22 3a 20
#>   [778] 22 44 54 79 70 65 50 6f 6c 69 63 79 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b 22 6e 61 6d 65 22 3a 20 22 66 6c
#>   [815] 6f 61 74 33 32 22 7d 2c 20 22 72 65 67 69 73 74 65 72 65 64 5f 6e 61 6d 65 22 3a 20 6e 75 6c 6c 7d 2c 20 22 75
#>   [852] 6e 69 74 73 22 3a 20 33 32 2c 20 22 61 63 74 69 76 61 74 69 6f 6e 22 3a 20 22 72 65 6c 75 22 2c 20 22 75 73 65
#>   [889] 5f 62 69 61 73 22 3a 20 74 72 75 65 2c 20 22 6b 65 72 6e 65 6c 5f 69 6e 69 74 69 61 6c 69 7a 65 72 22 3a 20 7b
#>   [926] 22 6d 6f 64 75 6c 65 22 3a 20 22 6b 65 72 61 73 2e 69 6e 69 74 69 61 6c 69 7a 65 72 73 22 2c 20 22 63 6c 61 73
#>   [963] 73 5f 6e 61 6d 65 22 3a 20 22 47 6c 6f 72 6f 74 55 6e 69 66 6f 72 6d 22 2c 20 22 63 6f 6e 66 69 67 22 3a 20 7b
#>  [1000] 22 73 65 65 64 22 3a 20 6e 75 6c 6c 7d 2c 20 22 72 65 67 69 73 74 65 72 65 64 5f 6e 61 6d 65 22 3a 20 6e 75 6c
#>  [1037] 6c 7d 2c 20 22 62 69 61 73 5f 69 6e 69 74 69 61 6c 69 7a 65 72 22 3a 20 7b 22 6d 6f 64 75 6c 65 22 3a 20 22 6b
#> 
#> ...
#> and 1674 more lines.
```

### Inspect Final Model

You can extract the underlying Keras model and its training history for further inspection.


``` r
# Extract the Keras model summary
final_penguin_fit |>
  extract_fit_parsnip() |>
  extract_keras_model() |>
  summary()
#> Model: "sequential_376"
#> ┌─────────────────────────────────────────────────────┬────────────────────────────────────────┬───────────────────────
#> │ Layer (type)                                        │ Output Shape                           │               Param # 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_1127 (Dense)                                  │ (None, 32)                             │                   256 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dropout_750 (Dropout)                               │ (None, 32)                             │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_1128 (Dense)                                  │ (None, 64)                             │                 2,112 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dropout_751 (Dropout)                               │ (None, 64)                             │                     0 
#> ├─────────────────────────────────────────────────────┼────────────────────────────────────────┼───────────────────────
#> │ dense_1129 (Dense)                                  │ (None, 3)                              │                   195 
#> └─────────────────────────────────────────────────────┴────────────────────────────────────────┴───────────────────────
#>  Total params: 7,691 (30.05 KB)
#>  Trainable params: 2,563 (10.01 KB)
#>  Non-trainable params: 0 (0.00 B)
#>  Optimizer params: 5,128 (20.04 KB)
```


``` r
# Plot the Keras model
final_penguin_fit |>
  extract_fit_parsnip() |>
  extract_keras_model() |>
  plot(show_shapes = TRUE)
```

![Model](images/model_plot_shapes_ws.png){fig-alt="A picture showing the model shape"}


``` r
# Plot the training history
final_penguin_fit |>
  extract_fit_parsnip() |>
  extract_keras_history() |>
  plot()
```

![plot of chunk inspect-final-keras-model-history](figure/inspect-final-keras-model-history-1.png)

## Make Predictions and Evaluate

Finally, we will make predictions on the test set and evaluate the model's performance.


``` r
# Make predictions on the test set
penguin_test_pred <- predict(final_penguin_fit, new_data = penguin_test)
#> 3/3 - 0s - 38ms/step
penguin_test_prob <- predict(
  final_penguin_fit,
  new_data = penguin_test,
  type = "prob"
)
#> 3/3 - 0s - 19ms/step

# Combine predictions with actuals
penguin_results <- penguin_test |>
  select(species) |>
  bind_cols(penguin_test_pred, penguin_test_prob)

print(head(penguin_results))
#> # A tibble: 6 × 5
#>   species .pred_class .pred_Adelie .pred_Chinstrap .pred_Gentoo
#>   <fct>   <fct>              <dbl>           <dbl>        <dbl>
#> 1 Adelie  Adelie             1.000        1.37e- 7     1.41e- 8
#> 2 Adelie  Adelie             1            1.11e-11     4.72e-12
#> 3 Adelie  Adelie             1            3.23e-11     3.48e-11
#> 4 Adelie  Adelie             1            7.12e-10     1.13e-10
#> 5 Adelie  Adelie             1            4.24e-11     1.11e-11
#> 6 Adelie  Adelie             0.899        1.01e- 1     5.16e- 5

# Evaluate performance using yardstick metrics
metrics_results <- metric_set(
  accuracy,
  roc_auc,
  f_meas
)(
  penguin_results,
  truth = species,
  estimate = .pred_class,
  .pred_Adelie,
  .pred_Chinstrap,
  .pred_Gentoo
)

print(metrics_results)
#> # A tibble: 3 × 3
#>   .metric  .estimator .estimate
#>   <chr>    <chr>          <dbl>
#> 1 accuracy multiclass         1
#> 2 f_meas   macro              1
#> 3 roc_auc  hand_till          1

# Confusion Matrix
conf_mat(penguin_results, truth = species, estimate = .pred_class) |>
  autoplot(type = "heatmap")
```

![plot of chunk predict-evaluate](figure/predict-evaluate-1.png)

## Saving and Reloading Your Model

`kerasnip` serializes the Keras model weights to bytes at fit time and stores them alongside the workflow object. This means plain `saveRDS()` / `readRDS()` **works out of the box** — the underlying Keras model is restored automatically the first time `predict()` is called on the reloaded object.


``` r
# Save
saveRDS(final_penguin_fit, "penguin_model.rds")

# Reload — no extra steps needed
final_penguin_fit_loaded <- readRDS("penguin_model.rds")

predict(final_penguin_fit_loaded, new_data = penguin_test) |> head()
#> 3/3 - 0s - 48ms/step
#> # A tibble: 6 × 1
#>   .pred_class
#>   <fct>      
#> 1 Adelie     
#> 2 Adelie     
#> 3 Adelie     
#> 4 Adelie     
#> 5 Adelie     
#> 6 Adelie
```

If you need a fully self-contained bundle suitable for deployment with `vetiver` or other MLOps tools, use `bundle::bundle()` instead:


``` r
library(bundle)

# Save as a portable bundle
bundled <- bundle(final_penguin_fit)
saveRDS(bundled, "penguin_model_bundle.rds")

# Reload in any R session
library(kerasnip)
library(bundle)

final_penguin_fit_loaded <- unbundle(readRDS("penguin_model_bundle.rds"))

predict(final_penguin_fit_loaded, new_data = penguin_test) |> head()
#> 3/3 - 0s - 56ms/step
#> # A tibble: 6 × 1
#>   .pred_class
#>   <fct>      
#> 1 Adelie     
#> 2 Adelie     
#> 3 Adelie     
#> 4 Adelie     
#> 5 Adelie     
#> 6 Adelie
```

See `vignette("saving_and_reloading")` for a detailed comparison of both approaches.

---
title: "Prediction Intervals with Conformal Inference"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Prediction Intervals with Conformal Inference}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



## Why neural networks do not produce native prediction intervals

Linear models derive prediction intervals analytically from the covariance structure of their parameter estimates. A neural network has no equivalent: it computes a single deterministic output for each input, with no internal distributional model to derive intervals from.

This is not a limitation specific to `kerasnip`. The entire tidymodels ecosystem reflects the same reality, the `parsnip` package does not register `conf_int` or `pred_int` prediction types for any neural network engine, including `brulee`, the official tidymodels neural network package.

The supported and recommended path for obtaining prediction intervals from neural networks in tidymodels is **conformal inference**, provided by the `probably` package. Conformal inference is:

- **Distribution-free**: it makes no assumptions about the shape of the outcome distribution.
- **Model-agnostic**: it treats the model as a black box; only `fit()` and `predict()` calls are required.
- **Guaranteed to have correct coverage**: under the assumption that training and test data are exchangeable (i.e. identically and independently distributed), the intervals contain the true outcome with at least the requested probability.

The `probably` package provides three conformal methods, described in the sections below.

## Setup


``` r
library(kerasnip)
library(tidymodels)
#> ── Attaching packages ────────────────────────────────────────────────────────────────────────────── tidymodels 1.5.0 ──
#> ✔ broom        1.0.12     ✔ recipes      1.3.2 
#> ✔ dials        1.4.3      ✔ rsample      1.3.2 
#> ✔ dplyr        1.2.1      ✔ tailor       0.1.0 
#> ✔ ggplot2      4.0.3      ✔ tidyr        1.3.2 
#> ✔ infer        1.1.0      ✔ tune         2.1.0 
#> ✔ modeldata    1.5.1      ✔ workflows    1.3.0 
#> ✔ parsnip      1.5.0      ✔ workflowsets 1.1.1 
#> ✔ purrr        1.2.2      ✔ yardstick    1.4.0
#> ── Conflicts ───────────────────────────────────────────────────────────────────────────────── tidymodels_conflicts() ──
#> ✖ purrr::discard() masks scales::discard()
#> ✖ dplyr::filter()  masks stats::filter()
#> ✖ dplyr::lag()     masks stats::lag()
#> ✖ recipes::step()  masks stats::step()
library(probably)
#> 
#> Attaching package: 'probably'
#> The following objects are masked from 'package:base':
#> 
#>     as.factor, as.ordered
```

## Defining a kerasnip regression model

All three methods work with any fitted `workflows::workflow()` object, so the setup is identical to any other kerasnip workflow.


``` r
input_block <- function(model, input_shape) {
  keras3::keras_model_sequential(input_shape = input_shape)
}
dense_block <- function(model, units = 32) {
  model |>
    keras3::layer_dense(units = units, activation = "relu")
}
output_block <- function(model) {
  model |> keras3::layer_dense(units = 1)
}

create_keras_sequential_spec(
  model_name = "conf_mlp",
  layer_blocks = list(
    input  = input_block,
    dense  = dense_block,
    output = output_block
  ),
  mode = "regression"
)

spec <- conf_mlp(dense_units = 32, fit_epochs = 30) |>
  set_engine("keras")

data <- modeldata::ames |>
  mutate(Sale_Price = as.numeric(Sale_Price)) |>
  select(
    Sale_Price,
    Gr_Liv_Area,
    Year_Built,
    Garage_Area,
    Total_Bsmt_SF
  )
rec <- recipe(Sale_Price ~ ., data = data) |>
  step_normalize(all_numeric_predictors())
wflow <- workflow(rec, spec)
```

---

## Method 1: Split conformal inference (`int_conformal_split`)

This is the simplest and fastest method. The workflow is:

1. Fit the model on a **training set**.
2. Pass a separate **calibration set** to `int_conformal_split()`. The function runs `predict()` once on the calibration set and records the residuals.
3. Call `predict()` on the resulting object to obtain intervals for new data.

The model is never re-fitted. The calibration residuals act as a reference distribution for deciding how wide to make the intervals.


``` r
set.seed(1)
split <- initial_split(data, prop = 0.75)
train_dat <- training(split)
cal_dat <- testing(split)

# Fit on the training set
fit_obj <- fit(wflow, data = train_dat)

# Build the conformal object from the calibration set
conformal_split <- int_conformal_split(fit_obj, cal_data = cal_dat)
#> 23/23 - 0s - 8ms/step
conformal_split
#> Split Conformal inference
#> preprocessor: recipe 
#> model: conf_mlp (engine = keras) 
#> calibration set size: 733 
#> 
#> Use `predict(object, new_data, level)` to compute prediction intervals

# Predict intervals for new observations
new_obs <- cal_dat[1:6, ]
predict(conformal_split, new_data = new_obs, level = 0.90)
#> 1/1 - 0s - 63ms/step
#> # A tibble: 6 × 3
#>    .pred .pred_lower .pred_upper
#>    <dbl>       <dbl>       <dbl>
#> 1 59969.    -168695.     288633.
#> 2 27399.    -201265.     256063.
#> 3 60180.    -168485.     288844.
#> 4 43338.    -185326.     272002.
#> 5 42925.    -185739.     271589.
#> 6 18101.    -210563.     246765.
```

**When to use this**: when training cost is non-trivial and you can afford to
set aside a calibration set. This is the recommended method for kerasnip
models in most practical situations.

---

## Method 2: Cross-validation conformal inference (`int_conformal_cv`)

This method uses **cross-validation** to generate out-of-fold predictions, then pools those residuals to calibrate the intervals. The workflow is:

1. Define cross-validation folds.
2. Run `tune::fit_resamples()` on the workflow with `save_pred = TRUE` and `extract = function(x) x` so that both predictions and the fitted model objects are retained for each fold.
3. Pass the resulting `resample_results` object to `int_conformal_cv()`. It reads the held-out predictions and uses them to calibrate the intervals.
4. Call `predict()` on the resulting object to obtain intervals for new data.

The final model is fitted once more on the full training data inside `int_conformal_cv()`.


``` r
set.seed(1)
folds <- vfold_cv(data, v = 5)

# Fit the workflow on each fold, retaining predictions and model objects
fitted_folds <- tune::fit_resamples(
  wflow,
  folds,
  control = tune::control_resamples(
    save_pred = TRUE,
    extract = function(x) x
  )
)
#> 19/19 - 0s - 8ms/step
#> 19/19 - 0s - 12ms/step
#> 19/19 - 0s - 9ms/step
#> 19/19 - 0s - 10ms/step
#> 19/19 - 0s - 16ms/step

# Build the conformal object from the resampling results
conformal_cv <- int_conformal_cv(fitted_folds)
conformal_cv
#> Conformal inference via CV+
#> preprocessor: recipe 
#> model: conf_mlp (engine = keras) 
#> number of models: 5 
#> training set size: 2,930 
#> 
#> Use `predict(object, new_data, level)` to compute prediction intervals

predict(conformal_cv, new_data = data[1:6, ], level = 0.90)
#> 1/1 - 0s - 122ms/step
#> 1/1 - 0s - 150ms/step
#> 1/1 - 0s - 76ms/step
#> 1/1 - 0s - 84ms/step
#> 1/1 - 0s - 86ms/step
#> # A tibble: 6 × 3
#>   .pred_lower  .pred .pred_upper
#>         <dbl>  <dbl>       <dbl>
#> 1    -192359. 25742.     243844.
#> 2    -198053. 20049.     238151.
#> 3    -200821. 17281.     235383.
#> 4    -150145. 67957.     286059.
#> 5    -185703. 32399.     250501.
#> 6    -186155. 31947.     250049.
```

**When to use this**: when you do not want to reserve a separate calibration set and are already running cross-validation to evaluate your model. The cross-validation residuals tend to produce slightly narrower intervals than the split method because they use more of the data. For kerasnip models, note that this refits the model once per fold, which multiplies training time by the number of folds.

---

## Method 3: Full conformal inference (`int_conformal_full`)

Full conformal inference is the most principled of the three methods in terms of statistical guarantees, but it is the most expensive computationally.

The algorithm works observation by observation. For each new test point, it appends that observation to the training set with a **candidate outcome value**, refits the model from scratch on the augmented dataset, and checks whether the residual for the candidate value is consistent with the training residuals. By sweeping over a grid of candidate values, it determines the interval of values that the algorithm would not reject.

For a kerasnip model, every refit call goes through the full build–compile–train cycle: the Keras architecture is reconstructed from the layer blocks, compiled, and trained from random weight initialisation. This means the total number of training runs is:

> *(number of new test observations)* × *(number of candidate grid points)*

For the example below with 6 new observations and a 20-point grid, that is 120 complete model training runs. Even with 30 epochs on a small dataset this takes meaningful time. In practice, `int_conformal_split` or `int_conformal_cv` are the recommended choices for kerasnip models. Full conformal inference is provided for completeness and for situations where the extra statistical rigour is required and training cost is acceptable.


``` r
# Use a small subset to keep runtime reasonable in this vignette
data_small <- data[1:100, ]
new_obs_small <- data[101:106, ]

fit_small <- fit(wflow, data = data_small)

conformal_full <- int_conformal_full(
  fit_small,
  train_data = data_small,
  control = control_conformal_full(
    method = "grid",
    trial_points = 20
  )
)
#> 4/4 - 0s - 34ms/step
conformal_full
#> Conformal inference
#> preprocessor: recipe 
#> model: conf_mlp (engine = keras) 
#> training set size: 100 
#> 
#> Use `predict(object, new_data, level)` to compute prediction intervals

predict(conformal_full, new_data = new_obs_small, level = 0.90)
#> 1/1 - 0s - 64ms/step
#> # A tibble: 6 × 2
#>   .pred_lower .pred_upper
#>         <dbl>       <dbl>
#> 1    -452504.     452607.
#> 2    -451266.     451368.
#> 3    -444009.     444106.
#> 4    -471366.     471481.
#> 5    -463133.     463398.
#> 6    -502665.     502800.
```

**When to use this**: when you need the strongest possible coverage guarantees and are willing to pay the training cost. For neural networks, prefer `int_conformal_split` unless you have a specific reason to use full conformal inference.

---

## Comparison of the three methods

| Method | Model refits | Coverage guarantee | Recommended for kerasnip |
|---|---|---|---|
| `int_conformal_split` | 1 (on training data only) | Marginal | Yes, lowest cost |
| `int_conformal_cv` | 1 per fold + 1 final | Marginal | Yes, if already cross-validating |
| `int_conformal_full` | *(n_test × n_grid)* | Marginal | With caution, expensive |

All three methods provide **marginal coverage**: across many test observations, at least the requested fraction will have their true outcome within the interval. None of them provide **conditional coverage** (intervals that are exactly correct for every specific input value), but in practice they are well-calibrated for most regression problems.

---

## Cleanup


``` r
remove_keras_spec("conf_mlp")
#> Removed from parsnip registry objects: conf_mlp, conf_mlp_args, conf_mlp_encoding, conf_mlp_fit, conf_mlp_modes, conf_mlp_pkgs, conf_mlp_predict
#> Removed 'conf_mlp' from parsnip:::get_model_env()$models
```

---
title: "Building Functional Models with kerasnip"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Building Functional Models with kerasnip}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

This vignette demonstrates how to use the `create_keras_functional_spec()` function to build complex, non-linear Keras models that integrate seamlessly with the `tidymodels` ecosystem.

## When to Use the Functional API

While `create_keras_sequential_spec()` is perfect for models that are a simple, linear stack of layers, many advanced architectures are not linear. The Keras Functional API is designed for these cases. You should use `create_keras_functional_spec()` when your model has:

*   Multiple input or multiple output layers.
*   Shared layers between different branches.
*   Residual connections (e.g., ResNets), where a layer's input is added to its output.
*   Any other non-linear topology.

`kerasnip` makes it easy to define these architectures by automatically connecting a graph of layer blocks.

## The Core Concept: Building a Graph

`kerasnip` builds the model's graph by inspecting the `layer_blocks` you provide. The connection logic is simple but powerful:

1.  The **names of the list elements** in `layer_blocks` define the names of the nodes in your graph (e.g., `main_input`, `dense_path`, `output`).
2.  The **names of the arguments** in each block function specify its inputs. A block function like `my_block <- function(input_a, input_b, ...)` declares that it needs input from the nodes named `input_a` and `input_b`.

There are two special requirements:

*   **Input Block**: The first block in the list is treated as the main input node. Its function should not take other blocks as input.
*   **Output Block**: Exactly one block must be named `"output"`. The tensor returned by this block is used as the final output of the Keras model.

Let's see this in action.

## Example 1: A Two-Input Regression Model

This model will take two distinct inputs, process them separately, and then concatenate their outputs before a final regression layer. This clearly demonstrates the functional API's ability to handle multiple inputs, which is not possible with the sequential API.

### Step 1: Load Libraries

First, we load the necessary packages.




``` r
library(kerasnip)
library(tidymodels)
#> ── Attaching packages ────────────────────────────────────────────────────────────────────────────── tidymodels 1.5.0 ──
#> ✔ broom        1.0.12     ✔ recipes      1.3.2 
#> ✔ dials        1.4.3      ✔ rsample      1.3.2 
#> ✔ dplyr        1.2.1      ✔ tailor       0.1.0 
#> ✔ ggplot2      4.0.3      ✔ tidyr        1.3.2 
#> ✔ infer        1.1.0      ✔ tune         2.1.0 
#> ✔ modeldata    1.5.1      ✔ workflows    1.3.0 
#> ✔ parsnip      1.5.0      ✔ workflowsets 1.1.1 
#> ✔ purrr        1.2.2      ✔ yardstick    1.4.0
#> ── Conflicts ───────────────────────────────────────────────────────────────────────────────── tidymodels_conflicts() ──
#> ✖ purrr::discard() masks scales::discard()
#> ✖ dplyr::filter()  masks stats::filter()
#> ✖ dplyr::lag()     masks stats::lag()
#> ✖ recipes::step()  masks stats::step()
library(keras3)
#> 
#> Attaching package: 'keras3'
#> The following object is masked from 'package:yardstick':
#> 
#>     get_weights
#> The following object is masked from 'package:infer':
#> 
#>     generate

# Silence the startup messages from remove_keras_spec
options(kerasnip.show_removal_messages = FALSE)
```

### Step 2: Define Layer Blocks

These are the building blocks of our model. Each function represents a node in the graph.


``` r
# Input blocks for two distinct inputs
input_block_1 <- function(input_shape) {
  layer_input(shape = input_shape, name = "input_1")
}

input_block_2 <- function(input_shape) {
  layer_input(shape = input_shape, name = "input_2")
}

# Dense paths for each input
dense_path_1 <- function(tensor, units = 16) {
  tensor |> layer_dense(units = units, activation = "relu")
}

dense_path_2 <- function(tensor, units = 16) {
  tensor |> layer_dense(units = units, activation = "relu")
}

# A block to join two tensors
concat_block <- function(input_a, input_b) {
  layer_concatenate(list(input_a, input_b))
}

# The final output block for regression
output_block_1 <- function(tensor) {
  layer_dense(tensor, units = 1, name = "output_1")
}

output_block_2 <- function(tensor) {
  layer_dense(tensor, units = 1, name = "output_2")
}
```

### Step 3: Create the Model Specification

Now we assemble the blocks into a graph. The `inp_spec()` helper simplifies connecting these blocks, eliminating the need for verbose anonymous functions. `inp_spec()` automatically creates a wrapper that renames the arguments of our blocks to match the node names defined in the `layer_blocks` list.


``` r
model_name <- "two_output_reg_spec" # Changed model name
# Clean up the spec when the vignette is done knitting
on.exit(remove_keras_spec(model_name), add = TRUE)

create_keras_functional_spec(
  model_name = model_name,
  layer_blocks = list(
    input_1 = input_block_1,
    input_2 = input_block_2,
    processed_1 = inp_spec(dense_path_1, "input_1"),
    processed_2 = inp_spec(dense_path_2, "input_2"),
    concatenated = inp_spec(
      concat_block,
      c(input_a = "processed_1", input_b = "processed_2")
    ),
    output_1 = inp_spec(output_block_1, "concatenated"), # New output block 1
    output_2 = inp_spec(output_block_2, "concatenated") # New output block 2
  ),
  mode = "regression" # Still regression, but will have two columns in y
)
```

### Step 4: Use and Fit the Model

The new function `two_input_reg_spec()` is now available. Its arguments (`processed_1_units`, `processed_2_units`) were discovered automatically from our block definitions.


``` r
# We can override the default `units` for each path.
spec <- two_output_reg_spec( # Changed spec name
  processed_1_units = 16,
  processed_2_units = 8,
  fit_epochs = 10,
  fit_verbose = 0 # Suppress fitting output in vignette
) |>
  set_engine("keras")

print(spec)
#> two output reg spec Model Specification (regression)
#> 
#> Main Arguments:
#>   num_input_1 = structure(list(), class = "rlang_zap")
#>   num_input_2 = structure(list(), class = "rlang_zap")
#>   num_processed_1 = structure(list(), class = "rlang_zap")
#>   num_processed_2 = structure(list(), class = "rlang_zap")
#>   num_concatenated = structure(list(), class = "rlang_zap")
#>   num_output_1 = structure(list(), class = "rlang_zap")
#>   num_output_2 = structure(list(), class = "rlang_zap")
#>   processed_1_units = 16
#>   processed_2_units = 8
#>   learn_rate = structure(list(), class = "rlang_zap")
#>   fit_batch_size = structure(list(), class = "rlang_zap")
#>   fit_epochs = 10
#>   fit_callbacks = structure(list(), class = "rlang_zap")
#>   fit_validation_split = structure(list(), class = "rlang_zap")
#>   fit_validation_data = structure(list(), class = "rlang_zap")
#>   fit_shuffle = structure(list(), class = "rlang_zap")
#>   fit_class_weight = structure(list(), class = "rlang_zap")
#>   fit_sample_weight = structure(list(), class = "rlang_zap")
#>   fit_initial_epoch = structure(list(), class = "rlang_zap")
#>   fit_steps_per_epoch = structure(list(), class = "rlang_zap")
#>   fit_validation_steps = structure(list(), class = "rlang_zap")
#>   fit_validation_batch_size = structure(list(), class = "rlang_zap")
#>   fit_validation_freq = structure(list(), class = "rlang_zap")
#>   fit_verbose = 0
#>   fit_view_metrics = structure(list(), class = "rlang_zap")
#>   compile_optimizer = structure(list(), class = "rlang_zap")
#>   compile_loss = structure(list(), class = "rlang_zap")
#>   compile_metrics = structure(list(), class = "rlang_zap")
#>   compile_loss_weights = structure(list(), class = "rlang_zap")
#>   compile_weighted_metrics = structure(list(), class = "rlang_zap")
#>   compile_run_eagerly = structure(list(), class = "rlang_zap")
#>   compile_steps_per_execution = structure(list(), class = "rlang_zap")
#>   compile_jit_compile = structure(list(), class = "rlang_zap")
#>   compile_auto_scale_loss = structure(list(), class = "rlang_zap")
#> 
#> Computational engine: keras

# Prepare dummy data with two inputs and two outputs
set.seed(123)
x_data_1 <- matrix(runif(100 * 5), ncol = 5)
x_data_2 <- matrix(runif(100 * 3), ncol = 3)
y_data_1 <- runif(100)
y_data_2 <- runif(100) # New second output

# For tidymodels, inputs and outputs need to be in a data frame,
# potentially as lists of matrices
train_df <- tibble::tibble(
  input_1 = lapply(
    seq_len(nrow(x_data_1)),
    function(i) x_data_1[i, , drop = FALSE]
  ),
  input_2 = lapply(
    seq_len(nrow(x_data_2)),
    function(i) x_data_2[i, , drop = FALSE]
  ),
  output_1 = y_data_1, # Named output 1
  output_2 = y_data_2 # Named output 2
)

rec <- recipe(output_1 + output_2 ~ input_1 + input_2, data = train_df)
wf <- workflow() |>
  add_recipe(rec) |>
  add_model(spec)

fit_obj <- fit(wf, data = train_df)

# Predict on new data
new_data_df <- tibble::tibble(
  input_1 = lapply(seq_len(5), function(i) matrix(runif(5), ncol = 5)),
  input_2 = lapply(seq_len(5), function(i) matrix(runif(3), ncol = 3))
)
predict(fit_obj, new_data = new_data_df)
#> 1/1 - 0s - 169ms/step
#> # A tibble: 5 × 2
#>   .pred_output_1 .pred_output_2
#>      <dbl[,1,1]>    <dbl[,1,1]>
#> 1        0.387 …        0.471 …
#> 2        0.641 …        0.591 …
#> 3        0.314 …        0.539 …
#> 4        0.525 …        0.583 …
#> 5        0.606 …        0.626 …
```

## A common debugging workflow: `compile_keras_grid()`

In the original Keras guide, a common workflow is to incrementally add layers and call `summary()` to inspect the architecture. With `kerasnip`, the model is defined declaratively, so we can't inspect it layer-by-layer in the same way.

However, `kerasnip` provides a powerful equivalent: `compile_keras_grid()`. This function checks if your `layer_blocks` define a valid Keras model and returns the compiled model structure, all without running a full training cycle. This is perfect for debugging your architecture.

Let's see this in action with the `two_input_reg_spec` model:


``` r
# Create a spec instance
spec <- two_output_reg_spec( # Changed spec name
  processed_1_units = 16,
  processed_2_units = 8
)

# Prepare dummy data with two inputs and two outputs
x_dummy_1 <- matrix(runif(10 * 5), ncol = 5)
x_dummy_2 <- matrix(runif(10 * 3), ncol = 3)
y_dummy_1 <- runif(10)
y_dummy_2 <- runif(10) # New second output

# For tidymodels, inputs and outputs need to be in a data frame,
# potentially as lists of matrices
x_dummy_df <- tibble::tibble(
  input_1 = lapply(
    seq_len(nrow(x_dummy_1)),
    function(i) x_dummy_1[i, , drop = FALSE]
  ),
  input_2 = lapply(
    seq_len(nrow(x_dummy_2)),
    function(i) x_dummy_2[i, , drop = FALSE]
  )
)
y_dummy_df <- tibble::tibble(output_1 = y_dummy_1, output_2 = y_dummy_2)

# Use compile_keras_grid to get the model
# A one-row grid with no extra hyperparameters builds the spec as-is.
compilation_results <- compile_keras_grid(
  spec = spec,
  grid = tibble::tibble(.rows = 1L),
  x = x_dummy_df,
  y = y_dummy_df
)

# Print the summary
compilation_results |>
  select(compiled_model) |>
  pull() |>
  pluck(1) |>
  summary()
#> Model: "functional_1"
#> ┌───────────────────────────────────┬──────────────────────────────┬───────────────────┬───────────────────────────────
#> │ Layer (type)                      │ Output Shape                 │           Param # │ Connected to                  
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ input_1 (InputLayer)              │ (None, 1, 5)                 │                 0 │ -                             
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ input_2 (InputLayer)              │ (None, 1, 3)                 │                 0 │ -                             
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ dense_2 (Dense)                   │ (None, 1, 16)                │                96 │ input_1[0][0]                 
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ dense_3 (Dense)                   │ (None, 1, 8)                 │                32 │ input_2[0][0]                 
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ concatenate_1 (Concatenate)       │ (None, 1, 24)                │                 0 │ dense_2[0][0], dense_3[0][0]  
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ output_1 (Dense)                  │ (None, 1, 1)                 │                25 │ concatenate_1[0][0]           
#> ├───────────────────────────────────┼──────────────────────────────┼───────────────────┼───────────────────────────────
#> │ output_2 (Dense)                  │ (None, 1, 1)                 │                25 │ concatenate_1[0][0]           
#> └───────────────────────────────────┴──────────────────────────────┴───────────────────┴───────────────────────────────
#>  Total params: 178 (712.00 B)
#>  Trainable params: 178 (712.00 B)
#>  Non-trainable params: 0 (0.00 B)
```


``` r
compilation_results |>
  select(compiled_model) |>
  pull() |>
  pluck(1) |>
  plot(show_shapes = TRUE)
```

![model](images/model_plot_shapes_fAPI.png){fig-alt="A picture showing the model shape"}

## When to use the functional API

In general, the functional API is higher-level, easier and safer, and has a number of features that subclassed models do not support.

However, model subclassing provides greater flexibility when building models that are not easily expressible as directed acyclic graphs of layers. For example, you could not implement a Tree-RNN with the functional API and would have to subclass `Model` directly.

### Functional API strengths

*   **Less verbose**: There is no `super$initialize()`, no `call = function(...)`, no `self$...`, etc.
*   **Model validation during graph definition**: In the functional API, the input specification (shape and dtype) is created in advance using `layer_input()`. Each time a layer is called, it validates that the input specification matches its assumptions, raising a helpful error message if not.
*   **A functional model is plottable and inspectable**: You can plot the model as a graph, and you can easily access intermediate nodes in this graph.
*   **A functional model can be serialized or cloned**: As a data structure rather than code, a functional model is safely serializable. It can be saved as a single file, allowing you to recreate the exact same model without needing the original code.

### Functional API weakness

*   **It does not support dynamic architectures**: The functional API treats models as DAGs of layers. This is true for most deep learning architectures, but not all – for example, recursive networks or Tree RNNs do not follow this assumption and cannot be implemented in the functional API.

## Conclusion

The `create_keras_functional_spec()` function provides a powerful and intuitive way to define, fit, and tune complex Keras models within the `tidymodels` framework. By defining the model as a graph of connected blocks, you can represent nearly any architecture while `kerasnip` handles the boilerplate of integrating it with `parsnip`, `dials`, and `tune`.

# print method works

    Code
      print(rec$steps[[1]])
    Output
      Collapse predictors into list-column (unprepped)

---

    Code
      print(prepped_rec$steps[[1]])
    Output
      Collapse predictors into list-column: x1, x2  ->  predictor_matrix 

# print method works for prepped recipe

    Code
      print(rec$steps[[1]])
    Output
      Collapse predictors into list-column: x1, x2  ->  predictor_matrix 

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keras_tools.R
\name{extract_keras_model}
\alias{extract_keras_model}
\title{Extract Keras Model from a Fitted Kerasnip Object}
\usage{
extract_keras_model(object)
}
\arguments{
\item{object}{A \code{model_fit} object produced by a \code{kerasnip} specification.}
}
\value{
The raw Keras model object (\code{keras_model}).
}
\description{
Extracts and returns the underlying Keras model object from a \code{parsnip}
\code{model_fit} object created by \code{kerasnip}.
}
\details{
Extract the Raw Keras Model from a Kerasnip Fit

This is useful when you need to work directly with the Keras model object for
tasks like inspecting layer weights, creating custom plots, or passing it to
other Keras-specific functions.
}
\seealso{
keras_evaluate, extract_keras_history
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keras_tools.R
\name{extract_keras_history}
\alias{extract_keras_history}
\title{Extract Keras Training History}
\usage{
extract_keras_history(object)
}
\arguments{
\item{object}{A \code{model_fit} object produced by a \code{kerasnip} specification.}
}
\value{
A \code{keras_training_history} object. You can call \code{plot()} on this
object to visualize the learning curves.
}
\description{
Extracts and returns the training history from a \code{parsnip} \code{model_fit} object
created by \code{kerasnip}.
}
\details{
Extract Keras Training History

The history object contains the metrics recorded during model training, such
as loss and accuracy, for each epoch. This is highly useful for visualizing
the training process and diagnosing issues like overfitting.
The returned object can be plotted directly.
}
\seealso{
keras_evaluate, extract_keras_model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_model_env}
\alias{get_model_env}
\title{Get Parsnip's Model Environment}
\usage{
get_model_env()
}
\value{
The \code{parsnip} model environment.
}
\description{
This is an internal helper function to retrieve the environment where
\code{parsnip} stores its model definitions. It is used to dynamically
interact with the \code{parsnip} infrastructure.
}
\examples{
\donttest{
model_env <- kerasnip::get_model_env()
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{register_keras_optimizer}
\alias{register_keras_optimizer}
\title{Register a Custom Keras Optimizer}
\usage{
register_keras_optimizer(name, optimizer_fn)
}
\arguments{
\item{name}{The name to register the optimizer under (character).}

\item{optimizer_fn}{The optimizer function. It should return a Keras
optimizer object.}
}
\value{
No return value, called for side effects.
}
\description{
Allows users to register a custom optimizer function so it can be used by
name within \code{kerasnip} model specifications and tuned with \code{dials}.
}
\details{
Registered optimizers are stored in an internal environment. When a model is
compiled, \code{kerasnip} will first check this internal registry for an optimizer
matching the provided name before checking the \code{keras3} package.

The \code{optimizer_fn} can be a simple function or a partially applied function
using \code{purrr::partial()}. This is useful for creating versions of Keras
optimizers with specific settings.
}
\examples{
if (requireNamespace("keras3", quietly = TRUE)) {
  # Register a custom version of Adam with a different default beta_1
  my_adam <- purrr::partial(keras3::optimizer_adam, beta_1 = 0.8)
  register_keras_optimizer("my_adam", my_adam)

  # Now "my_adam" can be used as a string in a model spec, e.g.,
  # my_model_spec(compile_optimizer = "my_adam")
}
}
\seealso{
\code{\link[=register_keras_loss]{register_keras_loss()}}, \code{\link[=register_keras_metric]{register_keras_metric()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{process_y_sequential}
\alias{process_y_sequential}
\title{Process Outcome Input for Keras}
\usage{
process_y_sequential(y, is_classification = NULL, class_levels = NULL)
}
\arguments{
\item{y}{A vector of outcomes.}

\item{is_classification}{Logical, optional. If \code{TRUE}, treats \code{y} as
classification. If \code{FALSE}, treats as regression. If \code{NULL} (default),
it's determined from \code{is.factor(y)}.}

\item{class_levels}{Character vector, optional. The factor levels for
classification outcomes. If \code{NULL} (default), determined from \code{levels(y)}.}
}
\value{
A list containing:
\itemize{
\item \code{y_proc}: The processed outcome data (matrix or one-hot encoded array).
\item \code{is_classification}: Logical, indicating if \code{y} was treated as
classification.
\item \code{num_classes}: Integer, the number of classes for classification, or
\code{NULL}.
\item \code{class_levels}: Character vector, the factor levels for classification,
or \code{NULL}.
}
}
\description{
Preprocesses outcome data (\code{y}) into a format suitable for Keras models.
Handles both regression (numeric) and classification (factor) outcomes,
including one-hot encoding for classification.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/axe_methods.R
\name{axe-kerasnip_model_fit}
\alias{axe-kerasnip_model_fit}
\alias{axe_data.kerasnip_model_fit}
\alias{axe_env.kerasnip_model_fit}
\alias{axe_call.kerasnip_model_fit}
\alias{axe_ctrl.kerasnip_model_fit}
\alias{axe_fitted.kerasnip_model_fit}
\title{Butcher axe methods for kerasnip_model_fit}
\usage{
\method{axe_data}{kerasnip_model_fit}(x, verbose = FALSE, ...)

\method{axe_env}{kerasnip_model_fit}(x, verbose = FALSE, ...)

\method{axe_call}{kerasnip_model_fit}(x, verbose = FALSE, ...)

\method{axe_ctrl}{kerasnip_model_fit}(x, verbose = FALSE, ...)

\method{axe_fitted}{kerasnip_model_fit}(x, verbose = FALSE, ...)
}
\arguments{
\item{x}{A \code{kerasnip_model_fit} object.}

\item{verbose}{Logical. Print information about memory released and
disabled functions. Default is \code{FALSE}.}

\item{...}{Not used.}
}
\value{
An axed \code{kerasnip_model_fit} object with the \code{butcher_kerasnip_model_fit}
class prepended.
}
\description{
These methods allow \code{butcher::butcher()} to reduce the memory footprint of
fitted kerasnip model objects. The Keras model itself (stored as raw bytes
in \verb{$fit$keras_bytes}) is always preserved so that \code{predict()} continues
to work after butchering.

The main saving comes from \code{axe_data()}, which removes the training history
object (\verb{$fit$history}). For long training runs this can be several MB.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{model_exists}
\alias{model_exists}
\title{Check if a Kerasnip Model Specification Exists}
\usage{
model_exists(model_name)
}
\arguments{
\item{model_name}{A character string giving the name of the model
specification function to check (e.g., "my_mlp").}
}
\value{
A logical value, \code{TRUE} if the model exists, \code{FALSE} otherwise.
}
\description{
This is an internal helper function to check if a model specification has
been registered in the \code{parsnip} model environment.
}
\examples{
\donttest{
if (requireNamespace("parsnip", quietly = TRUE)) {
  library(parsnip)

  # Check for a model that exists in parsnip
  model_exists("mlp")

  # Check for a model that does not exist
  model_exists("non_existent_model")
}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/step_collapse.R
\name{step_collapse}
\alias{step_collapse}
\title{Collapse Predictors into a single list-column}
\usage{
step_collapse(
  recipe,
  ...,
  role = "predictor",
  trained = FALSE,
  columns = NULL,
  new_col = "predictor_matrix",
  skip = FALSE,
  id = recipes::rand_id("collapse")
)
}
\arguments{
\item{recipe}{A recipe object. The step will be added to the sequence of
operations for this recipe.}

\item{...}{One or more selector functions to choose which variables are
affected by the step. See \verb{[selections()]} for more details. For the \code{tidy}
method, these are not currently used.}

\item{role}{For model terms created by this step, what analysis role should
they be assigned?. By default, the new columns are used as predictors.}

\item{trained}{A logical to indicate if the quantities for preprocessing
have been estimated.}

\item{columns}{A character string of the selected variable names. This is
\code{NULL} until the step is trained by \verb{[prep.recipe()]}.}

\item{new_col}{A character string for the name of the new list-column. The
default is "predictor_matrix".}

\item{skip}{A logical. Should the step be skipped when the recipe is
baked by \verb{[bake.recipe()]}? While all operations are baked when \code{prep} is
run, skipping when \code{bake} is run may be other times when it is desirable to
skip a processing step.}

\item{id}{A character string that is unique to this step to identify it.}
}
\value{
An updated version of \code{recipe} with the new step added to the
sequence of existing steps (if any). For the \code{tidy} method, a tibble with
columns \code{terms} (the selected column names), \code{value} (the name of the
destination list-column), and \code{id} (the step identifier).
}
\description{
\code{step_collapse()} creates a a \emph{specification} of a recipe step that will
convert a group of predictors into a single list-column. This is useful
for custom models that need the predictors in a different format.
}
\examples{
library(recipes)

# 2 predictors
dat <- data.frame(
  x1 = 1:10,
  x2 = 11:20,
  y = 1:10
)

rec <- recipe(y ~ ., data = dat) \%>\%
  step_collapse(x1, x2, new_col = "pred") \%>\%
  prep()

bake(rec, new_data = NULL)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compile_keras_grid.R
\name{compile_keras_grid}
\alias{compile_keras_grid}
\title{Compile Keras Models Over a Grid of Hyperparameters}
\usage{
compile_keras_grid(spec, grid, x, y)
}
\arguments{
\item{spec}{A \code{parsnip} model specification created by
\code{create_keras_sequential_spec()} or \code{create_keras_functional_spec()}.}

\item{grid}{A \code{tibble} or \code{data.frame} containing the grid of hyperparameters
to evaluate. Each row represents a unique model architecture to be
compiled. Must have at least one row. To build the model once using only
the arguments already set on \code{spec} (e.g. for architecture inspection),
pass \code{tibble::tibble(.rows = 1L)}.}

\item{x}{A data frame or matrix of predictors. This is used to infer the
\code{input_shape} for the Keras model.}

\item{y}{A vector or factor of outcomes. This is used to infer the output
shape and the default loss function for the Keras model.}
}
\value{
A \code{tibble} with the following columns:
\itemize{
\item Columns from the input \code{grid}.
\item \code{compiled_model}: A list-column containing the compiled Keras model
objects. If compilation failed, the element will be \code{NULL}.
\item \code{error}: A list-column containing \code{NA} for successes or a
character string with the error message for failures.
}
}
\description{
Pre-compiles Keras models for each hyperparameter combination in a grid.

This function is a powerful debugging tool to use before running a full
\code{tune::tune_grid()}. It allows you to quickly validate multiple model
architectures, ensuring they can be successfully built and compiled without
the time-consuming process of actually fitting them. It helps catch common
errors like incompatible layer shapes or invalid argument values early.
}
\details{
Compile and Validate Keras Model Architectures

The function iterates through each row of the provided \code{grid}. For each
hyperparameter combination, it attempts to build and compile the Keras model
defined by the \code{spec}. The process is wrapped in a \code{try-catch} block to
gracefully handle and report any errors that occur during model instantiation
or compilation.

The output is a tibble that mirrors the input \code{grid}, with additional columns
containing the compiled model object or the error message, making it easy to
inspect which architectures are valid.
}
\examples{
\donttest{
if (requireNamespace("keras3", quietly = TRUE)) {
library(keras3)
library(parsnip)
library(dials)

# 1. Define layer blocks
input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}
hidden_block <- function(model, units = 32) {
  model |> layer_dense(units = units, activation = "relu")
}
output_block <- function(model, num_classes) {
  model |> layer_dense(units = num_classes, activation = "softmax")
}

# 2. Define a kerasnip model specification
create_keras_sequential_spec(
  model_name = "my_mlp_grid",
  layer_blocks = list(
    input = input_block,
    hidden = hidden_block,
    output = output_block
  ),
  mode = "classification"
)

mlp_spec <- my_mlp_grid(
  hidden_units = tune(),
  compile_loss = "categorical_crossentropy",
  compile_optimizer = "adam"
)

# 3. Create a hyperparameter grid
# Include an invalid value (-10) to demonstrate error handling
param_grid <- tibble::tibble(
  hidden_units = c(32, 64, -10)
)

# 4. Prepare dummy data
x_train <- matrix(rnorm(100 * 10), ncol = 10)
y_train <- factor(sample(0:1, 100, replace = TRUE))

# 5. Compile models over the grid
compiled_grid <- compile_keras_grid(
  spec = mlp_spec,
  grid = param_grid,
  x = x_train,
  y = y_train
)

print(compiled_grid)
remove_keras_spec("my_mlp_grid")

# 6. Inspect the results
# The row with `hidden_units = -10` will show an error.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remove_keras_spec.R
\name{remove_keras_spec}
\alias{remove_keras_spec}
\title{Remove a Keras Model Specification and its Registrations}
\usage{
remove_keras_spec(model_name, env = parent.frame())
}
\arguments{
\item{model_name}{A character string giving the name of the model
specification function to remove (e.g., "my_mlp").}

\item{env}{The environment from which to remove the function and its
\code{update()} method. Defaults to the calling environment (\code{parent.frame()}).}
}
\value{
Invisibly returns \code{TRUE} after attempting to remove the objects.
}
\description{
This function completely removes a model specification that was previously
created by \code{\link[=create_keras_sequential_spec]{create_keras_sequential_spec()}} or
\code{\link[=create_keras_functional_spec]{create_keras_functional_spec()}}. It cleans up both the function in the
user's environment and all associated registrations within the \code{parsnip}
package.
}
\details{
This function is essential for cleanly unloading a dynamically created model.
It performs three main actions:
\enumerate{
\item It removes the model specification function (e.g., \code{my_mlp()}) and
its corresponding \code{update()} method from the specified environment.
\item It searches \code{parsnip}'s internal model environment for all objects
whose names start with the \code{model_name} and removes them. This purges
the fit methods, argument definitions, and other registrations.
\item It removes the model's name from \code{parsnip}'s master list of models.
}
This function uses the un-exported \code{get_model_env()} to perform
the cleanup.
}
\examples{
\donttest{
if (requireNamespace("keras3", quietly = TRUE)) {
  # First, create a dummy spec
  input_block <- function(model, input_shape) {
    keras3::keras_model_sequential(input_shape = input_shape)
  }
  dense_block <- function(model, units = 16) {
    model |> keras3::layer_dense(units = units)
  }
  create_keras_sequential_spec(
    "my_temp_model",
    list(
      input = input_block,
      dense = dense_block
    ),
    "regression"
  )

  # Check it exists in the environment and in parsnip
  exists("my_temp_model")
  "my_temp_model" \%in\% parsnip::show_engines("my_temp_model")$model

  # Now remove it
  remove_keras_spec("my_temp_model")

  # Check it's gone
  !exists("my_temp_model")
  !model_exists("my_temp_model")
}
}
}
\seealso{
\code{\link[=create_keras_sequential_spec]{create_keras_sequential_spec()}}, \code{\link[=create_keras_functional_spec]{create_keras_functional_spec()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_keras_spec_helpers.R
\name{inp_spec}
\alias{inp_spec}
\title{Remap Layer Block Arguments for Model Specification}
\usage{
inp_spec(block, input_map)
}
\arguments{
\item{block}{A function that defines a Keras layer or a set of layers. The
first arguments should be the input tensor(s).}

\item{input_map}{A single character string or a named character vector that
specifies how to rename/remap the arguments of \code{block}.}
}
\value{
A new function (a closure) that wraps the \code{block} function with
renamed arguments, ready to be used in a \code{layer_blocks} list.
}
\description{
Creates a wrapper function around a Keras layer block to rename its
arguments. This is a powerful helper for defining the \code{layer_blocks} in
\code{\link[=create_keras_functional_spec]{create_keras_functional_spec()}} and \code{\link[=create_keras_sequential_spec]{create_keras_sequential_spec()}},
allowing you to connect reusable blocks into a model graph without writing
verbose anonymous functions.
}
\details{
\code{inp_spec()} makes your model definitions cleaner and more readable. It
handles the metaprogramming required to create a new function with the
correct argument names, while preserving the original block's hyperparameters
and their default values.

The function supports two modes of operation based on \code{input_map}:
\enumerate{
\item \strong{Single Input Renaming}: If \code{input_map} is a single character string,
the wrapper function renames the \emph{first} argument of the \code{block} function
to the provided string. This is the common case for blocks that take a
single tensor input.
\item \strong{Multiple Input Mapping}: If \code{input_map} is a named character vector,
the \strong{names must match the argument names of \code{block}} and each value
must be the name of an upstream layer block whose output should be fed
into that argument. This orientation matches the
syntax (e.g., \code{c(numeric = "processed_numerical")}). This is used for
blocks with multiple inputs, like a concatenation layer.
}

\emph{Note}: Prior releases accepted the opposite orientation
(\code{c(processed_numerical = "numeric")}). Existing code written in that style
must flip the names/values when upgrading to this version.
}
\examples{
\donttest{
# --- Example Blocks ---
# A standard dense block with one input tensor and one hyperparameter.
dense_block <- function(tensor, units = 16) {
  tensor |> keras3::layer_dense(units = units, activation = "relu")
}

# A block that takes two tensors as input.
concat_block <- function(input_a, input_b) {
  keras3::layer_concatenate(list(input_a, input_b))
}

# An output block with one input.
output_block <- function(tensor) {
  tensor |> keras3::layer_dense(units = 1)
}

# --- Usage ---
layer_blocks <- list(
  main_input = keras3::layer_input,
  path_a = inp_spec(dense_block, "main_input"),
  path_b = inp_spec(dense_block, "main_input"),
  concatenated = inp_spec(
    concat_block,
    c(input_a = "path_a", input_b = "path_b")
  ),
  output = inp_spec(output_block, "concatenated")
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{register_keras_metric}
\alias{register_keras_metric}
\title{Register a Custom Keras Metric}
\usage{
register_keras_metric(name, metric_fn)
}
\arguments{
\item{name}{The name to register the metric under (character).}

\item{metric_fn}{The metric function.}
}
\value{
No return value, called for side effects.
}
\description{
Allows users to register a custom metric function so it can be used by name
within \code{kerasnip} model specifications.
}
\details{
Registered metrics are stored in an internal environment. When a model is
compiled, \code{kerasnip} will first check this internal registry for a metric
matching the provided name before checking the \code{keras3} package.
}
\seealso{
\code{\link[=register_keras_optimizer]{register_keras_optimizer()}}, \code{\link[=register_keras_loss]{register_keras_loss()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{data}
\name{keras_objects}
\alias{keras_objects}
\alias{keras_optimizers}
\alias{keras_losses}
\alias{keras_metrics}
\title{Dynamically Discovered Keras Objects}
\format{
An object of class \code{character} of length 13.

An object of class \code{character} of length 22.

An object of class \code{character} of length 32.
}
\usage{
keras_optimizers

keras_losses

keras_metrics
}
\description{
These exported vectors contain the names of optimizers, losses, and metrics
discovered from the installed \code{keras3} package when \code{kerasnip} is loaded.
This ensures that \code{kerasnip} is always up-to-date with your Keras version.
}
\details{
These objects are primarily used to provide the default \code{values} for the
\code{dials} parameter functions, \code{\link[=optimizer_function]{optimizer_function()}} and
\code{\link[=loss_function_keras]{loss_function_keras()}}. This allows for tab-completion in IDEs and
validation of optimizer and loss names when tuning models.

The discovery process in \code{.onLoad()} scrapes the \code{keras3} namespace for
functions matching \verb{optimizer_*}, \verb{loss_*}, and \verb{metric_*} patterns.
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_functional_fit.R
\name{generic_functional_fit}
\alias{generic_functional_fit}
\title{Internal Fitting Engine for Functional API Models}
\usage{
generic_functional_fit(formula, data, layer_blocks, ...)
}
\arguments{
\item{formula}{A formula specifying the predictor and outcome variables,
passed down from the \code{parsnip::fit()} call.}

\item{data}{A data frame containing the training data, passed down from the
\code{parsnip::fit()} call.}

\item{layer_blocks}{A named list of layer block functions. This is passed
internally from the \code{parsnip} model specification.}

\item{...}{Additional arguments passed down from the model specification.
These can include:
\itemize{
\item \strong{Layer Parameters:} Arguments for the layer blocks, prefixed
with the block name (e.g., \code{dense_units = 64}).
\item \strong{Architecture Parameters:} Arguments to control the number
of times a block is repeated, in the format \verb{num_\{block_name\}} (e.g.,
\code{num_dense = 2}).
\item \strong{Compile Parameters:} Arguments to customize model
compilation, prefixed with \code{compile_} (e.g., \code{compile_loss = "mae"},
\code{compile_optimizer = "sgd"}).
\item \strong{Fit Parameters:} Arguments to customize model fitting,
prefixed with \code{fit_} (e.g., \code{fit_callbacks = list(...)},
\code{fit_class_weight = list(...)}).
}}
}
\value{
A list containing the fitted model and other metadata. This list is
stored in the \code{fit} slot of the \code{parsnip} model fit object. The list
contains the following elements:
\itemize{
\item \code{fit}: The raw, fitted Keras model object.
\item \code{history}: The Keras training history object.
\item \code{lvl}: A character vector of the outcome factor levels (for
classification) or \code{NULL} (for regression).
}
}
\description{
This function serves as the internal engine for fitting \code{kerasnip} models
that are based on the Keras functional API. It is not intended to be called
directly by the user. The function is invoked by \code{parsnip::fit()} when a
\code{kerasnip} functional model specification is used.
}
\details{
Generic Fitting Function for Functional Keras Models

The function orchestrates the three main steps of the model fitting process:
\enumerate{
\item \strong{Build and Compile:} It calls
\code{build_compile_func_model()} to construct the Keras model
architecture based on the provided \code{layer_blocks} and hyperparameters.
\item \strong{Process Data:} It preprocesses the input (\code{x}) and output
(\code{y}) data into the format expected by Keras.
\item \strong{Fit Model:} It calls \code{keras3::fit()} with the compiled model
and processed data, passing along any fitting-specific arguments (e.g.,
\code{epochs}, \code{batch_size}, \code{callbacks}).
}
}
\examples{
# This function is not called directly by users.
# It is called internally by `parsnip::fit()`.
# For example:
\donttest{
library(parsnip)
# create_keras_functional_spec(...) defines my_functional_model

# spec <- my_functional_model(hidden_units = 128, fit_epochs = 10) |>
#   set_engine("keras")

# # This call to fit() would invoke generic_functional_fit() internally
# fitted_model <- fit(spec, y ~ x, data = training_data)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{register_keras_loss}
\alias{register_keras_loss}
\title{Register a Custom Keras Loss}
\usage{
register_keras_loss(name, loss_fn)
}
\arguments{
\item{name}{The name to register the loss under (character).}

\item{loss_fn}{The loss function.}
}
\value{
No return value, called for side effects.
}
\description{
Allows users to register a custom loss function so it can be used by name
within \code{kerasnip} model specifications and tuned with \code{dials}.
}
\details{
Registered losses are stored in an internal environment. When a model is
compiled, \code{kerasnip} will first check this internal registry for a loss
matching the provided name before checking the \code{keras3} package.
}
\seealso{
\code{\link[=register_keras_optimizer]{register_keras_optimizer()}}, \code{\link[=register_keras_metric]{register_keras_metric()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_keras_sequential_spec.R
\name{create_keras_sequential_spec}
\alias{create_keras_sequential_spec}
\title{Create a Custom Keras Sequential Model Specification for Tidymodels}
\usage{
create_keras_sequential_spec(
  model_name,
  layer_blocks,
  mode = c("regression", "classification"),
  ...,
  env = parent.frame()
)
}
\arguments{
\item{model_name}{A character string for the name of the new model
specification function (e.g., "custom_cnn"). This should be a valid R
function name.}

\item{layer_blocks}{A named, ordered list of functions. Each function defines
a "block" of Keras layers. The function must take a Keras model object as
its first argument and return the modified model. Other arguments to the
function will become tunable parameters in the final model specification.}

\item{mode}{A character string, either "regression" or "classification".}

\item{...}{Reserved for future use. Currently not used.}

\item{env}{The environment in which to create the new model specification
function and its associated \code{update()} method. Defaults to the calling
environment (\code{parent.frame()}).}
}
\value{
Invisibly returns \code{NULL}. Its primary side effect is to create a new
model specification function (e.g., \code{my_mlp()}) in the specified
environment and register the model with \code{parsnip} so it can be used within
the \code{tidymodels} framework.
}
\description{
This function acts as a factory to generate a new \code{parsnip} model
specification based on user-defined blocks of Keras layers using the
Sequential API. This is the ideal choice for creating models that are a
simple, linear stack of layers. For models with complex, non-linear
topologies, see \code{\link[=create_keras_functional_spec]{create_keras_functional_spec()}}.
}
\details{
This function generates all the boilerplate needed to create a custom,
tunable \code{parsnip} model specification that uses the Keras Sequential API.

The function inspects the arguments of your \code{layer_blocks} functions
(ignoring special arguments like \code{input_shape} and \code{num_classes})
and makes them available as arguments in the generated model specification,
prefixed with the block's name (e.g., \code{dense_units}).

The new model specification function and its \code{update()} method are created in
the environment specified by the \code{env} argument.
}
\section{Model Architecture (Sequential API)}{

\code{kerasnip} builds the model by applying the functions in \code{layer_blocks} in
the order they are provided. Each function receives the Keras model built by
the previous function and returns a modified version.
\enumerate{
\item The \strong{first block} must initialize the model (e.g., with
\code{keras_model_sequential()}). It can accept an \code{input_shape} argument,
which \code{kerasnip} will provide automatically during fitting.
\item \strong{Subsequent blocks} add layers to the model.
\item The \strong{final block} should add the output layer. For classification, it
can accept a \code{num_classes} argument, which is provided automatically.
}

A key feature of this function is the automatic creation of
\verb{num_\{block_name\}} arguments (e.g., \code{num_hidden}). This allows you to
control how many times each block is repeated, making it easy to tune the
depth of your network.
}

\section{Saving and Reloading Models}{

To save a fitted workflow and reload it in a new R session, use
\code{bundle::bundle()} before saving — this is required to preserve the Keras
model weights:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(bundle)
bundled <- bundle(fitted_workflow)
saveRDS(bundled, "model.rds")

# New session:
library(kerasnip); library(bundle)
fitted_workflow <- unbundle(readRDS("model.rds"))
predict(fitted_workflow, new_data = test_data)  # works
}\if{html}{\out{</div>}}

Plain \code{saveRDS()} without \code{bundle()} does not preserve Keras weights, but
\code{predict()} will still auto-register the parsnip model type from metadata
stored on the spec.
}

\examples{
\donttest{
if (requireNamespace("keras3", quietly = TRUE)) {
library(keras3)
library(parsnip)
library(dials)

# 1. Define layer blocks for a complete model.
# The first block must initialize the model. `input_shape` is passed
# automatically.
input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}
# A block for hidden layers. `units` will become a tunable parameter.
hidden_block <- function(model, units = 32) {
  model |> layer_dense(units = units, activation = "relu")
}

# The output block. `num_classes` is passed automatically for classification.
output_block <- function(model, num_classes) {
  model |> layer_dense(units = num_classes, activation = "softmax")
}

# 2. Create the spec, providing blocks in the correct order.
create_keras_sequential_spec(
model_name = "my_mlp_seq_spec",
  layer_blocks = list(
    input = input_block,
    hidden = hidden_block,
    output = output_block
  ),
  mode = "classification"
)

# 3. Use the newly created specification function!
# Note the new arguments `num_hidden` and `hidden_units`.
model_spec <- my_mlp_seq_spec(
  num_hidden = 2,
  hidden_units = 64,
  epochs = 10,
  learn_rate = 0.01
)

print(model_spec)
remove_keras_spec("my_mlp_seq_spec")
}
}
}
\seealso{
\code{\link[=remove_keras_spec]{remove_keras_spec()}}, \code{\link[parsnip:add_on_exports]{parsnip::new_model_spec()}},
\code{\link[=create_keras_functional_spec]{create_keras_functional_spec()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kerasnip_spec_methods.R
\name{fit.kerasnip_spec}
\alias{fit.kerasnip_spec}
\title{Fit Method for kerasnip Spec Objects}
\usage{
\method{fit}{kerasnip_spec}(object, ...)
}
\arguments{
\item{object}{A \code{kerasnip_spec} model specification.}

\item{...}{Passed to \code{parsnip::fit.model_spec()}.}
}
\value{
A \code{model_fit} object with the additional \code{kerasnip_model_fit} class
prepended to its class vector.
}
\description{
S3 method for \code{fit()} dispatched on \code{kerasnip_spec} objects. Delegates to
the standard parsnip \code{fit.model_spec()} and then tags the result with the
\code{kerasnip_model_fit} class so that \code{predict.kerasnip_model_fit()} is
dispatched on subsequent calls.

\code{kerasnip_spec} is stripped from the class before \code{NextMethod()} to prevent
parsnip's internal \code{specific_model()} helper from returning more than one
model-class entry, which would break registry lookups. The custom metadata
attributes remain on the object and are thus stored inside the resulting
\code{model_fit$spec}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kerasnip_spec_methods.R
\name{set_engine.kerasnip_spec}
\alias{set_engine.kerasnip_spec}
\title{set_engine Method for kerasnip Spec Objects}
\usage{
\method{set_engine}{kerasnip_spec}(object, engine, ...)
}
\arguments{
\item{object}{A \code{kerasnip_spec} model specification.}

\item{engine}{A character string naming the engine (e.g., \code{"keras"}).}

\item{...}{Additional engine-specific arguments passed to
\code{parsnip::set_engine.model_spec()}.}
}
\value{
A \code{model_spec} object with the \code{kerasnip_spec} class and metadata
attributes re-attached.
}
\description{
S3 method for \code{set_engine()} dispatched on \code{kerasnip_spec} objects.
\code{parsnip::set_engine.model_spec()} internally calls \code{new_model_spec()},
which re-creates the spec from scratch with only \code{c(cls, "model_spec")} as
the class vector — stripping \code{kerasnip_spec} and any custom attributes. This
wrapper preserves the \code{kerasnip_layer_blocks} and \code{kerasnip_functional}
metadata attributes and re-attaches them (along with the \code{kerasnip_spec}
class) after \code{NextMethod()} has done its work.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compile_keras_grid.R
\name{extract_valid_grid}
\alias{extract_valid_grid}
\title{Extract Valid Grid from Compilation Results}
\usage{
extract_valid_grid(compiled_grid)
}
\arguments{
\item{compiled_grid}{A tibble, the result of a call to
\code{compile_keras_grid()}.}
}
\value{
A tibble containing the subset of the original grid that resulted in
a successful model compilation. The \code{compiled_model} and \code{error} columns
are removed, leaving a clean grid ready for tuning.
}
\description{
This helper function filters the results from \code{compile_keras_grid()} to
return a new hyperparameter grid containing only the combinations that
compiled successfully.
}
\details{
Filter a Grid to Only Valid Hyperparameter Sets

After running \code{compile_keras_grid()}, you can use this function to remove
problematic hyperparameter combinations before proceeding to the full
\code{tune::tune_grid()}.
}
\examples{
\donttest{
if (requireNamespace("keras3", quietly = TRUE)) {
  library(keras3)
  library(parsnip)
  library(dials)

  # 1. Define layer blocks
  input_block <- function(model, input_shape) {
    keras_model_sequential(input_shape = input_shape)
  }
  hidden_block <- function(model, units = 32) {
    model |> layer_dense(units = units, activation = "relu")
  }
  output_block <- function(model, num_classes) {
    model |> layer_dense(units = num_classes, activation = "softmax")
  }

  # 2. Define a kerasnip model specification
  create_keras_sequential_spec(
    model_name = "my_mlp_grid_2",
    layer_blocks = list(
      input = input_block,
      hidden = hidden_block,
      output = output_block
    ),
    mode = "classification"
  )

  mlp_spec <- my_mlp_grid_2(
    hidden_units = tune(),
    compile_loss = "categorical_crossentropy",
    compile_optimizer = "adam"
  )

  # 3. Create a hyperparameter grid
  param_grid <- tibble::tibble(
    hidden_units = c(32, 64, -10)
  )

  # 4. Prepare dummy data
  x_train <- matrix(rnorm(100 * 10), ncol = 10)
  y_train <- factor(sample(0:1, 100, replace = TRUE))

  # 5. Compile models over the grid
  compiled_grid <- compile_keras_grid(
    spec = mlp_spec,
    grid = param_grid,
    x = x_train,
    y = y_train
  )

  # 6. Extract the valid grid
  valid_grid <- extract_valid_grid(compiled_grid)
  print(valid_grid)
  remove_keras_spec("my_mlp_grid_2")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dials.R
\name{optimizer_function}
\alias{optimizer_function}
\title{Dials Parameter for Keras Optimizers}
\usage{
optimizer_function(values = NULL)
}
\arguments{
\item{values}{A character vector of possible optimizers. Defaults to all
known optimizers (keras defaults + custom registered).}
}
\value{
A \code{dials} parameter object for Keras optimizers.
}
\description{
Dials Parameter for Keras Optimizers
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keras_tools.R
\name{keras_evaluate}
\alias{keras_evaluate}
\title{Evaluate a Kerasnip Model}
\usage{
keras_evaluate(object, x, y = NULL, ...)
}
\arguments{
\item{object}{A \code{model_fit} object produced by a \code{kerasnip} specification.}

\item{x}{A data frame or matrix of new predictor data.}

\item{y}{A vector or data frame of new outcome data corresponding to \code{x}.}

\item{...}{Additional arguments passed on to \code{keras3::evaluate()} (e.g.,
\code{batch_size}).}
}
\value{
A named list containing the evaluation results (e.g., \code{loss},
\code{accuracy}). The names are determined by the metrics the model was compiled
with.
}
\description{
This function provides an \code{kera_evaluate()} method for \code{model_fit} objects
created by \code{kerasnip}. It preprocesses the new data into the format expected
by Keras and then calls \code{keras3::evaluate()} on the underlying model to
compute the loss and any other metrics.
}
\details{
Evaluate a Fitted Kerasnip Model on New Data
}
\examples{
\donttest{
if (requireNamespace("keras3", quietly = TRUE)) {
library(keras3)
library(parsnip)

# 1. Define layer blocks
input_block <- function(model, input_shape) {
  keras_model_sequential(input_shape = input_shape)
}
hidden_block <- function(model, units = 32) {
  model |> layer_dense(units = units, activation = "relu")
}
output_block <- function(model, num_classes) {
  model |> layer_dense(units = num_classes, activation = "softmax")
}

# 2. Define and fit a model ----
create_keras_sequential_spec(
  model_name = "my_mlp_tools",
  layer_blocks = list(
    input = input_block,
    hidden = hidden_block,
    output = output_block
  ),
  mode = "classification"
)

mlp_spec <- my_mlp_tools(
  hidden_units = 32,
  compile_loss = "categorical_crossentropy",
  compile_optimizer = "adam",
  compile_metrics = "accuracy",
  fit_epochs = 5
) |> set_engine("keras")

x_train <- matrix(rnorm(100 * 10), ncol = 10)
y_train <- factor(sample(0:1, 100, replace = TRUE))
train_df <- data.frame(x = I(x_train), y = y_train)

fitted_mlp <- fit(mlp_spec, y ~ x, data = train_df)

# 3. Evaluate the model on new data ----
x_test <- matrix(rnorm(50 * 10), ncol = 10)
y_test <- factor(sample(0:1, 50, replace = TRUE))

eval_metrics <- keras_evaluate(fitted_mlp, x_test, y_test)
print(eval_metrics)

# 4. Extract the Keras model object ----
keras_model <- extract_keras_model(fitted_mlp)
summary(keras_model)

# 5. Extract the training history ----
history <- extract_keras_history(fitted_mlp)
plot(history)
remove_keras_spec("my_mlp_tools")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dials.R
\name{loss_function_keras}
\alias{loss_function_keras}
\title{Dials Parameter for Keras Loss Functions}
\usage{
loss_function_keras(values = NULL)
}
\arguments{
\item{values}{A character vector of possible loss functions. Defaults to all
known losses (keras defaults + custom registered).}
}
\value{
A \code{dials} parameter object for Keras loss.
}
\description{
Dials Parameter for Keras Loss Functions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keras_tools.R
\name{tidy.kerasnip_model_fit}
\alias{tidy.kerasnip_model_fit}
\title{Tidy a Fitted Kerasnip Model}
\usage{
\method{tidy}{kerasnip_model_fit}(x, ...)
}
\arguments{
\item{x}{A \code{kerasnip_model_fit} object.}

\item{...}{Not used.}
}
\value{
A tibble with columns \code{layer} (character), \code{class} (character),
and \code{n_params} (integer).
}
\description{
Returns a tibble with one row per layer of the underlying Keras model,
summarising the layer name, Python class, and parameter count.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kerasnip_spec_methods.R
\name{set_args.kerasnip_spec}
\alias{set_args.kerasnip_spec}
\title{set_args Method for kerasnip Spec Objects}
\usage{
\method{set_args}{kerasnip_spec}(object, ...)
}
\arguments{
\item{object}{A \code{kerasnip_spec} model specification.}

\item{...}{Named model arguments to update, passed to
\code{parsnip::set_args.model_spec()}.}
}
\value{
A \code{model_spec} object with the \code{kerasnip_spec} class and metadata
attributes re-attached.
}
\description{
S3 method for \code{set_args()} dispatched on \code{kerasnip_spec} objects.
\code{parsnip::set_args.model_spec()} calls \code{new_model_spec()}, which strips any
extra classes and attributes. This wrapper saves and re-attaches the
\code{kerasnip_layer_blocks} and \code{kerasnip_functional} metadata attributes (and
the \code{kerasnip_spec} class) after \code{NextMethod()} has done its work.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{process_x_sequential}
\alias{process_x_sequential}
\title{Process Predictor Input for Keras}
\usage{
process_x_sequential(x)
}
\arguments{
\item{x}{A data frame or matrix of predictors.}
}
\value{
A list containing:
\itemize{
\item \code{x_proc}: The processed predictor data (matrix or array).
\item \code{input_shape}: The determined input shape for the Keras model.
}
}
\description{
Preprocesses predictor data (\code{x}) into a format suitable for Keras models.
Handles both tabular data and list-columns of arrays (e.g., for images).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{process_x_functional}
\alias{process_x_functional}
\title{Process Predictor Input for Keras (Functional API)}
\usage{
process_x_functional(x)
}
\arguments{
\item{x}{A data frame or matrix of predictors.}
}
\value{
A list containing:
\itemize{
\item \code{x_proc}: The processed predictor data (matrix or array, or list of
arrays).
\item \code{input_shape}: The determined input shape(s) for the Keras model.
}
}
\description{
Preprocesses predictor data (\code{x}) into a format suitable for Keras models
built with the Functional API. Handles both tabular data and list-columns
of arrays (e.g., for images), supporting multiple inputs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kerasnip_spec_methods.R
\name{predict.kerasnip_model_fit}
\alias{predict.kerasnip_model_fit}
\title{Predict Method for kerasnip Model Fits}
\usage{
\method{predict}{kerasnip_model_fit}(object, new_data, ...)
}
\arguments{
\item{object}{A \code{kerasnip_model_fit} object.}

\item{new_data}{A data frame of predictors.}

\item{...}{Passed to the parsnip predict method.}
}
\value{
A tibble of predictions.
}
\description{
S3 method for \code{predict()} dispatched on \code{kerasnip_model_fit} objects.
Before delegating to the standard parsnip predict machinery, it checks
whether the underlying model type is registered in the current parsnip
session. If not (e.g. after loading a saved workflow in a new R session),
it transparently replays the full parsnip registration using metadata stored
on the spec object — requiring no manual step from the user.
}
\details{
The metadata needed for re-registration (\code{kerasnip_layer_blocks},
\code{kerasnip_functional}) is embedded on the spec object by the spec
constructor function at call time. This means it is preserved across
\code{saveRDS()}/\code{readRDS()} and \code{bundle()}/\code{unbundle()} round-trips.

For full model weight portability (i.e. to be able to \code{predict()} on new
data in a new R session), use \code{bundle::bundle()} before saving. Plain
\code{saveRDS()} preserves the spec structure and will auto-register, but the
underlying Keras model weights are not portable without bundling.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keras_tools.R
\name{glance.kerasnip_model_fit}
\alias{glance.kerasnip_model_fit}
\title{Glance at a Fitted Kerasnip Model}
\usage{
\method{glance}{kerasnip_model_fit}(x, ...)
}
\arguments{
\item{x}{A \code{kerasnip_model_fit} object.}

\item{...}{Not used.}
}
\value{
A one-row tibble with one column per compiled metric. Returns an
empty tibble if training history has been stripped (e.g. by butcher).
}
\description{
Returns a one-row tibble of summary statistics from the final training epoch:
every metric the model was compiled with (e.g. \code{loss}, \code{accuracy}).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_keras_functional_spec.R
\name{create_keras_functional_spec}
\alias{create_keras_functional_spec}
\title{Create a Custom Keras Functional API Model Specification for Tidymodels}
\usage{
create_keras_functional_spec(
  model_name,
  layer_blocks,
  mode = c("regression", "classification"),
  ...,
  env = parent.frame()
)
}
\arguments{
\item{model_name}{A character string for the name of the new model
specification function (e.g., "custom_resnet"). This should be a valid R
function name.}

\item{layer_blocks}{A named list of functions where each function defines a
"block" (a node) in the model graph. The list names are crucial as they
define the names of the nodes. The arguments of each function define how
the nodes are connected. See the "Model Graph Connectivity" section for
details.}

\item{mode}{A character string, either "regression" or "classification".}

\item{...}{Reserved for future use. Currently not used.}

\item{env}{The environment in which to create the new model specification
function and its associated \code{update()} method. Defaults to the calling
environment (\code{parent.frame()}).}
}
\value{
Invisibly returns \code{NULL}. Its primary side effect is to create a
new model specification function (e.g., \code{custom_resnet()}) in the
specified environment and register the model with \code{parsnip} so it can be
used within the \code{tidymodels} framework.
}
\description{
This function acts as a factory to generate a new \code{parsnip} model
specification based on user-defined blocks of Keras layers using the
Functional API. This allows for creating complex, tunable architectures
with non-linear topologies that integrate seamlessly with the \code{tidymodels}
ecosystem.
}
\details{
This function generates all the boilerplate needed to create a custom,
tunable \code{parsnip} model specification that uses the Keras Functional API.
This is ideal for models with complex, non-linear topologies, such as
networks with multiple inputs/outputs or residual connections.

The function inspects the arguments of your \code{layer_blocks} functions and
makes them available as tunable parameters in the generated model
specification, prefixed with the block's name (e.g., \code{dense_units}).
Common training parameters such as \code{epochs} and \code{learn_rate} are also added.
}
\section{Model Graph Connectivity}{

\code{kerasnip} builds the model's directed acyclic graph by inspecting the
arguments of each function in the \code{layer_blocks} list. The connection logic
is as follows:
\enumerate{
\item The \strong{names of the elements} in the \code{layer_blocks} list define the names
of the nodes in your graph (e.g., \code{main_input}, \code{dense_path}, \code{output}).
\item The \strong{names of the arguments} in each block function specify its inputs.
A block function like \verb{my_block <- function(input_a, input_b, ...)}
declares that it needs input from the nodes named \code{input_a} and
\code{input_b}. \code{kerasnip} will automatically supply the output tensors from
those nodes when calling \code{my_block}.
}

There are two special requirements:
\itemize{
\item \strong{Input Block}: The first block in the list is treated as the input
node. Its function should not take other blocks as input, but it can have
an \code{input_shape} argument, which is supplied automatically during
fitting.
\item \strong{Output Block}: Exactly one block must be named \code{"output"}. The tensor
returned by this block is used as the final output of the Keras model.
}

A key feature is the automatic creation of \verb{num_\{block_name\}} arguments
(e.g., \code{num_dense_path}). This allows you to control how many times a block
is repeated, making it easy to tune the depth of your network. A block can
only be repeated if it has exactly one input from another block in the graph.

The new model specification function and its \code{update()} method are created
in the environment specified by the \code{env} argument.
}

\section{Saving and Reloading Models}{

To save a fitted workflow and reload it in a new R session, use
\code{bundle::bundle()} before saving — this is required to preserve the Keras
model weights:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(bundle)
bundled <- bundle(fitted_workflow)
saveRDS(bundled, "model.rds")

# New session:
library(kerasnip); library(bundle)
fitted_workflow <- unbundle(readRDS("model.rds"))
predict(fitted_workflow, new_data = test_data)  # works
}\if{html}{\out{</div>}}

Plain \code{saveRDS()} without \code{bundle()} does not preserve Keras weights, but
\code{predict()} will still auto-register the parsnip model type from metadata
stored on the spec.
}

\examples{
\donttest{
if (requireNamespace("keras3", quietly = TRUE)) {
  library(keras3)
  library(parsnip)

  # 1. Define block functions. These are the building blocks of our model.
  # An input block that receives the data's shape automatically.
  input_block <- function(input_shape) layer_input(shape = input_shape)

  # A dense block with a tunable `units` parameter.
  dense_block <- function(tensor, units) {
    tensor |> layer_dense(units = units, activation = "relu")
  }

  # A block that adds two tensors together (for the residual connection).
  add_block <- function(input_a, input_b) layer_add(list(input_a, input_b))

  # An output block for regression.
  output_block_reg <- function(tensor) layer_dense(tensor, units = 1)

  # 2. Create the spec. The `layer_blocks` list defines the graph.
  create_keras_functional_spec(
    model_name = "my_resnet_spec",
    layer_blocks = list(
      # The names of list elements are the node names.
      main_input = input_block,

      # The argument `main_input` connects this block to the input node.
      dense_path = function(main_input, units = 32) {
        dense_block(main_input, units)
      },

      # This block's arguments connect it to the original input AND the dense
      # layer.
      add_residual = function(main_input, dense_path) {
        add_block(main_input, dense_path)
      },

      # This block must be named 'output'. It connects to the residual add
      # layer.
      output = function(add_residual) output_block_reg(add_residual)
    ),
    mode = "regression"
  )

  # 3. Use the newly created specification function!
  # The `dense_path_units` argument was created automatically.
  model_spec <- my_resnet_spec(dense_path_units = 64, epochs = 10)

  # You could also tune the number of dense layers since it has a single
  # input:
  # model_spec <- my_resnet_spec(num_dense_path = 2, dense_path_units = 32)

  print(model_spec)
  remove_keras_spec("my_resnet_spec")
  # tune::tunable(model_spec)
}
}
}
\seealso{
\code{\link[=remove_keras_spec]{remove_keras_spec()}}, \code{\link[parsnip:add_on_exports]{parsnip::new_model_spec()}},
\code{\link[=create_keras_sequential_spec]{create_keras_sequential_spec()}}
}
